/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import kumbang.configurator.client.plugin.Colors;
import kumbang.configurator.client.plugin.IKumbangPlugin;
import kumbang.configurator.client.plugin.Icons;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.configurator.client.plugin.PluginHandler;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedEvent;
import kumbang.configurator.client.plugin.preferences.PreferenceConstants;
import kumbang.configurator.server.core.ServerProperties;

import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.resource.ColorRegistry;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;

/**
 * The activator class controls the plug-in life cycle
 */
public class Activator extends AbstractUIPlugin implements IKumbangPlugin {

	// The plug-in ID
	public static final String PLUGIN_ID = "kumbang.configurator";

	
	// The shared instance
	private static Activator plugin;


	
	// Preference store for the Window->Preference-stuff
	private IPreferenceStore store;
	
	// Listener for preference changes
	private IPropertyChangeListener listener;
	
	private ImageRegistry imageRegistry;

	private ColorRegistry colorRegistry;
	
	private PluginHandler handler;
	private String eclipsePluginPath;
	
	
	/**
	 * The constructor
	 */
	public Activator() {
		plugin = this;
		handler = new PluginHandler(this);
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#start(org.osgi.framework.BundleContext)
	 */
	public void start(BundleContext context) throws Exception {
		super.start(context);
		store = getPreferenceStore();
		colorRegistry = JFaceResources.getColorRegistry();
		initializeColorRegistry();
		
		// Listener for the property changes in the preferences page
		this.listener =	new IPropertyChangeListener() {
		    /* Setup the listener for changes in the preferences */
		    public void propertyChange(PropertyChangeEvent e) {
		    	
		    	// Communication method was changed
		    	if (e.getProperty().equals(PreferenceConstants.P_COMMUNICATIONMETHOD)) {

		    		// We need to restart the connection to the server
		    		if (handler.getClientServices() != null) {
		    			System.out.println("Communication method changed, restarting connection");
		    			handler.fireOpenCloseEvent(ConfigurationOpenedClosedEvent.EVENT_CLOSE);
		    			handler.endSession();
		    		} else {
		    			// There was no client yet, so just get a new one
		    			handler.getClientServices();
		    		}	

		    		// If the method is set to local, load the properties from the preferenec page
		    		if (((String) e.getNewValue()).equals("local")) {
		    			loadServerProperties();		    		
		    		}	
		    		
		    	}
		    	// If the server address or port is changed, we need to restart the connection 
		    	// to the new address
		    	else if (e.getProperty().equals(PreferenceConstants.P_SERVERADDRESS)
						|| e.getProperty().equals(PreferenceConstants.P_SERVERPORT)) {
					System.out.println("Server location changed, restarting connection");
					handler.fireOpenCloseEvent(ConfigurationOpenedClosedEvent.EVENT_CLOSE);
					handler.endSession();
		    	} 
		    }

		};
		
		store.addPropertyChangeListener(listener);
 
		// If the server is set to local initially, set the default, working preferences now
		// These can't be changed at the moment, which is probably better when using
		// local server
		if (((String)store.getString(PreferenceConstants.P_COMMUNICATIONMETHOD)).equals("local")) {
			loadServerProperties();
		}

		
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext)
	 */
	public void stop(BundleContext context) throws Exception {
		plugin = null;
		super.stop(context);
		handler.endSession();
		handler.clearListeners();
		store.removePropertyChangeListener(listener);
	}

	/**
	 * Returns the shared instance
	 *
	 * @return the shared instance
	 */
	public static Activator getDefault() {
		return plugin;
	}
	
	protected void initializeDefaultPreferences(IPreferenceStore store) {
	    store.setDefault(PreferenceConstants.P_COMMUNICATIONMETHOD, "local");
	    store.setDefault(PreferenceConstants.P_SERVERADDRESS, "localhost");
	    store.setDefault(PreferenceConstants.P_SERVERPORT, "6969");
	    store.setDefault(PreferenceConstants.P_UML_INTERFACES, false);
	}

	private void initializeColorRegistry() {
		colorRegistry.put(Colors.COMPLETE_CONFIGURATION, new RGB(0,204,51));
		colorRegistry.put(Colors.INCOMPLETE_CONFIGURATION, new RGB(204,0,51));
		colorRegistry.put(Colors.COMPONENT, new RGB(205, 220, 235));
		colorRegistry.put(Colors.BLACK, new RGB(0,0,0));
		colorRegistry.put(Colors.WHITE, new RGB(255,255,255));
		colorRegistry.put(Colors.SELECTION, new RGB(230, 130, 20));
		colorRegistry.put(Colors.ACTION, new RGB(250, 179, 83));
		colorRegistry.put(Colors.GRAY, new RGB(125, 125, 125));
		colorRegistry.put(Colors.INTERFACE, new RGB(250, 250, 210));
		colorRegistry.put(Colors.TOOLBAR, new RGB(200, 183, 145));
		
	}
	
	public ColorRegistry getColorRegistry() {
		return colorRegistry;
	}
	
	private void loadServerProperties() {
		ServerProperties.getInstance().setParserClass(ServerProperties.DEFAULT_PARSER_CLASS);
	}
	
	/**
	 * Gets the ImageRegistry, which holds the images that
	 * are frequently used.
	 */
	public ImageRegistry getImageRegistry() {
		if (imageRegistry == null) {
			imageRegistry = createImageRegistry();
			initializeImageRegistry(imageRegistry);
		}
		return imageRegistry;
	}
	
	/**
	 * Initializes ImageRegistry and loads frequently used images into it.
	 */
	protected void initializeImageRegistry(ImageRegistry imageRegistry) {
		imageRegistry.put(Icons.ATTRIBUTE, getImageDescriptor(Icons.ATTRIBUTE));
		imageRegistry.put(Icons.BINDING_SOURCE, getImageDescriptor(Icons.BINDING_SOURCE));
		imageRegistry.put(Icons.BINDING_TARGET, getImageDescriptor(Icons.BINDING_TARGET));
		imageRegistry.put(Icons.COMPONENT, getImageDescriptor(Icons.COMPONENT));
        imageRegistry.put(Icons.CONNECT, getImageDescriptor(Icons.CONNECT));
        imageRegistry.put(Icons.CURSOR_CONNECT_1, getImageDescriptor(Icons.CURSOR_CONNECT_1));
        imageRegistry.put(Icons.CURSOR_CONNECT_2, getImageDescriptor(Icons.CURSOR_CONNECT_2));
        imageRegistry.put(Icons.CURSOR_DISCONNECT, getImageDescriptor(Icons.CURSOR_DISCONNECT));
        imageRegistry.put(Icons.CURSOR_ZOOM, getImageDescriptor(Icons.CURSOR_ZOOM));
        imageRegistry.put(Icons.DISCONNECT, getImageDescriptor(Icons.DISCONNECT));
		imageRegistry.put(Icons.DOWN, getImageDescriptor(Icons.DOWN));
		imageRegistry.put(Icons.FEATURE, getImageDescriptor(Icons.FEATURE));
		imageRegistry.put(Icons.PROVIDED_INTERFACE, getImageDescriptor(Icons.PROVIDED_INTERFACE));
		imageRegistry.put(Icons.REQUIRED_INTERFACE, getImageDescriptor(Icons.REQUIRED_INTERFACE));
		imageRegistry.put(Icons.REQUIRED_INTERFACE_UML, getImageDescriptor(Icons.REQUIRED_INTERFACE_UML));
		imageRegistry.put(Icons.PROVIDED_INTERFACE_UML, getImageDescriptor(Icons.PROVIDED_INTERFACE_UML));
		imageRegistry.put(Icons.LOGO, getImageDescriptor(Icons.LOGO));
		imageRegistry.put(Icons.NEW, getImageDescriptor(Icons.NEW));
		imageRegistry.put(Icons.NEW_ATTRIBUTE, getImageDescriptor(Icons.NEW_ATTRIBUTE));
		imageRegistry.put(Icons.NEW_COMPONENT, getImageDescriptor(Icons.NEW_COMPONENT));
		imageRegistry.put(Icons.NEW_FEATURE, getImageDescriptor(Icons.NEW_FEATURE));
		imageRegistry.put(Icons.NEW_INTERFACE, getImageDescriptor(Icons.NEW_INTERFACE));
		imageRegistry.put(Icons.PART, getImageDescriptor(Icons.PART));
		imageRegistry.put(Icons.SELECT, getImageDescriptor(Icons.SELECT));
		imageRegistry.put(Icons.STATUS_INCOMPLETE, getImageDescriptor(Icons.STATUS_INCOMPLETE));
		imageRegistry.put(Icons.STATUS_COMPLETE, getImageDescriptor(Icons.STATUS_COMPLETE));
		imageRegistry.put(Icons.TRASH_CAN, getImageDescriptor(Icons.TRASH_CAN));
		imageRegistry.put(Icons.ZOOM, getImageDescriptor(Icons.ZOOM));
		imageRegistry.put(Icons.ZOOM_OUT, getImageDescriptor(Icons.ZOOM_OUT));
	}
	
	/**
	 * Returns an image descriptor for the image file at the given
	 * plug-in relative path.
	 *
	 * @param name the file name (inside 'pics'-folder)
	 * @return the image descriptor
	 */
	public ImageDescriptor getImageDescriptor(String name) {
		String picPath = "pics/";
		
		/* get bundle with the specified id */
		Bundle bundle = Platform.getBundle(PLUGIN_ID);
		
		if( bundle == null ) {
			throw new RuntimeException("Could not resolve plugin: " + PLUGIN_ID);
		}
		
		/* resolve Bundle.getEntry to local URL */
		URL pluginURL = null;
		try {
			pluginURL = Platform.resolve(bundle.getEntry("/"));
		} catch (IOException e) {
			throw new RuntimeException("Could not get installation directory of the plugin: " + PLUGIN_ID);
		}
		
		URL url;
		try {
			url = new URL(pluginURL, picPath + name);
			return ImageDescriptor.createFromURL(url);
		} catch (MalformedURLException e) {
			// should not happen
			return ImageDescriptor.getMissingImageDescriptor();
		}
		
	}
	
		// TODO: t�t� ei v�ltt�m�tt� tarvitse
	   public String getPluginDir() {
	        return eclipsePluginPath;
	    }
		
		/**
		 * Returns the absolute plug-in path.
		 * @param pluginId
		 * @return Absolute plug-in path
		 */
		public String getPluginDir(String pluginId) {
			/* get bundle with the specified id */
			Bundle bundle = Platform.getBundle(pluginId);
			if( bundle == null ) {
				throw new RuntimeException("Could not resolve plugin: " + pluginId);
			}
			
			/* resolve Bundle.getEntry to local URL */
			URL pluginURL = null;
			try {
				pluginURL = Platform.resolve(bundle.getEntry("/"));
			} catch (IOException e) {
				throw new RuntimeException("Could not get installation directory of the plugin: " + pluginId);
			}
			String pluginInstallDir = pluginURL.getPath().trim();
			if( pluginInstallDir.length() == 0 )
				throw new RuntimeException("Could not get installation directory of the plugin: " + pluginId);
			
			/* since path returned by URL.getPath starts with a forward slash, that
			 * is not suitable to run commandlines on Windows-OS, but for Unix-based
			 * OSes it is needed. So strip one character for windows. There seems
			 * to be no other clean way of doing this. */
			if( Platform.getOS().compareTo(Platform.OS_WIN32) == 0 )
				pluginInstallDir = pluginInstallDir.substring(1);
			
			return pluginInstallDir;
		}
		
		public boolean isUMLPreferenceSet() {
			return this.store.getBoolean(PreferenceConstants.P_UML_INTERFACES);
		}
		
	    public boolean showCompletenessPopups() {
	        return this.store.getBoolean(PreferenceConstants.P_COMPLETENESS_POPUPS);
	    }
	    
		public String getCommunicationMethod() {
			IPreferenceStore store = getPreferenceStore();
			return store.getString(PreferenceConstants.P_COMMUNICATIONMETHOD);
		}
	    
	    public String getServerLocation() {
	        IPreferenceStore store = getPreferenceStore();
	        String address = store.getString(PreferenceConstants.P_SERVERADDRESS);
	        String port = store.getString(PreferenceConstants.P_SERVERPORT);
	        return "//" + address + ":" + port;
	    }
		
		public String getSaveLocation() {
			IPreferenceStore store = getPreferenceStore();
			return store.getString(PreferenceConstants.P_SAVEDIRECTORY);
		}

		public PluginHandler getHandler() {
			return handler;
		}

		public String getPluginID() {
			return PLUGIN_ID;
		}


		

}
