/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ConfigurationClient.java
 *
 * Created on 23. helmikuuta 2004, 11:37
 */
package kumbang.configurator.client.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;


import kumbang.configurator.client.core.exception.ClientException;
import kumbang.configurator.client.core.exception.OpenException;

import kumbang.configurator.server.core.ServerProperties; 
// TODO this import breaks layering rules (references only to the layer below)

import kumbang.configurator.server.services.ConfigurationServer;
import kumbang.configurator.server.services.ConfigurationServices;
import kumbang.configurator.server.services.RemoteSession;
import kumbang.configurator.server.services.ServerException;

import kumbang.core.configuration.core.ConfigurationException;
import kumbang.core.configuration.task.ClientTask;
import kumbang.core.configuration.task.ConfigurationResponse;

import kumbang.core.model.core.Model;
import kumbang.core.model.exception.ModelException;

/**
 * 
 * @author vmyllarn
 */
public class ConfigurationClient {

	protected ConfigurationHandler configHandler;

	protected TaskHandler taskHandler;

	protected ExportHandler exportHandler;

	protected SaveHandler saveHandler;

//   was private
	protected ConfigurationServices server;

    // was private
	protected RemoteSession session;

	//private boolean isAutoApply = true;

	public static final int METHOD_RMI = 0;
	public static final int METHOD_LOCAL = 0;
	public static final String RMI_LOCATION = "/KumbangServer";

	/** Creates a new instance of ConfigurationClient */
	public ConfigurationClient(String method) throws RemoteException {
		this(method, null);
	}

	/**
	 * Creates a new instance of Configuration Client
	 * 
	 * @param method communication method
	 * @param location server location (ip:port)
	 * @throws RemoteException
	 */
	public ConfigurationClient(String method, String location) throws RemoteException {
		initServerConnection(method, location);
		KumbangConfigurator.logConfig("Starting configuration client");
		initialiseHandlers();
	}
	
	/**
	 * This method must be separate from constructor and
	 * protected in order to allow custom handling for subclasses.
	 */
	protected void initialiseHandlers() {
		this.configHandler = new ConfigurationHandler();
		this.taskHandler = new TaskHandler();
		this.exportHandler = new ExportHandler();
		this.saveHandler = new SaveHandler(this.configHandler);
	}
	
	// ------------------------------------------------------------------
	// ------ PUBLIC METHODS --------------------------------------------
	// ------------------------------------------------------------------

	public void reset() {
		configHandler.reset();
		saveHandler.reset();
		exportHandler.reset();
	}

	public ConfigurationHandler getConfigurationHandler() {
		return configHandler;
	}

	public TaskHandler getTaskHandler() {
		return taskHandler;
	}

	public SaveHandler getSaveHandler() {
		return saveHandler;
	}

	public ExportHandler getExportHandler() {
		return exportHandler;
	}

	public String[] getModelNames() {
		try {
			return server.getModelNames();
		} catch (RemoteException re) {
			return new String[0];
		}
	}

	// ------------------------------------------------------------------
	// ------ BASIC OPERATIONS ------------------------------------------
	// ------ import, open, close, exit etc. ----------------------------

	public void clearAllModels() throws ClientException, ModelException {
		try {
			server.clearAllModels();
		} catch (RemoteException e) {
			throw new ClientException("Remote failure in accessing " + "configuration server.\n"
					+ e);
		}
	}

	public void removeModel(String modelName) throws ClientException, ModelException {
		try {
			server.removeModel(modelName);
		} catch (RemoteException e) {
			throw new ClientException("Remote failure in accessing " + "configuration server.\n"
					+ e);
		}

	}

	public int numberOfSessions() throws ClientException {
		try {
			return server.numberOfSessions();
		} catch (RemoteException e) {
			throw new ClientException("Remote failure in accessing " + "configuration server.\n"
					+ e);
		}
	}

	public void importModel(File file, String name) throws ClientException, ModelException {
		KumbangConfigurator.logConfig("Importing configuration model " + "from file " + file);
		try {
			String modelString = readToString(file);
			server.importModel(modelString, name);
		} catch (RemoteException re) {
			throw new ClientException("Remote failure in accessing " + "configuration server.\n"
					+ re);
		} catch (IOException io) {
			throw new ClientException("Cannot read model file.\n" + io);
		}
	}

	public void openModel(String modelName, String configurationName) throws ClientException,
			ModelException {
		KumbangConfigurator.logConfig("Opening configuration " + configurationName + " from model "
				+ modelName);
		try {
			closeConfiguration();
			Model resp = server.openModel(modelName, session);
			reset();
			configHandler.initConfiguration(modelName, resp, configurationName);
			computeConfigurationState(null);
		} catch (RemoteException re) {
			throw new ClientException("Cannot access configuration server.\n" + re);
		} catch (ServerException se) {
			throw new ClientException(se.getMessage());
		} catch (ConfigurationException ce) {
			throw new ClientException("Cannot initialize configuration based " + "on this model.");
		}
	}

	public void closeConfiguration() {
		KumbangConfigurator.logConfig("Closing configuration.");
		try {
			server.closeConfiguration(session);
			configHandler.reset();
		} catch (RemoteException re) {
			KumbangConfigurator.logConfig("Remote failure when closing " + "configuration: " + re);
		} catch (ServerException se) {
			KumbangConfigurator.logConfig("Server failure when closing " + "configuration: "
					+ se.getMessage());
		}
	}

	public void openConfiguration(File openFile) throws OpenException, ClientException {
		KumbangConfigurator.logConfig("Opening configuration from file " + openFile);
		try {
			reset();
			saveHandler.openConfiguration(openFile);
			server.openModel(configHandler.getModelName(), session);

			// add all instances and send configuration request
			// addAllInstancesToRequest();
			computeConfigurationState(null);
		} catch (RemoteException re) {
			throw new ClientException("Cannot access configuration server.\n" + re);
		} catch (ServerException se) {
			throw new ClientException(se.getMessage());
		} catch (ModelException me) {
			throw new ClientException("Model was not valid.\n" + me);
		} catch (ConfigurationException ce) {
			throw new ClientException("Cannot initialize configuration based on this model.");
		}
	}

	public void findCompleteConfiguration() throws ClientException {
		KumbangConfigurator.logConfig("Searching for a complete and consistent configuration.");
		try {
			findConfiguration();
		} catch (ConfigurationException ce) {
			throw new ClientException(ce.getMessage());
		}
	}

	public void exit() {
		closeConnection();
	}

	public void closeConnection() {
		try {
			KumbangConfigurator.logConfig("Exiting configuration client");
			this.server.exitClient(session);
		} catch (RemoteException re) {
			KumbangConfigurator.logConfig("Remote failure when exiting: " + re);
		}
	}

	// ------------------------------------------------------------------
	// ------ OPERATIONS THAT MODIFY CONFIGURATION WITH TASKS------------
	// ------------------------------------------------------------------

	public ClientTask applyTask(ClientTask task) throws ClientException, ConfigurationException {
		if (task.isMarkedError())
			throw new ConfigurationException(task.getErrorMsg());

		configHandler.applyTask(task);
		
		ConfigurationResponse response;
		try {
			response = computeConfigurationState(task);
		} catch (ClientException cle) {
			KumbangConfigurator.logConfig("Exception when handling task: " + cle.getMessage());
			task.markError(cle.getMessage());
			throw cle;
		} catch (ConfigurationException ce) {
			KumbangConfigurator.logConfig("Exception when handling task: " + ce.getMessage());
			task.markError(ce.getMessage());
			throw ce;
		}

		ClientTask whole = new ClientTask(false);
		whole.addSubTask(task);
		whole.addSubTask(response.getTask());
		
		return whole;
	}

	// ------------------------------------------------------------------
	// ------ PRIVATE METHODS -------------------------------------------
	// ------------------------------------------------------------------


	/**
	 * Initializes the server connection
	 * 
	 * @param method communication method
	 * @param location server location
	 * @throws RemoteException
	 */
	private void initServerConnection(String method, String location) throws RemoteException {
		method = method.toLowerCase();

		if (method.equals(ClientProperties.COMMUNICATION_METHOD_RMI)) {
            String name;
            if (location == null) {
                name = "//" + ServerProperties.DEFAULT_HOST + ":" + ServerProperties.DEFAULT_PORT + ConfigurationClient.RMI_LOCATION;
            } else {
                name = location + ConfigurationClient.RMI_LOCATION;
            }
			try {
				// Comment next one out, if security manager seems to be
				// too restrictive.

				// XXX This creates problems in Eclipse environment, if uncommented
				if (System.getSecurityManager() == null) {
					// System.setSecurityManager(new RMISecurityManager());
				}
				this.server = (ConfigurationServices) Naming.lookup(name);
				this.session = this.server.enterClient();
			} catch (NotBoundException nbe) {
				throw new RemoteException("Server is not bound at " + name + "\n" + nbe);
			} catch (MalformedURLException mue) {
				throw new RemoteException("Cannot reach URL " + name + "\n" + mue);
			}
		} else if (method.equals(ClientProperties.COMMUNICATION_METHOD_LOCAL)) {

			try {
				KumbangConfigurator.initLogger("fine");
			} catch (IOException e) {
				e.printStackTrace();
			}

			// Set the local server's properties
			ServerProperties.getInstance().setRootDir(ClientProperties.getInstance().getRootDir());



			this.server = new ConfigurationServer(KumbangConfigurator.logger);
			this.session = this.server.enterClient();
		} else { // communication method is neither 'rmi' or 'local'
			throw new IllegalArgumentException("Illegal communication method");
		}
	}


	// was private
	protected ConfigurationResponse computeConfigurationState(ClientTask task) throws ClientException,
			ConfigurationException {
		ConfigurationResponse response;
		try {
			response = server.computeConfigurationState(configHandler
					.getConfiguration(), session);
			handleResponse(response, task);
		} catch (ServerException se) {
			throw new ClientException("Server failure:\n" + se.getMessage());
		} catch (RemoteException re) {
			throw new ClientException("Cannot access configuration server.\n" + re);
		}
		
		return response;
	}

	private ConfigurationResponse findConfiguration() throws ClientException, ConfigurationException {
		ConfigurationResponse response;
		try {
			response = server.findConfiguration(configHandler.getConfiguration(), session);
			handleResponse(response, null);
		} catch (ServerException se) {
			throw new ClientException("Server failure:\n" + se.getMessage());
		} catch (RemoteException re) {
			throw new ClientException("Cannot access configuration server.\n" + re);
		}
		
		return response;
	}
	
	protected void handleResponse(ConfigurationResponse response, ClientTask task)
			throws ConfigurationException {
		if (!response.isConsistent()) {
			configHandler.revertTask(task);
			throw new ConfigurationException(response.getMessage());
		}

		//response.getTask().convertReferences(configHandler);
		response.getTask().convertReferences(configHandler.getConfiguration());
		
		configHandler.applyTask(response.getTask());
		configHandler.setComplete(response.isComplete());
	}

	private String readToString(File file) throws IOException {
		StringBuffer str = new StringBuffer();
		String line;
		BufferedReader buf = new BufferedReader(new FileReader(file));
		while ((line = buf.readLine()) != null) {
			str.append(line + "\n");
		}
		return str.toString();
	}
}
