/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ConfigurationHandler.java
 *
 * Created on 3. maaliskuuta 2004, 14:57
 */

package kumbang.configurator.client.core;

import java.io.Reader;
import java.io.StringReader;
import java.util.Iterator;

import kumbang.core.configuration.core.Configuration;
import kumbang.core.configuration.instance.AttributeInstance;
import kumbang.core.configuration.instance.AttributedInstance;
import kumbang.core.configuration.instance.BindingInstance;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.ComposableInstance;
import kumbang.core.configuration.instance.FeatureInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.PartInstance;
import kumbang.core.configuration.reference.AttributeReference;
import kumbang.core.configuration.reference.BindingReference;
import kumbang.core.configuration.reference.ComposableReference;
import kumbang.core.configuration.reference.InterfaceReference;
import kumbang.core.configuration.reference.Reference;
import kumbang.core.configuration.reference.ReferenceItem;
import kumbang.core.configuration.task.AddBindingTask;
import kumbang.core.configuration.task.AddComposableTask;
import kumbang.core.configuration.task.AddInstanceTask;
import kumbang.core.configuration.task.AddInterfaceTask;
import kumbang.core.configuration.task.ClientTask;
import kumbang.core.configuration.task.EditAttributesTask;
import kumbang.core.configuration.task.EditInterfacesTask;
import kumbang.core.configuration.task.RemoveInstanceTask;
import kumbang.core.configuration.task.SetAttributeTask;
import kumbang.core.language.parser.KumbangParser;
import kumbang.core.language.parser.ParseException;
import kumbang.core.language.syntaxtree.SyntaxConfigurationReference;
import kumbang.core.language.visitor.ConfigurationReferenceVisitor;

import kumbang.core.model.core.Model;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.PartDefinition;

/**
 * What does this class do? It seems to be a kind of wrapper for a Configuration object. 
 * In that case, it would make more sense to create a new ConfigurationHandler for each
 * new configuration. -toasikai
 * Not necessarily, since this object is passed to elements in the GUI once. If this
 * object was created from scratch every time when there is a new configuration, these
 * GUI elements would need to be aware of configuration initialisation and reinitialisation.
 * - vmyllarn
 * 
 * @author vmyllarn
 */
public class ConfigurationHandler {
	public static final int STATE_RESET = 0;
	public static final int STATE_INCOMPLETE = 1;
	public static final int STATE_COMPLETE = 2;

	private Configuration configuration;

	private int state;

	private boolean modified;

	private String configurationName;
	private String modelName;
	

	/**
	 * Creates a new instance of ConfigurationHandler.
	 */
	public ConfigurationHandler() {
		this.state = STATE_RESET;
		this.modified = false;
	}
	
	// -------------GENERAL PUBLIC METHODS---------------------------

	/**
	 * Returns the root of the component configuration.
	 * 
	 * @return ComponentInstance root
	 */
	public ComponentInstance getComponentRoot() {
		if (configuration == null)
			return null;
		
		return configuration.getComponentRoot();
	}

	/**
	 * Returns the root of the feature configuration.
	 * 
	 * @return FeatureInstance root
	 */
	public FeatureInstance getFeatureRoot() {
		if (configuration == null)
			return null;
		
		return configuration.getFeatureRoot();
	}

	public Configuration getConfiguration() {
		return this.configuration;
	}

	/**
	 * Initialises a new configuration based on the given model and configuration name.
	 * 
	 * @param modelName name of the model
	 * @param total total configuration that is used for initialising the configuration
	 * @param configurationName name of the configuration
	 */
	public void initConfiguration(String modelName, Model model, String configurationName) {
		this.modelName = modelName;
		
		configuration = new Configuration(model);

		this.state = STATE_INCOMPLETE;
		this.modified = true;
		this.configurationName = configurationName;
	}

	/**
	 * Initialises an existing configuration to the handler. Also saves configuration name and model
	 * name.
	 * 
	 * @param modelName name of the model
	 * @param configurationName name of the configuration
	 * @param configuration existing configuration
	 */
	public void initConfiguration(String modelName, String configurationName,
			Configuration configuration) {
		this.configuration = configuration;
		this.state = STATE_INCOMPLETE;
		this.modified = true;
		this.configurationName = configurationName;
        this.modelName = modelName;
	}

	/**
	 * Resets the current configuration. Clears configuration and total configuration.
	 * 
	 * TODO Wouldn't it make more sense to scrap the entire object and create a new one? -toasikai
	 */
	public void reset() {
		configuration = null;
		this.state = STATE_RESET;
		this.modified = false;
	}

	/**
	 * Sets the modified flag. The flag is used to check whether there are changes when the current
	 * configuration is about to get closed (open, close, etc.).
	 *
	 */
	private void setModified() {
		this.modified = true;
	}

	public boolean isModified() {
		return this.modified && this.state != STATE_RESET;
	}

	public int getState() {
		return this.state;
	}

	public void setComplete(boolean complete) {
		if (complete) {
			this.state = STATE_COMPLETE;
		} else {
			this.state = STATE_INCOMPLETE;
		}
	}

	public String getModelName() {
		return this.modelName;
	}

	public String getConfigurationName() {
		return this.configurationName;
	}

	// -------------METHODS FOR HANDLING TASKS -------------------------

	public void applyTask(ClientTask task) {
		//System.out.println(getClass().getName() + ": Apply task " + task);
		if (task instanceof AddComposableTask) {
			AddComposableTask act = (AddComposableTask) task;
			applyTask(act);
		} else if (task instanceof AddInterfaceTask) {
			applyTask((AddInterfaceTask) task);
		} else if (task instanceof AddBindingTask) {
			AddBindingTask abt = (AddBindingTask) task;
			applyTask(abt);
		} else if (task instanceof SetAttributeTask) {
			applyTask((SetAttributeTask) task);
		} else if (task instanceof EditInterfacesTask) {
			EditInterfacesTask ect = (EditInterfacesTask) task;
			applyTask(ect);
		} else if (task instanceof EditAttributesTask) {
			EditAttributesTask ect = (EditAttributesTask) task;
			applyTask(ect);
		} else if (task instanceof RemoveInstanceTask) {
			RemoveInstanceTask rit = (RemoveInstanceTask) task;
			applyTask(rit);
		} else {
			Iterator subTasks = task.getSubTasks();
			while (subTasks.hasNext()) {
				applyTask((ClientTask) subTasks.next());
			}
		}
		
		/*
		 * This will be called anyway in one of the methods called in this method, one can as
		 * well (or better) call it here. -toasikai 28.7.2006
		 */
		setModified();
	}

	public void applyTask(AddComposableTask task) {
		ComposableInstance instance = task.getParent().addInstance((ComposableType) task.getType(),
				task.isUserTask());
		
		task.setAddedInstance(instance);

		Iterator it = task.getSubTasks();
		while (it.hasNext()) {
			applyTask((ClientTask) it.next());
		}
	}

	public void applyTask(AddInterfaceTask task) {
		InterfaceInstance instance = null;
		ComponentInstance parent = null;
		ClientTask parentTask = task.getParentTask();
		
		if (parentTask instanceof AddComposableTask) {
			AddComposableTask act = (AddComposableTask) parentTask;
			parent = (ComponentInstance) act.getAddedInstance();
		} else if (parentTask instanceof EditInterfacesTask) {
			EditInterfacesTask ect = (EditInterfacesTask) parentTask;
			parent = (ComponentInstance) ect.getInstance();
		}
		
		instance = parent.addInterface(task.getDefininition(), task.getType(), task.isUserTask());
			
		task.setAddedInstance(instance);

		Iterator it = task.getSubTasks();
		while (it.hasNext()) {
			applyTask((ClientTask) it.next());
		}
	}

	public void applyTask(AddBindingTask task) {
		InterfaceInstance source = (InterfaceInstance) task.getSource();
		InterfaceInstance target = (InterfaceInstance) task.getTarget();
		
		BindingInstance binding = new BindingInstance(source, target, task.isUserTask());   
		
		task.setAddedBinding(binding);
	}

	public void applyTask(EditInterfacesTask task) {
		Iterator it = task.getSubTasks();
		while (it.hasNext()) {
			applyTask((ClientTask) it.next());
		}
	}
	
	public void applyTask(EditAttributesTask task) {
		Iterator it = task.getSubTasks();
		while (it.hasNext()) {
			applyTask((ClientTask) it.next());
		}
	}

	public void applyTask(SetAttributeTask task) {
		AttributedInstance parent = task.getParentInstance();
		
		AttributeInstance ai = parent.getAttribute(task.getName());
		task.setAttribute(ai);
		ai.setValue(task.getValue(), task.isUserTask());
	}

	public void applyTask(RemoveInstanceTask task) {
		Iterator it = task.getSubTasks();
		while (it.hasNext()) {
			applyTask((ClientTask) it.next());
		}
		
		Instance instance = task.getInstance();
		instance.remove();
	}
	
	/**
	 * This method can be used to revert or undo a task that has previously been applied to the
	 * configuration. 
	 * 
	 * @param task
	 */
	public void revertTask(ClientTask task) {
		if (task == null)
			return;
		
		/*
		 * TODO Must somehow check that the task has indeed been applied before
		 * one starts to revert it. -toasikai 2.8.2006
		 */
		if (task instanceof AddInstanceTask) {
			((AddInstanceTask) task).getAddedInstance().remove();
		} else if (task instanceof SetAttributeTask) {
			SetAttributeTask sat = (SetAttributeTask) task;
			sat.getAttribute().setValue(sat.getOldValue(), sat.wasUserSet());
		} else if (task instanceof RemoveInstanceTask) {
			((RemoveInstanceTask) task).getInstance().restore();
		} else {
			Iterator subTasks = task.getSubTasks();
			while (subTasks.hasNext()) {
				revertTask((ClientTask) subTasks.next());
			}
		}
	}
	
	public PartInstance getPart(String refString) {
		System.out.print("ConfigurationHandler - configuration:" +configuration.toString());
		return configuration.getPart(refString);
	}
	
	public Instance getInstance(String refString) {
		return configuration.getInstance(refString);
	}	
	
}
