/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ComponentDiagramPanel.java
 *
 * Created on 25. helmikuuta 2004, 10:05
 */

package kumbang.configurator.client.gui.core;

import kumbang.configurator.client.gui.diagram.*;

import kumbang.util.gui.Icons;
import kumbang.util.gui.Margins;

import kumbang.configurator.client.gui.event.ConfigurationModificationListener;
import kumbang.configurator.client.gui.event.ConfigurationModifiedEvent;
import kumbang.configurator.client.gui.event.ComponentSelectionListener;
import kumbang.configurator.client.gui.event.ComponentSelectedEvent;

import kumbang.core.configuration.instance.Instance;

import java.awt.Component;
import java.awt.Cursor;
import java.awt.Insets;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;

/**
 *
 * @author  vmyllarn
 */
public class ComponentDiagramPanel extends JPanel 
        implements ActionListener, MouseListener, 
        ConfigurationModificationListener, ComponentSelectionListener {
    
    /** state of this diagram */
    private int state;
    
    /** actual diagram area */
    private ComponentDiagram diagram;
    
    private JScrollPane diagramScroll;
   
    /** buttons to do the dirty work */
    private JButton up;
    private JButton down;
    private JButton select;
    private JButton move;
    private JButton connect;
    private JButton disconnect;
    
    private MainGUI mainGUI;
    
    private DiagramHandler diagramHandler;
    
    private ComponentElement rootElement;
    private DiagramElement selected;
    private BoxedElement inAction;
    private ComponentElement popupElement;
    
    private JPopupMenu popup;
    private JMenuItem navigateItem;
    private JMenuItem editAItem;
    private JMenuItem editIItem;
    private JMenuItem removeItem;
    
    //public static final Dimension DIAGRAM_SIZE = new Dimension(600,400);
    
    public static final int STATE_SELECTING     = 0;
    public static final int STATE_MOVING        = 1;
    public static final int STATE_GOING_DOWN    = 2;
    public static final int STATE_CONNECTING    = 3;
    public static final int STATE_DISCONNECTING = 4;
    
    
    public ComponentDiagramPanel(MainGUI mainGUI) {
        this.mainGUI = mainGUI;
        this.diagramHandler = mainGUI.getDiagramHandler();
        this.state = STATE_SELECTING;
        initComponents();
        mainGUI.addConfigurationModificationListener(this);
        mainGUI.addComponentSelectionListener(this);
    }
    
    
    public void init() {
        //load configuration level from diagram handler
        diagram.setDiagramLevel(diagramHandler.getCurrentComponentLevel());
        if (diagramHandler.getCurrentComponentLevel() != null) {
            rootElement = diagramHandler.getCurrentComponentLevel().getRootComponentElement();
        }
        updateView();
    }
    
    public void reset() {
        diagram.setDiagramLevel(null);
        selected = null;
        inAction = null;
        rootElement = null;
        popupElement = null;
        state = STATE_SELECTING;
        updateView();
    }
    
    public void configurationModified(ConfigurationModifiedEvent e) {
        updateView();
    }    

    public void componentSelected(ComponentSelectedEvent e) {
        selectInstance(e.getSelectedInstance());
    }
    
    public void showInstance(Instance instance) {
        ComponentDiagramLevel level 
            = diagramHandler.navigateToComponent(instance);
        if (level != null) {
            diagram.setDiagramLevel(level);
            updateView();
        }
    }
    
    public void showNames(boolean show) {
        diagram.showNames(show);
        diagram.repaint();
    }
    
    
    public void showTypes(boolean show) {
        diagram.showTypes(show);
        diagram.repaint();
    }
    
    
    public void showConnections(boolean show) {
        diagram.showConnections(show);
        diagram.repaint();
    }
    
    
    public void showPotential(boolean show) {
        diagram.showPotential(show);
        diagram.repaint();
    }
    
    
    private void initComponents() {
        setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
        add(Margins.createVerticalMarginXS());
        initButtons();
        add(Margins.createVerticalMarginXS());
        initDiagram();
        
        popup = new JPopupMenu("Component");
        editIItem = new JMenuItem("Edit interfaces of this component", Icons.INTERFACE);
        editAItem = new JMenuItem("Edit attributes of this component", Icons.ATTRIBUTE);
        navigateItem = new JMenuItem("Navigate levels down", Icons.DOWN);
        removeItem = new JMenuItem("Remove this component", Icons.DELETE);
        editIItem.addActionListener(this);
        editAItem.addActionListener(this);
        navigateItem.addActionListener(this);
        removeItem.addActionListener(this);
        popup.add(editIItem);
        popup.add(editAItem);
        popup.add(navigateItem);
        popup.add(removeItem);
    }
    
    private void initDiagram() {
        diagram = new ComponentDiagram();
        diagram.addMouseListener(this);
        
        diagramScroll = new JScrollPane(diagram, 
            JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, 
            JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        diagramScroll.getViewport().setBackground(diagram.getBackground());
        diagramScroll.setAlignmentX(JComponent.LEFT_ALIGNMENT);
        this.add(diagramScroll);
    }
    
    private void initButtons() {
        JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.LINE_AXIS));
        
        up = new JButton(Icons.UP);
        down = new JButton(Icons.DOWN);
        select = new JButton(Icons.SELECT);
        move = new JButton(Icons.MOVE);
        connect = new JButton(Icons.CONNECT);
        disconnect = new JButton(Icons.DISCONNECT);
        
        up.setMargin(new Insets(0,0,0,0));
        down.setMargin(new Insets(0,0,0,0));
        select.setMargin(new Insets(0,0,0,0));
        move.setMargin(new Insets(0,0,0,0));
        connect.setMargin(new Insets(0,0,0,0));
        disconnect.setMargin(new Insets(0,0,0,0));

        up.setToolTipText("Navigate component levels up");
        down.setToolTipText("Navigate component levels down");
        select.setToolTipText("Select elements in diagram");
        move.setToolTipText("Move components");
        connect.setToolTipText("Connect interfaces");
        disconnect.setToolTipText("Disconnect interfaces");
                
        up.addActionListener(this);
        down.addActionListener(this);
        select.addActionListener(this);
        move.addActionListener(this);
        connect.addActionListener(this);
        disconnect.addActionListener(this);
        
        buttonPanel.add(Margins.createHorizontalMarginXS());
        buttonPanel.add(up);
        buttonPanel.add(Margins.createHorizontalMarginXS());
        buttonPanel.add(down);
        buttonPanel.add(Margins.createHorizontalMarginM());
        buttonPanel.add(select);
        buttonPanel.add(Margins.createHorizontalMarginXS());
        buttonPanel.add(move);
        buttonPanel.add(Margins.createHorizontalMarginM());
        buttonPanel.add(connect);
        buttonPanel.add(Margins.createHorizontalMarginXS());
        buttonPanel.add(disconnect);
        buttonPanel.add(Box.createHorizontalGlue());
        
        buttonPanel.setAlignmentX(JComponent.LEFT_ALIGNMENT);
        add(buttonPanel);
    }
    
    
    public void actionPerformed(ActionEvent e) {
        Object o = e.getSource();
        if (o == up) {
            navigateLevelUp();
            state = STATE_SELECTING;
            diagram.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        }
        else if (o == down) {
            //go down!
            state = STATE_GOING_DOWN;
            diagram.setCursor(new Cursor(Cursor.HAND_CURSOR));
        }
        else if (o == select) {
            state = STATE_SELECTING;
            diagram.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        }
        else if (o == move) {
            state = STATE_MOVING;
            diagram.setCursor(new Cursor(Cursor.MOVE_CURSOR));
        }
        else if (o == connect) {
            state = STATE_CONNECTING;
            diagram.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
        }
        else if (o == disconnect) {
            state = STATE_DISCONNECTING;
            diagram.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
        }
        else if (o == editAItem) {
            mainGUI.editAttributes(popupElement.getComponentInstance());
        }
        else if (o == editIItem) {
            mainGUI.editInterfaces(popupElement.getComponentInstance());
        }
        else if (o == navigateItem) {
            navigateLevelDown(popupElement);
        }
        else if (o == removeItem) {
            mainGUI.removeComposable(popupElement.getComponentInstance());
        }
    }
   
    
    public void mouseClicked(MouseEvent e) {
        if (e.getSource() != diagram) {
            return;
        }
        int button = e.getButton();
        if (button != MouseEvent.BUTTON1 && state == STATE_SELECTING) {
            DiagramElement elem = diagram.getElementAt(e.getX(), e.getY());
            if (elem instanceof ComponentElement) {
                selectElement(elem, true);
                popupElement = (ComponentElement)elem;
                navigateItem.setEnabled(popupElement.isNavigable());
                editIItem.setEnabled(popupElement.areInterfacesEditable());
                editAItem.setEnabled(popupElement.areAttributesEditable());
                removeItem.setEnabled(!popupElement.isRootElement());
                popup.show(e.getComponent(), e.getX(), e.getY());
                diagram.repaint();
            }
        }
        else if (button == MouseEvent.BUTTON1 && state == STATE_SELECTING) {
            selectElementAt(e.getX(), e.getY());
            diagram.repaint();
            if (selected instanceof PartElement) {
                PartElement part = (PartElement) selected;
                mainGUI.addComposable(part.getPartInstance());
            }
            else if (selected instanceof ComponentElement && e.getClickCount() > 1) {
                ComponentElement c = (ComponentElement) selected;
                if (c.isNavigable()) {
                    navigateLevelDown(c);
                }
                else if (c.areAttributesEditable()) {
                    mainGUI.editAttributes(c.getComponentInstance());
                }
                else if (c.areInterfacesEditable()) {
                    mainGUI.editInterfaces(c.getComponentInstance());
                }
            } else if (selected instanceof InstantiatedElement) {
                mainGUI.selectInstance(
                        ((InstantiatedElement) selected).getInstance());
            }
        }
        else if (button == MouseEvent.BUTTON1 && state == STATE_DISCONNECTING) {
            DiagramElement elem = diagram.getElementAt(e.getX(), e.getY());
            if (elem instanceof BindingElement) {
                mainGUI.removeBinding(((BindingElement)elem).getBindingInstance());

            }
            diagram.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            state = STATE_SELECTING;
        }
        else if (button == MouseEvent.BUTTON1 && state == STATE_GOING_DOWN) {
            DiagramElement elem = diagram.getElementAt(e.getX(), e.getY());
            if (elem instanceof ComponentElement) {
                ComponentElement c = (ComponentElement)elem;
                if (c.isNavigable()) {
                    navigateLevelDown(c);
                }
            }
            diagram.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            state = STATE_SELECTING;
        }
    }
    
    public void mouseEntered(MouseEvent e) {
    }
    
    public void mouseExited(MouseEvent e) {
    }
    
    public void mousePressed(MouseEvent e) {
        if (e.getSource() != diagram) {
            return;
        }
        int button = e.getButton();
        if (button == MouseEvent.BUTTON1 && state == STATE_MOVING) {
            DiagramElement elem = diagram.getElementAt(e.getX(), e.getY());
            if (elem != rootElement && (elem instanceof ComponentElement || 
                                        elem instanceof PartElement)) {
                elem.setInAction(true);
                inAction = (BoxedElement) elem;
                diagram.repaint();
            }
        }
        else if (button == MouseEvent.BUTTON1 && state == STATE_CONNECTING) {
            DiagramElement elem = diagram.getElementAt(e.getX(), e.getY());
            if (selected != null) {
                selected.setSelected(false);
                selected = null;
            }
            if (elem instanceof InterfaceElement) {
                elem.setInAction(true);
                inAction = (BoxedElement)elem;
                diagram.repaint();
            }
            else {
                state = STATE_SELECTING;
                diagram.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));                
            }            
        }
         
    }
    
    public void mouseReleased(MouseEvent e) {
        if (e.getSource() != diagram) {
            return;
        }
        int button = e.getButton();
        if (button == MouseEvent.BUTTON1 && state == STATE_MOVING) {
            if (inAction != null) {
                DiagramLevel level = inAction.getLevel();
                inAction.setInAction(false);
                if (inAction instanceof BoxedElement) {
                    level.moveChildElement((BoxedElement)inAction, 
                                            e.getX(), e.getY());
                    updateView();
                }
                inAction = null;
                diagram.repaint();
            }
            state = STATE_SELECTING;
            diagram.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        }
        else if (button == MouseEvent.BUTTON1 && state == STATE_CONNECTING) {
            DiagramElement elem = diagram.getElementAt(e.getX(), e.getY());
            if (inAction != null && inAction instanceof InterfaceElement &&
                elem != null && elem instanceof InterfaceElement) {
                InterfaceElement intf1 = (InterfaceElement)inAction;
                InterfaceElement intf2 = (InterfaceElement)elem;
                mainGUI.addBinding(intf1.getInterfaceInstance(), 
                                        intf2.getInterfaceInstance());
            }
            inAction.setInAction(false);
            inAction = null;
            state = STATE_SELECTING;
            diagram.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            diagram.repaint();
        }
         
    }
    

    public Component getFirstFocusableComponent() {
        return up;
    }
    
    public Component getLastFocusableComponent() {
        return disconnect;
    }
    
    public Component getNextFocusableComponent(Component aComponent) {
        if (aComponent.equals(up)) {
            return down;
        } else if (aComponent.equals(down)) {
            return select;
        } else if (aComponent.equals(select)) {
            return move;
        } else if (aComponent.equals(move)) {
            return connect;
        } else if (aComponent.equals(connect)) {
            return disconnect;
        } else {
            return null;
        }
    }
    
    public Component getPreviousFocusableComponent(Component aComponent) {
        if (aComponent.equals(disconnect)) {
            return connect;
        } else if (aComponent.equals(connect)) {
            return move;
        } else if (aComponent.equals(move)) {
            return select;
        } else if (aComponent.equals(select)) {
            return down;
        } else if (aComponent.equals(down)) {
            return up;
        } else {
            return null;
        }
    }
    
    
    private void selectElementAt(int x, int y) {
        selectElement(diagram.getElementAt(x, y), true);
    }
    
    private void selectElement(DiagramElement element, boolean informOthers) {
        // deselect old selection
        if (this.selected != null) {
            this.selected.setSelected(false);
            this.selected = null;
        }
        if (element != null) {
            element.setSelected(true);
            this.selected = element;
        }
        if (informOthers && element instanceof InstantiatedElement) {
            mainGUI.selectInstance(((InstantiatedElement)selected).getInstance());
        }
    }
    
    
    private void selectInstance(Instance instance) {
        // first check that component levels are ok
        if (diagramHandler.getCurrentComponentLevel() == null) {
            return;
        }
        // check whether the instance is already selected
        if (selected instanceof InstantiatedElement &&
            ((InstantiatedElement)selected).getInstance().equals(instance)) {
            return; //no need to do anything
        }
        // check whether instance corresponds to root element
        if (rootElement.getComponentInstance().equals(instance)) {
            selectElement(rootElement, false);
            repaint();
            return;
        }
        // look for instance from child elements
        DiagramElement element = diagramHandler.getCurrentComponentLevel()
                                        .getChildElementForInstance(instance);
        // deselect old selection
        if (this.selected != null) {
            this.selected.setSelected(false);
            this.selected = null;
        }
        if (element != null) {
            element.setSelected(true);
            this.selected = element;
        }
        repaint();
    }
    
    private void updateView() {
        diagram.updateSize();
        diagramScroll.setViewportView(diagram);
    }
    
    private void navigateLevelUp() {
        //go up!
        ComponentDiagramLevel level = diagramHandler.navigateComponentLevelUp();
        if (level != null) {
            init();
            updateView();
            repaint();
        }
    }
    
    private void navigateLevelDown(ComponentElement element) {
        //go down!
        ComponentDiagramLevel level = diagramHandler.navigateComponentLevelDown(element);
        if (level != null) {
            init();
            updateView();
            repaint();
        }
    }
    
}
