/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ComponentInfoPanel.java
 *
 * Created on 23. helmikuuta 2004, 12:58
 */

package kumbang.configurator.client.gui.core;

import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.InterfaceInstance;

import kumbang.util.gui.Icons;
import kumbang.util.gui.Margins;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.util.Iterator;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;

/**
 *
 * @author  vmyllarn
 */
public class ComponentInfoPanel extends JPanel implements ActionListener {
    
    private static final Dimension LIST_SIZE = new Dimension(150, 100);
       
    /** Reference to main GUI */
    private MainGUI mainGUI;
    
    /** Current component instance that is showing or null */
    private ComponentInstance instance;
        
    //---------- COMPONENT INFO -----------------------------------------------
    
    /** label containing component name */
    private JLabel componentName;
    
    /** label containing component type */
    private JLabel componentType;
    
    /** list containing parts of the component */
    private JList parts;
    
    /** list containing interfaces of the component */
    private JList interfaces;
    
    /** list containing attributes of the component */
    private JList attributes;
    
    /** list model containing parts of the component */
    private DefaultListModel partModel;
    
    /** list model containing interfaces of the component */
    private DefaultListModel interfaceModel;
    
    /** list model containing attributes of the component */
    private DefaultListModel attributeModel;
    
    /** button for adding parts under the component */
    private JButton addPart;
    
    /** button for removing parts under the component */
    private JButton removePart;
    
    /** button for navigating into part of the component */
    private JButton gotoPart;
    
    /** button for adding interfaces to the component */
    private JButton addInterface;
    
    /** button for removing interfaces from the component */
    private JButton removeInterface;
    
    /** button for navigating into interface of the component */
    private JButton gotoInterface;
    
    /** button for editing attributes of the component */
    private JButton editAttribute;
    
    
    //---------- PUBLIC METHODS -----------------------------------------------
    
    
    /**
     * Creates a new instance of ComponentInfoPanel 
     */
    public ComponentInfoPanel(MainGUI mainGUI) {
        this.mainGUI = mainGUI;
        initComponents();
    }
    
    
    public void reset() {
        selectInstance(null);
    }
    
    public void selectInstance(ComponentInstance instance) {
        if (instance != null && this.instance == instance) {
            return;
        }
        this.instance = instance;
        extractComponentInfo();
    }
    
    public void updateInformation() {
        if (instance != null) {
            extractComponentInfo();
            repaint();
        }
    }
    
    private void extractComponentInfo() {
        if (instance == null) {
            return;
        }
        
        // extract component properties
        componentName.setText(instance.getDisplayName());
        componentType.setText(instance.getType().getDisplayName());

        // extract parts
        partModel.clear();
        Iterator it = instance.getPartInstances();
        while (it.hasNext()) {
            partModel.addElement(it.next());
        }
        parts.repaint();
        
        // extract interfaces
        interfaceModel.clear();
        it = instance.getInterfaces();
        while (it.hasNext()) {
            interfaceModel.addElement(it.next());
        }
        interfaces.repaint();
        
        // extract attributes
        attributeModel.clear();
        it = instance.getAttributes();
        while (it.hasNext()) {
            attributeModel.addElement(it.next());
        }
        attributes.repaint();
    }
    

    
    //---------- PRIVATE METHODS FOR INITING SWING COMPONENTS ----------------
        
    private void initComponents() {
        setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));

        // create panel for component properties (name, type etc)
        JPanel propertyPanel = createComponentPropertyPanel();
        propertyPanel.setAlignmentY(Component.TOP_ALIGNMENT);
        propertyPanel.setMaximumSize(propertyPanel.getPreferredSize());
        
        // create panel for parts
        JPanel partPanel = createComponentPartPanel();
        partPanel.setAlignmentY(Component.TOP_ALIGNMENT);
        partPanel.setMaximumSize(partPanel.getPreferredSize());
        
        // create panel for interfaces
        JPanel intPanel = createComponentInterfacePanel();
        intPanel.setAlignmentY(Component.TOP_ALIGNMENT);
        intPanel.setMaximumSize(intPanel.getPreferredSize());
        
        // create panel for attributes
        JPanel attrPanel = createComponentAttributePanel();
        attrPanel.setAlignmentY(Component.TOP_ALIGNMENT);
        attrPanel.setMaximumSize(attrPanel.getPreferredSize());
        
        //create box for header
        Box header = Box.createHorizontalBox();
        JLabel headerLabel = new JLabel("Component info", Icons.COMPONENT, 
                                        SwingConstants.CENTER);
        header.add(Margins.createHorizontalMarginM());
        header.add(headerLabel);
        header.add(Box.createHorizontalGlue());
        
        // create box for panels
        Box panels = Box.createHorizontalBox();
        panels.add(Margins.createHorizontalMarginXL());
        panels.add(propertyPanel);
        panels.add(Margins.createHorizontalMarginXL());
        panels.add(partPanel);
        panels.add(Margins.createHorizontalMarginXL());
        panels.add(intPanel);
        panels.add(Margins.createHorizontalMarginXL());
        panels.add(attrPanel);
        panels.add(Margins.createHorizontalMarginL());
        panels.add(Box.createHorizontalGlue());
        
        
        // add header and panel areas into component info panel
        add(Margins.createVerticalMarginM());
        add(header);
        add(Margins.createVerticalMarginL());
        add(panels);
        add(Margins.createVerticalMarginL());
        add(Box.createVerticalGlue());
    }

    /**
     * Creates a panel for general component info. This includes
     * info for name and type. Returns a panel that contains necessary
     * info.
     */
    private JPanel createComponentPropertyPanel() {
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints c = new GridBagConstraints();
        
        JPanel propertyPanel = new JPanel(gridbag);
        
        c.gridx = 0;
        c.gridy = 0;
        c.anchor = GridBagConstraints.FIRST_LINE_START;
        c.insets = new Insets(0, 0, Margins.S, Margins.M);

        JLabel tempLabel = new JLabel("Name");
        propertyPanel.add(tempLabel, c);
        
        c.gridy++;
        tempLabel = new JLabel("Type");
        propertyPanel.add(tempLabel, c);
        
        c.gridy = 0;
        c.gridx++;
        c.insets = new Insets(0, 0, Margins.S, 0);
        componentName = new JLabel("");
        propertyPanel.add(componentName, c);
        
        c.gridy++;
        componentType = new JLabel("");
        propertyPanel.add(componentType, c);
        
        c.gridx = 0;
        c.gridy++;
        c.gridwidth = 2;
        propertyPanel.add(Box.createVerticalGlue(), c);
         
        return propertyPanel;
    }
    
    private JPanel createComponentPartPanel() {
        JPanel partPanel = new JPanel();
        partPanel.setLayout(new BoxLayout(partPanel, BoxLayout.PAGE_AXIS));
        
        //initialise list
        partModel = new DefaultListModel();
        parts = new JList(partModel);
        JScrollPane listScroll = new JScrollPane(parts);
        listScroll.setPreferredSize(LIST_SIZE);
        listScroll.setMaximumSize(LIST_SIZE);
        
        //initilise buttons        
        addPart = new JButton(Icons.NEW_COMPONENT);
        addPart.setMargin(new Insets(0,0,0,0));
        addPart.setToolTipText("Add new subcomponent to this component");
        addPart.addActionListener(this);
        
        gotoPart = new JButton(Icons.RIGHT);
        gotoPart.setMargin(new Insets(0,0,0,0));
        gotoPart.setToolTipText("Go to the selected subcomponent of this component");
        gotoPart.addActionListener(this);
        
        removePart = new JButton(Icons.DELETE);
        removePart.setMargin(new Insets(0,0,0,0));
        removePart.setToolTipText("Remove selected subcomponent from this component");
        removePart.addActionListener(this);
        
        // lay buttons out in a box vertically (top to bottom)
        Box buttonBox = Box.createVerticalBox();
        buttonBox.add(gotoPart);
        buttonBox.add(Margins.createVerticalMarginXS());
        buttonBox.add(addPart);
        buttonBox.add(Margins.createVerticalMarginXS());
        buttonBox.add(removePart);
        buttonBox.add(Box.createVerticalGlue());
        
        // lay out list and buttons in a box
        Box listBox = Box.createHorizontalBox();
        listBox.add(listScroll);
        listBox.add(Margins.createHorizontalMarginS());
        listBox.add(buttonBox);
        listBox.add(Box.createHorizontalGlue());
        
        // add all parts into panel
        JLabel header = new JLabel("Subcomponents");
        header.setAlignmentX(Component.LEFT_ALIGNMENT);
        listBox.setAlignmentX(Component.LEFT_ALIGNMENT);
        partPanel.add(header);
        partPanel.add(Margins.createVerticalMarginS());
        partPanel.add(listBox);
        partPanel.add(Box.createVerticalGlue());
        
        return partPanel;
    }
    
    private JPanel createComponentInterfacePanel() {
        JPanel intPanel = new JPanel();
        intPanel.setLayout(new BoxLayout(intPanel, BoxLayout.PAGE_AXIS));

        interfaceModel = new DefaultListModel();
        interfaces = new JList(interfaceModel);
        JScrollPane listScroll = new JScrollPane(interfaces);
        listScroll.setPreferredSize(LIST_SIZE);
        listScroll.setMaximumSize(LIST_SIZE);
        
        addInterface = new JButton(Icons.NEW_INTERFACE);
        addInterface.setMargin(new Insets(0,0,0,0));
        addInterface.setToolTipText("Add new interface to this component");
        addInterface.addActionListener(this);

        gotoInterface = new JButton(Icons.RIGHT);
        gotoInterface.setMargin(new Insets(0,0,0,0));
        gotoInterface.setToolTipText("Go to the selected interface of this component");
        gotoInterface.addActionListener(this);

        removeInterface = new JButton(Icons.DELETE);
        removeInterface.setMargin(new Insets(0,0,0,0));
        removeInterface.setToolTipText("Remove selected interface from this component");
        removeInterface.addActionListener(this);

        // lay buttons out in a box vertically (top to bottom)
        Box buttonBox = Box.createVerticalBox();
        buttonBox.add(gotoInterface);
        buttonBox.add(Margins.createVerticalMarginXS());
        buttonBox.add(addInterface);
        buttonBox.add(Margins.createVerticalMarginXS());
        buttonBox.add(removeInterface);
        buttonBox.add(Box.createVerticalGlue());
        
        // lay out list and buttons in a box (left to right)
        Box listBox = Box.createHorizontalBox();
        listBox.add(listScroll);
        listBox.add(Margins.createHorizontalMarginS());
        listBox.add(buttonBox);
        listBox.add(Box.createHorizontalGlue());
        
        // add all parts into panel
        JLabel header = new JLabel("Interfaces");
        header.setAlignmentX(Component.LEFT_ALIGNMENT);
        listBox.setAlignmentX(Component.LEFT_ALIGNMENT);
        intPanel.add(header);
        intPanel.add(Margins.createVerticalMarginS());
        intPanel.add(listBox);
        intPanel.add(Box.createVerticalGlue());
        
        return intPanel;
    }
    
    
    private JPanel createComponentAttributePanel() {
        JPanel attrPanel = new JPanel();
        attrPanel.setLayout(new BoxLayout(attrPanel, BoxLayout.PAGE_AXIS));
        
        //initialise list
        attributeModel = new DefaultListModel();
        attributes = new JList(attributeModel);
        JScrollPane listScroll = new JScrollPane(attributes);
        listScroll.setPreferredSize(LIST_SIZE);
        listScroll.setMaximumSize(LIST_SIZE);
        
        //initilise buttons
        editAttribute = new JButton(Icons.ATTRIBUTE);
        editAttribute.setMargin(new Insets(0,0,0,0));
        editAttribute.setToolTipText("Edit attributes of this component");
        editAttribute.addActionListener(this);
        
        // lay buttons out in a box vertically (top to bottom)
        Box buttonBox = Box.createVerticalBox();
        buttonBox.add(editAttribute);
        buttonBox.add(Box.createVerticalGlue());
        
        // lay out list and buttons in a box
        Box listBox = Box.createHorizontalBox();
        listBox.add(listScroll);
        listBox.add(Margins.createHorizontalMarginS());
        listBox.add(buttonBox);
        listBox.add(Box.createHorizontalGlue());
        
        // add all parts into panel
        JLabel header = new JLabel("Attributes");
        header.setAlignmentX(Component.LEFT_ALIGNMENT);
        listBox.setAlignmentX(Component.LEFT_ALIGNMENT);
        attrPanel.add(header);
        attrPanel.add(Margins.createVerticalMarginS());
        attrPanel.add(listBox);
        attrPanel.add(Box.createVerticalGlue());
        
        return attrPanel;
    }
    
    

    
    public void actionPerformed(ActionEvent e) {
        Object source = e.getSource();
        
        /* parts for component */
        if (source == addPart) {
            if (instance != null) {
                mainGUI.addComposable(instance);
            }
        }
        else if (source == removePart) {
            ComponentInstance c = (ComponentInstance)parts.getSelectedValue();
            if (c != null) {
                mainGUI.removeComposable(c);
            }
        }
        else if (source == gotoPart) {
            ComponentInstance c = (ComponentInstance)parts.getSelectedValue();
            if (c != null) {
                mainGUI.selectInstance(c);
            }
        }
        /* interfaces for component */
        else if (source == addInterface) {
            if (instance != null) {
                mainGUI.editInterfaces(instance);
            }
        }
        else if (source == removeInterface) {
            InterfaceInstance i = (InterfaceInstance)interfaces.getSelectedValue();
            if (i != null) {
                mainGUI.removeInterface(i);
            }
        }
        else if (source == gotoInterface) {
            InterfaceInstance i = (InterfaceInstance)interfaces.getSelectedValue();
            if (i != null) {
                mainGUI.selectInstance(i);
            }
        }
        /* attributes for component */
        else if (source == editAttribute) {
            if (instance != null) {
                mainGUI.editAttributes(instance);
            }
        }
    }
    
}
