/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * FeatureInfoPanel.java
 *
 * Created on 28. tammikuuta 2005, 13:13
 */

package kumbang.configurator.client.gui.core;

import kumbang.core.configuration.instance.AttributedInstance;
import kumbang.core.configuration.instance.FeatureInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.task.AddComposableTask;
import kumbang.core.configuration.task.ClientTask;
import kumbang.core.configuration.task.RemoveInstanceTask;
import kumbang.core.configuration.task.SetAttributeTask;


import kumbang.configurator.client.gui.event.ConfigurationModificationListener;
import kumbang.configurator.client.gui.event.ConfigurationModifiedEvent;

import java.awt.CardLayout;
import java.awt.Component;

import java.util.HashMap;
import java.util.Iterator;

import javax.swing.JPanel;

/**
 * 
 * @author vmyllarn
 */
public class FeatureInfoPanel extends JPanel implements ConfigurationModificationListener {

	private static final String EMPTY = "empty";

	/** Reference to main GUI */
	private MainGUI mainGUI;

	/** Card layout holding different panels */
	private CardLayout cardLayout;

	/** Empty panel for showing no information */
	private JPanel emptyInfo;

	/** Hashmap for finding corresponding feature panels for features */
	private HashMap<String, FeaturePanel> panels;

	/** Current feature instance that is shown */
	private FeatureInstance currentInstance;

	/** Creates a new instance of FeatureSelectionPanel */
	public FeatureInfoPanel(MainGUI mainGUI) {
		this.mainGUI = mainGUI;
		this.mainGUI.addConfigurationModificationListener(this);
		this.panels = new HashMap<String, FeaturePanel>();
		emptyInfo = new JPanel();
		this.cardLayout = new CardLayout();
		setLayout(cardLayout);
		add(emptyInfo, EMPTY);
	}

	public void init(FeatureInstance featureRoot) {
		if (featureRoot != null) {
			addFeaturePanelsRecursively(featureRoot);
			showPanelForFeature(featureRoot);
		}
	}

	private void addFeaturePanelsRecursively(FeatureInstance feature) {
		addPanelForFeature(feature);
		Iterator it = feature.getPartInstances();
		while (it.hasNext()) {
			addPanelForFeature((FeatureInstance) it.next());
		}
	}

	public void reset() {
		removeAll();
		panels.clear();
		add(emptyInfo, EMPTY);
	}

	public void showPanelForFeature(FeatureInstance instance) {
		String key = getKey(instance);
		this.cardLayout.show(this, key);
		currentInstance = instance;
	}

	public FeatureInstance getCurrentShownFeature() {
		return currentInstance;
	}

	public void configurationModified(ConfigurationModifiedEvent e) {
		ClientTask task = e.getTask();
		handleTask(task);
	}

	private void handleTask(ClientTask task) {
		if (task instanceof AddComposableTask) {
			Instance instance = ((AddComposableTask) task).getAddedInstance();
			if (instance instanceof FeatureInstance) {
				addPanelForFeature((FeatureInstance) instance);
			}
		} else if (task instanceof RemoveInstanceTask) {
			Instance instance = ((RemoveInstanceTask) task).getInstance();
			if (instance instanceof FeatureInstance) {
				removePanelForFeature((FeatureInstance) instance);
			}
		} else if (task instanceof SetAttributeTask) {
			AttributedInstance instance = ((SetAttributeTask) task).getParentInstance();
			if (instance instanceof FeatureInstance) {
				updatePanelForFeature((FeatureInstance) instance);
			}
		}
		Iterator it = task.getSubTasks();
		while (it.hasNext()) {
			handleTask((ClientTask) it.next());
		}
	}

	private void addPanelForFeature(FeatureInstance instance) {
		FeaturePanel fPanel = new FeaturePanel(this, mainGUI, instance);
		String key = getKey(instance);
		panels.put(key, fPanel);
		add(fPanel, key);
		cardLayout.addLayoutComponent(fPanel, key);
		FeatureInstance parent = (FeatureInstance) instance.getParent();
		if (parent != null) {
			updatePanelForFeature(parent);
		}
	}

	private void removePanelForFeature(FeatureInstance instance) {
		String key = getKey(instance);
		FeaturePanel fPanel = panels.get(key);
		if (fPanel != null) {
			panels.remove(key);
			remove(fPanel);
		}
	}

	private void updatePanelForFeature(FeatureInstance instance) {
		String key = getKey(instance);
		FeaturePanel fPanel = panels.get(key);
		if (fPanel != null) {
			fPanel.update();
		}
	}

	private String getKey(FeatureInstance instance) {
		return instance.toString();
	}

	public Component getFirstFocusableComponent() {
		if (currentInstance != null) {
			String key = getKey(currentInstance);
			FeaturePanel fPanel = panels.get(key);
			if (fPanel != null) {
				return fPanel.getFirstFocusableComponent();
			}
		}
		return null;
	}

	public Component getLastFocusableComponent() {
		if (currentInstance != null) {
			String key = getKey(currentInstance);
			FeaturePanel fPanel = panels.get(key);
			if (fPanel != null) {
				return fPanel.getLastFocusableComponent();
			}
		}
		return null;
	}

	public Component getNextFocusableComponent(Component aComponent) {
		if (currentInstance != null) {
			String key = getKey(currentInstance);
			FeaturePanel fPanel = panels.get(key);
			if (fPanel != null) {
				return fPanel.getNextFocusableComponent(aComponent);
			}
		}
		return null;
	}

	public Component getPreviousFocusableComponent(Component aComponent) {
		if (currentInstance != null) {
			String key = getKey(currentInstance);
			FeaturePanel fPanel = panels.get(key);
			if (fPanel != null) {
				return fPanel.getPreviousFocusableComponent(aComponent);
			}
		}
		return null;
	}

}
