/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * InterfaceInfoPanel.java
 *
 * Created on 6. syyskuuta, 12:58
 */

package kumbang.configurator.client.gui.core;

import kumbang.util.gui.Icons;
import kumbang.util.gui.Margins;

import kumbang.core.configuration.instance.BindingInstance;
import kumbang.core.configuration.instance.InterfaceInstance;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.util.Iterator;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;

/**
 * 
 * @author vmyllarn
 */
public class InterfaceInfoPanel extends JPanel implements ActionListener {

	private static final Dimension LIST_SIZE = new Dimension(150, 100);

	/** Reference to main GUI */
	private MainGUI mainGUI;

	/** Current interface instance that is showing, or null */
	private InterfaceInstance instance;

	// ---------- INTERFACE INFO -----------------------------------------------

	/** label for interface name */
	private JLabel interfaceName;

	/** label for interface type */
	private JLabel interfaceType;

	/** label for interface direction (requires / provides) */
	private JLabel interfaceDirection;

	/** label for interface functions */
	// private JLabel interfaceFunctions;
	/** label for telling whether interface is optional or required */
	private JLabel interfaceRequired;

	/** label for telling whether interface is grounded or not */
	private JLabel interfaceGrounded;

	/** list for bindings from the interface */
	private JList bindings;

	/** list for functions of this interface */
	private JList functions;

	/** list for bindings from the interface */
	private DefaultListModel bindingModel;

	/** list for functions of this interface */
	private DefaultListModel functionModel;

	/** button for removing binding from interface */
	private JButton removeBinding;

	/** button for navigating into other interface behind the binding */
	private JButton gotoBinding;

	// ---------- PUBLIC METHODS -----------------------------------------------

	/**
	 * Creates a new instance of ComponentInfoPanel
	 */
	public InterfaceInfoPanel(MainGUI mainGUI) {
		this.mainGUI = mainGUI;
		initComponents();
	}

	public void reset() {
		selectInstance(null);
	}

	public void selectInstance(InterfaceInstance instance) {
		if (instance != null && this.instance == instance) {
			return;
		}
		this.instance = instance;
		extractInterfaceInfo();
	}

	public void updateInformation() {
		if (instance != null) {
			extractInterfaceInfo();
			repaint();
		}
	}

	private void extractInterfaceInfo() {
		if (instance == null) {
			return;
		}

		// extract interface properties
		interfaceName.setText(instance.getDisplayName());
		interfaceType.setText(instance.getType().getDisplayName());
		interfaceDirection.setText(instance.isProvided() ? "provided" : "required");
		interfaceRequired.setText(instance.isOptional() ? "no" : "yes");
		interfaceGrounded.setText(instance.isGrounded() ? "yes" : "no");

		// extract bindings
		bindingModel.clear();
		Iterator it = instance.getBindings();
		while (it.hasNext()) {
			bindingModel.addElement(it.next());
		}
		bindings.repaint();

		// extract functions
		functionModel.clear();
		it = instance.getFunctions();
		while (it.hasNext()) {
			functionModel.addElement(it.next());
		}
		functions.repaint();
	}

	// ---------- PRIVATE METHODS FOR INITING SWING COMPONENTS ----------------

	private void initComponents() {
		setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));

		// create panel for interface properties (name, type etc)
		JPanel propertyPanel = createInterfacePropertyPanel();
		propertyPanel.setAlignmentY(Component.TOP_ALIGNMENT);
		propertyPanel.setMaximumSize(propertyPanel.getPreferredSize());

		// create panel for bindings
		JPanel bindPanel = createInterfaceBindingPanel();
		bindPanel.setAlignmentY(Component.TOP_ALIGNMENT);
		bindPanel.setMaximumSize(bindPanel.getPreferredSize());

		// create panel for functions
		JPanel functionPanel = createInterfaceFunctionPanel();
		functionPanel.setAlignmentY(Component.TOP_ALIGNMENT);
		functionPanel.setMaximumSize(functionPanel.getPreferredSize());

		// create box for header
		Box header = Box.createHorizontalBox();
		JLabel headerLabel = new JLabel("Interface info", Icons.INTERFACE, SwingConstants.CENTER);
		header.add(Margins.createHorizontalMarginM());
		header.add(headerLabel);
		header.add(Box.createHorizontalGlue());

		// create box for panels
		Box panels = Box.createHorizontalBox();
		panels.add(Margins.createHorizontalMarginXL());
		panels.add(propertyPanel);
		panels.add(Margins.createHorizontalMarginXL());
		panels.add(bindPanel);
		panels.add(Margins.createHorizontalMarginXL());
		panels.add(functionPanel);
		panels.add(Margins.createHorizontalMarginL());
		panels.add(Box.createHorizontalGlue());

		// add header and panel areas into interface info panel
		add(Margins.createVerticalMarginM());
		add(header);
		add(Margins.createVerticalMarginL());
		add(panels);
		add(Margins.createVerticalMarginL());
		add(Box.createVerticalGlue());
	}

	/**
	 * Creates a panel for general interface info. This includes info for name and type. Returns a
	 * panel that contains necessary info.
	 */
	private JPanel createInterfacePropertyPanel() {
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();

		JPanel propertyPanel = new JPanel(gridbag);

		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.FIRST_LINE_START;
		c.insets = new Insets(0, 0, Margins.S, Margins.M);

		JLabel tempLabel = new JLabel("Name");
		propertyPanel.add(tempLabel, c);

		c.gridy++;
		tempLabel = new JLabel("Type");
		propertyPanel.add(tempLabel, c);

		c.gridy++;
		tempLabel = new JLabel("Direction");
		propertyPanel.add(tempLabel, c);

		c.gridy++;
		tempLabel = new JLabel("Required");
		propertyPanel.add(tempLabel, c);

		c.gridy++;
		tempLabel = new JLabel("Grounded");
		propertyPanel.add(tempLabel, c);

		c.gridy = 0;
		c.gridx++;
		c.insets = new Insets(0, 0, Margins.S, 0);
		interfaceName = new JLabel("");
		propertyPanel.add(interfaceName, c);

		c.gridy++;
		interfaceType = new JLabel("");
		propertyPanel.add(interfaceType, c);

		c.gridy++;
		interfaceDirection = new JLabel("");
		propertyPanel.add(interfaceDirection, c);

		c.gridy++;
		interfaceRequired = new JLabel("");
		propertyPanel.add(interfaceRequired, c);

		c.gridy++;
		interfaceGrounded = new JLabel("foobar foobar foobar foobar foobar foobar");
		propertyPanel.add(interfaceGrounded, c);

		c.gridx = 0;
		c.gridy++;
		c.gridwidth = 2;
		propertyPanel.add(Box.createVerticalGlue(), c);

		return propertyPanel;
	}

	private JPanel createInterfaceBindingPanel() {
		JPanel bindPanel = new JPanel();
		bindPanel.setLayout(new BoxLayout(bindPanel, BoxLayout.PAGE_AXIS));

		// initialise list
		bindingModel = new DefaultListModel();
		bindings = new JList(bindingModel);
		JScrollPane listScroll = new JScrollPane(bindings);
		listScroll.setPreferredSize(LIST_SIZE);
		listScroll.setMaximumSize(LIST_SIZE);

		// initilise buttons
		gotoBinding = new JButton(Icons.RIGHT);
		gotoBinding.setMargin(new Insets(0, 0, 0, 0));
		gotoBinding.setToolTipText("Go to the other end of selected binding");
		gotoBinding.addActionListener(this);

		removeBinding = new JButton(Icons.DELETE);
		removeBinding.setMargin(new Insets(0, 0, 0, 0));
		removeBinding.setToolTipText("Remove selected binding from this interface");
		removeBinding.addActionListener(this);

		// lay buttons out in a box vertically (top to bottom)
		Box buttonBox = Box.createVerticalBox();
		buttonBox.add(gotoBinding);
		buttonBox.add(Margins.createVerticalMarginXS());
		buttonBox.add(removeBinding);
		buttonBox.add(Box.createVerticalGlue());

		// lay out list and buttons in a box
		Box listBox = Box.createHorizontalBox();
		listBox.add(listScroll);
		listBox.add(Margins.createHorizontalMarginS());
		listBox.add(buttonBox);
		listBox.add(Box.createHorizontalGlue());

		// add all parts into panel
		JLabel header = new JLabel("Bindings");
		header.setAlignmentX(Component.LEFT_ALIGNMENT);
		listBox.setAlignmentX(Component.LEFT_ALIGNMENT);
		bindPanel.add(header);
		bindPanel.add(Margins.createVerticalMarginS());
		bindPanel.add(listBox);
		bindPanel.add(Box.createVerticalGlue());

		return bindPanel;
	}

	private JPanel createInterfaceFunctionPanel() {
		JPanel functionPanel = new JPanel();
		functionPanel.setLayout(new BoxLayout(functionPanel, BoxLayout.PAGE_AXIS));

		// initialise list
		functionModel = new DefaultListModel();
		functions = new JList(functionModel);
		JScrollPane listScroll = new JScrollPane(functions);
		listScroll.setPreferredSize(LIST_SIZE);
		listScroll.setMaximumSize(LIST_SIZE);

		// add all parts into panel
		JLabel header = new JLabel("Functions");
		header.setAlignmentX(Component.LEFT_ALIGNMENT);
		functions.setAlignmentX(Component.LEFT_ALIGNMENT);
		functionPanel.add(header);
		functionPanel.add(Margins.createVerticalMarginS());
		functionPanel.add(listScroll);
		functionPanel.add(Box.createVerticalGlue());

		return functionPanel;
	}

	public void actionPerformed(ActionEvent e) {
		Object source = e.getSource();
		/* bindings for interface */
		if (source == gotoBinding) {
			BindingInstance bind = (BindingInstance) bindings.getSelectedValue();
			if (instance != null && bind != null) {
//				InterfaceInstance i = instance.getInterfaceBindedWith(bind);
				
				InterfaceInstance i = null;
				if (!bind.getSourceInterface().equals(instance))
					i = bind.getSourceInterface();
				else
					i = bind.getTargetInterface();
				
				mainGUI.selectInstance(i);
			}
		} else if (source == removeBinding) {
			BindingInstance bind = (BindingInstance) bindings.getSelectedValue();
			if (bind != null) {
				mainGUI.removeBinding(bind);
			}
		}
	}

}
