/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * MainGUI.java
 *
 * Created on 23. helmikuuta 2004, 11:17
 */
package kumbang.configurator.client.gui.core;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Icon;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.WindowConstants;

import kumbang.configurator.client.core.ClientServices;
import kumbang.configurator.client.core.TaskHandler;
import kumbang.configurator.client.core.exception.ClientException;
import kumbang.configurator.client.core.exception.ExportException;
import kumbang.configurator.client.core.exception.OpenException;
import kumbang.configurator.client.core.exception.SaveException;
import kumbang.configurator.client.gui.diagram.ComponentElement;
import kumbang.configurator.client.gui.diagram.DiagramHandler;
import kumbang.configurator.client.gui.dialog.AddComposableDialog;
import kumbang.configurator.client.gui.dialog.EditAttributesDialog;
import kumbang.configurator.client.gui.dialog.EditInterfacesDialog;
import kumbang.configurator.client.gui.dialog.ExportConfigurationDialog;
import kumbang.configurator.client.gui.dialog.ImportModelDialog;
import kumbang.configurator.client.gui.dialog.OpenModelDialog;
import kumbang.configurator.client.gui.dialog.SaveConfigurationDialog;
import kumbang.configurator.client.gui.dialog.SelectPartDialog;
import kumbang.configurator.client.gui.event.ComponentSelectedEvent;
import kumbang.configurator.client.gui.event.ComponentSelectionListener;
import kumbang.configurator.client.gui.event.ConfigurationModificationListener;
import kumbang.configurator.client.gui.event.ConfigurationModifiedEvent;
import kumbang.configurator.client.gui.tree.ComponentTree;
import kumbang.configurator.client.gui.tree.FeatureTree;
import kumbang.configurator.client.gui.tree.TreeHandler;
import kumbang.core.configuration.core.ConfigurationException;
import kumbang.core.configuration.instance.AttributedInstance;
import kumbang.core.configuration.instance.BindingInstance;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.ComposableInstance;
import kumbang.core.configuration.instance.FeatureInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.PartInstance;
import kumbang.core.configuration.instance.TypedInstance;
import kumbang.core.configuration.task.AddBindingTask;
import kumbang.core.configuration.task.AddComposableTask;
import kumbang.core.configuration.task.ClientTask;
import kumbang.core.configuration.task.EditAttributesTask;
import kumbang.core.configuration.task.EditInterfacesTask;
import kumbang.core.configuration.task.RemoveInstanceTask;
import kumbang.core.model.exception.ModelException;
import kumbang.util.gui.Icons;
import kumbang.util.gui.KumbangFileChooser;

/**
 * 
 * @author vmyllarn
 */
public class MainGUI extends JFrame {

	/** configuration client */
	private ClientServices client;

	/** task handler */
	private TaskHandler taskHandler;

	/** consequence handler */
//	private ConsequenceHandler cnsqHandler;

	/** handler for diagram elements */
	private DiagramHandler diagramHandler;

	/** handler for tree elements */
	private TreeHandler treeHandler;

	/** filechooser for opening and saving */
	private KumbangFileChooser fileChooser;

	/** should configuration consequences applied automatically? */
	// private boolean checkConfigurationConsequences = true;
	// ------listeners-----------------------------------------
	/** a list of listeners for configuration modifications */
	private List<ConfigurationModificationListener> modificationListeners;

	/** a list of listeners for selecing elements in the configuration */
	private List<ComponentSelectionListener> selectionListeners;

	// ------graphical elements--------------------------------

	/** feature info panel */
	private FeatureInfoPanel featureInfo;

	/** component diagram */
	private ComponentDiagramPanel componentDiagram;

	/** diagram info panel */
	private DiagramInfoPanel diagramInfo;

	/** configuration info panel */
	private ConfigurationInfoPanel configurationInfo;

	/** feature tree */
	private FeatureTree featureTree;

	/** component tree */
	private ComponentTree componentTree;

	/** menu bar */
	private ClientMenuBar menubar;

	/** tab for trees */
	private JTabbedPane treeTabs;

	/** tab for diagrams */
	private JTabbedPane diagramTabs;

	protected static final int TAB_INDEX_FEATURE = 0;
	protected static final int TAB_INDEX_COMPONENT = 1;

	// ------------------------------------------------------------------
	// ------ PUBLIC METHODS --------------------------------------------
	// ------------------------------------------------------------------

	/**
	 * Creates a new instance of MainGUI
	 * 
	 * @param client reference to configuration client
	 */
	// public MainGUI(ConfigurationClient client) {
	public MainGUI(ClientServices client) {
		super("Kumbang Configuration Client");

		// System.out.println("MainGUI: Initing");
		this.client = client;
		// this.configHandler = client.getConfigurationHandler();
		this.taskHandler = client.getTaskHandler();
//		this.cnsqHandler = client.getConsequenceHandler();

		modificationListeners = new ArrayList<ConfigurationModificationListener>();
		selectionListeners = new ArrayList<ComponentSelectionListener>();

		// System.out.println("MainGUI: Init filechooser");
		this.fileChooser = new KumbangFileChooser();

		initComponents();

		setIconImage(Icons.COMPONENT.getImage());

		setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				exit();
			}
		});

		pack();
		setExtendedState(MAXIMIZED_BOTH);
	}

	// ------------------------------------------------------------------
	// ------ HANDLING LISTENERS ----------------------------------------
	// ------------------------------------------------------------------

	public void addConfigurationModificationListener(ConfigurationModificationListener listener) {
		modificationListeners.add(listener);
	}

	public void addComponentSelectionListener(ComponentSelectionListener listener) {
		selectionListeners.add(listener);
	}

	private void fireConfigurationModifiedEvent(ClientTask task) {
		ConfigurationModifiedEvent e = new ConfigurationModifiedEvent(task);
		for (ConfigurationModificationListener listener : modificationListeners) {
			listener.configurationModified(e);
		}
	}

	private void fireInstanceSelectedEvent(ComponentInstance instance) {
		ComponentSelectedEvent e = new ComponentSelectedEvent(instance);
		for (ComponentSelectionListener listener : selectionListeners) {
			listener.componentSelected(e);
		}
	}

	private void fireInstanceSelectedEvent(InterfaceInstance instance) {
		ComponentSelectedEvent e = new ComponentSelectedEvent(instance);
		for (ComponentSelectionListener listener : selectionListeners) {
			listener.componentSelected(e);
		}
	}

	private void fireInstanceSelectedEvent(BindingInstance instance) {
		ComponentSelectedEvent e = new ComponentSelectedEvent(instance);
		for (ComponentSelectionListener listener : selectionListeners) {
			listener.componentSelected(e);
		}
	}

	// ------------------------------------------------------------------
	// ------ GETTERS AND SETTERS ---------------------------------------
	// ------ Are these all necessary? ----------------------------------

	/**
	 * Returns the diagram handler
	 * 
	 * @return DiagramHandler
	 */
	public DiagramHandler getDiagramHandler() {
		return diagramHandler;
	}

	/**
	 * Returns the diagram handler
	 * 
	 * @return DiagramHandler
	 */
//	public ConsequenceHandler getConsequenceHandler() {
//		return cnsqHandler;
//	}

	/**
	 * Returns the kumbang file chooser
	 * 
	 * @return KumbangFileChooser
	 */
	public KumbangFileChooser getFileChooser() {
		return this.fileChooser;
	}

	/**
	 * Return the icon that is used for window icons
	 */
	public Icon getIcon() {
		return Icons.COMPONENT;
	}

	/**
	 * Return name of the current configuration
	 * 
	 * @return String name of the configuration
	 */
	public String getConfigurationName() {
		return client.getConfigurationName();
	}

	// ------------------------------------------------------------------
	// ------ BASIC OPERATIONS ------------------------------------------
	// ------ exit, open, close, save etc. ------------------------------

	/**
	 * Exits the client. Asks for a confirmation before exiting.
	 */
	public void exit() {
		/* check if there is an open, unsaved configuration */
		if (client.isConfigurationModified()) {
			int retValue = JOptionPane.showConfirmDialog(this, "Are you sure you want to exit?\n"
					+ "All changes you have made will be lost.", "Exit configurator",
					JOptionPane.YES_NO_OPTION);
			if (retValue == JOptionPane.NO_OPTION) {
				return;
			}
		}
		client.exit();
		dispose();
	}

	/**
	 * Opens a model and initializes the configuration. If there is a previous configuration open,
	 * asks for a confirmation.
	 */
	public void openModel() {
		OpenModelDialog dialog = new OpenModelDialog(this, client.getModelNames());
		dialog.setVisible(true);
		if (!dialog.getStatus()) {
			return;
		}
		String modelName = dialog.getModelName();
		String configurationName = dialog.getConfigurationName();
		if (client.isConfigurationModified()) {
			int retValue = JOptionPane.showConfirmDialog(this,
					"Do you want to open a new configuration?\n"
							+ "All changes you have made to this configuration will be lost.",
					"Open model", JOptionPane.YES_NO_OPTION);
			if (retValue == JOptionPane.NO_OPTION) {
				return;
			}
		}
		try {
			client.closeConfiguration();
			resetGUIContent();
			client.openModel(modelName, configurationName);
			initGUIContent();
		} catch (ModelException me) {
			JOptionPane.showMessageDialog(getContentPane(), me.getMessage(),
					"Model was not a valid Kumbang model", JOptionPane.ERROR_MESSAGE);
			client.closeConfiguration();
		} catch (ClientException ce) {
			JOptionPane.showMessageDialog(getContentPane(), ce.getMessage(),
					"Error when opening model", JOptionPane.ERROR_MESSAGE);
			client.closeConfiguration();
		}
	}

	/**
	 * Closes the current configuration. If there is a modified configuration open, asks for a
	 * confirmation before closing.
	 */
	public void closeConfiguration() {
		if (client.isConfigurationModified()) {
			int retValue = JOptionPane.showConfirmDialog(this,
					"Do you want to close this configuration?\n"
							+ "All changes you have made will be lost.", "Close configuration",
					JOptionPane.YES_NO_OPTION);
			if (retValue == JOptionPane.NO_OPTION) {
				return;
			}
		}
		client.closeConfiguration();
		resetGUIContent();
	}

	/**
	 * Saves the current configuration (to binary format). For a give configuration, the saving is
	 * always done in one particular file. If this is the first time one saves this configuration,
	 * this asks the name of the configuration.
	 */
	public void saveConfiguration() {
		if (client.isConfigurationStateReseted()) {
			return;
		}
		File saveFile = client.getSaveFile();
		if (saveFile == null) {
			SaveConfigurationDialog dialog = new SaveConfigurationDialog(this);
			dialog.setVisible(true);
			if (!dialog.getStatus()) {
				return;
			}
			saveFile = dialog.getSaveFile();
		}
		try {
			client.saveConfiguration(saveFile);
		} catch (SaveException se) {
			JOptionPane.showMessageDialog(this, se.getMessage(), "Cannot save configuration",
					JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Opens a previously saved configuration from a binary format.
	 */
	public void openConfiguration() {
		if (!fileChooser.showChooseConfigurationFileDialog(this)) {
			return;
		}
		File openFile = fileChooser.getSelectedFile();
		if (client.isConfigurationModified()) {
			int retValue = JOptionPane.showConfirmDialog(this,
					"Do you want to open a new configuration?\n"
							+ "All changes you have made will be lost.", "Open configuration",
					JOptionPane.YES_NO_OPTION);
			if (retValue == JOptionPane.NO_OPTION) {
				return;
			}
		}
		try {
			client.closeConfiguration();
			resetGUIContent();
			client.openConfiguration(openFile);
			initGUIContent();
		} catch (OpenException oe) {
			JOptionPane.showMessageDialog(this, oe.getMessage(), "Cannot open configuration",
					JOptionPane.ERROR_MESSAGE);
		} catch (ClientException ce) {
			JOptionPane.showMessageDialog(this, ce.getMessage(), "Cannot open configuration",
					JOptionPane.ERROR_MESSAGE);
		}
	}

	public void findCompleteConfiguration() {
		int retValue = JOptionPane.showConfirmDialog(this,
				"Are you sure you want to find a complete configuration\n"
						+ "based on the selections made so far? This will automatically\n"
						+ "add new instances in the configuration.", "Find complete configuration",
				JOptionPane.YES_NO_OPTION);
		if (retValue == JOptionPane.NO_OPTION) {
			return;
		}
		try {
			client.findCompleteConfiguration();
			// checkConfigurationConsequences();
		} catch (ClientException ce) {
			JOptionPane.showMessageDialog(this, ce.getMessage(), "Cannot find configuration",
					JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Imports a model to the server. This does not affect the current configuration.
	 */
	public void importModel() {
		ImportModelDialog dialog = new ImportModelDialog(this);
		dialog.setVisible(true);
		if (!dialog.getStatus()) {
			return;
		}
		try {
			client.importModel(dialog.getModelFile(), dialog.getName());
		} catch (ModelException me) {
			JOptionPane.showMessageDialog(this, me.getMessage(),
					"Model was not a valid Kumbang model", JOptionPane.ERROR_MESSAGE);
		} catch (ClientException ce) {
			JOptionPane.showMessageDialog(this, ce.getMessage(), "Error when importing model",
					JOptionPane.ERROR_MESSAGE);
		}

	}

	/**
	 * Exports the current configuration to a human-readable format.
	 */
	public void exportConfiguration() {
		if (client.isConfigurationStateReseted()) {
			return;
		}
		ExportConfigurationDialog dialog = new ExportConfigurationDialog(this);
		dialog.setVisible(true);
		if (!dialog.getStatus()) {
			return;
		}
		File exportFile = dialog.getExportFile();
		try {
			client.exportConfiguration(exportFile, dialog.getFormat());
		} catch (ExportException ee) {
			JOptionPane.showMessageDialog(this, ee.getMessage(),
					"Error when exporting configuration", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Opens a dialog for editing configuration preferences
	 */
	public void editConfigurationPreferences() {
	/*
	 * EditConfigurationPreferencesDialog dialog = new EditConfigurationPreferencesDialog(this,
	 * client.automaticallyApplyConfigurationConsequences()); dialog.show(); if
	 * (!dialog.getStatus()) { return; } boolean auto =
	 * dialog.automaticallyApplyConfigurationConsequences();
	 * client.applyConfigurationConsequences(auto);
	 */
	}

	// ------------------------------------------------------------------
	// ------ OPERATIONS THAT MODIFY CONFIGURATION WITH TASKS------------
	// ------------------------------------------------------------------

	// -------- ADD TASKS------------------------------------------------

	/**
	 * Adds a new component to the configuration, under the current component root that shows in the
	 * diagram.
	 */
	public void addComponent() {
		ComponentElement rootElement = diagramHandler.getCurrentComponentRoot();
		if (rootElement == null) {
			return;
		}
		diagramTabs.setSelectedIndex(TAB_INDEX_COMPONENT);
		addComposable(rootElement.getComponentInstance());
	}

	/**
	 * Adds a new feature to the configuration, under the current feature root that shows in the
	 * diagram.
	 */
	public void addFeature() {
		FeatureInstance inst = featureInfo.getCurrentShownFeature();
		if (inst != null) {
			diagramTabs.setSelectedIndex(TAB_INDEX_FEATURE);
			addComposable(inst);
		}
	}

	/**
	 * Adds a new composable instance to the configuration, under the given parent instance. Asks
	 * for the part that should be used for adding the instance.
	 * 
	 * @param parent parent under which one should add new composable instance
	 */
	public void addComposable(ComposableInstance parent) {
		SelectPartDialog dialog = new SelectPartDialog(this, parent);
		dialog.setVisible(true);
		if (!dialog.getStatus()) {
			return;
		}
		addComposable(dialog.getPart());
	}

	/**
	 * Adds a new composable instance to the given part instance. Opens a dialog for setting things
	 * for the new instance, such as selecting the type and selecting interfaces.
	 * 
	 * @param partInstance part
	 */
	public void addComposable(PartInstance partInstance) {
		AddComposableDialog dialog = new AddComposableDialog(this, partInstance);
		dialog.setVisible(true);

		AddComposableTask task = dialog.getTask(this.taskHandler);
		if (task == null) {
			return;
		}

		try {
			fireConfigurationModifiedEvent(client.handleTask(task));
		} catch (ConfigurationException ce) {
			JOptionPane.showMessageDialog(getContentPane(), ce.getMessage(), "Error when adding "
					+ partInstance.getComposedKind(), JOptionPane.ERROR_MESSAGE);
		} catch (ClientException cle) {
			JOptionPane.showMessageDialog(getContentPane(), cle.getMessage(), "Error when adding "
					+ partInstance.getComposedKind(), JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Adds a new binding instance between two given interface instances. Does not open a dialog.
	 * 
	 * @param instance1 interface instance number 1
	 * @param instance2 interface instance number 2
	 */
	public void addBinding(InterfaceInstance instance1, InterfaceInstance instance2) {
		try {
			AddBindingTask task = taskHandler.createAddBindingTask(instance1, instance2);
			fireConfigurationModifiedEvent(client.handleTask(task));
		} catch (ConfigurationException ce) {
			JOptionPane.showMessageDialog(getContentPane(), ce.getMessage(),
					"Error when adding binding", JOptionPane.ERROR_MESSAGE);
		} catch (ClientException cle) {
			JOptionPane.showMessageDialog(getContentPane(), cle.getMessage(),
					"Error when adding binding", JOptionPane.ERROR_MESSAGE);
		}
	}

	// -------- EDIT TASKS-----------------------------------------------

	/**
	 * Edits attributes of the given composable instance. Opens a dialog for setting attribute
	 * values.
	 * 
	 * @param instance attributed instance to be edited
	 */
	public void editAttributes(AttributedInstance instance) {
		// ComposableInstance totalInstance
		// = (ComposableInstance)configHandler.getTotalInstance(instance);
		EditAttributesDialog dialog = new EditAttributesDialog(this, instance);
		dialog.setVisible(true);
		EditAttributesTask task = dialog.getTask(this.taskHandler);
		if (task == null) {
			return;
		}
		try {
			fireConfigurationModifiedEvent(client.handleTask(task));
		} catch (ConfigurationException ce) {
			JOptionPane.showMessageDialog(getContentPane(), ce.getMessage(),
					"Error when editing attributes", JOptionPane.ERROR_MESSAGE);
		} catch (ClientException cle) {
			JOptionPane.showMessageDialog(getContentPane(), cle.getMessage(),
					"Error when editing attributes", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Edits interfaces of the given composable instance. Opens a dialog for setting interfaces on
	 * and off.
	 * 
	 * @param instance composable instance to be edited
	 */
	public void editInterfaces(ComponentInstance instance) {
		EditInterfacesDialog dialog = new EditInterfacesDialog(this, instance);
		dialog.setVisible(true);
		EditInterfacesTask task = dialog.getTask(this.taskHandler);
		if (task == null) {
			return;
		}
		try {
			fireConfigurationModifiedEvent(client.handleTask(task));
		} catch (ConfigurationException ce) {
			JOptionPane.showMessageDialog(getContentPane(), ce.getMessage(),
					"Error when editing interfaces", JOptionPane.ERROR_MESSAGE);
		} catch (ClientException cle) {
			JOptionPane.showMessageDialog(getContentPane(), cle.getMessage(),
					"Error when editing interfaces", JOptionPane.ERROR_MESSAGE);
		}
	}

	// -------- REMOVE TASKS---------------------------------------------

	/**
	 * Removes the given composable instance from the configuration. Confirms before starting actual
	 * removing.
	 * 
	 * @param instance composable instance to be removed
	 */
	public void removeComposable(ComposableInstance instance) {
		ComposableInstance parent = (ComposableInstance) instance.getParent();
		if (parent == null) {
			JOptionPane.showMessageDialog(this, "You cannot remove root of the configuration.\n"
					+ instance.toString(), "Cannot remove instance", JOptionPane.WARNING_MESSAGE);
			return;
		}
		int retValue = JOptionPane.showConfirmDialog(this, "Are you sure you want to remove this "
				+ instance.getInstanceKind() + "?\n\t\t" + instance + "\n"
				+ "This will remove all instances from the configuration that are under this "
				+ instance.getInstanceKind() + ".", "Remove " + instance.getInstanceKind(),
				JOptionPane.YES_NO_OPTION);
		if (retValue == JOptionPane.NO_OPTION) {
			return;
		}
		try {
			RemoveInstanceTask task = taskHandler.createRemoveInstanceTask(instance);
			fireConfigurationModifiedEvent(client.handleTask(task));
		} catch (ConfigurationException ce) {
			JOptionPane.showMessageDialog(getContentPane(), ce.getMessage(), "Error when removing "
					+ instance.getInstanceKind(), JOptionPane.ERROR_MESSAGE);
		} catch (ClientException cle) {
			JOptionPane.showMessageDialog(getContentPane(), cle.getMessage(),
					"Error when removing interface" + instance.getInstanceKind(),
					JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Removes the given interface instance from the configuration. Confirms before starting actual
	 * removing.
	 * 
	 * @param instance interface instance to be removed
	 */
	public void removeInterface(InterfaceInstance instance) {
		if (!instance.isOptional()) {
			JOptionPane.showMessageDialog(this,
					"You cannot remove a required interface from the configuration.\n"
							+ instance.toString(), "Cannot remove interface",
					JOptionPane.WARNING_MESSAGE);
			return;
		}
		int retValue = JOptionPane.showConfirmDialog(this,
				"Are you sure you want to remove this interface?\n\t\t" + instance
						+ "This will remove all bindings connected to this interface.",
				"Remove interface", JOptionPane.YES_NO_OPTION);
		if (retValue == JOptionPane.NO_OPTION) {
			return;
		}
		try {
			RemoveInstanceTask task = taskHandler.createRemoveInstanceTask(instance);
			fireConfigurationModifiedEvent(client.handleTask(task));
		} catch (ConfigurationException ce) {
			JOptionPane.showMessageDialog(getContentPane(), ce.getMessage(),
					"Error when removing interface", JOptionPane.ERROR_MESSAGE);
		} catch (ClientException cle) {
			JOptionPane.showMessageDialog(getContentPane(), cle.getMessage(),
					"Error when removing interface", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Removes the given binding instance from the configuration. Confirms before starting actual
	 * removing.
	 * 
	 * @param instance binding instance to be removed
	 */
	public void removeBinding(BindingInstance instance) {
		int retValue = JOptionPane.showConfirmDialog(this,
				"Are you sure you want to remove this binding?\n\t\t" + instance, "Remove binding",
				JOptionPane.YES_NO_OPTION);
		if (retValue == JOptionPane.NO_OPTION) {
			return;
		}
		try {
			RemoveInstanceTask task = taskHandler.createRemoveInstanceTask(instance);
			fireConfigurationModifiedEvent(client.handleTask(task));
		} catch (ConfigurationException ce) {
			JOptionPane.showMessageDialog(this, ce.getMessage(), "Error when removing binding",
					JOptionPane.ERROR_MESSAGE);
		} catch (ClientException cle) {
			JOptionPane.showMessageDialog(this, cle.getMessage(), "Error when removing binding",
					JOptionPane.ERROR_MESSAGE);
		}
	}

	// -------- ??????????-----------------------------------------------

	public void selectInstance(Instance instance) {
		if (instance instanceof FeatureInstance) {
			// What should this do? Maybe nothing?
		} else if (instance instanceof ComponentInstance) {
			fireInstanceSelectedEvent((ComponentInstance) instance);
		} else if (instance instanceof BindingInstance) {
			fireInstanceSelectedEvent((BindingInstance) instance);
		} else if (instance instanceof InterfaceInstance) {
			fireInstanceSelectedEvent((InterfaceInstance) instance);
		}
	}

	public void showInstanceInDiagram(Instance instance) {
		if (instance instanceof FeatureInstance) {
			featureInfo.showPanelForFeature((FeatureInstance) instance);
			diagramTabs.setSelectedIndex(TAB_INDEX_FEATURE);
		} else if (instance instanceof ComponentInstance) {
			componentDiagram.showInstance(instance);
			fireInstanceSelectedEvent((ComponentInstance) instance);
			diagramTabs.setSelectedIndex(TAB_INDEX_COMPONENT);
		} else if (instance instanceof BindingInstance) {
			componentDiagram.showInstance(instance);
			fireInstanceSelectedEvent((BindingInstance) instance);
			diagramTabs.setSelectedIndex(TAB_INDEX_COMPONENT);
		} else if (instance instanceof InterfaceInstance) {
			componentDiagram.showInstance(instance);
			fireInstanceSelectedEvent((InterfaceInstance) instance);
			diagramTabs.setSelectedIndex(TAB_INDEX_COMPONENT);
		}
	}

	public void showPotentialElements(boolean show) {
		componentDiagram.showPotential(show);
	}

	public void showNames(boolean show) {
		componentDiagram.showNames(show);
	}

	public void showTypes(boolean show) {
		componentDiagram.showTypes(show);
	}

	public void showConnections(boolean show) {
		componentDiagram.showConnections(show);
	}

	public void setLargeTreeIcons(boolean large) {
		treeHandler.setLargeIcons(large);
	}

	// ------------------------------------------------------------------
	// ------ PRIVATE METHODS -------------------------------------------
	// ------------------------------------------------------------------

	/**
	 * Initializes the components in the gui
	 */
	private void initComponents() {

		// init menubar
		menubar = new ClientMenuBar(this);
		getRootPane().setJMenuBar(menubar);

		// init trees
		treeHandler = new TreeHandler(this);
		componentTree = new ComponentTree(this);
		featureTree = new FeatureTree(this);
		treeHandler.setTrees(componentTree, featureTree);

		// init info panels
		featureInfo = new FeatureInfoPanel(this);
		diagramInfo = new DiagramInfoPanel(this);
		configurationInfo = new ConfigurationInfoPanel(this, client);

		// init diagram stuff
		diagramHandler = new DiagramHandler();
		addConfigurationModificationListener(diagramHandler);
		componentDiagram = new ComponentDiagramPanel(this);

		// init tabbed pane for trees
		treeTabs = new JTabbedPane(JTabbedPane.TOP);
		JScrollPane treeScroll = new JScrollPane(featureTree);
		treeTabs.addTab("Features", Icons.FEATURE, treeScroll);
		treeScroll = new JScrollPane(componentTree);
		treeTabs.addTab("Components", Icons.COMPONENT, treeScroll);

		// init split for diagram and diagram info
		JSplitPane diagramSplit = new JSplitPane(JSplitPane.VERTICAL_SPLIT, componentDiagram,
				diagramInfo);
		diagramSplit.setDividerLocation(0.7);
		diagramSplit.setResizeWeight(0.7);

		// init tabbed pane for feature info + component diagram
		diagramTabs = new JTabbedPane(JTabbedPane.TOP);
		diagramTabs.addTab("Features", Icons.FEATURE, featureInfo);
		diagramTabs.addTab("Components", Icons.COMPONENT, diagramSplit);

		// init tree split
		JSplitPane treeSplit = new JSplitPane(JSplitPane.VERTICAL_SPLIT, treeTabs, new JScrollPane(
				configurationInfo));
		treeSplit.setDividerLocation(0.7);
		treeSplit.setResizeWeight(0.7);

		// init main horizontal split
		JSplitPane horizontalSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, treeSplit,
				diagramTabs);
		horizontalSplit.setDividerLocation(0.3);
		horizontalSplit.setResizeWeight(0.3);

		getContentPane().add(horizontalSplit);

		// init focus traversal policy
		ClientFocusTraversalPolicy policy = new ClientFocusTraversalPolicy(this, treeTabs,
				featureTree, componentTree, diagramTabs, featureInfo, componentDiagram);
		setFocusTraversalPolicy(policy);
	}

	private void initGUIContent() {
		// boolean hasComponents = client.getComponentRoot() != null;
		// boolean hasFeatures = client.getFeatureRoot() != null;
		diagramHandler.init(client.getComponentRoot());
		treeHandler.init(client.getComponentRoot(), client.getFeatureRoot());
		featureInfo.init(client.getFeatureRoot());
		componentDiagram.init();
		diagramInfo.init();
		configurationInfo.init();
		// if (hasComponents && !hasFeatures) {
		// treeTabs.setSelectedIndex(TAB_INDEX_COMPONENT);
		// diagramTabs.setSelectedIndex(TAB_INDEX_COMPONENT);
		// } else if (!hasComponents && hasFeatures) {
		// treeTabs.setSelectedIndex(TAB_INDEX_FEATURE);
		// diagramTabs.setSelectedIndex(TAB_INDEX_FEATURE);
		// } else {
		// treeTabs.setSelectedIndex(TAB_INDEX_FEATURE);
		// diagramTabs.setSelectedIndex(TAB_INDEX_COMPONENT);
		// }
		repaint();
		featureTree.requestFocusInWindow();
		selectInstance(client.getComponentRoot());
	}

	private void resetGUIContent() {
		treeHandler.reset();
		diagramHandler.reset();
		componentDiagram.reset();
		featureInfo.reset();
		diagramInfo.reset();
		configurationInfo.reset();
		repaint();
	}

//	private void refreshGUIContent() {
//		componentDiagram.repaint();
//		diagramInfo.init();
//		configurationInfo.init();
//	}

}
