/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Connector.java
 *
 * Created on 26. helmikuuta 2004, 15:28
 */

package kumbang.configurator.client.gui.diagram;

//import java.awt.Color;
import java.awt.Graphics;

/**
 *
 * @author  vmyllarn
 */
public class ConnectorElement extends DiagramElement implements FilteredElement {
    
    public static final int MARGIN = 3;
    
    /** Foreground color of this connector element */
    //private Color foregroundColor = Color.BLACK;
    
    /** Selection color of this connector element */
    //private Color selectionColor = Color.BLUE;
    
    /** Connection points for this connector */
    private int x1;
    private int y1;
    
    private int x2;
    private int y2;
    
    /** Elements that this connector connects */
    private ConnectableElement element1;
    private ConnectableElement element2;
    
    /** Coefficients for the equation of this line, in format ax+by+c=0 */
    private int a;
    private int b;
    private int c;
    
    /** Creates a new instance of Connector */
    protected ConnectorElement(ConnectableElement elem1, ConnectableElement elem2) {
        this(elem1, elem2, 
                elem1.getConnectorLocationX(), elem1.getConnectorLocationY(), 
                elem2.getConnectorLocationX(), elem2.getConnectorLocationY());
    }
    
    /** Creates a new instance of Connector */
    protected ConnectorElement(ConnectableElement elem1, ConnectableElement elem2, int x1, int y1, int x2, int y2) {
        super("connects");
        this.element1 = elem1;
        this.element2 = elem2;
        this.x1 = x1;
        this.x2 = x2;
        this.y1 = y1;
        this.y2 = y2;
        calculateEquation();
    }
    
    public void paint(Graphics g, DiagramFilter filter) {
        if (filter != null && !filter.showsConnections()) {
            return;
        }
        if (isInAction()) {
            g.setColor(getActionColor());
        }
        else if (isSelected()) {
            g.setColor(getSelectionColor());
        }
        else {
            g.setColor(getForegroundColor());
        }
        g.drawLine(x1, y1, x2, y2);
    }
    
    public void paint(Graphics g) {
        paint(g, null);
    }
    
    /**
     * Calculate whether the given point is "inside" this connector.
     * If distance <= MARGIN, point is "inside", otherwise it is "outside".
     *
     * Equation of distance of point (x,y) from line ax + by + c = 0 is
     *    d^2 = (ax + by + c)^2 / (a^2 + b^2)
     */
    public boolean isInside(int x, int y) {
        /* if is outside of this lines scope, reject immediately */
        if ((x < x1-MARGIN && x < x2-MARGIN) || (y < y1-MARGIN && y < y2-MARGIN) ||
            (x > x1+MARGIN && x > x2+MARGIN) || (y > y1+MARGIN && y > y2+MARGIN)) {
                return false;
        }
        if (a == 0 && b == 0) {
            return false;
        }
        int distExp2 = 
            (a*x + b*y + c)*(a*x + b*y + c) / (a*a + b*b);
        /*
        System.out.println("Distance^2 of point "+x+","+y+" from line "
            +a+"x+"+b+"y+"+c+"=0 is "+distExp2);*/
        return (distExp2 <= MARGIN*MARGIN);
    }
    
    protected void setLocation(ConnectableElement element, int x, int y) {
        if (element == this.element1) {
            this.x1 = x;
            this.y1 = y;
        }
        else if (element == this.element2) {
            this.x2 = x;
            this.y2 = y;
        }
        calculateEquation();
    }
    
    /**
     * Equation of line that passes through points (x1,y1) and (x2,y2) is
     *    (x1 - x2)(y - y1) = (y1 - y2)(x - x1) 
     *    =>
     *    (y1 - y2)*x - (x1 - x2)*y + (x1 - x2)*y1 - (y1 - y2)*x1 = 0
     *    =>
     *    a = y1 - y2
     *    b = -(x1 - x2)
     *    c = (x1 - x2)*y1 - (y1 - y2)*x1
     */
    private void calculateEquation() {
        this.a = y1 - y2;
        this.b = -(x1 - x2);
        this.c = (x1 - x2)*y1 - (y1 - y2)*x1;
    }
    
    protected void removeFromDiagram() {
        element1.disconnectFromConnector(this);
        element2.disconnectFromConnector(this);
    }
    
    protected ConnectableElement getConnectedElement1() {
        return this.element1;
    }
    
    protected ConnectableElement getConnectedElement2() {
        return this.element2;
    }
    
    public boolean equals(Object o) {
        if (o instanceof ConnectorElement) {
            ConnectorElement con = (ConnectorElement)o;
            if (con.getConnectedElement1() == this.element1 &&
                con.getConnectedElement2() == this.element2) {
                    return true;
            }
        }
        return false;
    }
    
}
