/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * DiagramHandler.java
 *
 * Created on 4. maaliskuuta 2004, 9:27
 */
package kumbang.configurator.client.gui.diagram;


import kumbang.configurator.client.gui.event.ConfigurationModificationListener;
import kumbang.configurator.client.gui.event.ConfigurationModifiedEvent;

import kumbang.core.configuration.instance.BindingInstance;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.ComposableInstance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.PartInstance;
import kumbang.core.configuration.task.AddBindingTask;
import kumbang.core.configuration.task.AddComposableTask;
import kumbang.core.configuration.task.AddInterfaceTask;
import kumbang.core.configuration.task.ClientTask;
import kumbang.core.configuration.task.RemoveInstanceTask;
import kumbang.core.configuration.task.SetAttributeTask;

import java.util.HashMap;
import java.util.Iterator;

/**
 * 
 * @author vmyllarn
 */
public class DiagramHandler implements ConfigurationModificationListener {

	// private MainGUI mainGUI;

	/** The root level of the component diagram level hierarchy. */
	private ComponentDiagramLevel componentRootLevel;

	/** The current component level that is showing */
	private ComponentDiagramLevel currentComponentLevel;

	/**
	 * Contains mapping from instances to elements. Keys are obtained
	 */
	private HashMap<Instance, DiagramElement> elements;

	private HashMap<Instance, InstantiatedElement> rootElements;

	/**
	 * Creates a new instance of DiagramHandler.
	 */
	public DiagramHandler() {
		this.elements = new HashMap<Instance, DiagramElement>();
		this.rootElements = new HashMap<Instance, InstantiatedElement>();
		init();
		// mainGUI.addConfigurationModificationListener(this);
	}

	// --------------GETTERS FOR ROOTS AND LEVELS---------------------------

	/**
	 * Returns the current diagram level for component diagram.
	 * 
	 * @return ComponentDiagramLevel current level
	 */
	public ComponentDiagramLevel getCurrentComponentLevel() {
		return currentComponentLevel;
	}

	public ComponentElement getCurrentComponentRoot() {
		if (currentComponentLevel != null) {
			return (ComponentElement) currentComponentLevel.getRootElement();
		} else {
			return null;
		}
	}

	// --------------NAVIGATION BETWEEN DIAGRAM LEVELS----------------------

	/**
	 * Navigates up the diagram level hierarchy from the current component diagram level. If current
	 * diagram level is not defined, or there is no level above this level, this method returns
	 * null. Otherwise it returns the new current diagram level.
	 * 
	 * @return ComponentDiagramLevel new current level or null
	 */
	public ComponentDiagramLevel navigateComponentLevelUp() {
		ComponentDiagramLevel level;
		if (currentComponentLevel == null) {
			return null;
		}
		level = (ComponentDiagramLevel) currentComponentLevel.getParent();
		if (level == null) {
			return null;
		}
		currentComponentLevel = level;
		return level;
	}

	/**
	 * Navigates down the diagram level hierarchy to the level that has the given element as a root
	 * element. If there is no such level in the level hierarchy, this method returns null.
	 * Otherwise it returns the new current diagram level.
	 * 
	 * @return ComponentDiagramLevel new current level or null
	 */
	public ComponentDiagramLevel navigateComponentLevelDown(ComponentElement element) {
		// what is the level that this element is child in?
		DiagramLevel level = (ComponentDiagramLevel) element.getLevel();
		// look for level that has element as root element
		level = level.getChildLevelForRootElement(element);
		if (level == null) {
			return null;
		}
		currentComponentLevel = (ComponentDiagramLevel) level;
		return currentComponentLevel;
	}

	/**
	 * 
	 * @author vmyllarn
	 */
	public ComponentDiagramLevel navigateToComponent(Instance instance) {
		DiagramElement elem = getElementForInstance(instance);
		if (elem == null) {
			return null;
		}
		DiagramLevel level = elem.getLevel();
		if (level == null || !(level instanceof ComponentDiagramLevel)) {
			return null;
		}
		currentComponentLevel = (ComponentDiagramLevel) level;
		return currentComponentLevel;
	}

	// --------------INITING AND RESETING STRUCTURES-------------------------

	/**
	 * Public method to initialize component diagram using empty component root. Does also the
	 * layout for the diagram elements.
	 */
	private void init() {
		init(null);
	}

	/**
	 * Public method to initialize component diagram using the given component root. Does also the
	 * layout for the diagram elements.
	 */
	public void init(ComponentInstance rootComponent) {
		componentRootLevel = initDiagramLevel(null, rootComponent);
		currentComponentLevel = componentRootLevel;
		initLayout();
	}

	/**
	 * Resets the diagram elements from this handler.
	 */
	public void reset() {
		componentRootLevel = null;
		currentComponentLevel = null;
	}

	// --------------REACTING TO CONFIGURATION EVENTS-----------------------

	public void configurationModified(ConfigurationModifiedEvent e) {
		handleTask(e.getTask());
	}

	// ------------------------------------------------------------------
	// ------------------- PRIVATE METHODS ------------------------------
	// ------------------------------------------------------------------

	// --------------METHODS THAT HANDLE TASKS --------------------------

	private void handleTask(ClientTask task) {
		// System.out.println("DiagramHandler: handling "+task);
		if (task instanceof AddComposableTask) {
			addComposable((AddComposableTask) task);
			Iterator it = task.getSubTasks();
			while (it.hasNext()) {
				handleTask((ClientTask) it.next());
			}
		} else if (task instanceof AddInterfaceTask) {
			addInterface((AddInterfaceTask) task);
			Iterator it = task.getSubTasks();
			while (it.hasNext()) {
				handleTask((ClientTask) it.next());
			}
		} else if (task instanceof AddBindingTask) {
			addBinding((AddBindingTask) task);
		} else if (task instanceof SetAttributeTask) {
			// setAttribute((SetAttributeTask)task);
		} else if (task instanceof RemoveInstanceTask) {
			Iterator it = task.getSubTasks();
			while (it.hasNext()) {
				handleTask((ClientTask) it.next());
			}
			removeInstance((RemoveInstanceTask) task);
		} else {
			Iterator it = task.getSubTasks();
			while (it.hasNext()) {
				handleTask((ClientTask) it.next());
			}
		}
	}

	private void addComposable(AddComposableTask task) {
		ComposableInstance instance = (ComposableInstance) task.getAddedInstance();
		if (!(instance instanceof ComponentInstance)) {
			return;
		}
		if (getElementForInstance(instance) != null) {
			return;
		}
		PartInstance partInstance = instance.getParentPart();
		PartElement partElement = (PartElement) getElementForInstance(partInstance);
		ComponentDiagramLevel level = (ComponentDiagramLevel) partElement.getLevel();

		ComponentElement element = addNewElement((ComponentInstance) instance, level);
		level.updateLayoutForAddedElement(element, partElement);
		partElement.addInstanceElement(element);
		partElement.update();

		// initialize the whole diagram level below this element
		initLayout(initDiagramLevel(level, (ComponentInstance) instance));
	}

	private void addInterface(AddInterfaceTask task) {
		InterfaceInstance iInstance = (InterfaceInstance) task.getAddedInstance();
//		InterfaceInstance iInstance = task.getAddedInterface();
		if (getElementForInstance(iInstance) != null) {
			return;
		}
		ComponentInstance cInstance = iInstance.getComponent();
		ComponentElement cElement = (ComponentElement) getElementForInstance(cInstance);
		ComponentDiagramLevel level = (ComponentDiagramLevel) cElement.getLevel();
		InterfaceElement iElement = addNewElement(iInstance, level);
		cElement.addInterface(iElement);

		// check if we should add this interface to the root component element also
		cElement = (ComponentElement) getRootElementForInstance(cInstance);
		if (cElement != null) {
			level = (ComponentDiagramLevel) cElement.getLevel();
			if (!level.containsChildElementForInstance(iInstance)) {
				iElement = addNewElement(iInstance, level);
				cElement.addInterface(iElement);
				setRootElementForInstance(iInstance, iElement);
			}
		}
	}

	private void addBinding(AddBindingTask task) {
		BindingInstance binding = task.getAddedBinding();
		if (getElementForInstance(binding) != null) {
			return;
		}

		InterfaceInstance instance1 = binding.getSourceInterface();
		InterfaceInstance instance2 = binding.getTargetInterface();
		InterfaceElement iElement1 = (InterfaceElement) getElementForInstance(instance1);
		InterfaceElement iElement2 = (InterfaceElement) getElementForInstance(instance2);

		// check that interface elements are at the same level
		// if we bind interfaces that are contained in each other, getElementForInstance()
		// produces the wrong element (element that is one level too high)
		DiagramLevel level1 = iElement1.getLevel();
		DiagramLevel level2 = iElement2.getLevel();
		if (level1.getParent() == level2) {
			// level2 is parent of level1 -> element 2 is too high
			iElement2 = (InterfaceElement) level1.getChildElementForInstance(instance2);
			level2 = level1;
		} else if (level2.getParent() == level1) {
			// level1 is parent of level2 -> element 1 is too high
			iElement1 = (InterfaceElement) level2.getChildElementForInstance(instance1);
			level1 = level2;
		}
		addNewElement(binding, iElement1, iElement2);
	}

	private void removeInstance(RemoveInstanceTask task) {
		Instance instance = task.getInstance();
		// System.out.println(getClass().getName() + ": removing instance " + instance);
		if (instance instanceof ComposableInstance) {
			removeComposable((ComposableInstance) instance);
		} else if (instance instanceof InterfaceInstance) {
			removeInterface((InterfaceInstance) instance);
		} else if (instance instanceof BindingInstance) {
			removeBinding((BindingInstance) instance);
		}
	}

	// --------------METHODS FOR REMOVING ELEMENTS-----------------------

	private void removeComposable(ComposableInstance instance) {
		ComposableElement element = (ComposableElement) getElementForInstance(instance);
		if (element == null) {
			// nothing to be removed
			return;
		}
		ComposableElement rootElement = (ComposableElement) getRootElementForInstance(instance);
		DiagramLevel level = element.getLevel();

		// ComposableInstance parent = instance.getParent();
		PartInstance partInstance = instance.getParentPart();
		PartElement partElement = (PartElement) getElementForInstance(partInstance);

		partElement.update();
		partElement.removeInstanceElement(element);
		level.updateLayoutForRemovedElement(element, partElement);

		removeDiagramLevel(level, rootElement);
		level.removeChildElement(element);
		removeElementForInstance(instance);
	}

	private void removeBinding(BindingInstance bindInstance) {
		BindingElement bindElement = (BindingElement) getElementForInstance(bindInstance);
		if (bindElement == null) {
			// nothing to be removed
			return;
		}
		DiagramLevel level = bindElement.getLevel();
		level.removeChildElement(bindElement);
		// bindElement.removeFromDiagram();
		removeElementForInstance(bindInstance);
	}

	private void removeInterface(InterfaceInstance iInstance) {
		ComponentInstance cInstance = iInstance.getComponent();

		// first check if this interface is in the root of the diagram
		InterfaceElement rootElement = (InterfaceElement) getRootElementForInstance(iInstance);
		if (rootElement != null) {
			ComponentElement cElement = (ComponentElement) getRootElementForInstance(cInstance);
			DiagramLevel level = rootElement.getLevel();
			level.removeChildElement(rootElement);
			// iElement.removeFromDiagram();
			cElement.removeInterface(rootElement);
			removeRootElementForInstance(iInstance);
		}

		// then check if this interface is a child element in the diagram
		InterfaceElement iElement = (InterfaceElement) getElementForInstance(iInstance);
		if (iElement != null) {
			ComponentElement cElement = (ComponentElement) getElementForInstance(cInstance);
			DiagramLevel level = iElement.getLevel();
			level.removeChildElement(iElement);
			// iElement.removeFromDiagram();
			cElement.removeInterface(iElement);
			removeElementForInstance(iInstance);
		}

	}

	// -------------METHODS FOR LAYER INITIALISATION---------------------

	// private DiagramLevel initDiagramLevel(DiagramLevel parentLevel,
	// ComposableInstance rootInstance) {
	// if (parentLevel instanceof ComponentDiagramLevel &&
	// rootInstance instanceof ComponentInstance) {
	// return initComponentDiagramLevel((ComponentDiagramLevel)parentLevel,
	// (ComponentInstance)rootInstance);
	// }
	// else if (parentLevel instanceof FeatureDiagramLevel &&
	// rootInstance instanceof FeatureInstance) {
	// return initFeatureDiagramLevel((FeatureDiagramLevel)parentLevel,
	// (FeatureInstance)rootInstance);
	// }
	// else {
	// return null;
	// }
	// }

	/**
	 * Private method to create a new component diagram level that has the given component instance
	 * as a root instance and is unser parent diagram level.
	 * 
	 * @param parentLevel diagram level above the new diagram level, or null if this is to be the
	 *            first diagram level
	 * @param rootInstance component instance that is the root in the new diagram level
	 */
	private ComponentDiagramLevel initDiagramLevel(ComponentDiagramLevel parentLevel,
			ComponentInstance rootInstance) {
		if (rootInstance == null) {
			return null;
		}

		// create new diagram level and corresponding root element to it
		ComponentElement rootElement = new ComponentElement(rootInstance, true);
		ComponentDiagramLevel diagramLevel = new ComponentDiagramLevel(rootElement, parentLevel);
		setRootElementForInstance(rootInstance, rootElement);

		// inform the parent level, if it exists
		if (parentLevel != null) {
			parentLevel.addChildLevel(diagramLevel);
		}

		// iterate all interfaces of the root component
		Iterator intfs = rootInstance.getInterfaces();
		while (intfs.hasNext()) {
			InterfaceInstance intf = (InterfaceInstance) intfs.next();
			InterfaceElement iElement = addNewElement(intf, diagramLevel);
			rootElement.addInterface(iElement);
			setRootElementForInstance(intf, iElement);
		}

		// iterate all parts and create corresponding elements
		Iterator parts = rootInstance.getParts();
		while (parts.hasNext()) {
			PartInstance part = (PartInstance) parts.next();
			Iterator insts = part.getInstances();
			while (insts.hasNext()) {
				ComponentInstance comp = (ComponentInstance) insts.next();
				initComponentInComponentLevel(comp, diagramLevel);
				initDiagramLevel(diagramLevel, comp);
			}
			addNewPartElement(part, diagramLevel);
		}

		return diagramLevel;
	}

	private void initComponentInComponentLevel(ComponentInstance comp,
			ComponentDiagramLevel diagramLevel) {
		ComponentElement cElement = addNewElement(comp, diagramLevel);
		Iterator interfs = comp.getInterfaces();
		while (interfs.hasNext()) {
			InterfaceInstance intf = (InterfaceInstance) interfs.next();
			InterfaceElement iElement = addNewElement(intf, diagramLevel);
			cElement.addInterface(iElement);
			Iterator binds = intf.getBindings();
			while (binds.hasNext()) {
				BindingInstance bind = (BindingInstance) binds.next();
				// if there is no element for this binding, create one
				if (getElementForInstance(bind) == null) {
					InterfaceInstance instance1 = bind.getSourceInterface();
					InterfaceInstance instance2 = bind.getTargetInterface();
					InterfaceElement iElement1 = (InterfaceElement) diagramLevel
							.getChildElementForInstance(instance1);
					InterfaceElement iElement2 = (InterfaceElement) diagramLevel
							.getChildElementForInstance(instance2);
					// check that both interface elements have been created
					if (iElement1 != null && iElement2 != null) {
						addNewElement(bind, iElement1, iElement2);
					}
				}
			}
		}
	}

	private void initLayout() {
		if (componentRootLevel != null) {
			initLayout(componentRootLevel);
		}
	}

	private void initLayout(DiagramLevel level) {
		level.initLayout();
		Iterator children = level.getChildLevels();
		while (children.hasNext()) {
			DiagramLevel child = (DiagramLevel) children.next();
			initLayout(child);
		}
	}

	private void removeDiagramLevel(DiagramLevel parentLevel, ComposableElement rootElement) {
		DiagramLevel level = rootElement.getLevel();
		Instance rootInstance = rootElement.getInstance();
		removeRootElementForInstance(rootInstance);
		if (rootInstance instanceof ComponentInstance) {
			Iterator it = ((ComponentInstance) rootInstance).getInterfaces();
			while (it.hasNext()) {
				removeRootElementForInstance((Instance) it.next());
			}
		}
		level.removeAllChildElements();
		parentLevel.removeChildLevel(level);
	}

	// --------------METHODS FOR ADDING ELEMENTS-------------------------

	private ComponentElement addNewElement(ComponentInstance instance, ComponentDiagramLevel level) {
		ComponentElement element = new ComponentElement(instance);
		setElementForInstance(instance, element);
		level.addChildElement(element);
		return element;
	}

	private InterfaceElement addNewElement(InterfaceInstance instance, ComponentDiagramLevel level) {
		ComponentElement cElement = level.getRootComponentElement();
		boolean isBorderInterface = cElement != null
				&& cElement.getInstance().equals(instance.getParent());
		InterfaceElement element = new InterfaceElement(instance, isBorderInterface);
		if (!isBorderInterface) {
			setElementForInstance(instance, element);
		}
		level.addChildElement(element);
		return element;
	}

	private BindingElement addNewElement(BindingInstance binding, InterfaceElement iElement1,
			InterfaceElement iElement2) {
		DiagramLevel level = iElement1.getLevel();
		// create binding and add it to the level
		BindingElement con = new BindingElement(iElement1, iElement2, binding);
		setElementForInstance(binding, con);
		iElement1.connectWithConnector(con);
		iElement2.connectWithConnector(con);
		level.addChildElement(con);

		return con;
	}

	private PartElement addNewPartElement(PartInstance instance, ComponentDiagramLevel level) {
		PartElement partElement = new PartElement(instance);
		setElementForInstance(instance, partElement);
		level.addChildElement(partElement);
		return partElement;
	}

	// --------------OTHER HELPER METHODS-------------------------------

	/* usual diagram elements */

	private DiagramElement getElementForInstance(Instance instance) {
		return elements.get(instance);
	}

	private void setElementForInstance(Instance instance, DiagramElement element) {
		elements.put(instance, element);
	}

	private void removeElementForInstance(Instance instance) {
		elements.remove(instance);
	}

	/* root elements (includes also interfaces) */

	private InstantiatedElement getRootElementForInstance(Instance instance) {
		return rootElements.get(instance);
	}

	private void setRootElementForInstance(Instance instance, InstantiatedElement element) {
		rootElements.put(instance, element);
	}

	private void removeRootElementForInstance(Instance instance) {
		rootElements.remove(instance);
	}

}
