/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * InterfaceInstance.java
 *
 * Created on 24. helmikuuta 2004, 15:52
 */

package kumbang.configurator.client.gui.diagram;

import kumbang.util.gui.Colors;

import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.Instance;

import java.awt.Graphics;

import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import java.util.LinkedList;

/**
 *
 * @author  vmyllarn
 */
public class InterfaceElement extends BoxedElement 
        implements ConnectableElement, InstantiatedElement, FilteredElement {
    
    public static final int SIZE = 16;
    
    public static final int ORIENTATION_EAST = 1;
    public static final int ORIENTATION_WEST = 2;
    
    public static final int[] X_ARROW_LEFT = {-4, 5, -4};
    public static final int[] Y_ARROW_LEFT = {-5, 0, 5};
    
    public static final int[] X_ARROW_RIGHT = {-5, 4, 4};
    public static final int[] Y_ARROW_RIGHT = {0, -5, 5};

    //private ComponentElement parentElement;

    /** The interface instance that this element represents */
    private InterfaceInstance instance;
    
    /**
     * Tells where this interface is located in the component.
     * Possible values: ORIENTATION_EAST and ORIENTATION_WEST.
     */
    private int orientation;
    
    /** 
     * Tells whether this interface is required or provided.
     * Value TRUE means this is provided and value FALSE means
     * this is required.
     */
    private boolean provided;
    
    /**
     * Tells whether this interface is on the border of a component.
     * This means that it is an interface in a component that
     * is the root component in the current diagram level.
     */
    private boolean isBorderInterface;
    
    /** A list of all connectors this interface has */
    private List<ConnectorElement> connectors;
    
    /** 
     * Creates a new instance of InterfaceInstance
     * @param instance interface instance this element represents 
     */
    public InterfaceElement(InterfaceInstance instance,
                            boolean isBorderInterface) {
        super(instance.getName(), SIZE, SIZE, Colors.INTERFACE);
        this.instance = instance;
        this.provided = instance.isProvided();
        if (this.provided) {
            this.orientation = ORIENTATION_WEST;
        } else {
            this.orientation = ORIENTATION_EAST;
        }
        this.isBorderInterface = isBorderInterface;
        this.connectors = new ArrayList<ConnectorElement>();
    }
    
    /** 
     * Creates a new instance of InterfaceInstance
     * @param instance interface instance this element represents 
     */
    public InterfaceElement(InterfaceInstance instance) {
        this(instance, false);
    }
    
    /**
     * Returns the instance this element represents
     * @return Instance 
     */
    public Instance getInstance() {
        return this.instance;
    }
    
    /**
     * Returns the interface instance this element represents
     * @return InterfaceInstance 
     */
    public InterfaceInstance getInterfaceInstance() {
        return this.instance;
    }
    
    public Iterator<ConnectorElement> getConnectors() {
        return this.connectors.iterator();
    }
    
    /**
     * Connects a connector from this interface.
     * Simply adds the connector to this interface.
     */
    public void connectWithConnector(ConnectorElement con) {
        connectors.add(con);
    }
    
    /**
     * Disconnect a connector from this interface.
     * This is called by connector's removeFromDiagram() method.
     * Simply removes the connector from this interface.
     */
    public void disconnectFromConnector(ConnectorElement con) {
        connectors.remove(con);
    }
    
    public int getConnectorLocationX() {
        if ((orientation == ORIENTATION_WEST && !isBorderInterface)
                || (orientation == ORIENTATION_EAST && isBorderInterface)) {
            return super.getLocationX() - super.getWidth() / 2;
        } else {
            return super.getLocationX() + super.getWidth() / 2;
        }
    }
    
    public int getConnectorLocationY() {
        return super.getLocationY();
    }
    
    public int getOrientation() {
        return this.orientation;
    }
    
    public void paint(Graphics g, DiagramFilter filter) {
        boolean showsName = (filter != null ? filter.showsNames() : true);
        boolean showsType = (filter != null ? filter.showsTypes() : true);
        super.paint(g);
        g.translate(getLocationX(), getLocationY());
        g.drawString(getDescription(showsName, showsType),
                     -super.getWidth() / 2, -super.getHeight() / 2 - 1);
        if ((orientation == ORIENTATION_EAST && !provided)
                || (orientation == ORIENTATION_WEST && provided)) {
            g.fillPolygon(X_ARROW_LEFT, Y_ARROW_LEFT, 3);
        } else {
            g.fillPolygon(X_ARROW_RIGHT, Y_ARROW_RIGHT, 3);
        }
        g.translate(-getLocationX(), -getLocationY());
    }
    
    public void paint(Graphics g) {
        paint(g, null);
    }
    
    
    protected void setHeight(int height) {
        super.setHeight(SIZE);
    }
    
    protected void setWidth(int width) {
        super.setWidth(SIZE);
    }
    
    protected void setLocation(int centerX, int centerY) {
        super.setLocation(centerX, centerY);
        /* to move an interface, we have to move its connectors too */
        for (ConnectorElement con : connectors) {
            con.setLocation(this, getConnectorLocationX(),
                            getConnectorLocationY());
        }
    }    
    
    protected void removeFromDiagram() {
        List<ConnectorElement> removedConnectors = new LinkedList<ConnectorElement>();
        for (ConnectorElement con : connectors) {
            removedConnectors.add(con);
        }
        for (ConnectorElement con : removedConnectors) {
            con.removeFromDiagram();
        }
    }
    
    public boolean equals(Object o) {
        if (o instanceof InterfaceElement) {
            if (instance == ((InterfaceElement) o).getInstance()) {
                return true;
            }
        }
        return false;
    }

    public String toString() {
        return instance.getName() + ":" + instance.getType().getName();
    }

    private String getDescription(boolean name, boolean type) {
        String ret = "";
        if (!name && !type) {
            return ret;
        }
        if (name) {
            ret += instance.getDisplayName();
        }
        ret += ":";
        if (type) {
            ret += instance.getType().getDisplayName();
        }
        return ret;
    }
    
}
