/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ComponentTree.java
 *
 * Created on 24. helmikuuta 2004, 10:28
 */

package kumbang.configurator.client.gui.tree;

import kumbang.util.gui.Colors;
import kumbang.util.gui.Icons;

import kumbang.configurator.client.gui.core.MainGUI;

import kumbang.core.configuration.instance.AttributeInstance;
import kumbang.core.configuration.instance.AttributedInstance;
import kumbang.core.configuration.instance.BindingInstance;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.PartInstance;

import java.awt.Component;
import java.awt.Rectangle;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTree;

import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;

import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 *
 * @author  vmyllarn
 */
public class ComponentTree extends JTree implements TreeSelectionListener, 
                                                    MouseListener, 
                                                    ActionListener, 
                                                    KeyListener {
    
    private ConfigurationNode rootNode;
    
    private MainGUI mainGUI;
    
    private DefaultTreeModel treeModel;
    
    private HashMap nodes;
    
    private Instance popupInstance;
    
    private JPopupMenu componentPopup;
    private JMenuItem editCompAttributesItem;
    private JMenuItem editCompInterfacesItem;
    private JMenuItem addComponentItem;
    private JMenuItem removeComponentItem;
    private JMenuItem showComponentItem;
    
    private JPopupMenu partPopup;
    private JMenuItem addComponentToPartItem;
    
    private JPopupMenu attributePopup;
    private JMenuItem editAttributeItem;
    
    private JPopupMenu interfacePopup;
    private JMenuItem editInterfaceItem;
    private JMenuItem removeInterfaceItem;
    private JMenuItem showInterfaceItem;
    
    private JPopupMenu bindingPopup;
    private JMenuItem removeBindingItem;
    private JMenuItem showBindingItem;
    
    private boolean largeIcons;
    
    /** 
     * Creates a new instance of ComponentTree
     * @param mainGUI main GUI 
     */
    public ComponentTree(MainGUI mainGUI) {
        super();
        this.rootNode = new ConfigurationNode(ConfigurationNode.COMPONENT_ROOT);
        //this.treeHandler = treeHandler;
        this.mainGUI = mainGUI;
        this.nodes = new HashMap();

        initPopups();

        this.treeModel = new DefaultTreeModel(this.rootNode);
        super.setModel(this.treeModel);

        setCellRenderer(new ComponentTreeCellRenderer());
        setRootVisible(true);
        setShowsRootHandles(true);
        getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        setSelectionRow(0);

        addTreeSelectionListener(this);
        addMouseListener(this);
        addKeyListener(this);

        setBackground(Colors.COMPONENT_BACKGROUND);

        this.largeIcons = false;
    }
    
    
    //------------- PROTECTED METHODS TO TREE HANDLER ETC --------------------
    
    
    protected void setLargeIcons(boolean large) {
        if (large != this.largeIcons) {
            this.largeIcons = large;
            this.treeModel.reload();
        }
    }
    
    /**
     * Initializes the tree. Fetches component configuration root from
     * the tree handler, and creates nodes according to it.
     */
    protected void init(ComponentInstance rootInstance) {
        if (rootInstance != null) {
            createNodesForInstance(rootInstance, rootNode);
        }
        scrollRowToVisible(0);
        setSelectionRow(0);
    }
    
    /**
     * Resets the tree and clears all children from it.
     */
    protected void reset() {
        nodes.clear();
        this.rootNode.removeAllChildren();
        this.treeModel.reload();
    }
    
    /**
     * Adds a new node into this tree. The new node can represent a component,
     * an interface, or a binding. Creates recusrsively other nodes below
     * the new node.
     * @param instance new instance to be added
     */
    protected void addInstance(Instance instance) {
        if (containsNodeForInstance(instance)) {
            // this instance already has a node in the tree
            return;
        }
        if (instance instanceof ComponentInstance) {
            ComponentInstance cInstance = (ComponentInstance) instance;
            PartInstance part = cInstance.getParentPart();
            if (part == null) {
                return;
            }
            ConfigurationNode partNode = getNodeForInstance(part);
            if (partNode == null) {
                return;
            }
            createNodesForInstance(cInstance, partNode);
        } else if (instance instanceof InterfaceInstance) {
            InterfaceInstance iInstance = (InterfaceInstance) instance;
            ComponentInstance parent = iInstance.getComponent();
            ConfigurationNode parentNode = getNodeForInstance(parent);
            if (parentNode == null) {
                return;
            }
            createNodesForInstance(iInstance, parentNode);
        } else if (instance instanceof BindingInstance) {
            BindingInstance binding = (BindingInstance) instance;
            InterfaceInstance source = binding.getSourceInterface();
            InterfaceInstance target = binding.getTargetInterface();
            ConfigurationNode sourceNode = getNodeForInstance(source);
            ConfigurationNode targetNode = getNodeForInstance(target);
            if (sourceNode == null || targetNode == null) {
                return;
            }
            createNodesForInstance(binding, source, sourceNode);
            createNodesForInstance(binding, target, targetNode);
        }
    }
    
    /**
     * Removes the node corresponding to the given instance from this tree, and
     * all nodes below that node. If the given instance is binding instance,
     * removes all nodes corresponding to that instance from the tree.
     * 
     * @pararm instance instance to be removed
     */ 
    protected void removeInstance(Instance instance) {
        if (instance == null) {
            return;
        } else if (instance instanceof BindingInstance) {
            Iterator iter = getNodesForBinding((BindingInstance) instance);
            while (iter.hasNext()) {
                removeNodes((BindingNode) iter.next());
            }
        } else {
            removeNodes(getNodeForInstance(instance));
        }
    }
    
    /**
     * Selects the given instance from this tree.
     * 
     * @param instance
     *            instance to be selected
     */
    protected void selectInstance(Instance instance) {
        if (instance instanceof BindingInstance) {
            //            // EI TOIMI -> StackOverFlowError
            //            Iterator iter = getNodesForBinding((BindingInstance)instance);
            //            while (iter.hasNext()) {
            //                selectedNode = (ConfigurationNode)iter.next();
            //                System.out.println("Selecting node "+selectedNode);
            //                TreePath path = new TreePath(selectedNode.getPath());
            //                if (path != null) {
            //                    setSelectionPath(path);
            //                    scrollPathToVisible(path);
            //                }
            //            }
        } else {
            ConfigurationNode node = getNodeForInstance(instance);
            if (node != null) {
                TreePath path = new TreePath(node.getPath());
                if (path != null) {
                    setSelectionPath(path);
                    scrollPathToVisible(path);
                }
            }
        }
    }
    
    /**
     * Updates the given instance in this tree.
     * @param instance instance to be updated
     */
    protected void updateInstance(Instance instance) {
        if (instance == null) {
            return;
        } else if (instance instanceof BindingInstance) {
            Iterator iter = getNodesForBinding((BindingInstance) instance);
            while (iter.hasNext()) {
                ConfigurationNode node = (ConfigurationNode) iter.next();
                if (node != null) {
                    treeModel.reload(node);
                }
            }
        } else {
            ConfigurationNode node = getNodeForInstance(instance);
            if (node != null) {
                treeModel.reload(node);
            }
        }
    }
    
    //------- PRIVATE METHODS FOR CREATING AND REMOVING NODES -----------------
    
    /**
     * Creates a component node for the given component instance. Proceeds 
     * recursively to create nodes for attributes, interfaces and parts under
     * this component instance.
     * @param instance component instance
     * @param parentNode parent node of this new node
     */
    private void createNodesForInstance(ComponentInstance instance,
                                        ConfigurationNode parentNode) {
        ComponentNode cNode = new ComponentNode(instance);

        treeModel.insertNodeInto(cNode, parentNode, parentNode.getChildCount());
        parentNode.add(cNode);
        setNodeForInstance(cNode, instance);
        
        TreePath path = new TreePath(cNode.getPath());
        scrollPathToVisible(path);
        //setSelectionPath(path);

        /* add nodes for attributes */
        Iterator attrs = instance.getAttributes();
        while (attrs.hasNext()) {
            createNodesForInstance((AttributeInstance) attrs.next(), cNode);
        }

        /* add nodes for interfaces */
        Iterator interfaces = instance.getInterfaces();
        while (interfaces.hasNext()) {
            InterfaceInstance interf = (InterfaceInstance) interfaces.next();
            createNodesForInstance(interf, cNode);
        }

        /* add nodes for parts */
        Iterator parts = instance.getParts();
        while (parts.hasNext()) {
            createNodesForInstance((PartInstance) parts.next(), cNode);
        }
    }
    
    
    /**
     * Creates an interface node for the given interface instance. Proceeds 
     * recursively to create nodes for bindings from this interface instance.
     * @param instance interface instance
     * @param parentNode parent node of this new node
     */
    private void createNodesForInstance(InterfaceInstance instance,
                                        ConfigurationNode parentNode) {
        InterfaceNode intNode = new InterfaceNode(instance);
        treeModel.insertNodeInto(intNode, parentNode, 
                                 parentNode.getChildCount());
        scrollPathToVisible(new TreePath(intNode.getPath()));
        setNodeForInstance(intNode, instance);

        Iterator iter = instance.getBindings();
        while (iter.hasNext()) {
            createNodesForInstance((BindingInstance) iter.next(), instance,
                                   intNode);
        }
    }
    
    /**
     * Creates a binding node for the given binding instance, under the
     * given interface instance. Binding nodes are different from other nodes,
     * since one instance is created as two nodes: one for each interface node.
     * @param binding binding instance
     * @param interf interface instance
     * @param parentNode parent node of this new binding node
     */
    private void createNodesForInstance(BindingInstance binding,
                                        InterfaceInstance interf, 
                                        ConfigurationNode parentNode) {
        BindingNode bNode = new BindingNode(binding, interf);
        treeModel.insertNodeInto(bNode, parentNode, parentNode.getChildCount());
        //scrollPathToVisible(new TreePath(bNode.getPath()));
        setNodeForInstance(bNode, binding);
    }
    
    /**
     * Creates an attribute node for the given attribute instance. 
     * @param instance attribute instance
     * @param parentNode parent node of this new node
     */
    private void createNodesForInstance(AttributeInstance instance, 
                                        ConfigurationNode parentNode) {
        AttributeNode aNode = new AttributeNode(instance);
        
        treeModel.insertNodeInto(aNode, parentNode, parentNode.getChildCount());
        scrollPathToVisible(new TreePath(aNode.getPath()));
        parentNode.add(aNode);
        setNodeForInstance(aNode, instance);
    }
    
    /**
     * Creates a part node for the given part instance. Proceeds recursively
     * to create nodes for component instances under this part. 
     * @param instance part instance
     * @param parentNode parent node of this new node
     */
    private void createNodesForInstance(PartInstance instance,
                                        ConfigurationNode parentNode) {
        PartNode pNode = new PartNode(instance);

        treeModel.insertNodeInto(pNode, parentNode, parentNode.getChildCount());
        scrollPathToVisible(new TreePath(pNode.getPath()));
        parentNode.add(pNode);
        setNodeForInstance(pNode, instance);

        Iterator iter = instance.getInstances();
        while (iter.hasNext()) {
            createNodesForInstance((ComponentInstance) iter.next(), pNode);
        }
    }
    
    /**
     * Removes nodes from the tree. Removes this node, all its subchildren, and
     * removes the nodes from the hashmap.
     * 
     * @param node
     *            node to be removed
     */
    private void removeNodes(ConfigurationNode node) {
        if (node == null) {
            return;
        }
        Enumeration children = node.children();
        while (children.hasMoreElements()) {
            removeNodes((ConfigurationNode) children.nextElement());
        }
        if (node instanceof InstanceNode) {
            Instance instance = ((InstanceNode) node).getInstance();
            unsetNodeForInstance(instance);
        }
        treeModel.removeNodeFromParent(node);
    }
    

    
    //------------- PRIVATE METHODS FOR MAPPING NODES ------------------------
    
    public boolean containsNodeForInstance(Instance instance) {
        return nodes.containsKey(instance);
    }
    
    /**
     * Gets the node that corresponds to the given instance, or null
     * if no such instance exists in the tree. NOTE: bindings cannot be
     * fetched like this, since they have two nodes for one instance
     * @param instance instance
     * @return ConfigurationNode node corresponding to instance
     */
    private ConfigurationNode getNodeForInstance(Instance instance) {
        if (instance != null) {
            return (ConfigurationNode) nodes.get(instance);
        }
        return null;
    }
    
    /**
     * Gets the nodes that correspond to the given binding instance.
     * Returns an iterator of binding nodes, or an empty iterator if no such
     * binding instance exists in the tree.
     * @param binding binding instance
     * @return Iterator an iterator containing the binding nodes
     */
    private Iterator getNodesForBinding(BindingInstance binding) {
        List list = (List) nodes.get(binding);
        if (list != null) {
            return list.iterator();
        } else {
            return new LinkedList().iterator();
        }
    }
    
    /**
     * Sets a mapping between the given node and instance. If this is a binding
     * instance, takes care that there might be several nodes mapped to one
     * binding instance.
     * @param node
     * @param instance
     */
    private void setNodeForInstance(ConfigurationNode node, Instance instance) {
        if (node == null || instance == null) {
            return;
        }
        if (instance instanceof BindingInstance) {
            if (nodes.containsKey(instance)) {
                List list = (List) nodes.get(instance);
                list.add(node);
            } else {
                List list = new LinkedList();
                list.add(node);
                nodes.put(instance, list);
            }
        } else {
            nodes.put(instance, node);
        }
    }
      
    /**
     * Removes a mapping between the given instance and a node in the tree.
     * 
     * @param instance
     */
    private void unsetNodeForInstance(Instance instance) {
        if (instance != null) {
            nodes.remove(instance);
        }
    }

    
    //------------- PRIVATE METHODS FOR POPUPS -------------------------------
    
    private void initPopups() {
        /* init popup for component nodes */
        componentPopup = new JPopupMenu("Component");
        editCompAttributesItem 
            = new JMenuItem("Edit attributes of this component", 
                            Icons.ATTRIBUTE);
        editCompInterfacesItem 
            = new JMenuItem("Edit interfaces of this component", 
                            Icons.NEW_INTERFACE);
        addComponentItem 
            = new JMenuItem("Add new component under this component", 
                            Icons.NEW_COMPONENT);
        removeComponentItem 
            = new JMenuItem("Remove this component", Icons.DELETE);
        showComponentItem 
            = new JMenuItem("Show component in the diagram", Icons.DIAGRAM);
        editCompAttributesItem.addActionListener(this);
        editCompInterfacesItem.addActionListener(this);
        addComponentItem.addActionListener(this);
        removeComponentItem.addActionListener(this);
        showComponentItem.addActionListener(this);
        componentPopup.add(editCompAttributesItem);
        componentPopup.add(editCompInterfacesItem);
        componentPopup.add(addComponentItem);
        componentPopup.add(removeComponentItem);
        componentPopup.add(showComponentItem);
        
        /* init popup for part nodes */
        partPopup = new JPopupMenu("Part");
        addComponentToPartItem 
            = new JMenuItem("Add new component under this part", 
                            Icons.NEW_COMPONENT);
        addComponentToPartItem.addActionListener(this);
        partPopup.add(addComponentToPartItem);
        
        /* init popup for interface nodes */
        interfacePopup = new JPopupMenu("Interface");
        editInterfaceItem 
            = new JMenuItem("Edit interfaces", Icons.NEW_INTERFACE);
        removeInterfaceItem 
            = new JMenuItem("Remove this interface", Icons.DELETE);
        showInterfaceItem 
            = new JMenuItem("Show nterface in the diagram", Icons.DIAGRAM);
        editInterfaceItem.addActionListener(this);
        removeInterfaceItem.addActionListener(this);
        showInterfaceItem.addActionListener(this);
        interfacePopup.add(editInterfaceItem);
        interfacePopup.add(removeInterfaceItem);
        interfacePopup.add(showInterfaceItem);
        
        /* init popup for bindings */
        bindingPopup = new JPopupMenu("Binding");
        removeBindingItem = new JMenuItem("Remove this binding", Icons.DELETE);
        showBindingItem 
            = new JMenuItem("Show binding in the diagram", Icons.DIAGRAM);
        removeBindingItem.addActionListener(this);
        showBindingItem.addActionListener(this);
        bindingPopup.add(removeBindingItem);
        bindingPopup.add(showBindingItem);
        
        /* init popup for attribute nodes */
        attributePopup = new JPopupMenu("Attribute");
        editAttributeItem = new JMenuItem("Edit attributes", Icons.ATTRIBUTE);
        editAttributeItem.addActionListener(this);
        attributePopup.add(editAttributeItem);
    }
    
    
    //------------- METHODS IMPLEMENTING INTERFACES --------------------------
    
    /**
     * Reacts to TreeSelectionEvents by telling mainGUI to select the same
     * instance.
     * @param e tree selection event
     */
    public void valueChanged(TreeSelectionEvent e) {
        ConfigurationNode node 
            = (ConfigurationNode) getLastSelectedPathComponent();
        if (node != null && node instanceof InstanceNode) {
            mainGUI.selectInstance(((InstanceNode) node).getInstance());
        }
    }    
    
    public void actionPerformed(ActionEvent e) {
        Object source = e.getSource();
        if (source == editCompAttributesItem) {
            if (popupInstance != null
                    && popupInstance instanceof ComponentInstance) {
                mainGUI.editAttributes((ComponentInstance) popupInstance);
            }
        } else if (source == editCompInterfacesItem) {
            if (popupInstance != null
                    && popupInstance instanceof ComponentInstance) {
                mainGUI.editInterfaces((ComponentInstance) popupInstance);
            }
        } else if (source == addComponentItem) {
            if (popupInstance != null 
                    && popupInstance instanceof ComponentInstance) {
                mainGUI.addComposable(((ComponentInstance) popupInstance));
            }
        } else if (source == removeComponentItem) {
            if (popupInstance != null
                    && popupInstance instanceof ComponentInstance) {
                mainGUI.removeComposable((ComponentInstance) popupInstance);
            }
        } else if (source == showComponentItem) {
            if (popupInstance != null 
                    && popupInstance instanceof ComponentInstance) {
                mainGUI.showInstanceInDiagram(popupInstance);
            }
        } else if (source == addComponentToPartItem) {
            if (popupInstance != null && popupInstance instanceof PartInstance) {
                mainGUI.addComposable(((PartInstance) popupInstance));
            }
        } else if (source == editInterfaceItem) {
            if (popupInstance != null
                    && popupInstance instanceof InterfaceInstance) {
                mainGUI.editInterfaces(((InterfaceInstance) popupInstance)
                                       .getComponent());
            }
        } else if (source == removeInterfaceItem) {
            if (popupInstance != null
                    && popupInstance instanceof InterfaceInstance) {
                mainGUI.removeInterface((InterfaceInstance) popupInstance);
            }
        } else if (source == showInterfaceItem) {
            if (popupInstance != null 
                    && popupInstance instanceof InterfaceInstance) {
                mainGUI.showInstanceInDiagram(popupInstance);
            }
        } else if (source == removeBindingItem) {
            if (popupInstance != null
                    && popupInstance instanceof BindingInstance) {
                mainGUI.removeBinding((BindingInstance) popupInstance);
            }
        } else if (source == showBindingItem) {
            if (popupInstance != null 
                    && popupInstance instanceof BindingInstance) {
                mainGUI.showInstanceInDiagram(popupInstance);
            }
        } else if (source == editAttributeItem) {
            if (popupInstance != null
                    && popupInstance instanceof AttributeInstance) {
                mainGUI.editAttributes((AttributedInstance) 
                		((AttributeInstance) popupInstance).getParent());
            }
        }
    }
    
    public void keyTyped(KeyEvent e) { 
        if (e.getKeyChar() == 27) {
            showPopup(getSelectionPath());
        }
    }
    
    public void keyPressed(KeyEvent e) { }
    
    public void keyReleased(KeyEvent e) { }
    
    public void mousePressed(MouseEvent e) {
        if (e.isPopupTrigger()) {
            showPopup(getPathForLocation(e.getX(), e.getY()), 
                      e.getX(), e.getY());
        }
    }

    public void mouseReleased(MouseEvent e) {
        if (e.isPopupTrigger()) {
            showPopup(getPathForLocation(e.getX(), e.getY()), 
                      e.getX(), e.getY());
        }
    }
    
    public void mouseClicked(MouseEvent e) { }
    
    public void mouseEntered(MouseEvent e) { }

    public void mouseExited(MouseEvent e) { }
    
    private void showPopup(TreePath path) {
        if (path == null) {
            return;
        }
        Rectangle pathArea = getPathBounds(path);
        showPopup(path, 
                  pathArea.x + pathArea.width / 2, 
                  pathArea.y + pathArea.height / 2);
    }
    
    private void showPopup(TreePath path, int x, int y) {
        if (path == null) {
            return;
        }
        ConfigurationNode node 
            = (ConfigurationNode) path.getLastPathComponent();
        if (node instanceof ComponentNode) {
            popupInstance = ((ComponentNode) node).getComponentInstance();
            componentPopup.show(this, x, y);
        } else if (node instanceof PartNode) {
            popupInstance = ((PartNode) node).getPartInstance();
            partPopup.show(this, x, y);
        } else if (node instanceof InterfaceNode) {
            popupInstance = ((InterfaceNode) node).getInterfaceInstance();
            interfacePopup.show(this, x, y);
        } else if (node instanceof BindingNode) {
            popupInstance = ((BindingNode) node).getBindingInstance();
            bindingPopup.show(this, x, y);
        } else if (node instanceof AttributeNode) {
            popupInstance = ((AttributeNode) node).getAttributeInstance();
            attributePopup.show(this, x, y);
        }
    }
    
    
    //------------- INNER CLASS FOR RENDERING TREE --------------------------
    
    /**
     * An own renderer to show different icons
     */
    class ComponentTreeCellRenderer extends DefaultTreeCellRenderer {
        /**
         * Overriden to implement own icons.
         */
        public Component getTreeCellRendererComponent(JTree tree, Object value,
                boolean selected, boolean expanded, boolean leaf, int row,
                boolean hasFocus) {
            ConfigurationNode node = (ConfigurationNode) value;
            super.getTreeCellRendererComponent(tree, value, selected, expanded,
                                               leaf, row, hasFocus);
            if (largeIcons) {
                setIcon(node.getLargeIcon());
            } else {
                setIcon(node.getSmallIcon());
            }
            setText(node.toString());
            //setOpaque(true);
            setBackgroundNonSelectionColor(Colors.COMPONENT_BACKGROUND);
            return this;
        }
    }
}
