/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * FeatureTree.java
 *
 * Created on 23. helmikuuta 2004, 12:57
 */

package kumbang.configurator.client.gui.tree;

import kumbang.util.gui.Colors;
import kumbang.util.gui.Icons;

import kumbang.configurator.client.gui.core.MainGUI;

import kumbang.core.configuration.instance.AttributeInstance;
import kumbang.core.configuration.instance.AttributedInstance;
import kumbang.core.configuration.instance.FeatureInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.PartInstance;

import java.awt.Component;
import java.awt.Rectangle;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.HashMap;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTree;

import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;

import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 *
 * @author  vmyllarn
 */
public class FeatureTree extends JTree implements TreeSelectionListener, 
                                                  MouseListener, 
                                                  ActionListener, 
                                                  KeyListener {
    
    private ConfigurationNode rootNode;
    
    private MainGUI mainGUI;
    
    private DefaultTreeModel treeModel;
    
    private HashMap nodes;
    
    private Instance popupInstance;
    
    private JPopupMenu featurePopup;
    
    private JMenuItem showFeatureItem;
    
    private JMenuItem editFeatureItem;
    
    private JMenuItem removeFeatureItem;
    
    private JMenuItem addFeatureItem;
    
    private JPopupMenu partPopup;
    
    private JMenuItem addFeatureToPartItem;
    
    private JPopupMenu attributePopup;
    
    private JMenuItem editAttributeItem;
    
    private boolean largeIcons;

    /** 
     * Creates a new instance of FeatureTree.
     * @param mainGUI main GUI 
     */
    public FeatureTree(MainGUI mainGUI) {
        super();
        this.mainGUI = mainGUI;
        this.rootNode = new ConfigurationNode(ConfigurationNode.FEATURE_ROOT);
        this.nodes = new HashMap();
        
        initPopups();
        
        this.treeModel = new DefaultTreeModel(this.rootNode);
        super.setModel(this.treeModel);
        
        setCellRenderer(new FeatureTreeCellRenderer());
        setRootVisible(true);
        setShowsRootHandles(true);
        getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        setSelectionRow(0);
        
        addTreeSelectionListener(this);
        addMouseListener(this);
        addKeyListener(this);
        
        setBackground(Colors.FEATURE_BACKGROUND);
        
        this.largeIcons = false;
    }
    
    //------------- PROTECTED METHODS TO TREE HANDLER ETC --------------------
    
    
    protected void setLargeIcons(boolean large) {
        if (large != this.largeIcons) {
            this.largeIcons = large;
            this.treeModel.reload();
        }
    }
    
    /**
     * Initializes the tree. Fetches feature configuration root from
     * the tree handler, and creates nodes according to it.
     */
    protected void init(FeatureInstance rootInstance) {
        if (rootInstance != null) {
            createNodesForInstance(rootInstance, rootNode);
        }
        scrollRowToVisible(0);
        setSelectionRow(0);
    }
    
    /**
     * Resets the tree and clears all children from it.
     */
    protected void reset() {
        nodes.clear();
        this.rootNode.removeAllChildren();
        this.treeModel.reload();
    }
    
    /**
     * Adds a new node into this tree. The new node can represent a feature, 
     * other instances are ignored. Creates recusrsively other nodes below
     * the new node.
     * @param instance new instance to be added
     */
    protected void addInstance(Instance instance) {
        if (containsNodeForInstance(instance)) {
            // this instance already has a node in the tree
            return;
        }
        if (instance instanceof FeatureInstance) {
            FeatureInstance fInstance = (FeatureInstance) instance;
            PartInstance part = fInstance.getParentPart();
            if (part == null) {
                return;
            }
            ConfigurationNode partNode = getNodeForInstance(part);
            if (partNode == null) {
                return;
            }
            createNodesForInstance(fInstance, partNode);
        }
    }
    
    /**
     * Removes the node corresponding to the given instance from this tree, and
     * all nodes below that node.
     * @pararm instance instance to be removed
     */ 
    protected void removeInstance(Instance instance) {
        removeNodes(getNodeForInstance(instance));
    }
    
    /**
     * Selects the given instance from this tree.
     * @param instance instance to be selected
     */
    protected void selectInstance(Instance instance) {
        ConfigurationNode node = getNodeForInstance(instance);
        if (node != null) {
            TreePath path = new TreePath(node.getPath());
            if (path != null) {
                setSelectionPath(path);
                scrollPathToVisible(path);
            }
        }
    }
    
    
    /**
     * Updates the given instance in this tree.
     * @param instance instance to be updated
     */
    protected void updateInstance(Instance instance) {
        if (instance == null) {
            return;
        }
        ConfigurationNode node = getNodeForInstance(instance);
        if (node != null) {
            treeModel.reload(node);
        }
    }
    
    //------- PRIVATE METHODS FOR CREATING AND REMOVING NODES -----------------
    
    /**
     * Creates a feature node for the given feature instance. Proceeds 
     * recursively to create nodes for attributes and parts under
     * this feature instance.
     * @param instance feature instance
     * @param parentNode parent node of this new node
     */
    private void createNodesForInstance(FeatureInstance instance, 
                                        ConfigurationNode parentNode) {
        FeatureNode fNode = new FeatureNode(instance);
        
        treeModel.insertNodeInto(fNode, parentNode, parentNode.getChildCount());
        parentNode.add(fNode);
        setNodeForInstance(fNode, instance);
        
        TreePath path = new TreePath(fNode.getPath());
        scrollPathToVisible(path);
        //setSelectionPath(path);
        
        /* add nodes for attributes */
        Iterator attrs = instance.getAttributes();
        while (attrs.hasNext()) {
            createNodesForInstance((AttributeInstance) attrs.next(), fNode);
        }
        
        /* add nodes for parts */
        Iterator parts = instance.getParts();
        while (parts.hasNext()) {
            createNodesForInstance((PartInstance) parts.next(), fNode);
        }
    }
    
    /**
     * Creates an attribute node for the given attribute instance. 
     * @param instance attribute instance
     * @param parentNode parent node of this new node
     */
    private void createNodesForInstance(AttributeInstance instance, 
                                        ConfigurationNode parentNode) {
        AttributeNode aNode = new AttributeNode(instance);
        
        treeModel.insertNodeInto(aNode, parentNode, parentNode.getChildCount());
        scrollPathToVisible(new TreePath(aNode.getPath()));
        parentNode.add(aNode);
        setNodeForInstance(aNode, instance);
    }
    
    
    /**
     * Creates a part node for the given part instance. Proceeds recursively
     * to create nodes for component instances under this part. 
     * @param instance part instance
     * @param parentNode parent node of this new node
     */
    private void createNodesForInstance(PartInstance instance, 
                                        ConfigurationNode parentNode) {
        if (instance.getNumberOfPossibleTypes() == 1 &&
                instance.getLowerBound() == 1 &&
                instance.getUpperBound() == 1) {
            // no need for part node, directly attached to feature
            Iterator iter = instance.getInstances();
            while (iter.hasNext()) {
                createNodesForInstance((FeatureInstance) iter.next(), parentNode);
            }
            return;
        }
        PartNode pNode = new PartNode(instance);
        
        treeModel.insertNodeInto(pNode, parentNode, parentNode.getChildCount());
        scrollPathToVisible(new TreePath(pNode.getPath()));
        parentNode.add(pNode);
        setNodeForInstance(pNode, instance);
        
        Iterator iter = instance.getInstances();
        while (iter.hasNext()) {
            createNodesForInstance((FeatureInstance) iter.next(), pNode);
        }
    }
    
    /**
     * Removes nodes from the tree. Removes this node, all its subchildren,
     * and removes the nodes from the hashmap.
     * @param node node to be removed
     */
    private void removeNodes(ConfigurationNode node) {
        if (node == null) {
            return;
        }
        Enumeration children = node.children();
        while (children.hasMoreElements()) {
            removeNodes((ConfigurationNode) children.nextElement());
        }
        if (node instanceof InstanceNode) {
            Instance instance = ((InstanceNode) node).getInstance();
            unsetNodeForInstance(instance);
        }        
        treeModel.removeNodeFromParent(node);
    }
    
    //------------- PRIVATE METHODS FOR MAPPING NODES ------------------------
    
    public boolean containsNodeForInstance(Instance instance) {
        return nodes.containsKey(instance);
    }
    
    /**
     * Gets the node that corresponds to the given instance, or null
     * if no such instance exists in the tree.
     * @param instance instance
     * @return ConfigurationNode node corresponding to instance
     */
    private ConfigurationNode getNodeForInstance(Instance instance) {
        if (instance != null) {
            return (ConfigurationNode) nodes.get(instance);
        }
        return null;
    }
    
    /**
     * Sets a mapping between the given node and instance.
     * @param node
     * @param instance
     */
    private void setNodeForInstance(ConfigurationNode node, Instance instance) {
        if (node != null && instance != null) {
            nodes.put(instance, node);
        }
    }
    
    /**
     * Removes a mapping between the given instance and a node in the tree.
     * @param instance
     */
    private void unsetNodeForInstance(Instance instance) {
        if (instance != null) {
            nodes.remove(instance);
        }
    }
    
    //------------- PRIVATE METHODS FOR POPUPS -------------------------------
    
    private void initPopups() {
        /* init popup for feature nodes */
        featurePopup = new JPopupMenu("Feature");
        editFeatureItem = new JMenuItem("Edit attributes of this feature", 
                                        Icons.ATTRIBUTE);
        addFeatureItem = new JMenuItem("Add new feature under this feature", 
                                       Icons.NEW_FEATURE);
        removeFeatureItem = new JMenuItem("Remove this feature", Icons.DELETE);
        showFeatureItem = new JMenuItem("Show details of this feature", 
                                        Icons.ZOOM);
        editFeatureItem.addActionListener(this);
        addFeatureItem.addActionListener(this);
        removeFeatureItem.addActionListener(this);
        showFeatureItem.addActionListener(this);
        featurePopup.add(editFeatureItem);
        featurePopup.add(addFeatureItem);
        featurePopup.add(removeFeatureItem);
        featurePopup.add(showFeatureItem);
        
        /* init popup for part nodes */
        partPopup = new JPopupMenu("Part");
        addFeatureToPartItem = new JMenuItem("Add new feature under this part", 
                                             Icons.NEW_FEATURE);
        addFeatureToPartItem.addActionListener(this);
        partPopup.add(addFeatureToPartItem);
        
        /* init popup for attribute nodes */
        attributePopup = new JPopupMenu("Attribute");
        editAttributeItem = new JMenuItem("Edit attributes", Icons.ATTRIBUTE);
        editAttributeItem.addActionListener(this);
        attributePopup.add(editAttributeItem);
    }
    
    //------------- METHODS IMPLEMENTING INTERFACES --------------------------
    
    /**
     * Reacts to TreeSelectionEvents by telling mainGUI to select the same
     * instance.
     * @param e tree selection event
     */
    public void valueChanged(TreeSelectionEvent e) {
        ConfigurationNode node 
                = (ConfigurationNode) getLastSelectedPathComponent();
        if (node != null && node instanceof InstanceNode) {
            mainGUI.selectInstance(((InstanceNode) node).getInstance());
        }
    }
    
    public void actionPerformed(ActionEvent e) {
        Object source = e.getSource();
        if (source == editFeatureItem) {
            if (popupInstance != null
                    && popupInstance instanceof FeatureInstance) {
                mainGUI.editAttributes((FeatureInstance) popupInstance);
            }
        } else if (source == addFeatureItem) {
            if (popupInstance != null 
                    && popupInstance instanceof FeatureInstance) {
                mainGUI.addComposable(((FeatureInstance) popupInstance));
            }
        } else if (source == removeFeatureItem) {
            if (popupInstance != null
                    && popupInstance instanceof FeatureInstance) {
                mainGUI.removeComposable((FeatureInstance) popupInstance);
            }
        } else if (source == addFeatureToPartItem) {
            if (popupInstance != null 
                    && popupInstance instanceof PartInstance) {
                mainGUI.addComposable(((PartInstance) popupInstance));
            }
        } else if (source == editAttributeItem) {
            if (popupInstance != null 
                    && popupInstance instanceof AttributeInstance) {
                mainGUI.editAttributes((AttributedInstance)
                		((AttributeInstance) popupInstance).getParent());
            }
        } else if (source == showFeatureItem) {
            if (popupInstance != null 
                    && popupInstance instanceof FeatureInstance) {
                mainGUI.showInstanceInDiagram(popupInstance);
            }
        }
    }
        
    public void keyTyped(KeyEvent e) {
        if (e.getKeyChar() == 27) {
            showPopup(getSelectionPath());
        }
    }
    
    public void keyPressed(KeyEvent e) { }
    
    public void keyReleased(KeyEvent e) { }
    
    public void mousePressed(MouseEvent e) {
        if (e.isPopupTrigger())  {
            showPopup(getPathForLocation(e.getX(), e.getY()), 
                      e.getX(), e.getY());
        }
    }
    
    public void mouseReleased(MouseEvent e) {
        if (e.isPopupTrigger()) {
            showPopup(getPathForLocation(e.getX(), e.getY()), 
                      e.getX(), e.getY());
        }
    }
    
    public void mouseClicked(MouseEvent e) { }
    
    public void mouseEntered(MouseEvent e) { }
    
    public void mouseExited(MouseEvent e) { }
    
    private void showPopup(TreePath path) {
        if (path == null) {
            return;
        }
        Rectangle pathArea = getPathBounds(path);
        showPopup(path, 
                  pathArea.x + pathArea.width / 2, 
                  pathArea.y + pathArea.height / 2);
    }
    
    private void showPopup(TreePath path, int x, int y) {
        if (path == null) {
            return;
        }
        ConfigurationNode node 
            = (ConfigurationNode) path.getLastPathComponent();
        if (node instanceof FeatureNode) {
            popupInstance = ((FeatureNode) node).getFeatureInstance();
            featurePopup.show(this, x, y);
        } else if (node instanceof PartNode) {
            popupInstance = ((PartNode) node).getPartInstance();
            partPopup.show(this, x, y);
        } else if (node instanceof AttributeNode) {
            popupInstance = ((AttributeNode) node).getAttributeInstance();
            attributePopup.show(this, x, y);
        }
    }
    
    //------------- INNER CLASS FOR RENDERING TREE --------------------------
    
    /**
     * An own renderer to show different icons 
     */
    class FeatureTreeCellRenderer extends DefaultTreeCellRenderer {
        /**
         * Overriden to implement own icons.
         */
        public Component getTreeCellRendererComponent(JTree tree, Object value,
                boolean selected, boolean expanded, boolean leaf, int row,
                boolean hasFocus) {
            ConfigurationNode node = (ConfigurationNode) value;
            super.getTreeCellRendererComponent(tree, value, selected, expanded,
                                               leaf, row, hasFocus);
            if (largeIcons) {
                setIcon(node.getLargeIcon());
            } else {
                setIcon(node.getSmallIcon());
            }
            setText(node.toString());
            setBackgroundNonSelectionColor(Colors.FEATURE_BACKGROUND);
            return this;
        }
    }
}
