/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * TreeHandler.java
 *
 * Created on 26. maaliskuuta 2004, 17:33
 */

package kumbang.configurator.client.gui.tree;



import kumbang.configurator.client.gui.core.MainGUI;

import kumbang.configurator.client.gui.event.ConfigurationModificationListener;
import kumbang.configurator.client.gui.event.ConfigurationModifiedEvent;
import kumbang.configurator.client.gui.event.ComponentSelectionListener;
import kumbang.configurator.client.gui.event.ComponentSelectedEvent;

import kumbang.core.configuration.instance.AttributeInstance;
import kumbang.core.configuration.instance.BindingInstance;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.FeatureInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.task.AddInstanceTask;
import kumbang.core.configuration.task.ClientTask;
import kumbang.core.configuration.task.RemoveInstanceTask;
import kumbang.core.configuration.task.SetAttributeTask;

import java.util.Iterator;

/**
 *
 * @author  vmyllarn
 */
public class TreeHandler implements ConfigurationModificationListener, ComponentSelectionListener {
    
    private ComponentTree componentTree;
    
    private FeatureTree featureTree;
    
    
    /** Creates a new instance of TreeHandler */
    public TreeHandler(MainGUI mainGUI, ComponentTree cTree, FeatureTree fTree) {
        this.componentTree = cTree;
        this.featureTree = fTree;
        mainGUI.addConfigurationModificationListener(this);
        mainGUI.addComponentSelectionListener(this);
    }
    
    /** Creates a new instance of TreeHandler */
    public TreeHandler(MainGUI mainGUI) { 
        mainGUI.addConfigurationModificationListener(this);
        mainGUI.addComponentSelectionListener(this);
    }
    
    public void setTrees(ComponentTree cTree, FeatureTree fTree) {
        this.componentTree = cTree;
        this.featureTree = fTree;
    }
    
    public void init(ComponentInstance componentRoot, FeatureInstance featureRoot) {
        this.componentTree.init(componentRoot);
        this.featureTree.init(featureRoot);
    }
    
    public void reset() {
        this.componentTree.reset();
        this.featureTree.reset();
    }    
    
    public void configurationModified(ConfigurationModifiedEvent e) {
        handleTask(e.getTask());
    }
    
    public void setLargeIcons(boolean large) {
        featureTree.setLargeIcons(large);
        componentTree.setLargeIcons(large);
    }
    
    //--------------ADDING AND REMOVING INSTANCES ----------------------

    private void handleTask(ClientTask task) {
        
        // add instance task
        if (task instanceof AddInstanceTask) {
            Instance instance = ((AddInstanceTask)task).getAddedInstance();
            if (isFromComponentConfiguration(instance)) {
                componentTree.addInstance(instance);
            }
            else if (isFromFeatureConfiguration(instance)) {
                featureTree.addInstance(instance);
            }
        }
        // set attribute task
        else if (task instanceof SetAttributeTask) {
            SetAttributeTask sTask = (SetAttributeTask)task;
            AttributeInstance attribute = sTask.getAttribute();
            if (isFromComponentConfiguration(attribute)) {
                componentTree.updateInstance(attribute);
            }
            if (isFromFeatureConfiguration(attribute)) {
                featureTree.updateInstance(attribute);
            }
        }
        // remove instance task
        else if (task instanceof RemoveInstanceTask) {
            Iterator it = task.getSubTasks();
            while (it.hasNext()) {
                handleTask((ClientTask)it.next());
            }
            Instance instance = ((RemoveInstanceTask)task).getInstance();
            if (isFromComponentConfiguration(instance)) {
                componentTree.removeInstance(instance);
            }
            else if (isFromFeatureConfiguration(instance)) {
                featureTree.removeInstance(instance);
            }
        }
        // other tasks
        else {
            Iterator it = task.getSubTasks();
            while (it.hasNext()) {
                handleTask((ClientTask)it.next());
            }
        }
    }
    
    public void componentSelected(ComponentSelectedEvent e) {
        selectInstance(e.getSelectedInstance());
    }
    
    private void selectInstance(Instance instance) {
        if (isFromComponentConfiguration(instance)) {
            componentTree.selectInstance(instance);
        }
        else if (isFromFeatureConfiguration(instance)) {
            featureTree.selectInstance(instance);
        }
    }


    
    private boolean isFromComponentConfiguration(Instance instance) {
        return (instance instanceof ComponentInstance || 
                instance instanceof InterfaceInstance ||
                instance instanceof BindingInstance ||
                instance instanceof AttributeInstance);
    }
    
    private boolean isFromFeatureConfiguration(Instance instance) {
        return (instance instanceof FeatureInstance || 
                instance instanceof AttributeInstance);
    }
    

    
}
