/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EventListener;
import java.util.Iterator;
import java.util.List;

import kumbang.configurator.client.core.ClientProperties;
import kumbang.configurator.client.core.ClientServices;
import kumbang.configurator.client.core.ConfigurationClient;
import kumbang.configurator.client.plugin.event.ConfigurationModificationListener;
import kumbang.configurator.client.plugin.event.ConfigurationModifiedEvent;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedEvent;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedListener;
import kumbang.configurator.client.plugin.event.ConfigurationSaveListener;
import kumbang.configurator.client.plugin.event.SaveEvent;
import kumbang.configurator.client.plugin.preferences.PreferenceConstants;

import kumbang.configurator.server.core.ServerProperties;
import kumbang.core.configuration.task.ClientTask;
import kumbang.util.KumbangLogger;

import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.resource.ColorRegistry;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.IPerspectiveDescriptor;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;


/**
 * The main plugin class to be used in the desktop.
 * 
 */
public class KumbangConfiguratorPlugin extends AbstractUIPlugin implements IKumbangPlugin {  

	//The shared instance.
	private static KumbangConfiguratorPlugin plugin;
	
	// Some IDs
	public final static String PLUGIN_ID = "kumbang.configurator.client.plugin";
	public final static String FEATUREVIEW_ID = "KumbangTools.view1";
	public final static String COMPONENTVIEW_ID = "KumbangTools.view2";
	public final static String STATUSVIEW_ID = "KumbangTools.view3";
	public final static String DIAGRAMVIEW_ID = "KumbangTools.view4";
	
	// Preference store for the Window->Preference-stuff
	private IPreferenceStore store;
	
	// Listener for preference changes
	private IPropertyChangeListener listener;
	
	private ImageRegistry imageRegistry;
	private ClientServices client; // Kumbang Client Services
	
	
	private boolean serverAvailable;
	
	private ColorRegistry colorRegistry;
	
	private List<EventListener> modificationListeners;
	private List<EventListener> openCloseListeners;
	private List<EventListener> saveListeners;

	private String eclipsePluginPath;


	/**
	 * KumbangConfiguratorPlugin Constructor
	 *
	 */
	public KumbangConfiguratorPlugin() {
        // Apparently, after the new version of smodels-library,
        // the plugin will not work when using RMI, if Logger is
        // null... go figure
        
		KumbangLogger.initLogger("client");
		plugin = this;
		modificationListeners = Collections.synchronizedList(new ArrayList<EventListener>());
		openCloseListeners = Collections.synchronizedList(new ArrayList<EventListener>());
		saveListeners = Collections.synchronizedList(new ArrayList<EventListener>());
	}
	
	protected void initializeDefaultPreferences(IPreferenceStore store) {
	    store.setDefault(PreferenceConstants.P_COMMUNICATIONMETHOD, "local");
	    store.setDefault(PreferenceConstants.P_SERVERADDRESS, "localhost");
	    store.setDefault(PreferenceConstants.P_SERVERPORT, "6969");
	    store.setDefault(PreferenceConstants.P_UML_INTERFACES, false);
	}

	
	/**
	 * This method is called upon plug-in activation
	 */
	public void start(BundleContext context) throws Exception {
		super.start(context);
		// Tell that This is the Eclipse client
		ClientProperties.getInstance().setClient(ClientProperties.CLIENT_ECLIPSE);
		
		
		// Set the proper root directory, otherwise it would use the
		// Eclipse install directory
		//ClientProperties.getInstance().setRootDir(getPluginDir(PLUGIN_ID));
		
		// This works when using plugin-JAR and local communication method
		eclipsePluginPath = Platform.getInstallLocation().getURL().getPath() + "plugins/" + PLUGIN_ID + "/";
		if( Platform.getOS().compareTo(Platform.OS_WIN32) == 0 )
			eclipsePluginPath = eclipsePluginPath.substring(1);
		
		ClientProperties.getInstance().setRootDir(eclipsePluginPath);
		
		
		store = getPreferenceStore();
		colorRegistry = JFaceResources.getColorRegistry();
		initializeColorRegistry();
		
		// Listener for the property changes in the preferences page
		this.listener =	new IPropertyChangeListener() {
		    /* Setup the listener for changes in the preferences */
		    public void propertyChange(PropertyChangeEvent e) {
		    	
		    	// Communication method was changed
		    	if (e.getProperty().equals(PreferenceConstants.P_COMMUNICATIONMETHOD)) {

		    		// We need to restart the connection to the server
		    		if (KumbangConfiguratorPlugin.this.client != null) {
		    			System.out.println("Communication method changed, restarting connection");
		    			fireOpenCloseEvent(ConfigurationOpenedClosedEvent.EVENT_CLOSE);
		    			endSession();
		    		} else {
		    			// There was no client yet, so just get a new one
		    			KumbangConfiguratorPlugin.this.client = KumbangConfiguratorPlugin.this.getClientServices();
		    		}	

		    		// If the method is set to local, load the properties from the preferenec page
		    		if (((String) e.getNewValue()).equals("local")) {
		    			loadServerProperties();		    		
		    		}	
		    		
		    	}
		    	// If the server address or port is changed, we need to restart the connection 
		    	// to the new address
		    	else if (e.getProperty().equals(PreferenceConstants.P_SERVERADDRESS)
						|| e.getProperty().equals(PreferenceConstants.P_SERVERPORT)) {
					System.out.println("Server location changed, restarting connection");
					fireOpenCloseEvent(ConfigurationOpenedClosedEvent.EVENT_CLOSE);
					endSession();
		    	} 
		    }

		};
		
		store.addPropertyChangeListener(listener);
 
		// If the server is set to local initially, set the default, working preferences now
		// These can't be changed at the moment, which is probably better when using
		// local server
		if (((String)store.getString(PreferenceConstants.P_COMMUNICATIONMETHOD)).equals("local")) {
			loadServerProperties();
		}
		
		// Let's presume server is not available right away
		serverAvailable = false;
	}


    /*
     * Helper method reload properties from the preferenec store to the 
     * ServerProperties class
     */
	private void loadServerProperties() {
		ServerProperties.getInstance().setParserClass(ServerProperties.DEFAULT_PARSER_CLASS);
	}

	private void initializeColorRegistry() {
		colorRegistry.put(Colors.COMPLETE_CONFIGURATION, new RGB(0,204,51));
		colorRegistry.put(Colors.INCOMPLETE_CONFIGURATION, new RGB(204,0,51));
		colorRegistry.put(Colors.COMPONENT, new RGB(205, 220, 235));
		colorRegistry.put(Colors.BLACK, new RGB(0,0,0));
		colorRegistry.put(Colors.WHITE, new RGB(255,255,255));
		colorRegistry.put(Colors.SELECTION, new RGB(230, 130, 20));
		colorRegistry.put(Colors.ACTION, new RGB(250, 179, 83));
		colorRegistry.put(Colors.GRAY, new RGB(125, 125, 125));
		colorRegistry.put(Colors.INTERFACE, new RGB(250, 250, 210));
		colorRegistry.put(Colors.TOOLBAR, new RGB(200, 183, 145));
		
	}
	
	public ColorRegistry getColorRegistry() {
		return colorRegistry;
	}


	/**
	 * Ends the current session with the server by closing a open configuration
	 * and disconnecting from it.
	 *
	 */
	private void endSession() {
		if (this.client != null) {
			this.client.closeConfiguration();
			this.client.exit();
			this.client = null;
		}
	}
	/**
	 * This method is called when the plug-in is stopped
	 */
	public void stop(BundleContext context) throws Exception {
		super.stop(context);
		endSession();
		plugin = null;
		// Empty the listener list, they should be empty at this point 
		// though..
		modificationListeners.clear();
		openCloseListeners.clear();
		store.removePropertyChangeListener(listener);
	}

	
	/**
	 * Returns the shared instance.
	 */
	public static KumbangConfiguratorPlugin getDefault() {
		return plugin;
	}

	
	/**
	 * Gets the ImageRegistry, which holds the images that
	 * are frequently used.
	 */
	public ImageRegistry getImageRegistry() {
		if (imageRegistry == null) {
			imageRegistry = createImageRegistry();
			initializeImageRegistry(imageRegistry);
		}
		return imageRegistry;
	}
	
	
	/**
	 * Initializes ImageRegistry and loads frequently used images into it.
	 */
	protected void initializeImageRegistry(ImageRegistry imageRegistry) {
		imageRegistry.put(Icons.ATTRIBUTE, getImageDescriptor(Icons.ATTRIBUTE));
		imageRegistry.put(Icons.BINDING_SOURCE, getImageDescriptor(Icons.BINDING_SOURCE));
		imageRegistry.put(Icons.BINDING_TARGET, getImageDescriptor(Icons.BINDING_TARGET));
		imageRegistry.put(Icons.COMPONENT, getImageDescriptor(Icons.COMPONENT));
        imageRegistry.put(Icons.CONNECT, getImageDescriptor(Icons.CONNECT));
        imageRegistry.put(Icons.CURSOR_CONNECT_1, getImageDescriptor(Icons.CURSOR_CONNECT_1));
        imageRegistry.put(Icons.CURSOR_CONNECT_2, getImageDescriptor(Icons.CURSOR_CONNECT_2));
        imageRegistry.put(Icons.CURSOR_DISCONNECT, getImageDescriptor(Icons.CURSOR_DISCONNECT));
        imageRegistry.put(Icons.CURSOR_ZOOM, getImageDescriptor(Icons.CURSOR_ZOOM));
        imageRegistry.put(Icons.DISCONNECT, getImageDescriptor(Icons.DISCONNECT));
		imageRegistry.put(Icons.DOWN, getImageDescriptor(Icons.DOWN));
		imageRegistry.put(Icons.FEATURE, getImageDescriptor(Icons.FEATURE));
		imageRegistry.put(Icons.PROVIDED_INTERFACE, getImageDescriptor(Icons.PROVIDED_INTERFACE));
		imageRegistry.put(Icons.REQUIRED_INTERFACE, getImageDescriptor(Icons.REQUIRED_INTERFACE));
		imageRegistry.put(Icons.REQUIRED_INTERFACE_UML, getImageDescriptor(Icons.REQUIRED_INTERFACE_UML));
		imageRegistry.put(Icons.PROVIDED_INTERFACE_UML, getImageDescriptor(Icons.PROVIDED_INTERFACE_UML));
		imageRegistry.put(Icons.LOGO, getImageDescriptor(Icons.LOGO));
		imageRegistry.put(Icons.NEW, getImageDescriptor(Icons.NEW));
		imageRegistry.put(Icons.NEW_ATTRIBUTE, getImageDescriptor(Icons.NEW_ATTRIBUTE));
		imageRegistry.put(Icons.NEW_COMPONENT, getImageDescriptor(Icons.NEW_COMPONENT));
		imageRegistry.put(Icons.NEW_FEATURE, getImageDescriptor(Icons.NEW_FEATURE));
		imageRegistry.put(Icons.NEW_INTERFACE, getImageDescriptor(Icons.NEW_INTERFACE));
		imageRegistry.put(Icons.PART, getImageDescriptor(Icons.PART));
		imageRegistry.put(Icons.SELECT, getImageDescriptor(Icons.SELECT));
		imageRegistry.put(Icons.STATUS_INCOMPLETE, getImageDescriptor(Icons.STATUS_INCOMPLETE));
		imageRegistry.put(Icons.STATUS_COMPLETE, getImageDescriptor(Icons.STATUS_COMPLETE));
		imageRegistry.put(Icons.TRASH_CAN, getImageDescriptor(Icons.TRASH_CAN));
		imageRegistry.put(Icons.ZOOM, getImageDescriptor(Icons.ZOOM));
		imageRegistry.put(Icons.ZOOM_OUT, getImageDescriptor(Icons.ZOOM_OUT));
	}
	
	
	
	/**
	 * Returns an image descriptor for the image file at the given
	 * plug-in relative path.
	 *
	 * @param name the file name (inside 'pics'-folder)
	 * @return the image descriptor
	 */
	public ImageDescriptor getImageDescriptor(String name) {
		String picPath = "pics/";
		
		/* get bundle with the specified id */
		Bundle bundle = Platform.getBundle(PLUGIN_ID);
		
		if( bundle == null ) {
			throw new RuntimeException("Could not resolve plugin: " + PLUGIN_ID);
		}
		
		/* resolve Bundle.getEntry to local URL */
		URL pluginURL = null;
		try {
			pluginURL = Platform.resolve(bundle.getEntry("/"));
		} catch (IOException e) {
			throw new RuntimeException("Could not get installation directory of the plugin: " + PLUGIN_ID);
		}
		
		URL url;
		try {
			url = new URL(pluginURL, picPath + name);
			return ImageDescriptor.createFromURL(url);
		} catch (MalformedURLException e) {
			// should not happen
			return ImageDescriptor.getMissingImageDescriptor();
		}
		
	}
	
    public String getPluginDir() {
        return eclipsePluginPath;
    }
	
	/**
	 * Returns the absolute plug-in path.
	 * @param pluginId
	 * @return Absolute plug-in path
	 */
	public String getPluginDir(String pluginId) {
		/* get bundle with the specified id */
		Bundle bundle = Platform.getBundle(pluginId);
		if( bundle == null ) {
			throw new RuntimeException("Could not resolve plugin: " + pluginId);
		}
		
		/* resolve Bundle.getEntry to local URL */
		URL pluginURL = null;
		try {
			pluginURL = Platform.resolve(bundle.getEntry("/"));
		} catch (IOException e) {
			throw new RuntimeException("Could not get installation directory of the plugin: " + pluginId);
		}
		String pluginInstallDir = pluginURL.getPath().trim();
		if( pluginInstallDir.length() == 0 )
			throw new RuntimeException("Could not get installation directory of the plugin: " + pluginId);
		
		/* since path returned by URL.getPath starts with a forward slash, that
		 * is not suitable to run commandlines on Windows-OS, but for Unix-based
		 * OSes it is needed. So strip one character for windows. There seems
		 * to be no other clean way of doing this. */
		if( Platform.getOS().compareTo(Platform.OS_WIN32) == 0 )
			pluginInstallDir = pluginInstallDir.substring(1);
		
		return pluginInstallDir;
	}
	
	public boolean isUMLPreferenceSet() {
		return this.store.getBoolean(PreferenceConstants.P_UML_INTERFACES);
	}
	
    public boolean showCompletenessPopups() {
        return this.store.getBoolean(PreferenceConstants.P_COMPLETENESS_POPUPS);
    }
    
	private String getCommunicationMethod() {
		IPreferenceStore store = getPreferenceStore();
		return store.getString(PreferenceConstants.P_COMMUNICATIONMETHOD);
	}
    
    private String getServerLocation() {
        IPreferenceStore store = getPreferenceStore();
        String address = store.getString(PreferenceConstants.P_SERVERADDRESS);
        String port = store.getString(PreferenceConstants.P_SERVERPORT);
        return "//" + address + ":" + port;
    }
	
	public ClientServices getClientServices() {
        if (this.client == null) {
            String comMethod = this.getPreferenceStore().getString(
                    PreferenceConstants.P_COMMUNICATIONMETHOD);
            System.out.println("Communication method: "
                    + this.getPreferenceStore().getString(
                            PreferenceConstants.P_COMMUNICATIONMETHOD));
            System.out
                    .println("KumbangConfigurator: Initializing configuration client");
            try {
                if (comMethod.equals("local")) {
                    ConfigurationClient client = new ConfigurationClient(
                            getCommunicationMethod());
                    this.client = new ClientServices(client);
                } else if (comMethod.equals("rmi")) {
                    ConfigurationClient client = new ConfigurationClient(
                            getCommunicationMethod(), getServerLocation());
                    this.client = new ClientServices(client);
                }

                this.serverAvailable = true;
                return this.client;
            } catch (RemoteException re) {
                System.out.println("Cannot initialize connection "
                        + "to configuration server:\n" + re);
                this.serverAvailable = false;
                return null;
            }
        } else
            return this.client;
    }
	
	public String saveLocation() {
		IPreferenceStore store = getPreferenceStore();
		return store.getString(PreferenceConstants.P_SAVEDIRECTORY);
	}
	
	public boolean isServerAvailable() {
		// Server hasn't been contacted yet
		if (this.client == null) {
			this.getClientServices();
			return serverAvailable;
		} else {
			return serverAvailable;
		}
	
	}

// This should no longer be used since we've moved to using listeners	

	/*	public void refreshViews() {
		FeatureView featureView = (FeatureView)(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().findView(KumbangConfiguratorPlugin.FEATUREVIEW_ID));
		ComponentView componentView = (ComponentView)(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().findView(KumbangConfiguratorPlugin.COMPONENTVIEW_ID));
		StatusView statusView = (StatusView)(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().findView(KumbangConfiguratorPlugin.STATUSVIEW_ID));
		ComponentDiagramView diagramView = (ComponentDiagramView)(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().findView(KumbangConfiguratorPlugin.DIAGRAMVIEW_ID));
		
		if (featureView != null) {
			featureView.refresh();
		}
		if (componentView != null) {
			componentView.refresh();
		}
		if (statusView != null) {
			statusView.refresh();
		}
		if (diagramView != null) {
			diagramView.refresh();
		}
	}
*/	
	public void openKumbangPerspective() {
		IWorkbench workbench = PlatformUI.getWorkbench();
		IWorkbenchWindow window = workbench.getActiveWorkbenchWindow();
		IWorkbenchPage page = window.getActivePage();
		IPerspectiveDescriptor perspective = workbench.getPerspectiveRegistry().findPerspectiveWithId("KumbangTools.perspective1");
		page.setPerspective(perspective);
	}
	
	// Listener related methods
	
    public void addConfigurationModificationListener(ConfigurationModificationListener listener) {
        if (modificationListeners.contains(listener))
        	return; // don't add duplicates
        
    	modificationListeners.add(listener);
    }
    
    public void removeConfigurationModificationListener(ConfigurationModificationListener listener) {
    	modificationListeners.remove(listener);
    }
    
    public void addOpenCloseListener(ConfigurationOpenedClosedListener l) {
    	openCloseListeners.add(l);
    }
    
    public void removeOpenCloseListener(ConfigurationOpenedClosedListener l) {
    	openCloseListeners.remove(l);
    }
    
    public void addSaveListener(ConfigurationSaveListener l) {
    	saveListeners.add(l);
    }
    
    public void removeSaveListener(ConfigurationSaveListener l) {
    	saveListeners.remove(l);
    }
    
    
    public void fireOpenCloseEvent(String eventType) {
    	ConfigurationOpenedClosedEvent e = new ConfigurationOpenedClosedEvent(eventType);
    	synchronized (openCloseListeners) {
    		Iterator it = openCloseListeners.iterator();
    		while (it.hasNext()) {
    			ConfigurationOpenedClosedListener l =
    				(ConfigurationOpenedClosedListener)it.next();
    			l.handleOpenCloseEvent(e);
    		}
    	}
    }
    
    public void fireConfigurationModifiedEvent(ClientTask task) {
    	ConfigurationModifiedEvent e = new ConfigurationModifiedEvent(task);
    	synchronized (modificationListeners) {
    		Iterator it = modificationListeners.iterator();
    		while (it.hasNext()) {
    			ConfigurationModificationListener listener 
    			= (ConfigurationModificationListener)it.next();
    			listener.configurationModified(e);
    		}
    	}
    }
    
    public void fireSaveEvent(File saveFile) {
    	SaveEvent e = new SaveEvent(saveFile);
    	synchronized (saveListeners) {
    		Iterator it = saveListeners.iterator();
    		while (it.hasNext()) {
				ConfigurationSaveListener listener = (ConfigurationSaveListener) it.next();
				listener.handleSaveEvent(e);
			}

		}
    }

	public PluginHandler getHandler() {
		// TODO Auto-generated method stub
		return null;
	}

	public String getSaveLocation() {
		// TODO Auto-generated method stub
		return null;
	}

	public String getPluginID() {
		// TODO Auto-generated method stub
		return null;
	}
}
