/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin;

import java.io.File;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EventListener;
import java.util.Iterator;
import java.util.List;

import kumbang.configurator.Activator;
import kumbang.configurator.client.core.ClientProperties;
import kumbang.configurator.client.core.ClientServices;
import kumbang.configurator.client.core.ConfigurationClient;
import kumbang.configurator.client.plugin.event.ConfigurationModificationListener;
import kumbang.configurator.client.plugin.event.ConfigurationModifiedEvent;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedEvent;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedListener;
import kumbang.configurator.client.plugin.event.ConfigurationSaveListener;
import kumbang.configurator.client.plugin.event.SaveEvent;
import kumbang.configurator.client.plugin.preferences.PreferenceConstants;
import kumbang.core.configuration.task.ClientTask;
import kumbang.util.KumbangLogger;

import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.ui.IPerspectiveDescriptor;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;

public class PluginHandler {

private static final String PLUGIN_ID = "kumbang.configurator";

	private ClientServices client; // Kumbang Client Services

	protected boolean serverAvailable;

	private List<EventListener> modificationListeners;
	private List<EventListener> openCloseListeners;
	private List<EventListener> saveListeners;

	private String eclipsePluginPath;

	private static IKumbangPlugin activator;
	private static PluginHandler instance;
	
	public PluginHandler(IKumbangPlugin activator) {
        // Apparently, after the new version of smodels-library,
        // the plugin will not work when using RMI, if Logger is
        // null... go figure
		this.activator = activator;
        instance = this;
//		KumbangLogger.initLogger("client");
		modificationListeners = Collections.synchronizedList(new ArrayList<EventListener>());
		openCloseListeners = Collections.synchronizedList(new ArrayList<EventListener>());
		saveListeners = Collections.synchronizedList(new ArrayList<EventListener>());
		start();
	}
	
	public static PluginHandler getInstance() {
		if (instance != null)
			return instance;
		else {
			throw new NullPointerException("PluginHandler instance was null");
		}
	}
	
	public static IKumbangPlugin getActivator() {
		return activator;
	}
	
	
	private void start() {
		// Tell that This is the Eclipse client
		ClientProperties.getInstance().setClient(ClientProperties.CLIENT_ECLIPSE);
		
		
		// Set the proper root directory, otherwise it would use the
		// Eclipse install directory
		ClientProperties.getInstance().setRootDir(activator.getPluginDir(PLUGIN_ID));
		
		// This works when using plugin-JAR and local communication method
		eclipsePluginPath = Platform.getInstallLocation().getURL().getPath() + "plugins/" + PLUGIN_ID + "/";
		if( Platform.getOS().compareTo(Platform.OS_WIN32) == 0 )
			eclipsePluginPath = eclipsePluginPath.substring(1);
		
		ClientProperties.getInstance().setRootDir(eclipsePluginPath);

		
		// Let's presume server is not available right away
		serverAvailable = false;
	}

		




	public void endSession() {
		if (this.client != null) {
			this.client.closeConfiguration();
			this.client.exit();
			this.client = null;
		}
	}
	
	public void clearListeners() {
		modificationListeners.clear();
		openCloseListeners.clear();
	}
	
	public ClientServices getClientServices() {
        if (this.client == null) {
            String comMethod = getCommunicationMethod();
            System.out.println("Communication method: " + comMethod);
            System.out
                    .println("KumbangConfigurator: Initializing configuration client");
            try {
                if (comMethod.equals("local")) {
                    ConfigurationClient client = new ConfigurationClient(
                            getCommunicationMethod());
                    this.client = new ClientServices(client);
                } else if (comMethod.equals("rmi")) {
                    ConfigurationClient client = new ConfigurationClient(
                            getCommunicationMethod(), getServerLocation());
                    this.client = new ClientServices(client);
                }

                this.serverAvailable = true;
                return this.client;
            } catch (RemoteException re) {
                System.out.println("Cannot initialize connection "
                        + "to configuration server:\n" + re);
                this.serverAvailable = false;
                return null;
            }
        } else
            return this.client;
    }
	
	protected String getServerLocation() {
		return Activator.getDefault().getServerLocation();
	}

	protected String getCommunicationMethod() {
		return Activator.getDefault().getCommunicationMethod();
	}

	public boolean isServerAvailable() {
		// Server hasn't been contacted yet
		if (this.client == null) {
			this.getClientServices();
			return serverAvailable;
		} else {
			return serverAvailable;
		}
	
	}
	
	public void openKumbangPerspective() {
		IWorkbench workbench = PlatformUI.getWorkbench();
		IWorkbenchWindow window = workbench.getActiveWorkbenchWindow();
		IWorkbenchPage page = window.getActivePage();
		IPerspectiveDescriptor perspective = workbench.getPerspectiveRegistry().findPerspectiveWithId("KumbangTools.perspective1");
		page.setPerspective(perspective);
	}
	
	// Listener related methods
	
    public void addConfigurationModificationListener(ConfigurationModificationListener listener) {
        if (modificationListeners.contains(listener))
        	return; // don't add duplicates
        
    	modificationListeners.add(listener);
    }
    
    public void removeConfigurationModificationListener(ConfigurationModificationListener listener) {
    	modificationListeners.remove(listener);
    }
    
    public void addOpenCloseListener(ConfigurationOpenedClosedListener l) {
    	openCloseListeners.add(l);
    }
    
    public void removeOpenCloseListener(ConfigurationOpenedClosedListener l) {
    	openCloseListeners.remove(l);
    }
    
    public void addSaveListener(ConfigurationSaveListener l) {
    	saveListeners.add(l);
    }
    
    public void removeSaveListener(ConfigurationSaveListener l) {
    	saveListeners.remove(l);
    }
    
    
    public void fireOpenCloseEvent(String eventType) {
    	ConfigurationOpenedClosedEvent e = new ConfigurationOpenedClosedEvent(eventType);
    	synchronized (openCloseListeners) {
    		Iterator it = openCloseListeners.iterator();
    		while (it.hasNext()) {
    			ConfigurationOpenedClosedListener l =
    				(ConfigurationOpenedClosedListener)it.next();
    			l.handleOpenCloseEvent(e);
    		}
    	}
    }
    
    public void fireConfigurationModifiedEvent(ClientTask task) {
    	ConfigurationModifiedEvent e = new ConfigurationModifiedEvent(task);
    	synchronized (modificationListeners) {
    		Iterator it = modificationListeners.iterator();
    		while (it.hasNext()) {
    			ConfigurationModificationListener listener 
    			= (ConfigurationModificationListener)it.next();
    			listener.configurationModified(e);
    		}
    	}
    }
    
    public void fireSaveEvent(File saveFile) {
    	SaveEvent e = new SaveEvent(saveFile);
    	synchronized (saveListeners) {
    		Iterator it = saveListeners.iterator();
    		while (it.hasNext()) {
				ConfigurationSaveListener listener = (ConfigurationSaveListener) it.next();
				listener.handleSaveEvent(e);
			}

		}
    }
    
    
	
}
