/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.diagrams;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kumbang.configurator.client.plugin.diagrams.BindingElement;
import kumbang.configurator.client.plugin.diagrams.BoxedElement;
import kumbang.configurator.client.plugin.diagrams.ComponentElement;
import kumbang.configurator.client.plugin.diagrams.DiagramElement;
import kumbang.configurator.client.plugin.diagrams.DiagramLevel;
import kumbang.configurator.client.plugin.diagrams.InterfaceElement;
import kumbang.configurator.client.plugin.diagrams.PartElement;
import kumbang.core.configuration.instance.PartInstance;

public class ComponentDiagramLevel extends DiagramLevel {
    
    /** what is the distance of root component from diagram border */
    public static final int ROOT_MARGIN = 20;
    
    /** what is the initial x-distance of child components from root border */
    public static final int INNER_MARGIN_X = 100;
    
    /** what is the initial y-distance of child components from root border */
    public static final int INNER_MARGIN_Y = 70;
    
    /** what is the minimum distance of child components from root border */
    public static final int INNER_MARGIN_MIN = 30;
    
    /** what is the distance of two child components in x-direction */
    public static final int COMPONENT_MARGIN_X = 150;
    
    /** what is the distance of two child components in y-direction */
    public static final int COMPONENT_MARGIN_Y = 70;
    
    /** what is the distance of two interfaces from each other */
    public static final int INTERFACE_MARGIN = 5;
    
    /** what is the minimum width of root */
    public static final int MIN_ROOT_WIDTH = 200;
    
    /** what is the minimum height of root */
    public static final int MIN_ROOT_HEIGHT = 200;
    
    /** root component of this component diagram level */
    private ComponentElement rootElement;
    
   
    /** what are the min/max coordinates of this root element */
    private int minRootX;
    private int minRootY;
    private int maxRootX;
    private int maxRootY;
    
    /** what are the current min/max coordinates that contain a child element */
    private int minChildX;
    private int minChildY;
    private int maxChildX;
    private int maxChildY;
    
    /** what are the corresponding element that are at min/max positions */
    private BoxedElement minXChild;
    private BoxedElement minYChild;
    private BoxedElement maxXChild;
    private BoxedElement maxYChild;
    
    
    /** Creates a new instance of ComponentDiagramLevel */
    public ComponentDiagramLevel(ComponentElement rootElement) {
        this(rootElement, null);
    }
    
    /** 
     * Creates a new instance of ComponentDiagramLevel 
     */
    public ComponentDiagramLevel(ComponentElement rootElement, 
                                 ComponentDiagramLevel parentLevel) {
        super(parentLevel);
        this.rootElement = rootElement;
        rootElement.setLevel(this);
    }
    
    public void addChildElement(DiagramElement element) {
        if (element instanceof ComponentElement
                || element instanceof PartElement 
                || element instanceof InterfaceElement
                || element instanceof BindingElement) {
            super.addChildElement(element);
        }
    }
        
    public DiagramElement getRootElement() {
        return rootElement;
    }
    
    public ComponentElement getRootComponentElement() {
        return rootElement;
    }
    
    public void initLayout() {
        resetChildBoundaries();
        initChildLayout();
        initRootLayout();
    }
    
    public int getPreferredHeight() {
        return this.maxRootY - this.minRootY + 2 * ROOT_MARGIN;
    }
    
    public int getPreferredWidth() {
        return this.maxRootX - this.minRootX + 2 * ROOT_MARGIN +
            // this is for ensuring that interface names fit in
            rootElement.getLongestEasternName() * WIDTH_PER_LETTER;
    }

    
    protected void updateLayoutForElement(BoxedElement element) {
        if (!element.isVisible()) {
            return;
        }
        if (updateBoundariesForChild(element)) {
            updateBoundariesForRoot();
        }
    }
    
    //------------------------------------------------------------------
    //------------------- PRIVATE METHODS ------------------------------
    //------------------------------------------------------------------
    

    /**
     * This method initialises the layout of child elements in this level.
     * Note that this method is called only when level layout is initalised,
     * not afterwards, since user can move components around.
     * Starts component layout from location 
     * (ROOT_MARGIN + INNER_MARGIN, ROOT_MARGIN + INNER_MARGIN) and
     * proceeds down towards the corner. Components are laid out so that
     * components in the same part form one column.
     */
    private void initChildLayout() {        
        //initialise counter for component positions
        int positionX = ROOT_MARGIN + INNER_MARGIN_X;
        int positionY = ROOT_MARGIN + INNER_MARGIN_Y;
        
        Iterator parts = rootElement.getComponentInstance().getParts();
        while (parts.hasNext()) {
            // do layout for next column (components in the same part)
            PartInstance part = (PartInstance) parts.next();
            Iterator columnElements = getChildElementsWithName(part.getName());
            // start y positioning from the beginning (first row)
            positionY = ROOT_MARGIN + INNER_MARGIN_Y;
            // advance x positioning to accommodate the width of the column
            int maxWidth = getMaxWidthOfChildElementsWithName(part.getName());
            positionX += maxWidth / 2;
            while (columnElements.hasNext()) {
                DiagramElement elem = (DiagramElement)columnElements.next();
                if (!(elem instanceof BoxedElement)) {
                    continue;
                }
                // do layout for one element in the column 
                BoxedElement e = (BoxedElement)elem;
                positionY += e.getHeight() / 2;
                e.setLocation(positionX, positionY);
                updateBoundariesForChild(e);
                positionY += e.getHeight() / 2;
                positionY += COMPONENT_MARGIN_Y;
            }
            positionX += maxWidth / 2;
            positionX += COMPONENT_MARGIN_X;
        }
    }
    
    private void initRootLayout() {
        int rootWidth = maxChildX - minChildX + 2 * INNER_MARGIN_X;
        int rootHeight = maxChildY - minChildY + 2 * INNER_MARGIN_Y;
        if (rootWidth < MIN_ROOT_WIDTH) {
            rootWidth = MIN_ROOT_WIDTH;
        }
        if (rootHeight < MIN_ROOT_HEIGHT) {
            rootHeight = MIN_ROOT_HEIGHT;
        }
        this.minRootX = ROOT_MARGIN;
        this.minRootY = ROOT_MARGIN;
        this.maxRootX = minRootX + rootWidth;
        this.maxRootY = minRootY + rootHeight;
        //System.out.println("Inited root boundaries for "+this+" ("
        //    +minRootX+","+minRootY+") and ("+maxRootX+","+maxRootY+")");
        rootElement.setWidth(rootWidth);
        rootElement.setHeight(rootHeight);
        rootElement.setLocation(ROOT_MARGIN + rootElement.getWidth() / 2, 
                                ROOT_MARGIN + rootElement.getHeight() / 2);
        //System.out.println("Inited root "+rootElement+" to width="+rootWidth+
        //                    ", height="+rootHeight);
    }
    
    private boolean updateBoundariesForChild(BoxedElement element) {
        boolean updated = false;
        int x = element.getLocationX();
        int y = element.getLocationY();
        int halfWidth = element.getWidth() / 2;
        int halfHeight = element.getHeight() / 2;
        if (x - halfWidth < minChildX) {
            minChildX = x - halfWidth;
            minXChild = element;
            updated = true;
        }
        if (x + halfWidth > maxChildX) {
            maxChildX = x + halfWidth;
            maxXChild = element;
            updated = true;
        }
        if (y - halfHeight < minChildY) {
            minChildY = y - halfHeight;
            minYChild = element;
            updated = true;
        }
        if (y + halfHeight > maxChildY) {
            maxChildY = y + halfHeight;
            maxYChild = element;
            updated = true;
        }
        return updated;
    }
    
    private void updateBoundariesForRoot() {
        boolean updated = false;
        if (minChildX < minRootX + INNER_MARGIN_MIN) {
            minRootX = minChildX - INNER_MARGIN_MIN;
            updated = true;
        }
        if (maxChildX > maxRootX - INNER_MARGIN_MIN) {
            maxRootX = maxChildX + INNER_MARGIN_MIN;
            updated = true;
        }
        if (minChildY < minRootY + INNER_MARGIN_MIN) {
            minRootY = minChildY - INNER_MARGIN_MIN;
            updated = true;
        }
        if (maxChildY > maxRootY - INNER_MARGIN_MIN) {
            maxRootY = maxChildY + INNER_MARGIN_MIN;
            updated = true;
        }
        if (updated) {
            //System.out.println("Updated root boundaries for "+this+" ("
            //    +minRootX+","+minRootY+") and ("+maxRootX+","+maxRootY+")");
            int width = Math.abs(maxRootX) - Math.abs(minRootX);
            int height = Math.abs(maxRootY) - Math.abs(minRootY);
            rootElement.setWidth(width);
            rootElement.setHeight(height);
            rootElement.setLocation(minRootX + width / 2, 
                                    minRootY + (maxRootY - minRootY) / 2);
            //System.out.println("Updated root "+rootElement+" to width="+
            //                width+", height="+height);
        }
        // if diagram boundaries have moved to negative values,
        // we must translate all elements towards the center
        // this is a stupid hack to enable showing elements that
        // have negative coordinates!!
        int minDiagramX = minRootX - ROOT_MARGIN;
        int minDiagramY = minRootY - ROOT_MARGIN;
        int transX = (minDiagramX < 0) ? -minDiagramX : 0;
        int transY = (minDiagramY < 0) ? -minDiagramY : 0;
        if (transX != 0 || transY != 0) {
            translateAllElements(transX, transY);
        }
        
    }
    
    
    private void resetChildBoundaries() {
        minChildX = Integer.MAX_VALUE;
        minChildY = Integer.MAX_VALUE;
        maxChildX = Integer.MIN_VALUE;
        maxChildY = Integer.MIN_VALUE;
        minXChild = null;
        minYChild = null;
        maxXChild = null;
        maxYChild = null;
    }
    
    private Iterator getChildElementsWithName(String name) {
        List<DiagramElement> elements = new LinkedList<DiagramElement>();
        Iterator allElements = getChildElements();
        while (allElements.hasNext()) {
            DiagramElement elem = (DiagramElement)allElements.next();
            if (elem.getName().equals(name)) {
                elements.add(elem);
            }
        }
        return elements.iterator();
    }
    
    private int getMaxWidthOfChildElementsWithName(String name) {
        int maxWidth = 0;
        Iterator allElements = getChildElements();
        while (allElements.hasNext()) {
            DiagramElement elem = (DiagramElement) allElements.next();
            if (elem.getName().equals(name) && elem instanceof BoxedElement) {
                int width = ((BoxedElement) elem).getWidth();
                if (width > maxWidth) {
                    maxWidth = width;
                }
            }
        }
        return maxWidth;
    }


    private void translateAllElements(int deltaX, int deltaY) {
        //System.out.println("Translating elements in "+this+" with ("+deltaX+","+deltaY+")");
        Iterator it = getChildElements();
        while (it.hasNext()) {
            DiagramElement elem = (DiagramElement) it.next();
            if (elem instanceof ComponentElement || 
                elem instanceof PartElement) {
                // don't move interfaces, they are moved along the components
                ((BoxedElement) elem).addToLocation(deltaX, deltaY);
            }
        }
        rootElement.addToLocation(deltaX, deltaY);
        this.minChildX += deltaX;
        this.maxChildX += deltaX;
        this.minChildY += deltaY;
        this.maxChildY += deltaY;
        this.minRootX += deltaX;
        this.maxRootX += deltaX;
        this.minRootY += deltaY;
        this.maxRootY += deltaY;
    }
    
    public boolean equals(Object obj) {
    	if (!(obj instanceof ComponentDiagramLevel)) return false;
    	
    	ComponentDiagramLevel l = (ComponentDiagramLevel)obj;
    	if (l.toString().equals(this.toString()))
    		return true;
    	else
    		return false;
    }
    
}

