/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.diagrams;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kumbang.configurator.Activator;
import kumbang.configurator.client.plugin.Colors;
import kumbang.configurator.client.plugin.IKumbangPlugin;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.configurator.client.plugin.diagrams.ComposableElement;
import kumbang.configurator.client.plugin.diagrams.DiagramLevel;
import kumbang.configurator.client.plugin.diagrams.InterfaceElement;
import kumbang.core.configuration.instance.ComponentInstance;


public class ComponentElement extends ComposableElement {
    
    /** Contains all interface elements that belong to this component */
    private List interfaces;
    
    /** All interfaces that have eastern orientation */
    private List easternInterfaces;
    
    /** All interfaces that have western orientation */
    private List westernInterfaces;
    
    /** 
     * This variable measures the longest name of eastern interfaces.
     * This is used for evaluating space needed besides this component.
     */
    private int longestEasternName = 0;
    
    /** 
     * This variable measures the longest name of western interfaces.
     * This is used for evaluating space needed besides this component.
     */
    private int longestWesternName = 0;
    
    public static final int DEFAULT_ROOT_WIDTH = 380;
    public static final int DEFAULT_ROOT_HEIGHT = 380;
    
    public static final int MARGIN_INTERFACE = 20;
    
    /** 
     * Creates a new instance of ComponentElement that is not the root 
     * component in the component diagram level.
     * @param instance component instance that this element represents
     */
    public ComponentElement(ComponentInstance instance) {
        this(instance, false);
    }
    
    /** 
     * Creates a new instance of ComponentElement. 
     */
    public ComponentElement(ComponentInstance instance, boolean isRoot) {
        super(instance, (isRoot ? getPlugin().getColorRegistry().get(Colors.WHITE) : getPlugin().getColorRegistry().get(Colors.COMPONENT)), isRoot);
        if (isRoot) {
            setWidth(DEFAULT_ROOT_WIDTH);
            setHeight(DEFAULT_ROOT_HEIGHT);
        }
        this.interfaces = new LinkedList();
        this.easternInterfaces = new LinkedList();
        this.westernInterfaces = new LinkedList();
    }
    
    /**
     * Returns the component instance this element represents
     * @return ComponentInstance 
     */
    public ComponentInstance getComponentInstance() {
        return (ComponentInstance)getInstance();
    }
    
    /**
     * Adds interface to this component element. Adjusts also size,
     * so that all interfaces fit in.
     * @param intf interface element to be added
     */
    protected void addInterface(InterfaceElement intf) {
        interfaces.add(intf);
        if (intf.getOrientation() == InterfaceElement.ORIENTATION_EAST) {
            easternInterfaces.add(intf);
            if (intf.toString().length() > longestEasternName) {
                longestEasternName = intf.toString().length();
            }
        }
        else {
            westernInterfaces.add(intf);
            if (intf.toString().length() > longestWesternName) {
                longestWesternName = intf.toString().length();
            }
        }
        adjustSize();
        doInterfaceLayout();
    }
    
    protected void removeInterface(InterfaceElement intf) {
        if (interfaces.contains(intf)) {
            interfaces.remove(intf);
            if (intf.getOrientation() == InterfaceElement.ORIENTATION_EAST) {
                easternInterfaces.remove(intf);
            }
            else {
                westernInterfaces.remove(intf);
            }
            adjustSize();
            doInterfaceLayout();
        }
    }
    
    public Iterator getInterfaces() {
        return interfaces.iterator();
    }
    
    public boolean areInterfacesEditable() {
        return getComponentInstance().getComponentType().hasInterfaces();
    }
    
    public int getNumberOfInterfaces() {
        return interfaces.size();
    }
    
    protected int getLongestEasternName() {
        return this.longestEasternName;
    }
    
    private int getNumberOfEasternInterfaces() {
        return easternInterfaces.size();
    }
    
    private int getNumberOfWesternInterfaces() {
        return westernInterfaces.size();
    }
    
    private void doInterfaceLayout() {
        int centerX = getLocationX();
        int centerY = getLocationY();
        int interfaceX, interfaceY;
        
        /* lay out eastern interfaces */
        Iterator it = easternInterfaces.iterator();
        interfaceX = centerX + getWidth()/2;
        int interfaceHeight = (easternInterfaces.size()-1) * 
                                (InterfaceElement.SIZE + MARGIN_INTERFACE);
        int distanceFromCorner = (getHeight() - interfaceHeight) / 2;
        while (it.hasNext()) {
            InterfaceElement inter = (InterfaceElement)it.next();
            interfaceY = centerY - getHeight()/2 + distanceFromCorner;
            inter.setLocation(interfaceX, interfaceY);
            distanceFromCorner += MARGIN_INTERFACE + InterfaceElement.SIZE;
        }
        
        /* lay out western interfaces */
        it = westernInterfaces.iterator();
        interfaceX = centerX - getWidth()/2;
        interfaceHeight = (westernInterfaces.size()-1) * 
                                (InterfaceElement.SIZE + MARGIN_INTERFACE);
        distanceFromCorner = (getHeight() - interfaceHeight) / 2;
        while (it.hasNext()) {
            InterfaceElement inter = (InterfaceElement)it.next();
            interfaceY = centerY - getHeight()/2 + distanceFromCorner;
            inter.setLocation(interfaceX, interfaceY);
            distanceFromCorner += MARGIN_INTERFACE + InterfaceElement.SIZE;
        }
    }
    
    protected void setLocation(int centerX, int centerY) {
        super.setLocation(centerX, centerY);
        doInterfaceLayout();
    }
    
    protected void removeFromDiagram() {
        List removedInterfaces = new LinkedList();
        Iterator it = getInterfaces();
        while (it.hasNext()) {
            removedInterfaces.add(it.next());
        }
        interfaces.clear();
        easternInterfaces.clear();
        westernInterfaces.clear();
        it = removedInterfaces.iterator();
        while (it.hasNext()) {
            InterfaceElement interf = (InterfaceElement)it.next();
            interf.removeFromDiagram();
        }
    }
    
    /**
     * Adjusts the size of this component element. Since component
     * element should include all interface elements along its sides,
     * the size of the element should be big enough. This also adjusts
     * width to accommodate long names for interfaces.
     */
    private void adjustSize() {
        int requiredHeight; 
        if (getNumberOfEasternInterfaces() > getNumberOfWesternInterfaces()) {
            requiredHeight = MARGIN_INTERFACE + 
                getNumberOfEasternInterfaces() * (InterfaceElement.SIZE + MARGIN_INTERFACE);
        }
        else {
            requiredHeight = MARGIN_INTERFACE + 
                getNumberOfWesternInterfaces() * (InterfaceElement.SIZE + MARGIN_INTERFACE);
        }
        if (getHeight() < requiredHeight) {
            setHeight(requiredHeight);
        }
        // if interface name is long enough, and there are interfaces
        // at both sides, put some extra into the width
        int width = longestWesternName * DiagramLevel.WIDTH_PER_LETTER + MARGIN_INTERFACE;
        if (width > getWidth() && getNumberOfEasternInterfaces() > 0) {
            setWidth(width);
        }
    }
    
    
}

