/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.diagrams;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import kumbang.configurator.client.plugin.diagrams.BoxedElement;
import kumbang.configurator.client.plugin.diagrams.ComposableElement;
import kumbang.configurator.client.plugin.diagrams.DiagramElement;
import kumbang.configurator.client.plugin.diagrams.InstantiatedElement;
import kumbang.configurator.client.plugin.diagrams.PartElement;
import kumbang.core.configuration.instance.Instance;


public abstract class DiagramLevel {
    
    public static final int DISTANCE_Y_ELEMENT_TO_POTENTIAL = 100;
    
    public static final int WIDTH_PER_LETTER = 6;
    
    private DiagramLevel parentLevel;
    
    private List childLevels;
    
    private List childElements;
    
    /** Creates a new instance of DiagramLevel */
    public DiagramLevel() {
        this(null);
    }
    
    /** Creates a new instance of DiagramLevel */
    public DiagramLevel(DiagramLevel parentLevel) {
        this.parentLevel = parentLevel;
        this.childLevels = new LinkedList();
        this.childElements = new LinkedList();
    }
    
    public DiagramLevel getParent() {
        return this.parentLevel;
    }
    
    public void addChildElement(DiagramElement element) {
        childElements.add(element);
        element.setLevel(this);
        // updates for layout
    }
    
    public void removeChildElement(DiagramElement element) {
        childElements.remove(element);
        element.setLevel(null);
        // updates for layout
    }
    
    public void removeAllChildElements() {
        Iterator it = getChildElements();
        while (it.hasNext()) {
            DiagramElement elem = (DiagramElement)it.next();
            elem.setLevel(null);
        }
        childElements.clear();
    }
    
    public void moveChildElement(BoxedElement element, int locationX, int locationY) {
        if (element.getLevel() != this) {
            return;
        }
        element.setLocation(locationX, locationY);
        // updates for layout
        updateLayoutForElement(element);
    }
    
    public void updateLayoutForAddedElement(ComposableElement element, PartElement part) {
        if (element.getLevel() != this) {
            return;
        }
        // move the potential element and add the new element to its place
        int locX = part.getLocationX();
        int locY = part.getLocationY();
        element.setLocation(locX, locY);
        part.setLocation(locX, locY + DISTANCE_Y_ELEMENT_TO_POTENTIAL);
        // updates for layout
        updateLayoutForElement(element);
        updateLayoutForElement(part);
    }
    
    public void updateLayoutForRemovedElement(ComposableElement element, PartElement part) {
        if (element.getLevel() != this || part.getLevel() != this) {
            return;
        }
        part.setLocation(element.getLocationX(), element.getLocationY());
        // updates for layout
        updateLayoutForElement(part);
    }
    
    public Iterator getChildElements() {
        return childElements.iterator();
    }
    
    public ListIterator getChildElementsFromEnd() {
        return childElements.listIterator(childElements.size());
    }
    
    public int getNumberOfChildElements() {
        return childElements.size();
    }
    
    public boolean isEmpty() {
        return childElements.isEmpty();
    }
    
    public boolean containsChildElementForInstance(Instance instance) {
        Iterator it = childElements.iterator();
        while (it.hasNext()) {
            DiagramElement element = (DiagramElement)it.next();
            if (element instanceof InstantiatedElement) {
                return ((InstantiatedElement)element).getInstance() == instance;
            }
        }
        return false;
    }
    
    public DiagramElement getChildElementForInstance(Instance instance) {
        Iterator it = childElements.iterator();
        while (it.hasNext()) {
            DiagramElement element = (DiagramElement)it.next();
            if (element instanceof InstantiatedElement &&
                    ((InstantiatedElement)element).getInstance().equals(instance)) {
                return element;
            }
        }
        return null;
    }
    
    
    
    public void addChildLevel(DiagramLevel child) {
        childLevels.add(child);
    }
    
    public void removeChildLevel(DiagramLevel child) {
        childLevels.remove(child);
    }
    
    public Iterator getChildLevels() {
        return childLevels.iterator();
    }
    
    public DiagramLevel getChildLevelForRootElement(DiagramElement element) {
        Iterator it = getChildLevels();
        while (it.hasNext()) {
            DiagramLevel level = (DiagramLevel)it.next();
            if (level.getRootElement().equals(element)) {
                return level;
            }
        }
        return null;
    }
    
    public DiagramLevel getChildLevelForRootInstance(Instance instance) {
        Iterator it = getChildLevels();
        while (it.hasNext()) {
            DiagramLevel level = (DiagramLevel)it.next();
            DiagramElement element = level.getRootElement();
            if (element instanceof InstantiatedElement &&
                    ((InstantiatedElement)element).getInstance().equals(instance)) {
                return level;
            }
        }
        return null;
    }
    
    public abstract DiagramElement getRootElement();
    
    public abstract void initLayout();
    
    protected abstract void updateLayoutForElement(BoxedElement element);
    
    public abstract int getPreferredWidth();
    
    public abstract int getPreferredHeight();
    
    public String toString() {
        return "level (root "+getRootElement().toString()+")";
    }
}

