/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.diagrams;

//import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.draw2d.Graphics;

import kumbang.configurator.client.plugin.Colors;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.configurator.client.plugin.diagrams.BoxedElement;
import kumbang.configurator.client.plugin.diagrams.ConnectableElement;
import kumbang.configurator.client.plugin.diagrams.ConnectorElement;
import kumbang.configurator.client.plugin.diagrams.DiagramFilter;
import kumbang.configurator.client.plugin.diagrams.FilteredElement;
import kumbang.configurator.client.plugin.diagrams.InstantiatedElement;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.InterfaceInstance;

public class InterfaceElement extends BoxedElement implements
		ConnectableElement, InstantiatedElement, FilteredElement {

	public static final int SIZE = 16;

	public static final int ORIENTATION_EAST = 1;

	public static final int ORIENTATION_WEST = 2;

//	public static final int[] X_ARROW_LEFT = { -4, 5, -4 };
//
//	public static final int[] Y_ARROW_LEFT = { -5, 0, 5 };
	
	public static final int[] ARROW_LEFT = { -4, -5, 5, 0, -4, 5 };

//	public static final int[] X_ARROW_RIGHT = { -5, 4, 4 };
//
//	public static final int[] Y_ARROW_RIGHT = { 0, -5, 5 };
	
	public static final int[] ARROW_RIGHT = { -5, 0, 4, -5, 4, 5 };

	// private ComponentElement parentElement;

	/** The interface instance that this element represents */
	private InterfaceInstance instance;

	/**
	 * Tells where this interface is located in the component. Possible values:
	 * ORIENTATION_EAST and ORIENTATION_WEST.
	 */
	private int orientation;

	/**
	 * Tells whether this interface is required or provided. Value TRUE means
	 * this is provided and value FALSE means this is required.
	 */
	private boolean provided;

	/**
	 * Tells whether this interface is on the border of a component. This means
	 * that it is an interface in a component that is the root component in the
	 * current diagram level.
	 */
	private boolean isBorderInterface;

	/** A list of all connectors this interface has */
	private List connectors;

	/**
	 * Creates a new instance of InterfaceInstance
	 * 
	 * @param instance
	 *            interface instance this element represents
	 */
	public InterfaceElement(InterfaceInstance instance,
			boolean isBorderInterface) {
		super(instance.getName(), SIZE, SIZE, getPlugin().getColorRegistry().get(Colors.INTERFACE));
		this.instance = instance;
		this.provided = instance.isProvided();
		if (this.provided) {
			this.orientation = ORIENTATION_WEST;
		} else {
			this.orientation = ORIENTATION_EAST;
		}
		this.isBorderInterface = isBorderInterface;
		this.connectors = new ArrayList();
	}

	/**
	 * Creates a new instance of InterfaceInstance
	 * 
	 * @param instance
	 *            interface instance this element represents
	 */
	public InterfaceElement(InterfaceInstance instance) {
		this(instance, false);
	}

	/**
	 * Returns the instance this element represents
	 * 
	 * @return Instance
	 */
	public Instance getInstance() {
		return this.instance;
	}

	/**
	 * Returns the interface instance this element represents
	 * 
	 * @return InterfaceInstance
	 */
	public InterfaceInstance getInterfaceInstance() {
		return this.instance;
	}

	public Iterator getConnectors() {
		return this.connectors.iterator();
	}

	/**
	 * Connects a connector from this interface. Simply adds the connector to
	 * this interface.
	 */
	public void connectWithConnector(ConnectorElement con) {
		connectors.add(con);
	}

	/**
	 * Disconnect a connector from this interface. This is called by connector's
	 * removeFromDiagram() method. Simply removes the connector from this
	 * interface.
	 */
	public void disconnectFromConnector(ConnectorElement con) {
		connectors.remove(con);
	}

	public int getConnectorLocationX() {
		if ((orientation == ORIENTATION_WEST && !isBorderInterface)
				|| (orientation == ORIENTATION_EAST && isBorderInterface)) {
			return super.getLocationX() - super.getWidth() / 2;
		} else {
			return super.getLocationX() + super.getWidth() / 2;
		}
	}

	public int getConnectorLocationY() {
		return super.getLocationY();
	}

	public int getOrientation() {
		return this.orientation;
	}

	public void paint(Graphics g, DiagramFilter filter) {
		boolean showsName = (filter != null ? filter.showsNames() : true);
		boolean showsType = (filter != null ? filter.showsTypes() : true);
		super.paint(g);
		g.translate(getLocationX(), getLocationY());
		g.drawString(getDescription(showsName, showsType),
				-super.getWidth() / 2, -super.getHeight() / 2 - 12);
		if ((orientation == ORIENTATION_EAST && !provided)
				|| (orientation == ORIENTATION_WEST && provided)) {
			g.setBackgroundColor(getForegroundColor());
			g.fillPolygon(ARROW_LEFT);
		} else {
			g.setBackgroundColor(getForegroundColor());
			g.fillPolygon(ARROW_RIGHT);
		}
		g.translate(-getLocationX(), -getLocationY());
	}

	public void paint(Graphics g) {
		paint(g, null);
	}

	protected void setHeight(int height) {
		super.setHeight(SIZE);
	}

	protected void setWidth(int width) {
		super.setWidth(SIZE);
	}

	protected void setLocation(int centerX, int centerY) {
		super.setLocation(centerX, centerY);
		/* to move an interface, we have to move its connectors too */
		Iterator it = connectors.iterator();
		while (it.hasNext()) {
			ConnectorElement con = (ConnectorElement) it.next();
			con.setLocation(this, getConnectorLocationX(),
					getConnectorLocationY());
		}
	}

	protected void removeFromDiagram() {
		List removedConnectors = new LinkedList();
		Iterator it = getConnectors();
		while (it.hasNext()) {
			removedConnectors.add((ConnectorElement) it.next());
		}
		it = removedConnectors.iterator();
		while (it.hasNext()) {
			ConnectorElement con = (ConnectorElement) it.next();
			con.removeFromDiagram();
		}
	}

	public boolean equals(Object o) {
		if (o instanceof InterfaceElement) {
			if (instance == ((InterfaceElement) o).getInstance()) {
				return true;
			}
		}
		return false;
	}

	public String toString() {
		return instance.getName() + ":" + instance.getType().getName();
	}

	private String getDescription(boolean name, boolean type) {
		String ret = "";
		if (!name && !type) {
			return ret;
		}
		if (name) {
			ret += instance.getDisplayName();
		}
		ret += ":";
		if (type) {
			ret += instance.getType().getDisplayName();
		}
		return ret;
	}

}
