/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.diagrams;

import org.eclipse.draw2d.Graphics;
import org.eclipse.swt.graphics.Image;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kumbang.configurator.client.plugin.Colors;
import kumbang.configurator.client.plugin.Icons;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.configurator.client.plugin.diagrams.BoxedElement;
import kumbang.configurator.client.plugin.diagrams.ComposableElement;
import kumbang.configurator.client.plugin.diagrams.DiagramFilter;
import kumbang.configurator.client.plugin.diagrams.FilteredElement;
import kumbang.configurator.client.plugin.diagrams.InstantiatedElement;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.PartInstance;

public class PartElement extends BoxedElement implements InstantiatedElement, FilteredElement {
    
    public static final int DEFAULT_WIDTH = 80;
    public static final int DEFAULT_HEIGHT = 40;
    
    public static final int ICONS_IN_ROW = 5;
    
    public static final int MAX_ICONS = 9;
    
    private PartInstance instance;
    
    private Image newIcon;
    
    private Image clickIcon;
    
    private List instanceElements;
    
    /** Creates a new instance of PartElement */
    public PartElement(PartInstance instance) {
        this(instance, DEFAULT_WIDTH, DEFAULT_HEIGHT);
    }
    
    public PartElement(PartInstance instance, int width, int height) {
        super(instance.getName(), width, height, getPlugin().getColorRegistry().get(Colors.GRAY), false);
        this.instance = instance;
        this.clickIcon = getPlugin().getImageRegistry().get(Icons.SELECT);
        if (instance.getParent() instanceof ComponentInstance) {
            newIcon = getPlugin().getImageRegistry().get(Icons.NEW_COMPONENT);
        } else {
            newIcon = getPlugin().getImageRegistry().get(Icons.NEW_FEATURE);
        }
        this.instanceElements = new LinkedList();
        update();
    }
    
    public Instance getInstance() {
        return this.instance;
    }
    
    public PartInstance getPartInstance() {
        return this.instance;
    }
    
    public Iterator getInstanceElements() {
        return instanceElements.iterator();
    }
    
    public void addInstanceElement(ComposableElement element) {
        instanceElements.add(element);
    }
    
    public void removeInstanceElement(ComposableElement element) {
        instanceElements.remove(element);
    }
    
    /**
     * This method is called every time when potential element should
     * check the instance and update its information.
     */
    public void update() {
        // set visible, based on whether one can add more instances
        setVisible(!instance.isFull());
        // adjust width and height
        if (getNumberOfIcons() < ICONS_IN_ROW) {
            setWidth(getNumberOfIcons() * Icons.ICON_SIZE);
        } else {
            setWidth(ICONS_IN_ROW * Icons.ICON_SIZE);
        }
        setHeight(Icons.ICON_SIZE * (getNumberOfIcons() / ICONS_IN_ROW + 1));
    }
    
    
    
    public boolean isInside(int xCoordinate, int yCoordinate) {
        if (!isVisible()) {
            return false;
        }
        return super.isInside(xCoordinate, yCoordinate);
    }
    
    public void setSelected(boolean selected) {
        //new elements cannot be selected
        super.setSelected(false);
    }
    
    
    public boolean equals(Object o) {
        if (o instanceof PartElement) {
            if (((PartElement) o).getInstance().equals(instance)) {
                return true;
            }
        }
        return false;
    }
    
    public String toString() {
        return instance.toString();
    }
    
    public String getDescription() {
        return instance.getDescription();
    }
    
    public void paint(Graphics g, DiagramFilter filter) {
        if (filter != null && !filter.showsPotential()) {
            return;
        }
        if (!isVisible()) {
            return;
        }
        super.paint(g);
        g.translate(getLocationX(), getLocationY());
        int numberOfIcons = getNumberOfIcons();
        int locX = - getWidth() / 2;
        int locY = - getHeight() / 2;
        String str = instance.getName() + "[" + instance.getLowerBound();
        if (instance.getLowerBound() != instance.getUpperBound()) {
            str += "-" + instance.getUpperBound() + "]";
        } else {
            str += "]";
        }
        g.drawString(str, locX, locY-12);
        for (int i = 0; i < numberOfIcons && i < MAX_ICONS; i++) {
            g.drawImage(newIcon, locX, locY);
            if (i % ICONS_IN_ROW != 0 || i == 0) {
                locX += Icons.ICON_SIZE;
            } else {
                locX = - getWidth() / 2;
                locY += Icons.ICON_SIZE;
            }
        }
        if (numberOfIcons > MAX_ICONS) {
            g.fillRectangle(locX, locY,  1, 1);
            g.fillRectangle(locX + 2, locY,  1, 1);
            g.fillRectangle(locX + 4, locY,  1, 1);
        }
        if (locX != -getWidth() / 2) { // start of new row
            locX = -getWidth() / 2;
            locY += Icons.ICON_SIZE;
        }
        g.drawImage(clickIcon, locX, locY);
        g.translate(-getLocationX(), -getLocationY());
    }
    
    public void paint(Graphics g) {
        paint(g, null);
    }
    
//    protected void setNewIcon (ImageIcon icon) {
//        newIcon = icon;
//    }
    
    private int getNumberOfIcons() {
        return instance.getUpperBound() - instance.getNumberOfInstances();
    }
    
    protected void removeFromDiagram() {
        instanceElements.clear();
    }
}
