/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.dialogs;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kumbang.configurator.Activator;
import kumbang.configurator.client.core.TaskHandler;
import kumbang.configurator.client.core.exception.ClientException;
import kumbang.configurator.client.plugin.IKumbangPlugin;
import kumbang.configurator.client.plugin.Icons;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.core.configuration.core.ConfigurationException;
import kumbang.core.configuration.instance.PartInstance;
import kumbang.core.configuration.task.AddComposableTask;
import kumbang.core.model.type.AttributeDefinition;
import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.InterfaceDefinition;
import kumbang.core.model.type.InterfaceType;
import kumbang.core.model.type.Name;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.IMessageProvider;
import org.eclipse.jface.dialogs.TitleAreaDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StackLayout;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;


public class AddComposableDialog extends TitleAreaDialog {
	
	// Parent shell
	private Shell parent;
	
	// PartInstance of the TreeNode
	private PartInstance partInstance;
	
    // List of PossibleItemTypes (i.e. the Types that can be selected
    // from the dropdown menu
	private List possibleTypes;
	
	// Object that contains the type drop down menu and its state
	private TypeSelectionItem typeSelectionItem;
	
	private boolean hasInterfaces;
	
	private boolean hasAttributes;
    
    private List interfaceList;
    
    private List attributeList;
    
    private final StackLayout stackInterfaceLayout = new StackLayout();
    
    private final StackLayout stackAttributeLayout = new StackLayout();
    
    private Composite stackInterfaceComposite;
    
    private Composite stackAttributeComposite;
	
	
	public AddComposableDialog(Shell parentShell, PartInstance p) {
		super(parentShell);
		this.parent = parentShell;
		this.partInstance = p;
		this.possibleTypes = new LinkedList();
        this.interfaceList = new LinkedList();
        this.attributeList = new LinkedList();
		
		Iterator it = partInstance.getPossibleTypes();
		while (it.hasNext()) {
			PossibleTypeItem item = new PossibleTypeItem((ComposableType) it.next());
			if (!hasAttributes && item.getNumberOfAttributes() > 0) {
				this.hasAttributes = true;
			}
			if (!hasInterfaces && item.getNumberOfInterfaces() > 0) {
				this.hasInterfaces = true;
			}
			this.possibleTypes.add(item);
		}
	}
	
	
	/////// CONTENT CREATION METHODS
	
	/**
	 * Create the top area of the dialog. The message differs if the part is full
	 */
	protected Control createContents(Composite parent) {
		Control contents = super.createContents(parent);
		setTitle("Add a new " + partInstance.getComposedKind());
		setMessage("Please select the type of the component, interfaces to be included,\n" +
				" and initial values of attributes.",	IMessageProvider.INFORMATION);
		
		return contents;
	}
	
	/** 
	 * Create the actual dialog
	 */
	protected Control createDialogArea(Composite parent) {
		// Sets the proper margins etc
		
		Composite composite = (Composite) super.createDialogArea(parent);
		composite.getShell().setMinimumSize(DialogLayout.DIALOG_MIN_WIDTH, DialogLayout.DIALOG_MIN_HEIGHT);
		Composite innerComposite = new Composite(composite, SWT.NONE);
		innerComposite.setLayout(DialogLayout.innerCompositeLayout());
		GridData gridData = new GridData();
		gridData.horizontalAlignment = SWT.FILL;
		innerComposite.setLayoutData(gridData);
		
        // create the properties-group
		createNameTypeGroup(innerComposite);
		
		// Shows the logo
		Label imageLabel = new Label(innerComposite, SWT.BORDER);
		imageLabel.setLayoutData(DialogLayout.logoLayoutData());
		imageLabel.setImage(getPlugin().getImageRegistry().get(Icons.LOGO));
		
        
        // create the interface-group
        createInterfaceGroup(innerComposite);
        
        // creates the attribute-group
        createAttributeGroup(innerComposite);
        
        // updates the stacklayout (shows the correct groups)
        updateLayouts();
        
		return composite;
	}
	
	/**
     * Creates the area for that holds the feature's name and type selection
     * box.
     * 
     * @param parent
     *            Parent composite
     * @return the group container that was created
     */
    private Group createNameTypeGroup(Composite parent) {
		
		Group area = new Group(parent, SWT.SHADOW_ETCHED_IN);
		area.setText("Properties");
		GridLayout gl = new GridLayout(2,false);
		area.setLayout(gl);
		
		Label lbl = new Label(area, SWT.NONE);
		lbl.setText("Name:");
		lbl = new Label(area, SWT.NONE);
		lbl.setText(partInstance.getName());
		
		typeSelectionItem = new TypeSelectionItem(this.partInstance, area);
		
		lbl = new Label(area, SWT.NONE);
		lbl.setText("Parent name:");
		
		GridData gd = new GridData();
		gd.verticalIndent = 1;
		lbl.setLayoutData(gd);
		lbl = new Label(area, SWT.NONE);
        gd = new GridData();
        gd.verticalIndent = 1;
		lbl.setLayoutData(gd);
		lbl.setText(partInstance.getParent().getDisplayName());

		lbl = new Label(area, SWT.NONE);
		gd = new GridData();
		gd.verticalIndent = 2;
		lbl.setLayoutData(gd);
		lbl.setText("Parent type:");
		lbl = new Label(area, SWT.NONE);

		gd = new GridData();
		gd.verticalIndent = 2;
		lbl.setLayoutData(gd);

		lbl.setText(partInstance.getParent().getType().getDisplayName());
		return area;
	}
    
    /**
     * Creates the interface-groups that are stored inside the stacklayout
     * 
     * @param parent
     *            Parent composite
     */
    private void createInterfaceGroup(Composite parent) {
        if(this.hasInterfaces) {
            this.stackInterfaceComposite = new Composite(parent, SWT.NONE);
            stackInterfaceComposite.setLayout(stackInterfaceLayout);
            
            for (Iterator it = possibleTypes.iterator(); it.hasNext();) {
                
                Group iGroup = new Group(stackInterfaceComposite, SWT.NONE);
                iGroup.setLayout(new GridLayout());
                iGroup.setText("Interfaces");
                
                PossibleTypeItem item = (PossibleTypeItem)it.next();
                Group g = new Group(iGroup, SWT.SHADOW_ETCHED_OUT);
                g.setLayout(new GridLayout(2,false));
                g.setText(item.getType().toString());
                Iterator it2 = item.getInterfaceSelections();
                int i = 0;
                while (it2.hasNext()) {
                    InterfaceSelectionItem isi = (InterfaceSelectionItem) it2.next();
                    isi.deploy(g);
                    i++;
                }
                if (i > 0)
                    this.interfaceList.add(iGroup);
            }
        }
    }
    
    /**
     * Creates the attribute-groups that are stored inside the stacklayout
     * 
     * @param parent
     *            Parent composite
     */
    private void createAttributeGroup(Composite parent) {
        if (this.hasAttributes) {
            this.stackAttributeComposite = new Composite(parent, SWT.NONE);
            stackAttributeComposite.setLayout(stackAttributeLayout);
            
            for (Iterator it = possibleTypes.iterator(); it.hasNext();) {
                
                Group aGroup = new Group(stackAttributeComposite, SWT.NONE);
                aGroup.setLayout(new GridLayout(2,false));
                aGroup.setText("Attributes");
                
                PossibleTypeItem item = (PossibleTypeItem)it.next();
                Group g = new Group(aGroup, SWT.SHADOW_ETCHED_OUT);
                g.setLayout(new GridLayout(2, false));
                g.setText(item.getType().toString());
                Iterator it2 = item.getAttributeSelections();
                int i = 0;
                while (it2.hasNext()) {
                    AttributeSelectionItem asi = (AttributeSelectionItem) it2.next();
                    asi.deploy(g);
                    i++;
                }
                if (i > 0)
                    this.attributeList.add(aGroup);
            }
        }
    }
    
	
	protected void createButtonsForButtonBar(Composite parent) {
		createButton(parent, IDialogConstants.OK_ID, IDialogConstants.OK_LABEL,
				true);
		createButton(parent, IDialogConstants.CANCEL_ID,
				IDialogConstants.CANCEL_LABEL, false);
	}
	
	/**
	 * Set the title
	 * 
	 */
	protected void configureShell(Shell newShell) {
		super.configureShell(newShell);
		newShell.setText("Add " + partInstance.getComposedKind());
	}
	
	
	////////////////////////////////////////////// EVENT HANDLING METHODS
	/**
	 * Handles a button press.
	 * 
	 * @param buttonId
	 *            the ID of the pressed button
	 */
	protected void buttonPressed(int buttonId) {
		
		if (buttonId == IDialogConstants.OK_ID) {
			
			// TODO: Disabling the OK button would probably be better..
			if (partInstance.isFull()) {
				errorMessage("Error", "You have selected a part that has already\n" +
						" the maximum number of " + partInstance.getComposedKind() + "s.\n" +
						"Thus you cannot add a new " + partInstance.getComposedKind() + " to this part.\n" +
				"Please select another part.");
				return;
			}
			
			
			AddComposableTask task = getTask(getPlugin().getHandler().getClientServices().getTaskHandler());
			if (task == null) {
				close();
			}
			
			try {
				getPlugin().getHandler().fireConfigurationModifiedEvent(getPlugin().getHandler().getClientServices().handleTask(task));
				close();
				
			} catch (ConfigurationException ce) {
				errorMessage("Error when adding " + partInstance.getComposedKind(),
						ce.getMessage());
			} catch (ClientException cle) {
				errorMessage("Error when adding " + partInstance.getComposedKind(), 
						cle.getMessage());
			}
			
		} else {
			// Cancel-button has been pressed
			close();
		}
	}
	
	/**
	 * Closes the dialog
	 */
	public boolean close() {
		return super.close();
	}
	
	/**
     * Creates the task from the selections in the dialog
	 */
	private AddComposableTask getTask(TaskHandler taskHandler) {
		AddComposableTask task = null;
		
		// Fetch type and total instance to be created
		PossibleTypeItem item = typeSelectionItem.getSelectedItem();
		ComposableType type = item.getType();
		task = taskHandler.createAddComposableTask(this.partInstance, type);
		
		// If creating a component, create subtasks for interfaces
		if (type instanceof ComponentType) {
			Iterator it = item.getInterfaceSelections();
			while (it.hasNext()) {
				InterfaceSelectionItem  isi = (InterfaceSelectionItem ) it.next();
				if (isi.isSelected()) {
					taskHandler.addInterfaceToAddComposableTask(task,
							isi.getInterface(), isi.getInterfaceType());
				}
			}
		}
		
		/* Create subtasks for setting attribute avalues */
		Iterator it = item.getAttributeSelections();
		while (it.hasNext()) {
			AttributeSelectionItem asi = (AttributeSelectionItem) it.next();
			AttributeDefinition attr = asi.getAttribute();
			task.setAttributeValue(attr.getName(), asi.getSelectedValue());
		}
		return task;
	}
    
    /**
     * Updates the stacklayouts (interface and attribute) according to
     * the selected type.
     *
     */
    public void updateLayouts() {
        
        if (AddComposableDialog.this.hasInterfaces) {
            boolean isActiveI = false;
            for (int i = 0; i < AddComposableDialog.this.interfaceList.size(); i++) {
                Group g = (Group) ((Group)AddComposableDialog.this.interfaceList.get(i)).getChildren()[0];
                if (g.getText().equals(typeSelectionItem.combo.getText())) {
                    stackInterfaceLayout.topControl = (Control) AddComposableDialog.this.interfaceList.get(i);
                    stackInterfaceComposite.layout();
                    isActiveI = true;
                }
            }
            if (isActiveI == false) {
                stackInterfaceLayout.topControl = null;
                stackInterfaceComposite.layout();
            }
        }
        if (AddComposableDialog.this.hasAttributes) {
            boolean isActiveA = false;
            for (int i = 0; i < AddComposableDialog.this.attributeList.size(); i++) {
                Group g = (Group) ((Group)AddComposableDialog.this.attributeList.get(i)).getChildren()[0];
                if (g.getText().equals(typeSelectionItem.combo.getText())) {
                    System.out.println(g.getText());
                    stackAttributeLayout.topControl = (Control) AddComposableDialog.this.attributeList.get(i);
                    stackAttributeComposite.layout();
                    isActiveA = true;
                }
            }
            if (isActiveA == false) {
                System.out.println("topcontrol is null");
                stackAttributeLayout.topControl = null;
                stackAttributeComposite.layout();
            }
        }
    }
    
	
	private void errorMessage(String text, String message) {
		MessageBox msgbox = new MessageBox(this.parent, SWT.ICON_ERROR | SWT.OK);
		msgbox.setText(text);
		msgbox.setMessage(message);
		msgbox.open();
	}
	
	//////////////////////////////////////// HELPER CLASSES
	
	class PossibleTypeItem {
		private ComposableType type;
		private List interfaceSelections;
		private List attributeSelections;
		
		public PossibleTypeItem(ComposableType type) {
			this.type = type;
			this.interfaceSelections = new LinkedList();
			this.attributeSelections = new LinkedList();
			initInterfaceSelections();
			initAttributeSelections();
		}
		
		public String toString() {
			return type.getDisplayName();
		}
		public ComposableType getType() {
			return type;
		}
		public Iterator getInterfaceSelections() {
			return interfaceSelections.iterator();
		}
		public Iterator getAttributeSelections() {
			return attributeSelections.iterator();
		}
		public int getNumberOfAttributes() {
			return attributeSelections.size();
		}
		public int getNumberOfInterfaces() {
			return interfaceSelections.size();
		}
		
		private void initInterfaceSelections() {
			if (type instanceof ComponentType) {
				Iterator it = ((ComponentType) type).getInterfaces();
				while (it.hasNext()) {
					InterfaceDefinition intDef = (InterfaceDefinition) it.next();
					interfaceSelections.add(new InterfaceSelectionItem(intDef));
				}
			}
		}
		private void initAttributeSelections() {
			/*
			 * It is good to include the inherited attributes as well. -toasikai 13.9.2006
			 */
			Iterator it = type.getAttributes(); //getDirectAttributes();
			while (it.hasNext()) {
				AttributeDefinition attrDef = (AttributeDefinition) it.next();
				attributeSelections.add(new AttributeSelectionItem(attrDef));
			}
		}
	}
	
	
	class InterfaceSelectionItem {
		private Button check;
		private Combo combo;
		private InterfaceDefinition interfaceDef;
		
		public InterfaceSelectionItem(InterfaceDefinition iDef) {
			this.interfaceDef = iDef;
		}
		
		public void deploy(Composite parent) {
			this.check = new Button(parent, SWT.CHECK);
			this.check.setText(interfaceDef.getDisplayName() + " (" +
					(interfaceDef.isProvides() ? "provided" : "required") + ")");
			this.combo = new Combo(parent, SWT.DROP_DOWN | SWT.READ_ONLY);
			
			Iterator it = interfaceDef.getPossibleTypes();
			while (it.hasNext()) {
				InterfaceType type = (InterfaceType) it.next();
				this.combo.add(type.toString());
			}
			this.combo.setText(combo.getItem(0));
			this.check.setSelection(true);
		}
		
		public boolean isSelected() {
			return check.getSelection();
		}
		
		public InterfaceDefinition getInterface() {
			return this.interfaceDef;
		}
		
		public InterfaceType getInterfaceType() {
			String str = combo.getText();
			Iterator it = interfaceDef.getPossibleTypes();
			while (it.hasNext()) {
				InterfaceType type = (InterfaceType) it.next();
				if (type.toString().equals(str))
					return type;
			}
			return null;
		}
	}
	
	class AttributeSelectionItem {
		private Combo combo;
		private AttributeDefinition attributeDef;
		private HashMap valueMap;
		
		public AttributeSelectionItem(AttributeDefinition aDef) {
			this.attributeDef = aDef;
			this.valueMap = new HashMap();
		}
		
		public void deploy(Composite parent) {
            
            new Label(parent, SWT.NONE).setText(attributeDef.toString());
            
			this.combo = new Combo(parent, SWT.DROP_DOWN | SWT.READ_ONLY);
			Iterator it = attributeDef.getType().getValues();
			while (it.hasNext()) {
				Name value = (Name)it.next();
				valueMap.put(value.getDisplayName(), value.getName());
				combo.add(value.getDisplayName());
			}
			this.combo.setText(combo.getItem(0));
		}
		
		public AttributeDefinition getAttribute() {
			return this.attributeDef;
		}
		
		public Combo getCombo() {
			return this.combo;
		}
		
		public String getSelectedValue() {
			String displayName = combo.getText();
			return (String)valueMap.get(displayName);
		}
	}
	
	public IKumbangPlugin getPlugin() {
		return Activator.getDefault();
	}
	
	class TypeSelectionItem {
		// combo box for selecting part
		private Combo combo;
		
		// label for showing part name
		private Label label;
		
		
		
		/**
		 * Creates comboboxes based on the given attribute definition
		 * 
		 * @param attribute
		 *            attribute instance
		 * @param composite
		 *            parent composite
		 */
		public TypeSelectionItem(PartInstance instance,
				Composite composite) {
			
			this.label = new Label(composite, SWT.BEGINNING | SWT.CENTER);
			label.setText("Type:");
			GridData gd = new GridData(GridData.HORIZONTAL_ALIGN_BEGINNING);
			this.combo = new Combo(composite, SWT.READ_ONLY);
			this.combo.setLayoutData(gd);
			this.combo.addSelectionListener(new SelectionListener() {
				// TODO: Also disable the components within the group
				public void widgetSelected(SelectionEvent e) {
                    updateLayouts();
				}
				public void widgetDefaultSelected(SelectionEvent e) {
					// TODO Auto-generated method stub
				}
			});
			
			Iterator it = instance.getPossibleTypes();
			while (it.hasNext()) {
				ComposableType c = (ComposableType)it.next();
				this.combo.add(c.getDisplayName());
			}
			if (this.combo.getText().equals(""))
				this.combo.setText(combo.getItem(0));
		}
		
		public PossibleTypeItem getSelectedItem() {
			String str = combo.getText();
			Iterator it = AddComposableDialog.this.possibleTypes.iterator();
			while (it.hasNext()) {
				PossibleTypeItem item = (PossibleTypeItem) it.next();
				if (item.getType().getDisplayName().equals(str))
					return item;
			}
			return null;
		}
	}
}
