/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.dialogs;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kumbang.configurator.client.core.TaskHandler;
import kumbang.configurator.client.core.exception.ClientException;
import kumbang.configurator.client.plugin.Icons;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.configurator.client.plugin.PluginHandler;
import kumbang.core.configuration.core.ConfigurationException;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.PartInstance;
import kumbang.core.configuration.task.EditInterfacesTask;
import kumbang.core.model.type.AttributeDefinition;
import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.InterfaceDefinition;
import kumbang.core.model.type.InterfaceType;
import kumbang.core.model.type.Name;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.TitleAreaDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;

public class EditInterfacesDialog extends TitleAreaDialog {
	
	private Shell parent;
	
	private ComponentInstance componentInstance;
	
	private List selectionItems;
	
	private boolean hasInterfaces;
	
	private Group interfaceGroup;

	// Used to track if the user tries
	// to remove a non-optional interface
	private boolean removeError;
	
	
	public EditInterfacesDialog(Shell parentShell, ComponentInstance cInstance) {
		super(parentShell);
		
		this.parent = parentShell;
		this.componentInstance = cInstance;
			
		this.selectionItems = new LinkedList();
		
		this.removeError = false;
		
		PossibleTypeItem item = new PossibleTypeItem((ComposableType) cInstance.getComponentType());
	
	
		Iterator it = this.componentInstance.getComponentType().getInterfaces();
		while (it.hasNext()) {
			InterfaceDefinition intDef = (InterfaceDefinition) it.next();
			selectionItems.add(new InterfaceSelectionItem(intDef, this.componentInstance));
		}
		
		if (!hasInterfaces && item.getNumberOfInterfaces() > 0) {
			this.hasInterfaces = true;
		}
	}
	
/////// CONTENT CREATION METHODS
	
	/**
	 * Create the top area of the dialog. The message differs if the part is full
	 */
	protected Control createContents(Composite parent) {
		Control contents = super.createContents(parent);
		setTitle("Edit interfaces");
		setMessage("Edit interfaces on component " + this.componentInstance.toString());
		
		return contents;
	}
	
	protected Control createDialogArea(Composite parent) {
		Composite composite = (Composite) super.createDialogArea(parent);
		Composite innerComposite = new Composite(composite, SWT.NONE);
		innerComposite.setLayout(DialogLayout.innerCompositeLayout());
		GridData gridData = new GridData();
		gridData.horizontalAlignment = SWT.FILL;
		innerComposite.setLayoutData(gridData);
		
		createNameTypeGroup(innerComposite);
				
		// Shows the logo
		
		Label imageLabel = new Label(innerComposite, SWT.BORDER);
		imageLabel.setLayoutData(DialogLayout.logoLayoutData());
		imageLabel.setImage(PluginHandler.getActivator().getImageRegistry().get(Icons.LOGO));
		
		if (hasInterfaces) {
			// Create area for setting interfaces on / off
			interfaceGroup = createInterfaceGroup(innerComposite);
		}
			
		return composite;
	}
	
	
	
	protected void createButtonsForButtonBar(Composite parent) {
		createButton(parent, IDialogConstants.OK_ID, IDialogConstants.OK_LABEL,
				true);
		createButton(parent, IDialogConstants.CANCEL_ID,
				IDialogConstants.CANCEL_LABEL, false);
	}
	
	/**
	 * Set the title
	 * 
	 */
	protected void configureShell(Shell newShell) {
		super.configureShell(newShell);
		newShell.setText("Edit interfaces");
	}
	
/////// EVENT HANDLING METHODS
	/**
     * Handles a button press.
     * 
     * @param buttonId
     *            the ID of the pressed button
     */

    protected void buttonPressed(int buttonId) {
		
		if (buttonId == IDialogConstants.OK_ID) {
			
			// TODO: Disabling the OK button would probably be better..
			
			EditInterfacesTask task = getTask(PluginHandler.getInstance().getClientServices().getTaskHandler());
			if (task == null) {
				close();
			}
			
			try {
				PluginHandler.getInstance().fireConfigurationModifiedEvent(PluginHandler.getInstance().getClientServices().handleTask(task));
				if (removeError)
					MessageDialog.openError(this.parent, "Interface cannot be removed", "The interface you tried to remove is not optional and therefore wasn't removed.");
				close();
				
			} catch (ConfigurationException ce) {
				MessageDialog.openError(this.parent, ce.getMessage(),
						"Error when editing " + this.componentInstance.getInstanceKind());
			} catch (ClientException cle) {
				MessageDialog.openError(this.parent, cle.getMessage(),
						"Error when editing " + this.componentInstance.getInstanceKind());
			}
			
		} else {
			// Cancel-button has been pressed
			close();
		}
	}
	
	public EditInterfacesTask getTask(TaskHandler taskHandler) {
		EditInterfacesTask task = null;
		
		task = new EditInterfacesTask(this.componentInstance, true);
		Iterator it = selectionItems.iterator();
		while (it.hasNext()) {
			InterfaceSelectionItem isi = (InterfaceSelectionItem) it.next();
			InterfaceDefinition def = isi.getInterface();
			InterfaceType type = isi.getInterfaceType();
			InterfaceInstance inst = isi.getInterfaceInstance();
			if (isi.isSelected() && inst == null) {
				task.addInterface(def, type);
				//taskHandler.addInterfaceToEditComposableTask(task, def, type);
			} else if (!isi.isSelected() && inst != null) {
				task.removeInterface(inst);
				if (!inst.isOptional()) {
					this.removeError = true;
				}
			}
		}
		
		return task;
	}
	
	private Group createNameTypeGroup(Composite parent) {
		Group area = new Group(parent, SWT.SHADOW_ETCHED_IN);
		area.setText("Properties");
		GridLayout gl = new GridLayout(2,false);
		gl.marginWidth = DialogLayout.MARGIN_INNER;
		area.setLayout(gl);
		
		Label lbl = new Label(area, SWT.NONE);
		lbl.setText("Name:");
		lbl = new Label(area, SWT.NONE);
		lbl.setText(this.componentInstance.getName());
		
		
		lbl = new Label(area, SWT.NONE);
		lbl.setText("Parent name:");
		lbl = new Label(area, SWT.NONE);
		lbl.setText(componentInstance.getParent().getDisplayName());
		lbl = new Label(area, SWT.NONE);
		lbl.setText("Parent type:");
		lbl = new Label(area, SWT.NONE);
		lbl.setText(componentInstance.getParent().getType().getDisplayName());
		
		return area;
	}
	
	private Group createInterfaceGroup(Composite parent) {
		Group area = new Group(parent, SWT.SHADOW_ETCHED_IN);
		area.setText("Interfaces");
		GridLayout gl = new GridLayout(1,false);
		gl.marginLeft = DialogLayout.MARGIN_INNER;
		area.setLayout(gl);
		
		PossibleTypeItem item = new PossibleTypeItem(this.componentInstance.getComponentType());
		createInterfaceGroupForType(area, item);
		
		return area;
	}
	
	private Group createInterfaceGroupForType(Composite parent, PossibleTypeItem item) {
		Group area = new Group(parent, SWT.SHADOW_ETCHED_OUT);
		GridLayout gl = new GridLayout(2, false);
		gl.marginLeft = DialogLayout.MARGIN_INNER;
		area.setLayout(gl);
		area.setText(item.getType().toString());
		
		Iterator it = selectionItems.iterator();
		while (it.hasNext()) {
			InterfaceSelectionItem isi = (InterfaceSelectionItem) it.next();
			isi.deploy(area);
		}
		return area;
	}
	
	
	
//	-------------- ITEMS FOR SELECTING INTERFACES ---------------------------
	
///// HELPER CLASSES
	
	class PossibleTypeItem {
		private ComposableType type;
		private List interfaceSelections;
		private List attributeSelections;
		
		public PossibleTypeItem(ComposableType type) {
			this.type = type;
			this.interfaceSelections = new LinkedList();
			this.attributeSelections = new LinkedList();
			initInterfaceSelections();
			initAttributeSelections();
		}
		
		public String toString() {
			return type.getDisplayName();
		}
		public ComposableType getType() {
			return type;
		}
		public Iterator getInterfaceSelections() {
			return interfaceSelections.iterator();
		}
		public Iterator getAttributeSelections() {
			return attributeSelections.iterator();
		}
		public int getNumberOfAttributes() {
			return attributeSelections.size();
		}
		public int getNumberOfInterfaces() {
			return selectionItems.size();
		}
		
		private void initInterfaceSelections() {
			if (type instanceof ComponentType) {
				Iterator it = ((ComponentType) type).getInterfaces();
				while (it.hasNext()) {
					InterfaceDefinition intDef = (InterfaceDefinition) it.next();
				}
			}
		}
        
		private void initAttributeSelections() {
			Iterator it = type.getDirectAttributes();
			while (it.hasNext()) {
				AttributeDefinition attrDef = (AttributeDefinition) it.next();
				attributeSelections.add(new AttributeSelectionItem(attrDef));
			}
		}
	}
	
	
	class InterfaceSelectionItem {
		private Button check;
		private Combo combo;
		private InterfaceDefinition interfaceDef;
		private InterfaceInstance instance;
		
		public InterfaceSelectionItem(InterfaceDefinition iDef,
                ComponentInstance parent) {
            this.interfaceDef = iDef;
            this.instance = parent.getInterface(interfaceDef);
        }
		
		public void deploy(Composite parent) {
			this.check = new Button(parent, SWT.CHECK);
			this.check.setText(interfaceDef.getDisplayName() + " (" +
					(interfaceDef.isProvides() ? "provided" : "required") + ")");
			this.combo = new Combo(parent, SWT.DROP_DOWN);
			
			Iterator it = interfaceDef.getPossibleTypes();
			while (it.hasNext()) {
				InterfaceType type = (InterfaceType) it.next();
				this.combo.add(type.toString());
			}
			this.combo.setText(combo.getItem(0));
			this.check.setSelection(instance != null);
		}
		
		 
		
		public boolean isSelected() {
			return check.getSelection();
		}
		
		public InterfaceDefinition getInterface() {
			return this.interfaceDef;
		}
		
		public InterfaceInstance getInterfaceInstance() {
			return this.instance;
		}
		
		public InterfaceType getInterfaceType() {
			String str = combo.getText();
			Iterator it = interfaceDef.getPossibleTypes();
			while (it.hasNext()) {
				InterfaceType type = (InterfaceType) it.next();
				if (type.toString().equals(str))
					return type;
			}
			return null;
		}
	}
	
	class AttributeSelectionItem {
		private Combo combo;
		private AttributeDefinition attributeDef;
		private HashMap valueMap;
		
		public AttributeSelectionItem(AttributeDefinition aDef) {
			this.attributeDef = aDef;
			this.valueMap = new HashMap();
		}
		
		public void deploy(Composite parent) {
			this.combo = new Combo(parent, SWT.DROP_DOWN);
			combo.setText(attributeDef.toString());
			Iterator it = attributeDef.getType().getValues();
			while (it.hasNext()) {
				Name value = (Name)it.next();
				valueMap.put(value.getDisplayName(), value.getName());
				combo.add(value.getDisplayName());
			}
			this.combo.setText(combo.getItem(0));
		}
		
		public AttributeDefinition getAttribute() {
			return this.attributeDef;
		}
		
		public Combo getCombo() {
			return this.combo;
		}
		
		public String getSelectedValue() {
			String displayName = combo.getText();
			return (String)valueMap.get(displayName);
		}
		
	}
    
	class TypeSelectionItem {
		// combo box for selecting part
		private Combo combo;
		
		// label for showing part name
		private Label label;
		
		/**
		 * Creates comboboxes based on the given attribute definition
		 * 
		 * @param attribute
		 *            attribute instance
		 * @param composite
		 *            parent composite
		 */
		public TypeSelectionItem(PartInstance instance,
				Composite composite) {

			this.label = new Label(composite, SWT.CENTER | SWT.BOTTOM);
			label.setText("Type");
			GridData gd = new GridData(GridData.HORIZONTAL_ALIGN_CENTER);
			label.setLayoutData(gd);
			this.combo = new Combo(composite, SWT.READ_ONLY);
			this.combo.setLayoutData(gd);
			
			Iterator it = instance.getPossibleTypes();
			while (it.hasNext()) {
				ComposableType c = (ComposableType)it.next();
				this.combo.add(c.getDisplayName());
			}
			if (this.combo.getText().equals(""))
				this.combo.setText(combo.getItem(0));
		}
		
	}
}
