/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.views;


import java.util.List;

import kumbang.configurator.Activator;
import kumbang.configurator.client.core.ClientServices;
import kumbang.configurator.client.core.TaskHandler;
import kumbang.configurator.client.core.exception.ClientException;
import kumbang.configurator.client.plugin.Colors;
import kumbang.configurator.client.plugin.IKumbangPlugin;
import kumbang.configurator.client.plugin.Icons;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.configurator.client.plugin.PluginHandler;
import kumbang.configurator.client.plugin.dialogs.AddComposableDialog;
import kumbang.configurator.client.plugin.dialogs.EditAttributesDialog;
import kumbang.configurator.client.plugin.dialogs.EditInterfacesDialog;
import kumbang.configurator.client.plugin.dialogs.SelectPartDialog;
import kumbang.configurator.client.plugin.event.ConfigurationModifiedEvent;
import kumbang.configurator.client.plugin.event.ConfigurationModificationListener;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedEvent;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedListener;
import kumbang.configurator.client.plugin.preferences.PreferenceConstants;
import kumbang.configurator.client.plugin.tree.AttributeNode;
import kumbang.configurator.client.plugin.tree.BindingNode;
import kumbang.configurator.client.plugin.tree.ComponentNode;
import kumbang.configurator.client.plugin.tree.FeatureNode;
import kumbang.configurator.client.plugin.tree.InterfaceNode;
import kumbang.configurator.client.plugin.tree.PartNode;
import kumbang.configurator.client.plugin.tree.TreeNode;
import kumbang.core.configuration.core.ConfigurationException;
import kumbang.core.configuration.instance.BindingInstance;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.ComposableInstance;
import kumbang.core.configuration.instance.FeatureInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.PartInstance;
import kumbang.core.configuration.task.ClientTask;
import kumbang.core.configuration.task.RemoveInstanceTask;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.resource.ColorRegistry;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TreeItem;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.PageBook;
import org.eclipse.ui.part.ViewPart;

/**
 * Base class for the Feature and Component Views, does most
 * of the View-related initialization and updating stuff
 * 
 * @author jlemmett
 * 
 */



public class KumbangTreeView extends ViewPart 
implements ConfigurationModificationListener, ConfigurationOpenedClosedListener	 {
	
	public final static String DIAGRAMVIEW_ID = "KumbangTools.view4";
	
	protected ClientServices client;
	
	public TreeViewer treeViewer;
	private Label lblRefresh;
	private Composite fNoModelComposite;
	private Composite fTreeComposite;
	protected Composite fParent; // from private to protected
	protected Composite fNoTreeComposite;
	protected Label fLblNoTree;
	protected TaskHandler taskHandler;
	
	// PageBook contains the different composites
	private PageBook fPagebook;
	
	protected static final int PAGE_EMPTY = 0;
	protected static final int PAGE_TREE = 1;
	protected static final int PAGE_NOTREE = 2;
	
	private Action actionEditAttributes;
	private Action actionAddFeatures;
	private Action actionAddComponents;
	private Action actionEditInterfaces;
	private Action actionRemoveFeatures;
	private Action actionRemoveComponents;
	private Action actionRemoveInterfaces;
	private Action actionRemoveBindings;
	private Action actionShowInDiagram;
	private Action actionDisableParts;
	
	private Color background;

	private boolean bindingFound;
	
	// Listener for preference changes
	protected IPropertyChangeListener listener;
	
	public KumbangTreeView() {
		super();
		
		
	}
	
	
	public void createPartControl(Composite parent) {
		ColorRegistry cr = getPlugin().getColorRegistry();
		background = cr.get(Colors.WHITE);
		
		fParent = parent;
		fPagebook = new PageBook(parent, SWT.NONE);
		
		// Create the normal view
		createEnabledView(fPagebook);
		
		// Create the disabled view
		createDisabledView(fPagebook);
		
		// Create no component/feature tree view
		createNoTreeView(fPagebook);
		
		// Show the empty page initially
		showPage(PAGE_EMPTY);
		
		// check if a model is already open and react by showing the tree page
		updateView();
		
		// Start listening for configuration init/close events and conf mod events
		registerListeners();
		
		// 
		this.listener =	new IPropertyChangeListener() {
		    /* Setup the listener for changes in the preferences */
		    public void propertyChange(PropertyChangeEvent e) {
		        if (e.getProperty().equals(PreferenceConstants.P_COMMUNICATIONMETHOD)) {
		                treeViewer.setSelection(null);
		            }
		        }
		    };


		    getPlugin().getPreferenceStore().addPropertyChangeListener(listener);

	}

	protected void registerListeners() {
		getPlugin().getHandler().addOpenCloseListener(this);
	}
	
	
	/**
	 * Shows either the empty page or the tree page in the view
	 * 
	 * @param page
	 */
	protected void showPage(int page) {
		if (page == PAGE_EMPTY) {
			fPagebook.showPage(fNoModelComposite);
		} else if (page == PAGE_TREE) {
			fPagebook.showPage(fTreeComposite);
		} else if (page == PAGE_NOTREE) {
			fPagebook.showPage(fNoTreeComposite);
		}
	}
	
	/**
	 * Checks if a model has been opened and switches the view from the empty
	 * nag page to the actual tree. The extending classes must implement this
	 * because the procedure differs a bit depending on the tree
	 */
	protected void updateView() {
	}
	
	/**
	 * A public method for others (e.g. the initialization wizard) to call when
	 * they've done something that requires updating here
	 * 
	 * (Is this necessary, would it be enough to make updateView() public ?
	 */
	public void refresh() {
		updateView();
	}
	
	/**
	 * Setup the tree view
	 * 
	 * @param parent
	 */
	private void createEnabledView(Composite parent) {
		fTreeComposite = new Composite(parent, SWT.WRAP);
		fTreeComposite.setBackground(background);
		GridLayout layout = new GridLayout();
		fTreeComposite.setLayout(layout);
		treeViewer = new TreeViewer(fTreeComposite, SWT.SINGLE);
		treeViewer.setAutoExpandLevel(TreeViewer.ALL_LEVELS);
		GridData gd = new GridData(GridData.FILL_BOTH);
		treeViewer.getTree().setLayoutData(gd);
		
		getSite().setSelectionProvider(treeViewer);
		
	}
	
	// TODO: Move the hard coded text to some utility class..
	/**
	 * Setup the empty page
	 * 
	 * @param parent
	 */
	private void createDisabledView(Composite parent) {
		this.fNoModelComposite = new Composite(parent, SWT.WRAP);
		//fNoModelComposite.setBackground(background);
		GridLayout layout = new GridLayout();
		fNoModelComposite.setLayout(layout);
		lblRefresh = new Label(fNoModelComposite, SWT.NULL);
		//lblRefresh.setBackground(background);
		lblRefresh.setText("No model initialized.");
	}
	
	protected void createNoTreeView(Composite parent) {
		this.fNoTreeComposite = new Composite(parent, SWT.WRAP);
		//fNoTreeComposite.setBackground(background);
		fNoTreeComposite.setLayout(new GridLayout());
		fLblNoTree = new Label(fNoTreeComposite, SWT.NULL);
		//fLblNoTree.setBackground(background);
		// Set the label text in the implementing classes
	}
	
	/**
	 * Creates an invisible root node and hooks the actual feature tree to it
	 * Overridden in the child classes with the specific root. This allows
	 * for the actual root to show in the tree, otherwise it wouldn't be
	 * visible.
	 * 
	 * @return the root node as the TreeViewers input
	 */
	protected Object initializeTree() {
		return null;
	}
	
	protected void initAndShowTreeViewer() {
		treeViewer.setContentProvider(new KumbangTreeContentProvider());
		treeViewer.setLabelProvider(new KumbangTreeLabelProvider());
		treeViewer.setInput(initializeTree());
		
		makeActions();
		MenuManager menuManager = new MenuManager();
		menuManager.setRemoveAllWhenShown(true);
		
		menuManager.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				// get selection from tree and only add what's required
				Object obj = treeViewer.getTree().getSelection()[0].getData();
				
				// if tree-selection is attribute
				if (obj instanceof AttributeNode) {
					manager.add(actionEditAttributes);
					actionEditAttributes.setText("Edit attributes");
					return;
				}
				
				// check if tree-selection has attributes as children
				List children = ((TreeNode) obj).getChildren();
				boolean objectHasAttributes = false;
				for (int i = 0; i < children.size(); i++) {
					if (children.get(i) instanceof AttributeNode)
						objectHasAttributes = true;
				}
				if (objectHasAttributes) {
					manager.add(actionEditAttributes);
					actionEditAttributes.setText("Edit Attributes of this " + getTreeType());
				}
				
				// if tree-selection is part, have to be checked if the active tree is component or feature tree
				if (obj instanceof PartNode) {
					if (getTreeType().equals("component")) {
						manager.add(actionAddComponents);
						actionAddComponents.setText("Add new component under this part");
					}
					else  {
						manager.add(actionAddFeatures); 
						actionAddFeatures.setText("Add new feature under this part");
					}
					manager.add(actionDisableParts);
					PartNode p = (PartNode) obj;
					PartInstance pInstance = (PartInstance) p.getInstance();
					if (pInstance.getLowerBound() == 0)
						actionDisableParts.setEnabled(true);
//					if (pInstance.hasInstances())
//					actionDisableParts.setEnabled(false);
					else
						actionDisableParts.setEnabled(false);
					
					return;
				}
				
				if (obj instanceof FeatureNode) {
					manager.add(actionAddFeatures);
					actionAddFeatures.setText("Add new feature under this feature");
					manager.add(actionRemoveFeatures);
					
					TreeNode node = (TreeNode) obj;
					FeatureInstance fInstance = (FeatureInstance) node.getInstance();
					if (isRemovableInstance(fInstance) == false)
						actionRemoveFeatures.setEnabled(false);
					else
						actionRemoveFeatures.setEnabled(true);
				}
				
				if (obj instanceof ComponentNode) {
					manager.add(actionAddComponents);
					actionAddComponents.setText("Add new component under this component");
					manager.add(actionEditInterfaces);
					actionEditInterfaces.setText("Edit interfaces of this component");
					manager.add(actionRemoveComponents);
					manager.add(actionShowInDiagram);
					
					TreeNode node = (TreeNode) obj;
					ComponentInstance cInstance = (ComponentInstance) node.getInstance();
					if (cInstance.getParent() == null)
						actionEditInterfaces.setEnabled(false);		
					else
						actionEditInterfaces.setEnabled(true);		
					if (isRemovableInstance(cInstance) == false) 
						actionRemoveComponents.setEnabled(false);
					else
						actionRemoveComponents.setEnabled(true);
				}
				
				if (obj instanceof InterfaceNode) {
					TreeNode node = (TreeNode) obj;
					InterfaceNode iNode = (InterfaceNode) node;
					InterfaceInstance i = (InterfaceInstance) iNode.getInterfaceInstance();
					
					manager.add(actionEditInterfaces);
					actionEditInterfaces.setText("Edit interfaces");
					manager.add(actionRemoveInterfaces);
					manager.add(actionShowInDiagram);
					
					if (!i.isOptional())
						actionRemoveInterfaces.setEnabled(false);
					else
						actionRemoveInterfaces.setEnabled(true);
				}
				
				if (obj instanceof BindingNode) {
					manager.add(actionRemoveBindings);
					manager.add(actionShowInDiagram);
				}
			}
		});
		treeViewer.getTree().setMenu(
				menuManager.createContextMenu(treeViewer.getTree()));
		
		showPage(PAGE_TREE);
		getPlugin().getHandler().addConfigurationModificationListener(this);
	}
	
	
	public void setFocus() {
		fPagebook.setFocus();
	}
	
	
	protected void makeActions() {
		
		// Edit attributes
		actionEditAttributes = new Action() {
			public void run() {
				
				// Give EditAttributeDialog the proper node
				TreeNode node = (TreeNode) treeViewer.getTree().getSelection()[0].getData();
				if (node instanceof AttributeNode) {
					node = (TreeNode) node.getParent();
				}
				
				EditAttributesDialog dlg = new EditAttributesDialog(treeViewer.getControl().getShell(), node);
				
				dlg.open();
			}
		};
		
		
		
		actionEditAttributes.setToolTipText("Edit Attributes");
		actionEditAttributes.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.ATTRIBUTE));
		
		// Add features
		
		actionAddFeatures = new Action() {
			public void run() {
				TreeNode node = (TreeNode)treeViewer.getTree().getSelection()[0].getData();
				if (node instanceof FeatureNode) {
					ComposableInstance c = (ComposableInstance) node.getInstance();
					if (!c.hasParts()){
						errorMessage("Cannot add feature", "There are no parts under this feature, and thus \n" +
						"no features can be added to this feature.");
						return;
					}
					PartInstance p = null;
					SelectPartDialog dlg = new SelectPartDialog(fParent.getShell(), node);
					dlg.open();
					p = dlg.getSelectedPart();
					
					if (p != null) {
						AddComposableDialog addDlg = new AddComposableDialog(fParent.getShell(), p);
						addDlg.open();
					} 
				} 
				if (node instanceof PartNode) {
					PartInstance p = (PartInstance)node.getInstance();
					if (p.isFull()) {
						errorMessage("Cannot add " + p.getComposedKind(), "You have selected a part that has already\n" +
								" the maximum number of " + p.getComposedKind() + "s.\n" +
								"Thus you cannot add a new " + p.getComposedKind() + " to this part.\n" +
						"Please select another part.");
						return;
					}
					if (p != null) {
						AddComposableDialog addDlg = new AddComposableDialog(fParent.getShell(), p);
						addDlg.open();
					}
				}
			}
		};
		actionAddFeatures.setToolTipText("New feature");
		actionAddFeatures.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.FEATURE));
		
		// Add components
		
		actionAddComponents = new Action() {
			public void run() {
				TreeNode node = (TreeNode)treeViewer.getTree().getSelection()[0].getData();
				if (node instanceof ComponentNode) {
					ComposableInstance c = (ComposableInstance) node.getInstance();
					if (!c.hasParts()){
						errorMessage("Cannot add component", "There are no parts under this component, and thus " +
						"no component can be added to this component.");
						return;
					}
					PartInstance p = null;
					SelectPartDialog dlg = new SelectPartDialog(fParent.getShell(), node);
					dlg.open();
					p = dlg.getSelectedPart();
					
					if (p != null) {
						AddComposableDialog addDlg = new AddComposableDialog(fParent.getShell(), p);
						addDlg.open();
					} 
				} 
				if (node instanceof PartNode) {
					PartInstance p = (PartInstance)node.getInstance();
					if (p.isFull()) {
						errorMessage("Cannot add " + p.getComposedKind(), "You have selected a part that has already\n" +
								" the maximum number of " + p.getComposedKind() + "s.\n" +
								"Thus you cannot add a new " + p.getComposedKind() + " to this part.\n" +
						"Please select another part.");
						return;
					}
					if (p != null) {
						AddComposableDialog addDlg = new AddComposableDialog(fParent.getShell(), p);
						addDlg.open();
					}
				}
			}
		};
		
		actionAddComponents.setToolTipText("New component");
		actionAddComponents.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.COMPONENT));
		
		// Edit interfaces
		
		actionEditInterfaces = new Action() {
			public void run() {
				TreeNode node = (TreeNode)treeViewer.getTree().getSelection()[0].getData();
				if (node instanceof ComponentNode) {
					ComponentInstance c = (ComponentInstance) node.getInstance();
					EditInterfacesDialog dlg = new EditInterfacesDialog(fParent.getShell(), c);
					dlg.open();
				} 
				
				if (node instanceof InterfaceNode) {
					InterfaceNode iNode = (InterfaceNode) node;
					InterfaceInstance i = (InterfaceInstance) iNode.getInterfaceInstance();
					
//					System.out.println(i.getName());
					ComponentInstance c = (ComponentInstance) i.getParent();
					EditInterfacesDialog dlg = new EditInterfacesDialog(fParent.getShell(), c);
					dlg.open();
				} 
			}
		};
		
		actionEditInterfaces.setToolTipText("Edit interfaces");
		actionEditInterfaces.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.REQUIRED_INTERFACE));
		
		// Remove features
		
		actionRemoveFeatures = new Action() {
			public void run() {
				
				TreeNode node = (TreeNode) treeViewer.getTree().getSelection()[0].getData();
				
				// Opens a messagedialog for confirming the removing
				
				Shell parent = treeViewer.getControl().getShell();
				MessageBox msgbox = new MessageBox(parent, SWT.ICON_QUESTION | SWT.YES | SWT.NO);
				msgbox.setText("Removing feature");
				msgbox.setMessage("Are you sure you want to remove this feature?\n" + node.getInstance().toString());
				int answer = msgbox.open();
				
				if (answer == SWT.YES)
					removeComposableInstance((FeatureInstance) node.getInstance());
				
			}
		};
		
		
		actionRemoveFeatures.setText("Remove this feature");
		actionRemoveFeatures.setToolTipText("Remove features");
		actionRemoveFeatures.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.TRASH_CAN));
		
//		Remove components
		
		actionRemoveComponents = new Action() {
			public void run() {
				TreeNode node = (TreeNode) treeViewer.getTree().getSelection()[0].getData();
				
				// Opens a messagedialog for confirming the removing
				
				Shell parent = treeViewer.getControl().getShell();
				MessageBox msgbox = new MessageBox(parent, SWT.ICON_QUESTION | SWT.YES | SWT.NO);
				msgbox.setText("Removing component");
				msgbox.setMessage("Are you sure you want to remove this component?\n" + node.getInstance().toString());
				int answer = msgbox.open();
				
				if (answer == SWT.YES)
					removeComposableInstance((ComponentInstance) node.getInstance());
			}
		};
		
		actionRemoveComponents.setText("Remove this component");
		actionRemoveComponents.setToolTipText("Remove components");
		actionRemoveComponents.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.TRASH_CAN));
		
//		Remove interfaces
		actionRemoveInterfaces = new Action() {
			public void run() {
				TreeNode node = (TreeNode) treeViewer.getTree().getSelection()[0].getData();
				
				if (node instanceof InterfaceNode) {
					InterfaceNode iNode = (InterfaceNode) node; 
					InterfaceInstance i = (InterfaceInstance) iNode.getInterfaceInstance();
					
					// Checks if the interface is required
					
					Shell parent = treeViewer.getControl().getShell();	
					
					if (!i.isOptional()) {
						MessageBox error = new MessageBox(parent, SWT.ICON_ERROR | SWT.OK);
						error.setText("Error");
						error.setMessage("You cannot remove a required interface from the configuration.\n"+
								i.toString());
						error.open();
						return;
					}
					
					// Opens a messagedialog for confirming the removing
					
					MessageBox msgbox = new MessageBox(parent, SWT.ICON_QUESTION | SWT.YES | SWT.NO);
					msgbox.setText("Removing interface");
					msgbox.setMessage("Are you sure you want to remove this interface?\n" + i.toString());
					int answer = msgbox.open();
					
					if (answer == SWT.YES)
						removeInterfaceInstance(i);
				}
				
			}
		};
		
		actionRemoveInterfaces.setText("Remove this interface");
		actionRemoveInterfaces.setToolTipText("Remove interfaces");
		actionRemoveInterfaces.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.TRASH_CAN));
		
		// Remove bindings
		
		actionRemoveBindings = new Action() {
			public void run() {
				
				TreeNode node = (TreeNode) treeViewer.getTree().getSelection()[0].getData();
				
				// Opens a messagedialog for confirming the removing
				
				Shell parent = treeViewer.getControl().getShell();
				MessageBox msgbox = new MessageBox(parent, SWT.ICON_QUESTION | SWT.YES | SWT.NO);
				msgbox.setText("Removing binding");
				msgbox.setMessage("Are you sure you want to remove this binding?\n" + node.getInstance().toString() + "?");
				int answer = msgbox.open();
				
				if (answer == SWT.YES)
					removeBindingInstance((BindingInstance) node.getInstance());
			}
		};
		
		actionRemoveBindings.setText("Remove this binding");
		actionRemoveBindings.setToolTipText("Remove bindings");
		actionRemoveBindings.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.TRASH_CAN));
		
		
		// Show element in the diagram
		actionShowInDiagram = new Action() {
			public void run() {
				
				TreeNode node = (TreeNode) treeViewer.getTree().getSelection()[0].getData();
				Instance instance = node.getInstance();
				ComponentDiagramView diagramView = (ComponentDiagramView)(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().findView(getDiagramViewID()));
				
				if (instance instanceof ComponentInstance ||
						instance instanceof InterfaceInstance ||
						instance instanceof BindingInstance) {
					diagramView.showInstance(instance);
				}
				
			}

            protected String getDiagramViewID() {
               return DIAGRAMVIEW_ID;
            }
		};
		actionShowInDiagram.setText("Show this on diagram");
		actionShowInDiagram.setToolTipText("Shows element on the diagram");
		actionShowInDiagram.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.ZOOM));
		
		// Disable unnecessary parts
		actionDisableParts = new Action() {
			public void run() {
				// TODO disable parts	
			}
		};
		actionDisableParts.setText("Disable this part");
		actionDisableParts.setToolTipText("Disable unnecessary parts");
		actionDisableParts.setImageDescriptor(getPlugin()
				.getImageDescriptor(Icons.PART));
		
		
	}
	
	
	
	/**
	 * React to configuration modification events
	 */
	public void configurationModified(ConfigurationModifiedEvent e) {
		handleTask(e.getTask());
	}
	
	private void handleTask(ClientTask task) {
		treeViewer.refresh();
	}
	
	/**
	 * Unregister any remaining listeners at disposal
	 */
	public void dispose() {
		removeListeners();
		super.dispose();
	}

	private void removeListeners() {
		PluginHandler.getInstance().removeConfigurationModificationListener(this);
		PluginHandler.getInstance().removeOpenCloseListener(this);
		PluginHandler.getActivator().getPreferenceStore().removePropertyChangeListener(listener);
	}
	
	protected String getTreeType() {
		String treeType = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActivePart().getTitle();
		if (treeType.equals("Components"))
			treeType = "component";
		if (treeType.equals("Features"))
			treeType = "feature";
		
		return treeType;
	}
	
	protected boolean isRemovableInstance(ComposableInstance cInstance) {
		boolean isRemovable = true;
		if (cInstance.getParent() == null) {
			isRemovable = false;
			return isRemovable;
		}
		return isRemovable;
	}
	
	protected void removeComposableInstance(ComposableInstance cInstance) {
		
		if (cInstance.getParent() == null) {
			return;
		}
		try {
			taskHandler = client.getTaskHandler();
			RemoveInstanceTask task = taskHandler.createRemoveInstanceTask(cInstance);
			getPlugin().getHandler().fireConfigurationModifiedEvent(client.handleTask(task));
		}
		catch (ClientException e) {
			System.out.println("Error when removing instance: " + e.getMessage());
		}
		catch (ConfigurationException ce) {
			System.out.println("Error when removing instance: " + ce.getMessage());
			
		}	
	}
	
	private void removeBindingInstance(BindingInstance bInstance) {
		
		try {
			taskHandler = client.getTaskHandler();
			RemoveInstanceTask task = taskHandler.createRemoveInstanceTask(bInstance);
			getPlugin().getHandler().fireConfigurationModifiedEvent(client.handleTask(task));
		}
		catch (ClientException e) {
			System.out.println("Error when removing instance: " + e.getMessage());
		}
		catch (ConfigurationException ce) {
			System.out.println("Error when removing instance: " + ce.getMessage());
		}
	}
	
	private void removeInterfaceInstance(InterfaceInstance iInstance) {
		
		try {
			taskHandler = client.getTaskHandler();
			RemoveInstanceTask task = taskHandler.createRemoveInstanceTask(iInstance);
			getPlugin().getHandler().fireConfigurationModifiedEvent(client.handleTask(task));
		}
		catch (ClientException e) {
			System.out.println("Error when removing instance: " + e.getMessage());
		}
		catch (ConfigurationException ce) {
			System.out.println("Error when removing instance: " + ce.getMessage());
		}
	}
	
	/** 
	 * React to Init Configuration and Close Configuration events
	 */
	public void handleOpenCloseEvent(ConfigurationOpenedClosedEvent e) {
		String et = e.getEventType();
		if (et.equals(ConfigurationOpenedClosedEvent.EVENT_INITIALIZE) ||
				et.equals(ConfigurationOpenedClosedEvent.EVENT_SAVEOPENED)) {
			treeViewer.setSelection(null);
			updateView();
			// Start listening for configuration modifications
			
		} else if (et.equals(ConfigurationOpenedClosedEvent.EVENT_CLOSE)) {
			treeViewer.setSelection(null);
			showPage(PAGE_EMPTY);
			getPlugin().getHandler().removeConfigurationModificationListener(this);
		}
		
	}
	
	public void errorMessage(String text, String message) {
		Shell parent = treeViewer.getControl().getShell();
		MessageBox msgbox = new MessageBox(parent, SWT.ICON_ERROR | SWT.OK);
		msgbox.setText(text);
		msgbox.setMessage(message);
		msgbox.open();
	}
	
    
	/**
     * This method is for seeing the same selection in the component tree as in
     * the component diagram
     */
	public void selectElement(Instance instance) {
	    initAndShowTreeViewer();
		
	    TreeItem root = treeViewer.getTree().getItems()[0];
	    
	    if (instance instanceof ComponentInstance) {
            ComponentInstance cInstance = (ComponentInstance) instance;
            // checks whether the selected instanse is the component root
            if (cInstance.getParent() == null) {
                //treeViewer.getTree().setSelection(new TreeItem[] { root });
                treeViewer.setSelection(new StructuredSelection(new ComponentNode(cInstance)),true);
                return;
            }
            searchComponent((ComponentInstance) instance, root);
        }
	    if (instance instanceof InterfaceInstance)
	        searchInterface((InterfaceInstance)instance, root);
	    if (instance instanceof BindingInstance)  {
	        bindingFound = false;
	        searchBinding((BindingInstance)instance, root);
	    }
	}
	
	/**
     * Finds the selected component recursively from the tree
	 */
	private void searchComponent(ComponentInstance c, TreeItem parent) {
		TreeItem[] children = parent.getItems();
		//
		for (int i = 0; i < children.length; i++) {
			if (children[i].getData() instanceof ComponentNode) {
				ComponentNode node = (ComponentNode) children[i].getData(); 
				ComponentInstance cInst = (ComponentInstance) node.getInstance(); 
				if (c.equals(cInst)) {
					parent.setExpanded(true);
					treeViewer.setSelection(new StructuredSelection(node),true);
					return;
				}
			}
			searchComponent(c, children[i]);
		}
	}
	
    
	/**
     * Finds the selected interface recursively from the tree
     * @param inst
     * @param parent
	 */
	private void searchInterface(InterfaceInstance inst, TreeItem parent) {
		TreeItem[] children = parent.getItems();
		for (int i = 0; i < children.length; i++) {
			if (children[i].getData() instanceof InterfaceNode) {
				InterfaceNode node = (InterfaceNode) children[i].getData(); 
				InterfaceInstance iInst =  node.getInterfaceInstance();//getInstance(); 
				if (inst.equals(iInst)) {
					System.out.println(inst + " equals " + iInst);
					treeViewer.setSelection(new StructuredSelection(node),true);/*new InterfaceNode((ITreeNode)parent.getData(),inst))*/
					parent.setExpanded(true);
					return;
				}
			}
			searchInterface(inst, children[i]);
		}
	}
	
	/**
     * Finds the selected binding recursively from the tree
     * @param inst
     * @param parent
	 */
	private void searchBinding(BindingInstance inst, TreeItem parent) {
		TreeItem[] children = parent.getItems();
		for (int i = 0; i < children.length; i++) {
			// only one binding is selected (there may be duplicates in the tree)			
			if (bindingFound) {
				return;
			}
			if (children[i].getData() instanceof BindingNode) {
				BindingNode node = (BindingNode) children[i].getData(); 
				BindingInstance bInst = node.getBindingInstance(); 
				if (inst.equals(bInst)) {
					bindingFound = true;
					treeViewer.setSelection(new StructuredSelection(node),true);
					parent.setExpanded(true);
					return;
				}
			}
			searchBinding(inst, children[i]);
		}
	}
	
	public void selectNodeAgain() {
		Object obj = treeViewer.getTree().getSelection()[0].getData();
		if (obj instanceof ComponentNode)
			treeViewer.setSelection(new StructuredSelection((ComponentNode) obj));
		if (obj instanceof AttributeNode)
			treeViewer.setSelection(new StructuredSelection((AttributeNode) obj));
		if (obj instanceof FeatureNode)
			treeViewer.setSelection(new StructuredSelection((FeatureNode) obj));
	}
	
	public void resetSelection() {
		treeViewer.setSelection(null);
	}
	
    protected IKumbangPlugin getPlugin() {
        return Activator.getDefault();
    }
    

	
}
