/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.configurator.client.plugin.wizards;

import java.lang.reflect.InvocationTargetException;

import kumbang.configurator.Activator;
import kumbang.configurator.client.core.ClientServices;
import kumbang.configurator.client.core.exception.ClientException;
import kumbang.configurator.client.plugin.KumbangConfiguratorPlugin;
import kumbang.configurator.client.plugin.event.ConfigurationOpenedClosedEvent;
import kumbang.core.model.exception.ModelException;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.ui.INewWizard;
import org.eclipse.ui.IWorkbench;


/**
 * The Wizard for opening and initializing new 
 * configurations of models on the server. Provides also the
 * possibility to import new models to the server.
 * @author jlemmett
 *
 */
public class InitWizard extends Wizard implements INewWizard {
	
	private WizardPage mainPage;
	protected ClientServices client;
	
	public InitWizard() {
		super();
		setNeedsProgressMonitor(true);
	}

	public void init(IWorkbench workbench, IStructuredSelection selection) {
		setWindowTitle("New Kumbang Configuration");
	}

	public void addPages() {
		if (Activator.getDefault().getHandler().isServerAvailable()) {
			mainPage = new InitWizardPage(this);
			addPage(mainPage);
		} else {
			mainPage = new CannotInitPage();
			addPage(mainPage);
		}
	}

	/* I've commented out the container stuff, it might be useful 
	 * but I'm not sure how yet (non-Javadoc)
	 * @see org.eclipse.jface.wizard.IWizard#performFinish()
	 */
	public boolean performFinish() {
		
		//final String containerName = mainPage.getContainerName();
		final String fileName = ((InitWizardPage)mainPage).getFileName();
		final String modelName = ((InitWizardPage)mainPage).getModelname();
		IRunnableWithProgress op = new IRunnableWithProgress() {
			public void run(IProgressMonitor monitor) throws InvocationTargetException {
				try {
					//doFinish(containerName, fileName, modelName, monitor);
					doFinish(fileName, modelName, monitor);
				} catch (CoreException e) {
					throw new InvocationTargetException(e);
				} finally {
					monitor.done();
				}
			}
		};
		try {
			// Changed the first parameter (fork) to false to avoid
			// Invalid Thread Access exception
			getContainer().run(false, false, op);
		} catch (InterruptedException e) {
			return false;
		} catch (InvocationTargetException e) {
			Throwable realException = e.getTargetException();
			MessageDialog.openError(getShell(), "Error", realException.toString() + ":" + realException.getMessage());
			realException.printStackTrace();
			return false;
		}
		return true;
	}

	//private void doFinish(String containerName, String fileName, String modelName, IProgressMonitor monitor) throws CoreException {
	protected void doFinish(String fileName, String modelName, IProgressMonitor monitor) throws CoreException {
		// Here we should probably do something with the container and filename, maybe
		// create a new file or something?
		
		// For now, just do the initialization things like the old client does
		// Create a ConfigurationClient
		// Open and initialize a model from the server
		
		if (Activator.getDefault().getHandler().isServerAvailable()) {
			
			/* Ask for confirmation if there's an old config open */
			if (Activator.getDefault().getHandler().getClientServices().isConfigurationModified()) {
				if (!MessageDialog.openConfirm(this.getShell(), "Confirm Initialization", "Are you sure you want to initialize a new model, all unsaved changes to the current configuration will be lost")) {
					return;
				}
			}

			this.client = Activator.getDefault().getHandler().getClientServices();
			try {
				client.openModel(modelName, fileName);
				MessageDialog.openInformation(this.getShell(), "Success", "Opened and initialized model " + modelName + " as new configuration " + fileName);
				// Open the Kumbang perspective by force.. 
				// Maybe DialogSettings will help with the
				// confirmation dialog..
				Activator.getDefault().getHandler().openKumbangPerspective();
			}  catch (ModelException me) {
				MessageDialog.openError(this.getShell(), "Failure", "The model was not a valid Kumbang model.\nFilename: " + fileName + "\nModel name: " + modelName);
				client.closeConfiguration();
			} catch (ClientException ce) {
				MessageDialog.openError(this.getShell(), "Failure", "Error opening the model\nClientException: " + ce.getMessage());
				client.closeConfiguration();
			} finally {
				Activator.getDefault().getHandler().fireOpenCloseEvent(ConfigurationOpenedClosedEvent.EVENT_INITIALIZE);
			}
		} else {
			MessageDialog.openError(this.getShell(), "Server not available.", "The configuration server is not available");
		}
	}
	
}
