/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * SmTranslator.java
 *
 * Created on 30. lokakuuta 2004, 19:18
 */
package kumbang.configurator.server.smodels;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import kumbang.core.model.constraint.AttributeExpression;
import kumbang.core.model.constraint.AttributeReference;
import kumbang.core.model.constraint.BinaryExpression;
import kumbang.core.model.constraint.BooleanExpression;
import kumbang.core.model.constraint.CardinalityExpression;
import kumbang.core.model.constraint.ComparisonExpression;
import kumbang.core.model.constraint.ConjunctiveExpression;
import kumbang.core.model.constraint.ConnectedExpression;
import kumbang.core.model.constraint.ConstantExpression;
import kumbang.core.model.constraint.EquivalenceExpression;
import kumbang.core.model.constraint.HasInstancesExpression;
import kumbang.core.model.constraint.HasPartOfTypeExpression;
import kumbang.core.model.constraint.ImplicationExpression;
import kumbang.core.model.constraint.IndexedExpression;
import kumbang.core.model.constraint.InstanceOfExpression;
import kumbang.core.model.constraint.IntegerExpression;
import kumbang.core.model.constraint.PresentExpression;
import kumbang.core.model.constraint.Quantification;
import kumbang.core.model.constraint.QuantificationStack;
import kumbang.core.model.constraint.QuantifiedExpression;
import kumbang.core.model.constraint.ReferenceExpression;
import kumbang.core.model.constraint.StructureReference;
import kumbang.core.model.constraint.Variable;
import kumbang.core.model.core.AttributedModel;
import kumbang.core.model.core.ComponentedModel;
import kumbang.core.model.core.FeaturedModel;
import kumbang.core.model.core.Model;
import kumbang.core.model.type.AttributeDefinition;
import kumbang.core.model.type.AttributeValueType;
import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.Connection;
import kumbang.core.model.type.FeatureType;
import kumbang.core.model.type.InterfaceDefinition;
import kumbang.core.model.type.InterfaceType;
import kumbang.core.model.type.PartDefinition;

/**
 * A class for translating Kumbang (Forfamel, Koalish) models into WCRL programs.
 * 
 * @author toasikai
 */
public class Translator {
	private static final String DECLARATIONS = "%% Hide all other predicates except in(X), cn(X,Y) and hasattr(X,Y,Z)\n"
			+ "#hide. \n#show in(X).\n#show cn(X,Y).\n#show hasattr(X,Y,Z).\n"
			+ "#show cnf. #show cll. #show cff. #show cf(I,X).\n\n";

	private static final String FORFAMEL_ONTOLOGY = "% There must be a feature root in the configuration\n"
			+ "1 { in(F) : froot(F) } 1.\n\n";

	private static final String KOALISH_INTERFACES = "% The same for an interface\n"
			+ "in(I) :- hasint(C, I, N), pint(C, I, N), in(C).\n\n"

			+ "% A possible connection between interfaces justifies connecting them\n"
			+ "0 { cn(I1, I2) } 1 :- pcn(I1, I2), in(I1), in(I2).\n\n"

			+ "% A function is compatible with itself.\n"
			+ "% cmbf(F, F) :- intf(I, F).\n\n"

			+ "notsubtypeof(T1, T2) :- hasfunc(T2, F), not hasfunc(T1, F), inttype(T1).\n"
			+ "subtype(T1, T2) :- not notsubtypeof(T1, T2), inttype(T1), inttype(T2).\n\n"

			+ "% Connections between functions are implied by a connections between interfaces.\n"
			+ "% cnf(I1, F, I2, F) :- cn(I1, I2), pcn(I1, I2), intf(I1, F).\n\n"

			+ "% When can individual functions be bound.\n"
			+ "% 0 { cnf(I1, F1, I2, F2) } 1 :- cnb(I1, I2), intf(I1, F1), intf(I2, F2), cmbf(F1, F2), in(I1), in(I2), not grounded(I1).\n\n"

			+ "% A possible connection implies topological compatibility\n"
			+ "cnb(I1, I2) :- pint(C1, I1, N1), pint(C2, I2, N2), ppart(C, C1, M1, O1), ppart(C, C2, M2, O2), C1 != C2, req(I1), prov(I2).\n"
			+ "cnb(I1, I2) :- pint(C1, I1, N1), pint(C, I2, N2), ppart(C, C1, M1, O1), req(I1), req(I2).\n"
			+ "cnb(I2, I1) :- pint(C1, I1, N1), pint(C, I2, N2), ppart(C, C1, M1, O1), prov(I1), prov(I2).\n"
			+ "pcn(I1, I2) :- subtype(T2, T1), cnb(I1, I2), instance(I1, T1), instance(I2, T2).\n\n"

			+ "% The requirement for a single binding from each function in a non-grounded interface.\n"
			+ "cll(I1) :- 0 { cn(I1, I2) : pcn(I1, I2) } 0, pint(X, I1, N), in(I1), not grounded(I1).\n"
			+ "cll(I1) :- 2 { cn(I1, I2) : pcn(I1, I2) }, pint(X, I1, N), in(I1), not grounded(I1).\n"
			+ "cll :- cll(I1, F1), intf(I1, F1).\n" + "cll :- cll(I1), pint(X, I1, N).\n\n";

	private static final String KOALISH_ONTOLOGY = "% There must be a component root in the configuration\n"
			+ "1 { in(C) : croot(C) } 1.\n\n";

	private static final int NO_PARENT = -1;

	private static final String ONTOLOGY = "%% Ontological definitions\n\n"
			+ "% A component and a feature is in the configuration if it if part of something\n"
			+ "in(C2) :- haspart(C1, C2, N), ppart(C1, C2, N, I).\n\n"

			+ "% Track parts for instances of types with has_part_of_type constraints\n"
			+ "ppartt(X, X1) :- ppart(X, X1, N, I), trackParts(X).\n"
			+ "ppartt(X, X2) :- ppartt(X, X1), ppart(X1, X2, N, I), composable(X).\n\n"

			+ "% A direct instance is also a regular instance\n"
			+ "instance(X, T) :- directInstance(X, T).\n\n"// , composable(X).\n\n"

			+ "% An instance is a composable\n" + "composable(X) :- instance(X, T).\n\n"// ,
			// componentType(T).\n"
			// + "composable(X) :- instance(X, T), featureType(T).\n\n"

			+ "% Each composable instance is indexed\n"
			+ "index(X1, I) :- ppart(X0, X1, N, I).\n\n";

	/**
	 * Translates the model given as a parameter into a WCRL program.
	 * 
	 * @param model the model to be translated
	 * @return a String object containing the WCRL program
	 */
	public static String Translate(Model model) {
		Translator inner = new Translator(model);
		return inner.getModelString();
	}
	/**
	 * Index for the next connection constraint to be generated.
	 */
	private int connIndex = 0;

	/**
	 * Index for the next constraint to be generated.
	 */
	private int constrIndex = 0;
	private int id = 0;

	private boolean isForfamel;

	private boolean isKoalish;

	private StringBuffer modelString;

	/** Creates a new instance of SmTranslator */
	private Translator(Model model) {
		modelString = new StringBuffer();
		
		isKoalish = (model instanceof ComponentedModel);
		isForfamel = (model instanceof FeaturedModel);

		modelString.append(DECLARATIONS);
		modelString.append(ONTOLOGY);

		if (isForfamel)
			modelString.append(FORFAMEL_ONTOLOGY);

		if (isKoalish) {
			modelString.append(KOALISH_ONTOLOGY);
			modelString.append(KOALISH_INTERFACES);
		}

		modelString.append("\n%%%% Definitions from the model\n");
		translate(model);

		modelString.append("\n%%%% Possible instances in the configuration\n\n");
		instantiate(model);
	}

	private void addRule(String rule) {
		modelString.append(rule + "\n");
	}

	/**
	 * Returns the predicate corresponding to constraint cons given as parameter, including the
	 * variables defined outside the context of cons, in the variables of QuantificationStack rem,
	 * and in Quantification previous.
	 * 
	 * @param cons	
	 * @param remaining
	 * @param previous
	 * @param index
	 * @return
	 */
	private String generalConstraint(BooleanExpression cons, QuantificationStack rem,
			Quantification previous, int index) {
		String ret = "constr" + index + "(X";

		for (Iterator it = cons.appearingVariables(rem, previous).iterator(); it.hasNext();) {
			Variable var = (Variable) it.next();
			ret += ", V" + var.getIndex();
		}

		ret += ")";
		return ret;
	}

	private String generalDomains(BooleanExpression cons, QuantificationStack rem,
			Quantification previous, ComposableType type) {
		String ret = "";

		for (Iterator it = cons.appearingVariables(rem, previous).iterator(); it.hasNext();) {
			Variable var = (Variable) it.next();
			StructureReference ref = var.getReference();
			ret += ", " + referenceText(ref, 0, ref.getIndex(), type);
			ret += ", V" + var.getIndex() + " = X" + ref.getIndex();
		}

		if (ret.length() == 0)
			ret = ", " + NameMap.getSmName(type) + "(X)";

		return ret;
	}

	private int getNextId() {
		return id++;
	}

	private void instantiate(Model model) {
		/*
		 * IMPORTANT Components must be instantiated first, the id resolution algorithm
		 * relies on that.
		 */
		if (model instanceof ComponentedModel) {
			ComponentedModel cModel = (ComponentedModel) model;
			if (cModel.getComponentRoot() != null) {
				ComponentType rootType = cModel.getComponentRoot();
				instantiate(rootType, null, NO_PARENT, 0);				
			}
		}

		if (model instanceof FeaturedModel) {
			FeaturedModel fModel = (FeaturedModel) model;
			if (fModel.getFeatureRoot() != null) {
				
				FeatureType rootType = fModel.getFeatureRoot();
				instantiate(rootType, null, NO_PARENT, 0);				
			}
		}
	}

	private void instantiate(InterfaceType type, InterfaceDefinition def, int parentID) {
		String smName = NameMap.getSmName(type);
		int id = getNextId();

		addRule(smName + "(i" + id + ").");

		addRule("pint(i" + parentID + ", i" + id + ", " + NameMap.getSmName(def) + "). ");

		if (def.isGrounded())
			addRule("grounded(i" + id + "). ");

		if (def.isProvides())
			addRule("prov(i" + id + ").");
		else
			addRule("req(i" + id + ").");
	}

	/**
	 * Instantiates a composable type, i.e., instantiates recursively the instances that it may
	 * have as its parts and the interfaces it may contain.
	 * 
	 * @param type
	 * @param name
	 * @param parent
	 * @param rank
	 */
	private void instantiate(ComposableType type, PartDefinition part, int parentID, int rank) {
		int id = getNextId();

		String smName = NameMap.getSmName(type);

		// Create a rule for the instance itself
		addRule("direct" + smName + "(i" + id + ").");
		if (parentID != NO_PARENT) {
			String rule = "ppart(i" + parentID + ", i" + id + ", ";
			rule += NameMap.getSmName(part) + ", " + rank + ").";
			addRule(rule);
		}

		// Iterate through the part definitions
		for (Iterator parts = type.getParts(); parts.hasNext();) {
			PartDefinition def = (PartDefinition) parts.next();

			int maxCardinality = def.getUpperBound();
			for (Iterator types = def.getAllPossibleTypesIterator(false); types.hasNext();) {
				ComposableType possibleType = (ComposableType) types.next();
				if (def.getSimilarity() != PartDefinition.simDifferent)
					for (int i = 0; i < maxCardinality; i++)
						instantiate(possibleType, def, id, i + 1);
				else
					instantiate(possibleType, def, id, 1);
			}
		}
		
		// Only instantiate interfaces for component types
		if (!(type instanceof ComponentType))
			return;

		// Iterate through interface definition
		for (Iterator d = ((ComponentType) type).getInterfaces(); d.hasNext(); ) {
			InterfaceDefinition intDef = (InterfaceDefinition) d.next(); 
			for (Iterator t = intDef.getPossibleTypes(); t.hasNext(); ) {
				instantiate((InterfaceType) t.next(), intDef, id);
			}
		}
	}

	/**
	 * Returns the predicate corresponding to constraint cons given as parameter, including the
	 * variables defined in the context of the parent BooleanExpression.
	 * 
	 * @param cons
	 * @param parent
	 * @param index
	 * @return
	 */
	private String parentConstraint(BooleanExpression cons, BooleanExpression parent, int index) {
		return generalConstraint(cons, parent.addedVariables(), null, index);
	}

	private String parentDomains(BooleanExpression cons, BooleanExpression parent, ComposableType type) {
		return generalDomains(cons, parent.addedVariables(), null, type);
	}

	/**
	 * Returns the predicate corresponding to constraint cons given as parameter, including
	 * only the variables that are defined outside the context of the constraint.
	 * 
	 * constr(X, ...) :- compoType(X), 
	 * 
	 * @param cons
	 * @param type
	 * @param index
	 * @return
	 */
	private String publicConstraint(BooleanExpression cons, int index) {
		return generalConstraint(cons, cons.addedVariables(), null, index);
	}

	private String publicDomains(BooleanExpression cons, ComposableType type) {
		return generalDomains(cons, cons.addedVariables(), null, type);
	}

	private String referenceText(StructureReference ref, int level, int index) {
		return referenceText(ref, level, index, null);
	}

	/**
	 * Returns the predicate representing the reference in WCRL program.
	 * ref[index](X, variables, X0) 
	 * 
	 * @param ref The Structure reference being translated
	 * @param level The level of the reference, 0 for the main level
	 * @param index Index of the final variable
	 * @return
	 */
	private String referenceText(StructureReference ref, int level, int index, ComposableType type) {
		// "ref" + index + "(X, " + var.getReference().referenceText() + "X0)";
		// "ref" + index + "_" + level + "(X, " + referenceText(sr) + "X0, X0) :- ";

		if (ref == null)
			return "X" + index + " = X, " + NameMap.getSmName(type) + "(X)";

		String ret = "ref" + ref.getIndex();
		if (level > 0)
			ret += "_" + level;

		ret += "(X, " + variableString(ref) + "X" + index + ")";

		return ret;
	}

	private String getModelString() {
		return modelString.toString();
	}

	private void translate(AttributedModel model) {
		Iterator attrValueTypes = model.getAttributeValueTypes();
		while (attrValueTypes.hasNext()) {
			translate((AttributeValueType) attrValueTypes.next());
		}
	}
	
	/**
	 * 
	 * 
	 * @param ar the attribute reference to be tranlated
	 * @param id the identifier of the variable capturing the value
	 * @return 
	 */
	private String translate(AttributeExpression ar, ComposableType type, int id) {
		String rule = "hasattr(X" + id + ", " + NameMap.getAttributeName(ar.getName()) + ", N" + id + "), ";
		rule += referenceText(ar.getStructureReference(), 0, id, type);

		/*
		 * Domain for the attribute. For attributes that have unambiguous types
		 * unions of individual domains are used.
		 */
		if (ar.getAttributeReference().possibleTypeCount() == 1) {
			AttributeValueType avt = ar.getAttributeReference().getType();
			rule += ", " + NameMap.getSmName(avt) + "(N" + id + ")";
		} else
			rule += ", ut" + ar.getAttributeReference().getIndex() + "(N" + id + ")";

		return rule;
	}

	/**
	 * 
	 * 
	 * @param ar
	 */
	private void translate(AttributeReference ar) {
		//	addRule("\n% Attribute value type union: " + sr.getOrigin() + " -> " + sr.toString());

		for (Iterator it = ar.getPossibleTypes(); it.hasNext();)
			addRule("ut" + ar.getIndex() + "(V) :- "
					+ NameMap.getSmName((AttributeValueType) it.next()) + "(V).");
	}

	private void translate(AttributeValueType valueType) {
		String attrName = NameMap.getSmName(valueType);
		Iterator values = valueType.getValues();

		addRule("\n%% Attribute value type: " + valueType.getName() + " (" + attrName + ")\n");

		String current = null;
		while (values.hasNext()) {
			current = NameMap.getSmNameForAttributeValue(values.next().toString());
			addRule(attrName + "(" + current + ").");
		}
	}

	private String translate(BinaryExpression be, ComposableType type) {
		return "(" + translate(be.getLeft(), type) + " " + be.getOperator() + " "
				+ translate(be.getRight(), type) + ")";
	}

	/**
	 * Selects the correct translation method for a boolean expression
	 * 
	 * @param be BooleanExpression to be translated
	 * @param type ComposableType containing the constraint
	 * @return index of the resulting constr-rule; -1 signals failure
	 */
	private int translate(BooleanExpression be, BooleanExpression parent, ComposableType type) {
		int index;

		if (be instanceof ComparisonExpression)
			index = translate((ComparisonExpression) be, type);
		else if (be instanceof ConjunctiveExpression)
			index = translate((ConjunctiveExpression) be, type);
		else if (be instanceof ConnectedExpression)
			index = translate((ConnectedExpression) be, type);
		else if (be instanceof EquivalenceExpression)
			index = translate((EquivalenceExpression) be, type);
		else if (be instanceof HasInstancesExpression)
			index = translate((HasInstancesExpression) be, type);
		else if (be instanceof HasPartOfTypeExpression)
			index = translate((HasPartOfTypeExpression) be, type);
		else if (be instanceof ImplicationExpression)
			index = translate((ImplicationExpression) be, type);
		else if (be instanceof QuantifiedExpression)
			index = translate((QuantifiedExpression) be, type);
//		else if (be instanceof PresentInterfaceExpression)
//			index = translate((PresentInterfaceExpression) be, type);
		else if (be instanceof ReferenceExpression)
			index = translate((ReferenceExpression) be, type);
		else
			throw new Error("This kind of constraints are not supported.");

		if (be.isNegated()) {
			int prev = index;
			index = constrIndex++;

			String rule = "";

			if (parent == null) {
				rule = generalConstraint(be, null, null, index) + " :- not ";
				rule += generalConstraint(be, null, null, prev);
				rule += generalDomains(be, null, null, type) + ".";
			} else {
				rule = parentConstraint(be, parent, index) + " :- not ";
				rule += parentConstraint(be, parent, prev);
				rule += parentDomains(be, parent, type) + ".";
			}

			addRule(rule);
		}

		return index;
	}

	/**
	 * @param constraint
	 * @param type
	 * @return
	 */
	private String translate(CardinalityExpression cc, ComposableType type, int id) {
		String rule = "";
		StructureReference sr = cc.getReference();

		rule += "card(X, " + sr.getIndex() + ", " + variableString(sr) + "N" + id + "), ";
		rule += "poscard(" + sr.getIndex() + ", N" + id + ")";

		return rule;
	}

	private int translate(ComparisonExpression cc, ComposableType type) {
		int index = constrIndex++;

		Rule rule = new Rule(publicConstraint(cc, index));
		rule.addToBody(NameMap.getSmName(type) + "(X)");

		//	Add the primitives to the rule.
//		for (Iterator it = cc.getPrimitives().iterator(); it.hasNext();)
//			rule += translate((IndexedExpression) it.next(), type) + ", ";
		for (Iterator it = cc.getPrimitives().iterator(); it.hasNext();)
			rule.addToBody(translate((IndexedExpression) it.next(), type));

		//	Add the comparison itself
//		rule += translate(cc.getLeft(), type) + " ";
//		rule += cc.getOperator() + " ";
//		rule += translate(cc.getRight(), type) + ".";
		rule.addToBody(translate(cc.getLeft(), type) + " " + cc.getOperator() + " "
				+ translate(cc.getRight(), type));

		addRule(rule.toString());

		return index;
	}

	private void translate(ComponentType component, ComponentedModel model) {
		String componentName = NameMap.getSmName(component);

		addRule("\n%%% Component type: " + component.getName() + " (" + componentName + ")\n");

		addRule("componentType(" + componentName + ").");

		// Add root definition
		if (model.getComponentRoot() == component)
			addRule("%% The root component type\ncroot(X) :- " + componentName + "(X).");

		// common to composable types (parts, attributes)
		translate(component);

		// interfaces
		Iterator interfaces = component.getInterfaces();
		while (interfaces.hasNext()) {
			InterfaceDefinition idef = (InterfaceDefinition) interfaces.next();
			String intName = NameMap.getSmName(idef);
			int lower = 1;
			if (idef.isOptional())
				lower = 0;

			//	0/1 { hasint(C, I, name) : pint(C, I, name) } :- type(C).	
			addRule(lower + " { hasint(C, I, " + intName + ") : pint(C, I, " + intName
					+ ") } 1 :- " + componentName + "(C).");
		}

		/*
		 * connections
		 * TODO Can there be connections? Should be handled togethe with constraints.
		 */
		Iterator connections = component.getConnections();
		while (connections.hasNext()) {
			Connection c = (Connection) connections.next();
			translate(c, component);
		}
	}

	private void translate(ComposableType composable) {
		String composableName = NameMap.getSmName(composable);
		
		// supertypes
		Iterator supertypes = composable.getSupertypes();
		
		addRule(composableName + "(X) :- direct" + composableName + "(X).");

		if (supertypes.hasNext())
			addRule("\n%% Supertypes:\n");

		while (supertypes.hasNext()) {
			ComposableType supertype = (ComposableType) supertypes.next();
			
			String rule = NameMap.getSmName(supertype) + "(X) :- ";
			rule += composableName + "(X).";
			addRule(rule);
		}

		// attributes
		Iterator attributes = composable.getDirectAttributes();

		if (attributes.hasNext())
			addRule("\n%% Attributes\n");

		while (attributes.hasNext()) {
			AttributeDefinition ad = (AttributeDefinition) attributes.next();
			String attrName = NameMap.getSmName(ad); //	name clashes
			String attrType = NameMap.getSmName(ad.getType());

			String rule = "1 { hasattr(X, " + attrName + ", V) : " + attrType + "(V) } 1 ";
			rule += ":- in(X), " + composableName + "(X).";

			addRule(rule);
		}

		// parts
		Iterator parts = composable.getDirectParts();

		if (parts.hasNext())
			addRule("\n%% Part/subfeature definitions:\n");

		while (parts.hasNext()) {
			PartDefinition part = (PartDefinition) parts.next();
			int lower = part.getLowerBound();
			int upper = part.getUpperBound();
			String partName = NameMap.getSmName(part);

			String main = " { haspart(X1, X2, " + partName + ") : " + "ppart(X1, X2, " + partName + ", I) } ";

			addRule(lower + main + upper + " :- " + composableName + "(X1), in(X1).");
			
			if (part.getSimilarity() == PartDefinition.simSame) {
				Rule rule = new Rule();
				rule.addToBody(composableName + "(X)");
				rule.addToBody("ppart(X, X0, " + partName + ", 1)");
				rule.addToBody("directInstance(X0, T0)");
				rule.addToBody("ppart(X, X1, " + partName + ", N)");
				rule.addToBody("directInstance(X1, T1)");
				rule.addToBody("in(X0), in(X1), T0 != T1");
				
				addRule(rule.toString()); 
			}
		}

		// structure references
		Iterator references = composable.getStructureReferences();
		if (references.hasNext())
			addRule("\n%% References:");

		while (references.hasNext())
			translate((StructureReference) references.next(), composable);

		// constraints
		Iterator constraints = composable.getConstraints();

		if (constraints.hasNext())
			addRule("\n%% Constraints:");

		while (constraints.hasNext()) {
			BooleanExpression ee = (BooleanExpression) constraints.next();

			addRule("\n% Constraint: " + ee.toString() + "\n");

			int index = translate(ee, null, composable);
			String rule = "cf(" + index + ", X) :- ";
			rule += NameMap.getSmName(composable) + "(X), in(X), not constr" + index + "(X).";
			addRule(rule);

			rule = "cff :- cf(" + index + ", X), " + NameMap.getSmName(composable) + "(X).";
			addRule(rule);
		}

		// attribute references
		Iterator attributesRefs = composable.getAttributeReferences();
		boolean first = true;

		while (attributesRefs.hasNext()) {
			AttributeReference ar = (AttributeReference) attributesRefs.next();
			if (ar.possibleTypeCount() > 1) {
				if (first) {
					addRule("\n%% Attribute type unions:");
					first = false;
				}

				translate(ar);
			}
		}
	}

	/**
	 * Translates a conjunctive or disjunctive expression.
	 * 
	 * The complexity in this method is due mostly to quantifications;
	 * translating conjunctive (disjunctive) constraints as such
	 * is not really this complex. 
	 * 
	 * @param cc ConjunctiveExpression to be translated
	 * @return index of the resulting constr rule
	 */
	private int translate(ConjunctiveExpression cc, ComposableType type) {
		int index;

		ArrayList<Integer> indeces = new ArrayList<Integer>();
		int count = 0; //	The number of primitive constraints concatenated.

		for (Iterator it = cc.getPrimitives(); it.hasNext();) {
			indeces.add(new Integer(translate((BooleanExpression) it.next(), cc, type)));
			count++;
		}

		if (count == 1)
			throw new Error("ConjunctiveExpressions must have at least two elements.");

		index = constrIndex++;
		String rule = new String();
		int op = cc.getOperator();
		if (op == ConjunctiveExpression.opAnd) {
			rule = publicConstraint(cc, index) + " :- ";
			rule += NameMap.getSmName(type) + "(X), ";

			for (Iterator it = cc.getPrimitives(), in = indeces.iterator(); it.hasNext() && in.hasNext();) {
				BooleanExpression cp = (BooleanExpression) it.next();
				rule += parentConstraint(cp, cc, ((Integer) in.next()).intValue());
				rule += parentDomains(cp, cc, type);
				if (it.hasNext())
					rule += ", ";
			}

			rule += ".";
			addRule(rule);
		} else if (op == ConjunctiveExpression.opOr) {
			for (Iterator it = cc.getPrimitives(), in = indeces.iterator(); it.hasNext() && in.hasNext();) {
				BooleanExpression cp = (BooleanExpression) it.next();

				rule = publicConstraint(cc, index) + " :- ";
				rule += parentConstraint(cp, cc, ((Integer) in.next()).intValue());
				rule += parentDomains(cp, cc, type) + publicDomains(cc, type) + ".";
				addRule(rule);
			}
		} else
			throw new Error("Illegal operator.");

		return index;
	}

	private int translate(ConnectedExpression ce, ComposableType type) {
		int index = constrIndex++;

		String rule = publicConstraint(ce, index) + " :- ";
//		rule += interfaceReferenceText(ce.getFromReference(), 1) + ", ";
		rule += referenceText(ce.getFromReference(), 0, 1) + ", ";
		
		if (ce.getToReference() == null) {
			String second = new String(rule);
			rule += "pcn(X1, X2), cn(X1, X2)."; 
			addRule(rule);
			
			second += "pcn(X2, X1), cn(X2, X1).";
			addRule(second);
			
			return index;
		}
		
//		rule += interfaceReferenceText(ce.getToReference(), 2) + ", ";
		rule += referenceText(ce.getToReference(), 0, 2) + ", ";
		rule += "cn(X1, X2).";
		
		addRule(rule);
		
		return index;
	}

	/**
	 * A connection is much like a constraint.
	 *  
	 * @param c the connection to be translated
	 * @param type the type defining the connection
	 */
	private void translate(Connection c, ComposableType type) {
		int index = connIndex++;
		String rule;
		addRule("\n% Connection: ");

		rule = "conn" + index + "(C)" + " :- cn(I1, I2), ";

		// The source interface (#1)
		if (c.getSourcePart() == null) {
			rule += "pint(C, I1, ";
		} else {
			rule += "ppart(C, C1, " + NameMap.getSmName(c.getSourcePart()) + ", I), pint(C1, I1, ";
		}

		rule += NameMap.getSmName(c.getSourceInterface()) + "), ";

		// The target interface (#2)
		if (c.getTargetPart() == null) {
			rule += "pint(C, I2, ";
		} else {
			rule += "ppart(C, C2, " + NameMap.getSmName(c.getTargetPart()) + ", I), pint(C2, I2, ";
		}

		rule += NameMap.getSmName(c.getTargetInterface()) + ").";
		addRule(rule);

		if (c.getSourcePart() != null) {
			rule = "conn" + index + "(C) :- ppart(C, C1, ";
			rule += NameMap.getSmName(c.getSourcePart());
			rule += ", I), not in(C1).";
			addRule(rule);
		}

		addRule("cnf :- " + NameMap.getSmName(type) + "(C), in(C), not conn" + index + "(C).");
	}

	/**
	 * @param ee
	 * @param composable
	 */
	private int translate(EquivalenceExpression ee, ComposableType type) {
		BooleanExpression left = ee.getLeft(), right = ee.getRight();
		int index, leftIndex, rightIndex;
		String rule;

		leftIndex = translate(left, ee, type);
		if (right == null) {
			throw new Error("Equivalence constraint without a right side.");
		}

		rightIndex = translate(right, ee, type);
		index = constrIndex++;

		rule = publicConstraint(ee, index) + " :- ";
		rule += parentConstraint(left, ee, leftIndex) + ", ";
		rule += parentConstraint(right, ee, rightIndex);
		rule += parentDomains(left, ee, type) + parentDomains(right, ee, type) + ".";
		addRule(rule);

		rule = publicConstraint(ee, index) + " :- not ";
		rule += parentConstraint(left, ee, leftIndex) + ", not ";
		rule += parentConstraint(right, ee, rightIndex);
		rule += parentDomains(left, ee, type) + parentDomains(right, ee, type) + ".";
		addRule(rule);

		return index;
	}

	private void translate(FeatureType feature, FeaturedModel model) {
		String featureName = NameMap.getSmName(feature);

		addRule("\n%%% Feature type: " + feature.getName() + " (" + featureName + ")\n");

		addRule("featureType(" + NameMap.getSmName(feature) + ").");

		// add root definition
		if (model.getFeatureRoot() == feature)
			addRule("\n%% The root feature type\n\nfroot(X) :- " + featureName + "(X).");

		// common to composable types (parts, attributes)
		translate(feature);
	}

	private void translate(FeaturedModel model) {
		Iterator features = model.getFeatureTypes();
		while (features.hasNext())
			translate((FeatureType) features.next(), model);
	}

	/**
	 * @param exp
	 * @param type
	 * @return
	 */
	private int translate(HasInstancesExpression exp, ComposableType type) {
		int index = constrIndex++;

		String rule = publicConstraint(exp, index) + " :- ";
		rule += "in(X0), " + NameMap.getSmName(exp.getComparedType()) + "(X0), ";
		rule += NameMap.getSmName(type) + "(X).";

		addRule(rule);
		return index;
	}

	/**
	 * @param cp
	 * @param type
	 * @return
	 */
	private int translate(HasPartOfTypeExpression cp, ComposableType type) {
		int index = constrIndex++;
		String rule = "";

		// TODO Check if referenceText could be such that the cases would not be necessary here.
		if (cp.getReference() != null) {
			rule = "trackParts(X0) :- ";
			rule += referenceText(cp.getReference(), 0, 0, type) + ".";
			addRule(rule);
		} else {
			rule += "trackParts(X) :- " + NameMap.getSmName(type) + "(X).";
			addRule(rule);
		}

		rule = publicConstraint(cp, index) + " :- ";
		rule += referenceText(cp.getReference(), 0, 1, type) + ", ";
		rule += NameMap.getSmName(type) + "(X), ";
		rule += "ppartt(X1, X0), in(X0), " + NameMap.getSmName(cp.getType()) + "(X).";
		addRule(rule);

		return index;
	}

	private int translate(ImplicationExpression ic, ComposableType type) {
		BooleanExpression head, tail;
		head = ic.getHead();
		tail = ic.getTail();

		String rule;
		int index;

		int headIndex = -1, tailIndex = -1;

		if (tail != null)
			tailIndex = translate(ic.getTail(), ic, type);
		else
			throw new Error("This should really not occur: tail is null in an implication constraint.");

		if (head != null) {
			headIndex = translate(ic.getHead(), ic, type);
			index = constrIndex++;
			rule = publicConstraint(ic, index) + " :- ";
			rule += parentConstraint(head, ic, headIndex);
			rule += parentDomains(head, ic, type) + ".";
			addRule(rule);

			rule = publicConstraint(ic, index) + " :- not ";
			rule += parentConstraint(tail, ic, tailIndex);
			rule += parentDomains(tail, ic, type) + ".";
			addRule(rule);
		} else
			throw new Error("Implication constraint without a head.");

		return index;
	}

	private String translate(IndexedExpression iv, ComposableType type) {
		if (iv instanceof CardinalityExpression)
			return translate((CardinalityExpression) iv, type, iv.getIndex());
		if (iv instanceof AttributeExpression)
			return translate((AttributeExpression) iv, type, iv.getIndex());
		else
			throw new Error("Unknown subtype of IndexedValue.");
	}
	
	private int translate(InstanceOfExpression ioe, ComposableType type) {
		int index = constrIndex++;

		// constr(X, ...) :-
		String rule = publicConstraint(ioe, index) + " :- ";
		rule += referenceText(ioe.getReference(), 0, 0, type);
		
		//	TODO Check if the instance-thing is indeed needed.
		rule += ", " + NameMap.getSmName(ioe.getType()) + "(X0), in(X0).";

		addRule(rule);

		return index;
	}

	private String translate(IntegerExpression ae, ComposableType type) {
		if (ae instanceof BinaryExpression)
			return translate((BinaryExpression) ae, type);

		if (ae instanceof IndexedExpression)
			return "N" + ((IndexedExpression) ae).getIndex();

		if (ae instanceof ConstantExpression) {
			if (((ConstantExpression) ae).isInteger())
				return ae.toString();
			else
				return NameMap.getSmNameForAttributeValue(ae.toString());
		}
				
//			return NameMap.getSmName(ae.toString());

		throw new Error("Unknown kind of IntegerExpression.");
	}

//	private void translate(InterfaceReference reference, ComposableType type) {
//		String rule = "";
//		
//		rule = interfaceReferenceText(reference) + " :- ";
//		rule += referenceText(reference.getReference(), 0, 0, type) + ", ";
//		rule += "pint(X0, I0, " + NameMap.getInterfaceName(reference.getName()) + ").";
////		rule += "pint(X0, I0, " + reference.getName() + ").";
//
//		addRule(rule);
//	}

	private void translate(InterfaceType interf) {
		String intName = NameMap.getSmName(interf);
		
	//	addRule("int(I) :- " + intName + "(I).");	Not needed at the moment.

		addRule("\n%% Interface type: " + interf.getName() + " (" + intName + ")\n");
		
		addRule("inttype(" + intName + ").");
		addRule("instance(I, " + intName + ") :- " + intName + "(I).");
	
		Iterator functions = interf.getFunctions();

		//	This rule is supposed to tell that connecting all functions implies connecting
		//	the entire interface. Not added to the program at the moment, though.
	//	String rule = "cn(I1, I2) :- ";
		while (functions.hasNext()) {
			String func = NameMap.getSmName(functions.next().toString());

			addRule("hasfunc(" + intName + ", " + func + ").");// :- " + intName + "(I).");
		//	rule += "cnf(I1, " + func + ", I2, " + func + "), ";
		}
		
		// attributes
		Iterator attributes = interf.getDirectAttributes();

		if (attributes.hasNext())
			addRule("\n%% Attributes\n");

		while (attributes.hasNext()) {
			AttributeDefinition ad = (AttributeDefinition) attributes.next();
			String attrName = NameMap.getSmName(ad); //	name clashes
			String attrType = NameMap.getSmName(ad.getType());

			String rule = "1 { hasattr(X, " + attrName + ", V) : " + attrType + "(V) } 1 ";
			rule += ":- in(X), " + intName + "(X).";

			addRule(rule);
		}		
	}

	private void translate(ComponentedModel model) {
		Iterator intfTypes = model.getInterfaceTypes();
		while (intfTypes.hasNext())
			translate((InterfaceType) intfTypes.next());

		Iterator components = model.getComponentTypes();
		while (components.hasNext())
			translate((ComponentType) components.next(), model);
	}
	
	private void translate(Model model) {

		if (model instanceof AttributedModel) {
			translate((AttributedModel) model);
		}
		if (model instanceof ComponentedModel) {
			translate((ComponentedModel) model);
		}
		if (model instanceof FeaturedModel) {
			translate((FeaturedModel) model);
		}
	}
	
	/**
	 * 
	 * 
	 * @param pe
	 * @param type not really needed here, as the ref()() is a domain predicate.
	 * @return the index of the constraint generated
	 */
	private int translate(PresentExpression pe, ComposableType type) {
		int index = constrIndex++;

		// constr(X, ...) :-
		String rule = publicConstraint(pe, index) + " :- ";
		rule += referenceText(pe.getReference(), 0, 0, type) + ", in(X0).";

		addRule(rule);

		return index;
	}

	/**
	 * @param expression
	 * @param type
	 * @return
	 */
//	private int translate(PresentInterfaceExpression pie, ComposableType type) {
//		int index = constrIndex++;
//
//		String rule = publicConstraint(pie, index) + " :- ";
//		rule += interfaceReferenceText(pie.getReference()) + ", ";
//		rule += "in(I).";
//		addRule(rule);
//
//		return index;
//	}
	
	private int translate(QuantifiedExpression qe, ComposableType type) {
		boolean innerNegated, outerNegated = false;
		int prev, curr;
		String rule = "";

		prev = translate(qe.getExpression(), qe, type);

		QuantificationStack temp = new QuantificationStack(qe.addedVariables());
		innerNegated = qe.getExpression().isNegated();
		while (!temp.isEmpty()) {
			Quantification quant = temp.pop();
			boolean universal = quant.isUniversal();

			if (universal && innerNegated) {
				innerNegated = false;
				outerNegated = true;
			} else if (universal && !innerNegated) {
				innerNegated = true;
				outerNegated = true;
			} else if (!universal && innerNegated) {
				innerNegated = true;
				outerNegated = false;
			} else { // if (!universal && !innerNegated) {
				innerNegated = false;
				outerNegated = false;
			}

			//	If necessary, negate the inner constraint
//			if (innerNegated) {
//				curr = constrIndex++;
//
//				rule = generalConstraint(qe, temp, quant, curr) + " :- not ";
//				rule += generalConstraint(qe, temp, quant, prev);
//				rule += generalDomains(qe, temp, quant, type) + ".";
//				prev = curr;
//				addRule(rule);
//			}

			curr = constrIndex++;
			
			/*
			 * Remove the variable. - This could be combined with possible negating the inner
			 * constraint. In fact, wouldn't it be just possible to add a not right after the :-? 
			 */			
			rule = generalConstraint(qe, temp, null, curr) + " :- ";
			if (innerNegated)
				rule += "not ";
			rule += generalConstraint(qe, temp, quant, prev);
			rule += generalDomains(qe, temp, quant, type) + ".";
			addRule(rule);

			prev = curr;
			
			/*
			 * Handle the possible negation of the quantifier.
			 * A possible double negation is eliminated.
			 */
			if (quant.isNegated())
				outerNegated = !outerNegated;
				
			innerNegated = outerNegated;
		}

		if (outerNegated) { //	Negated the outmost constraint if necessary
			curr = constrIndex++;
			rule = generalConstraint(qe, temp, null, curr) + " :- not ";
			rule += generalConstraint(qe, temp, null, prev);
			rule += generalDomains(qe, temp, null, type) + ".";
			addRule(rule);

			prev = curr;
		}

		return prev;
	}

	private int translate(ReferenceExpression re, ComposableType type) {
		if (re instanceof InstanceOfExpression)
			return translate((InstanceOfExpression) re, type);
		if (re instanceof PresentExpression)
			return translate((PresentExpression) re, type);

		throw new Error("Unknown type of ReferenceExpression.");
	}

	/**
	 * Generates rules capturing the StructureReference given as a parameter. Most
	 * importantly, the function creates the ref(I, X, ...) and card(I, X, ...) for
	 * the StructureReference (the latter if required).
	 *  
	 * @param sr
	 * @param ct
	 */
	private void translate(StructureReference sr, ComposableType ct) {
		int index = sr.getIndex(), level = 1;
		String rule;
		String composableName = NameMap.getSmName(ct);

		rule = "\n% Reference: id = " + sr.getIndex();
		rule += ", type of origin = " + sr.getOrigin();
		rule += ", path = " + sr.toString() + "\n";
		addRule(rule);
		
		if (sr.isVoid()) {
			addRule("% This reference cannot refer to any instances.\n");
			return;
		}

		// The cardinality represents the maximum cardinality of the references,
		// i.e., the maximum number of instances that may occur as parts pointed
		// by the reference.
		int card = sr.getCardinality();
		rule = "poscard(" + index + ", 0.." + card + ").";
		addRule(rule);

		if (sr.isLocal() && sr.getVariable() == null) {
			if (!sr.isInterfaceReference()) {
				rule = "ref" + index + "(X, X) :- ";
				rule += composableName + "(X).";
			} else {
				rule = "ref" + index + "(X, I) :- ";
				rule += "pint(X, I, " + NameMap.getInterfaceName(sr.getInterfaceName()) + ").";
			}

			addRule(rule);
			return;
		}

		//	The first rule
		if (sr.getVariable() != null) {
			Variable var = sr.getVariable();

			if (sr.isLocal()) {
				if (!sr.isInterfaceReference()) {
					//	ref(index, X, ...) :- ref(varindex, X, ...), domains.
					rule = referenceText(sr, 0, 0) + " :- ";
					rule += referenceText(var.getReference(), 0, 0) + ", ";
					rule += "V" + var.getIndex() + " = X0.";
				} else {
					//	ref(index, X, ...) :- ref(varindex, X, ...), domains.
					rule = referenceText(sr, 0, 0) + " :- ";
					rule += referenceText(var.getReference(), 0, 1) + ", ";
					rule += "V" + var.getIndex() + " = X1, pint(X1, X0, ";
					rule += NameMap.getInterfaceName(sr.getInterfaceName()) + ")";
					
					if (sr.getInterfaceType() != null)
						rule += ", " + NameMap.getSmName(sr.getInterfaceType()) + "X1).";
					else
						rule += ".";
				}

				addRule(rule);

				return;
			}

			rule = referenceText(sr, level, 0) + " :- ";
			rule += referenceText(var.getReference(), 0, 0) + ", ";
			rule += "V" + var.getIndex() + " = X0.";
			addRule(rule);
		} else {
			if (sr.isImplementation())
				rule = "ref" + index + "_" + level + "(X, X0) :- croot(C), ppart(C, X0, ";
			else
				rule = "ref" + index + "_" + level + "(X, X0) :- ppart(X, X0, ";

			rule += NameMap.getPartName(sr.getName(level - 1)) + ", I), ";
			rule += composableName + "(X)";
			
			if (sr.getType(level - 1) != null)
				rule += ", " + NameMap.getSmName(sr.getType(level - 1)) + "(X0)";
			
			rule += ".";

			addRule(rule);
		}

		level++;
		for (; level <= sr.getDepth(); level++) {
			rule = referenceText(sr, level, 1) + " :- ";
			rule += referenceText(sr, level - 1, 0) + ", ";
			rule += "ppart(X0, X1, " + NameMap.getPartName(sr.getName(level - 1)) + ", I)";
			
			if (sr.getType(level - 1) != null)
				rule += ", " + NameMap.getSmName(sr.getType(level - 1)) + "(X1).";
			else
				rule += ".";

			addRule(rule);
		}

		if (!sr.isInterfaceReference()) { 
			rule = referenceText(sr, 0, 0) + " :- ";
			rule += referenceText(sr, level - 1, 0) + ".";
		} else {	//	An interface reference.
			rule = referenceText(sr, 0, 1) + " :- ";
			rule += referenceText(sr, level - 1, 0) + ", pint(X0, X1, ";
			rule += NameMap.getInterfaceName(sr.getInterfaceName()) + ")";
			
			if (sr.getInterfaceType() != null)
				rule += ", " + NameMap.getSmName(sr.getInterfaceType()) + "(X1).";
			else
				rule += ".";
		}
		
		addRule(rule);

		
		if (!sr.isCardinality())
			return;
		
		//	Define the cardinality predicate for the reference	
		rule = "card(X, " + index + ", " + variableString(sr) + "N) :- ";
		rule += "N { in(X0) : " + referenceText(sr, 0, 0) + " } N, ";
		rule += composableName + "(X), poscard(" + index + ", N).";
		addRule(rule);

		/*
		 *	Cardinality specific things continue.
		 *	The rule that the index of a part may not exceed the cardinality.
		 *	Really helps only if there is only one allowed part type. 
		 */
		rule = " :- card(X, " + index + ", " + variableString(sr) + "N), ";
		rule += referenceText(sr, 0, 0) + ", ";
		rule += "in(X0), index(X0, M), M > N, poscard(" + index + ", N), ";
		rule += composableName + "(X).";
		addRule(rule);
	}

	private String variableString(StructureReference ref) {
		if (ref == null || ref.getVariable() == null)
			return "";
		
		Variable var = ref.getVariable();
		String ret = "";
		
		while (var != null) {
			ret += "V" + var.getIndex() + ", ";
			var = var.getReference().getVariable();
		}

		return ret;
	}
	
	private class Rule {
		private String head;
		private List<String> body;
		
		public Rule() {
			body = new ArrayList<String>();
		}
		
		public Rule(String head) {
			this();
			this.head = head;
		}
		
		public void setHead(String head) {
			if (head == null)
				this.head = head;
		}
		
		public void addToBody(String atom) {
			if (!body.contains(atom))
				body.add(atom);
		}
		
		public String toString() {
			StringBuffer ret = new StringBuffer();
			
			if (head == null && body == null)
				return ret.toString();
			
			if (head != null)
				ret.append(head);
			
			if (body.isEmpty())
				ret.append(".").toString();
			
			ret.append(" :- ");
			
			for (Iterator it = body.iterator(); it.hasNext(); ) {
				ret.append(it.next());
				if (it.hasNext())
					ret.append(", ");
				else
					ret.append(".");
			}
			
			return ret.toString();
		}
	}
}
