/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Configuration.java
 *
 * Created on 3. maaliskuuta 2004, 9:02
 */

package kumbang.core.configuration.core;

import java.io.Serializable;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Iterator;


import kumbang.core.configuration.instance.AttributeInstance;
import kumbang.core.configuration.instance.ComponentInstance;
import kumbang.core.configuration.instance.ComposableInstance;
import kumbang.core.configuration.instance.FeatureInstance;
import kumbang.core.configuration.instance.Instance;
import kumbang.core.configuration.instance.InterfaceInstance;
import kumbang.core.configuration.instance.PartInstance;
import kumbang.core.configuration.reference.AttributeReference;
import kumbang.core.configuration.reference.BindingReference;
import kumbang.core.configuration.reference.ComposableReference;
import kumbang.core.configuration.reference.InterfaceReference;
import kumbang.core.configuration.reference.Reference;
import kumbang.core.configuration.reference.ReferenceItem;
import kumbang.core.language.parser.KumbangParser;
import kumbang.core.language.parser.ParseException;
import kumbang.core.language.syntaxtree.SyntaxConfigurationReference;
import kumbang.core.language.visitor.ConfigurationReferenceVisitor;
import kumbang.core.model.core.ComponentedModel;
import kumbang.core.model.core.FeaturedModel;
import kumbang.core.model.core.Model;
import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.FeatureType;
import kumbang.core.model.type.PartDefinition;

/**
 * Configuration is nothing but the component and feature roots. It may include some kind of search 
 * functions, though. But even these should be delegated to the roots.
 * 
 * TODO Consider adding the model as a member here. -toasikai
 * 
 * @author vmyllarn
 */
public class Configuration implements Serializable {

	private static final long serialVersionUID = -3914568156081109234L;
	
	private ComponentInstance componentRoot;
	private FeatureInstance featureRoot;
	
	/**
	 * The Model this configuration is based on�.
	 */
	protected Model model; 

	public Configuration(Model model) {
		this.model = model;
		instantiateRoots(); // possibility to override
	}

	/**
	 * Instantiate roots of the configuration. Called from the constructor.
	 * Note that this method must be protected and separate from constructor,
	 * since subclasses of this class may want to override this method
	 * to implement custom roots.
	 */
	protected void instantiateRoots() {
		if (model instanceof FeaturedModel) {
			FeaturedModel fModel = (FeaturedModel) model;
			if (fModel.getFeatureRoot() != null) {
				FeatureType rootType = fModel.getFeatureRoot();
				featureRoot = new FeatureInstance(rootType);		
			}
		}
		if (model instanceof ComponentedModel) {
			ComponentedModel cModel = (ComponentedModel) model;
			if (cModel.getComponentRoot() != null) {	
				ComponentType rootType = cModel.getComponentRoot();
				componentRoot = new ComponentInstance(rootType);				
			}
		}
	}

	/**
	 * Set feature root of the configuration.
	 * Note that this method must be protected, since subclasses of 
	 * this class may want to to implement custom roots.
	 * @param instance
	 */	
	protected void setFeatureRoot(FeatureInstance instance) {
		featureRoot = instance;
	}
	
	/**
	 * Set component root of the configuration.
	 * Note that this method must be protected, since subclasses of 
	 * this class may want to to implement custom roots.
	 * @param instance
	 */		
	protected void setComponentRoot(ComponentInstance instance) {
		componentRoot = instance;
	}
	
	/**
	 * Returns the root of the component configuration.
	 * 
	 * @return ComponentInstance root
	 */
	public ComponentInstance getComponentRoot() {
		return this.componentRoot;
	}

	/**
	 * Returns the root of the feature configuration.
	 * 
	 * @return FeatureInstance root
	 */
	public FeatureInstance getFeatureRoot() {
		return this.featureRoot;
	}

	public Iterator getInstances() {
		ArrayList list = new ArrayList();
		
		if (componentRoot != null)
			list.addAll(componentRoot.getInstanceList(null, true));
		
		if (featureRoot != null)
			list.addAll(featureRoot.getInstanceList(null, true));
		
		return list.iterator();
	}

	/**
	 * Resets the configuration.
	 */
	public void reset() {
		this.componentRoot = null;
		this.featureRoot = null;
	}

	public Model getModel() {
		return model;
	}

	
	/**
	 * Retrieves a part instance from the configuration based on the reference given as argument.
	 * 
	 * @param refString a reference describing an instance in the configuration 
	 * @return the part instance corresponding to ref if one exists, null otherwise
	 */
	public PartInstance getPart(String refString) {
		Reference ref = parseReference(refString);
		if(ref == null)
			return null;
		try {
			ComposableReference cr = (ComposableReference) ref;
			cr.setPartReference();
			return (PartInstance) getInstance(cr);
		} catch (ClassCastException e) {
			System.out.println("class cast :(");
			return null;
		}
	}
	
	/**
	 * Retrieves an instance from the configuration based on the reference given as argument.
	 * 
	 * @param refString a reference describing an instance in the configuration 
	 * @return the instance corresponding to ref if one exists, null otherwise
	 */
	public Instance getInstance(String refString) {
		if (refString == null)
			return null;
		
		Reference ref = parseReference(refString);
		if (ref == null)
			return null;
		
		return getInstance(ref);
	}
	
	public Instance getInstance(Reference ref) {
		ComposableInstance pointer;
		if (ref.isComponentReference())
			pointer = getComponentRoot();
		else
			pointer = getFeatureRoot();
		
		if (ref instanceof BindingReference) {
			BindingReference br = (BindingReference) ref;
			InterfaceReference sourceRef = br.getSource();
			InterfaceReference targetRef = br.getTarget();
			
			ComponentInstance sourceComp, targetComp;
			InterfaceInstance sourceInt, targetInt;
			
			sourceComp = (ComponentInstance) getComposableInstance(pointer, sourceRef.getComposable());
			targetComp = (ComponentInstance) getComposableInstance(pointer, targetRef.getComposable());
			
			if (sourceComp == null || targetComp == null)
				return null;
			
			sourceInt = sourceComp.getInterface(sourceRef.getName());
			targetInt = targetComp.getInterface(targetRef.getName());
			
			if (sourceInt == null || targetInt == null)
				return null;
			
			return  sourceInt.getBindingWith(targetInt);
		}
		
		ComposableReference compRef = ref.getComposable();
		
		if (compRef.isPartReference())
			return getComposableInstance(pointer, compRef);
		
		ComposableInstance comp = (ComposableInstance) getComposableInstance(pointer, compRef);
		
		if (comp == null)
			return null;
		
		if (ref instanceof ComposableReference)
			return comp;
		
		if (ref instanceof AttributeReference) {
			AttributeReference attr = (AttributeReference) ref;
			AttributeInstance ai = comp.getAttribute(attr.getName());
			
			if (ai == null)
				return null;
			
			if (attr.getValue() == null || ai.getValue().equals(attr.getValue()))
				return ai;
			
			return null;
		}
		
		if (ref instanceof InterfaceReference) {
			if (!ref.isComponentReference())
				return null;
			
			InterfaceReference ir = (InterfaceReference) ref;
			InterfaceInstance ii = ((ComponentInstance) comp).getInterface(ir.getName());
			if (ii == null)
				return null;
			
			if (ir.getType() == null || ii.getType().toString().equals(ir.getType()))
				return ii;
			
			return null;
		}
		
		return null;
	}

	private Instance getComposableInstance(ComposableInstance pointer, ComposableReference compRef) {
		// The root feature/component
		if (!compRef.getItems().hasNext())
			return pointer;
		
		for (Iterator it = compRef.getItems(); it.hasNext();) {
			ReferenceItem ri = (ReferenceItem) it.next();
			String typeString = ri.getType();
			ComposableType reqType = (ComposableType) getModel().getType(typeString,
					pointer.getInstanceKind());
			int reqIndex = ri.getIndex();

			ComposableType type = (ComposableType) pointer.getType();
			PartDefinition def = type.getPart(ri.getName());
			if (def == null)
				return null;

			PartInstance pi = pointer.getPart(def);
			if (compRef.isPartReference() && !it.hasNext())
				return pi;

			ComposableInstance test = null;
			int i = 0;
			for (Iterator pit = pi.getInstances(); pit.hasNext();) {
				test = (ComposableInstance) pit.next();
				if (test.getType().equals(reqType) || reqType == null) {
					// Found an instance of the right type
					if (reqIndex == -1 || reqIndex == i) {
						// The index matches as well
						if (it.hasNext()) {
							// Proceed to the next ReferenceItem
							pointer = test;
							break;
						} else
							return test;
					} else if (reqIndex != -1)
						i++;
				}
			}

			/*
			 * A hack, what we would have liked to do is to continue the outer loop while still in
			 * the inner loop.
			 */
			if (pointer == test)
				continue;

			// Couldn't find the right kind of part instance.
			return null;
		}

		return null;
	}

	public static Reference parseReference(String refString) {

		Reference ref = null;
		try {
			// This throws a nullpointer if parser hasn't been
			// previously initialized
			KumbangParser.ReInit(new StringReader(refString));
		} catch (NullPointerException e) {
			KumbangParser kp = new KumbangParser(new StringReader(refString));
		}

		try {
			SyntaxConfigurationReference scr = KumbangParser.SyntaxConfigurationReference();
			ref = ConfigurationReferenceVisitor.getReference(scr);
		} catch (ParseException e) {
			e.printStackTrace();
			return null;
		}
		return ref;
	}
	
	
}
