/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * BindingInstance.java
 *
 * Created on 2. maaliskuuta 2004, 12:47
 */

package kumbang.core.configuration.instance;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import kumbang.core.configuration.reference.BindingReference;
import kumbang.core.configuration.reference.InterfaceReference;
import kumbang.core.configuration.reference.Reference;

/**
 * 
 * @author vmyllarn
 */
public class BindingInstance extends Instance {

	/**
	 * This is an interface that is either required, or provided interface that is connected with a
	 * provided interface contained as a part in the component.
	 */
	private InterfaceInstance source;

	/**
	 * This is an interface that is either provided, or required interface that is connected with a
	 * required interface contained as a part in the component.
	 */
	private InterfaceInstance target;

	/**
	 * This contains all the functions that are binded by this binding. This list can only contain
	 * functions that are common to both interfaces.
	 * 
	 * The current modelling method does not include the notion of function binding. -toasikai
	 */
	private List<String> functions;

	/**
	 * Creates a new instance of BindingInstance that binds two given interface instances together.
	 * 
	 * @param target target interface
	 * @param source source interface
	 */
	public BindingInstance(InterfaceInstance source, InterfaceInstance target, boolean userAdded) {
		super(null, userAdded);
		
		this.source = source;
		this.target = target;
		
		this.functions = new ArrayList<String>();
		 
		source.addBinding(this);
		target.addBinding(this);
	}
	
	/*
	 * TODO Refactor to getSource
	 * 
	 * Something to ponder: How would this go should the bindings not be directed?
	 */
	public InterfaceInstance getSourceInterface() {
		return source;
	}

	/*
	 * TODO Refactor to getTarget
	 */
	public InterfaceInstance getTargetInterface() {
		return target;
	}

	/**
	 * Returns the other end interface that this binding connects.
	 * @param end 
	 * @return Other interface connected by this binding, or null if the given interface
	 * is not bound by this binding.
	 */
	public InterfaceInstance getOtherEnd(InterfaceInstance end) {
		if (end == source) {
			return target;
		} else if (end == target) {
			return source;
		} else {
			return null;
		}
	}
	
	public Iterator getFunctions() {
		return functions.iterator();
	}

	public void addFunction(String f) {
		if (source.getInterfaceType().hasFunction(f) && target.getInterfaceType().hasFunction(f)) {
			functions.add(f);
		}
	}

	public void removeFunction(String f) {
		functions.remove(f);
	}
	
	/**
	 * Tests whether this instance is equal to another instance
	 * 
	 * @param instance
	 */
	public boolean equals(Object o) {
		if (o instanceof BindingInstance) {
			BindingInstance bi = (BindingInstance) o;
			return bi != null && source.equals(bi.source) && target.equals(bi.target);
		}

		return false;
	}

	public String toString() {
		return source.getParent().toString() + " . " + source.toString() + " - "
				+ target.getParent().toString() + " . " + target.toString();
	}

	public String getDescription() {
		return source.getParent().getDescription() + "." + source.getDescription() + " -> "
				+ target.getParent().getDescription() + "." + target.getDescription();
	}

	public String getInstanceKind() {
		return "binding";
	}

	/**
	 * Removes this BindingInstance form the configuration.
	 */
	public void remove() {
		source.removeBinding(this);
		target.removeBinding(this);
	}
	
	public void restore() {
		// TODO Is this alright, or should the binding lists be sorted as well?
		source.addBinding(this);
		source.addBinding(this);
	}
	
	public Reference getPath() {
		return new BindingReference((InterfaceReference) source.getPath(),
				(InterfaceReference) target.getPath());
	}
}
