/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ComposableInstance.java
 *
 * Created on 2. maaliskuuta 2004, 11:55
 */

package kumbang.core.configuration.instance;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kumbang.core.configuration.reference.ComposableReference;
import kumbang.core.configuration.reference.Reference;
import kumbang.core.configuration.reference.ReferenceItem;
import kumbang.core.model.type.AttributeDefinition;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ElementaryType;
import kumbang.core.model.type.PartDefinition;

/**
 * TODO Are there instances that are typed but not identified? Combine the superclass and
 * implemented interface? -toasikai
 * 
 * Yes there are, attribute instances should typed but not identified. -vmyllarn
 * 
 * @author vmyllarn
 */
public abstract class ComposableInstance extends TypedInstance implements AttributedInstance {
	/**
	 * Parent part instance of this composable instance. This instance is contained in this part
	 * instance. Parent part is null if this is the root instance in the hierarchy.
	 */
	private PartInstance parentPart;
	
	/** The direct type of this composable instance */
	private ComposableType type;

	/**
	 * List of PartInstance objects under this ComposableInstance object.
	 */
	protected List<PartInstance> parts;

	/** 
	 * List of attribute instances, it is.
	 */
	protected List<AttributeInstance> attributes;
	
	/**
	 * Creates a new instance of ComposableInstance. This also informs the parent that is has a new
	 * child as part. Instantiates parts and attributes.
	 * 
	 * @param type type of this instance
	 * @param parent of this instance (as in part relationship)
	 */
	public ComposableInstance(ComposableType type, PartInstance parent, boolean userAdded) {
		super(parent != null ? (ComposableInstance) parent.getParent() : null, userAdded);
		this.type = type;
		this.parentPart = parent;
		
		if (parentPart != null)
			parentPart.addInstance(this);

		this.attributes = new ArrayList<AttributeInstance>();
		this.parts = new ArrayList<PartInstance>();
		
		initialiseAttributes();
		initialiseParts();
	}

	/**
	 * This method initialises attributes in the composable instance.
	 * Is called from constructor.
	 * Note: this method must be protected and separate from constructor
	 * in order to allow subtypes to implement custom attributes.
	 */
	protected void initialiseAttributes() {
		for (Iterator iter = type.getAttributes(); iter.hasNext(); ) {
			AttributeDefinition attrDef = (AttributeDefinition) iter.next();
			AttributeInstance attr = new AttributeInstance(this, attrDef, false);
			attributes.add(attr);
		}		
	}
	
	/**
	 * This method initialises parts in the composable instance.
	 * Is called from constructor.
	 * Note: this method must be protected and separate from constructor
	 * in order to allow subtypes to implement custom parts.
	 */	
	protected void initialiseParts() {
		for (Iterator iter = type.getParts(); iter.hasNext(); ) {
			PartDefinition def = (PartDefinition) iter.next();
			PartInstance part = new PartInstance(this, def);
			parts.add(part);
		}			
	}
	
	public String getName() {
		if (getDefinition() != null) {
			return getDefinition().getName();
		} else {
			return "root";
		}
	}

	public String getDisplayName() {
		if (getDefinition() != null) {
			return getDefinition().getDisplayName();
		} else {
			return "root " + getInstanceKind();
		}
	}

	private String getIndex() {
		// This happens if this is a root instance.
		if (parentPart == null)
			return "";
		
		int index = parentPart.getIndex(this);
		if (index == 0 && parentPart.getSize() == 1)
			return "";
		
		return "[" + parentPart.getIndex(this) + "]";
	}
	
	/*
	 * This index stuff does really not belong here. As said before, indeces are not
	 * something for instances to know, not their property.
	 * However, they are needed for display purposes. How to separate indexes so
	 * that GUI can get this information easily?
	 */
	public String getDisplayNameWithIndex() {
		return getDisplayName() + getIndex(); 
	}

	public String getNameWithIndex() {
		return getName() + getIndex();
	}

	public String toString() {
		return getDisplayNameWithIndex() + ":" + getType().getDisplayName();
	}

	public String getDescription() {
		return type.getName();
	}

	public PartInstance getParentPart() {
		return this.parentPart;
	}

	public ComposableType getComposableType() {
		return type;
	}

	/*
	 * If ComposableType is a subtype of Elementary type, this should not be overriden. 
	 */
//	public ComposableType getType() {
//		return type;
//	}

	/**
	 * Returns the part instance with the part definition, or null if no such part instance exists.
	 * 
	 * TODO Convert parts to (PartDefinition, PartInstance) pairs. -toasikai 25.7.2006
	 * 
	 * @param partDefinition part definition for the part
	 * @return PartInstance part instance there is part in this composable instance, or null if no
	 *         such part exists
	 */
	public PartInstance getPart(PartDefinition partDefinition) {
		for (Iterator it = parts.iterator(); it.hasNext(); ) {
			PartInstance pi = (PartInstance) it.next();
			PartDefinition def = pi.getDefinition();
			if (def.equals(partDefinition))
				return pi;
		}
		
		// Not found.
		return null;
	}

	/**
	 * Returns an iterator ranging over the parts in this composable instance, i.e., lists.
	 * 
	 * @return Iterator an iterator containing part objects in this composable instance
	 */
	public Iterator getParts() {
		return parts.iterator();
	}

	public boolean hasParts() {
		return !parts.isEmpty();
	}

	public int getNumberOfParts() {
		return parts.size();
	}

	public boolean hasInstancesAsPart() {
		Iterator pit = getParts();
		while (pit.hasNext()) {
			PartInstance part = (PartInstance) pit.next();
			if (part.hasInstances()) {
				return true;
			}
		}
		
		return false;
	}

	private List getPartInstanceList() {
		List parts = new ArrayList();
		
		Iterator pit = getParts();
		while (pit.hasNext()) {
			PartInstance partList = (PartInstance) pit.next();
			Iterator it = partList.getInstances();
			while (it.hasNext())
				parts.add(it.next());
		}
		
		return parts;
	}
	
	/**
	 * Returns an iterator of all composable instances that are as parts in this composable
	 * instance. This is a convenience method for browsing trough all instances under this
	 * composable instance.
	 * 
	 * @return Iterator an iterator containing composable instance objects that are as parts in this
	 *         composable instance
	 */
	public Iterator getPartInstances() {
		return getPartInstanceList().iterator();
	}

	/**
	 * Returns an iterator of all composable instances that are as part in this composable instance
	 * and have directly the specified type. Implementation is the same as in getPartsOfType(type,
	 * false).
	 * 
	 * @param type type of the composable instance
	 * @return Iterator an iterator containing composable instances
	 */
	public Iterator getInstancesAsPartsOfType(ComposableType type) {
		return getInstanceList(type, false).iterator();
		//return getInstancesAsPart(type, false);
	}


//	public Iterator getInstancesAsPart(ComposableType type, boolean recursive) {
//		return getInstanceList(type, recursive).iterator();
//	}	
	
	public List getInstanceList(ComposableType type, boolean recursive) {
		List partList = new ArrayList();
				
		// this can also be included in the list. 
		if (this.type == type || type == null)
			partList.add(this);
		
		//	This repeats the logic of the getPartInstanceList method.
		for (Iterator it = getPartInstances(); it.hasNext(); ) {
			ComposableInstance instance = (ComposableInstance) it.next();
			
			// TODO This does not account for a non-dag part-of relation.
			if (recursive) //	instance.getType() == type
				partList.addAll(instance.getInstanceList(type, recursive));
			else if (instance.getType() == type || type == null)
				partList.add(instance);
		}
		
		return partList;
	}


		
	public boolean hasAttributes() {
		return !attributes.isEmpty();
	}

	public int getNumberOfAttributes() {
		return attributes.size();
	}

	/**
	 * Returns an iterator of all attributes in this instance. Implementation can be changed if
	 * hashmaps are to be used.
	 */
	public Iterator<AttributeInstance> getAttributes() {
		return attributes.iterator();
	}

	/**
	 * Returns an attribute with the given name, or null if no such attribute exists. Implementation
	 * can be changed if hashmaps are to be used.
	 */
	public AttributeInstance getAttribute(String attributeName) {
		for (AttributeInstance attr : attributes) {
			if (attr.getName().equals(attributeName)) {
				return attr;
			}
		}
		return null;
	}

//	/**
//	 * Returns an attribute with the given name, or null if no such attribute exists. Implementation
//	 * can be changed if hashmaps are to be used.
//	 */
//	private AttributeInstance getAttribute(String attributeName, boolean ignoreCase) {
//		for (AttributeInstance attr : attributes) {
//			if (ignoreCase) {
//				if (attr.getName().equalsIgnoreCase(attributeName)) {
//					return attr;
//				}
//			} else {
//				if (attr.getName().equals(attributeName)) {
//					return attr;
//				}
//			}
//		}
//		return null;
//	}

	public void removeComposable(ComposableInstance instance) {
		PartInstance parent = instance.getParentPart();
		parent.removeInstance(instance);
	}
	
	private PartDefinition getDefinition() {
		if (parentPart != null)
			return parentPart.getDefinition();
		
		return null;
	}
	
	public void remove() {
		((ComposableInstance) getParent()).removeComposable(this);
	}
	
	/**
	 * Restores a previously removed instance to the configuration.
	 * 
	 * TODO Add some checks, so that it is not possible to restore instances that have not
	 * previously deleted.
	 */
	public void restore() {
		parentPart.addInstance(this);
	}
	
	public ElementaryType getType() {
		return type;
	}
	
	/**
	 * Returns a Reference object that refers to this entity in the configuration
	 * it belongs to.
	 * 
	 * @return see above.
	 */
	public Reference getPath() {
		if (parentPart == null)
			return new ComposableReference(getType() instanceof ComponentType,
					new LinkedList());
		
		return new ComposableReference((ComposableReference) parentPart.getParent().getPath(),
				new ReferenceItem(getName(), parentPart.getIndex(this), getType().toString()),
				false);
	}
	
	/*
	 * vmyllarn
	 * This is a stupid hack that can be avoided if interfaces are used instead of
	 * inheritance in creating type hierarchy for instances.
	 */
	public TypedInstance getAsTypedInstance() {
		return this;
	}
}
