/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * InterfaceInstance.java
 *
 * Created on 2. maaliskuuta 2004, 12:21
 */

package kumbang.core.configuration.instance;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import kumbang.core.configuration.reference.ComposableReference;
import kumbang.core.configuration.reference.InterfaceReference;
import kumbang.core.configuration.reference.Reference;
import kumbang.core.model.type.AttributeDefinition;
import kumbang.core.model.type.ElementaryType;
import kumbang.core.model.type.InterfaceDefinition;
import kumbang.core.model.type.InterfaceType;

/**
 * 
 * @author vmyllarn
 */
public class InterfaceInstance extends TypedInstance implements AttributedInstance {
	/** Parent component that this interfaces belongs to */
	private ComponentInstance parent;

	/**
	 * Interface definition (needed for naming, for example)
	 */
	private InterfaceDefinition definition;

	/** Interface type (separate from definition) */
	private InterfaceType type;

	/** Tells whether this interface is requires or provides */
	private boolean provides;

	/** Tells whether this interface is grounded */
	private boolean grounded;

	/** A list of all bindings of this interface */
	private List<BindingInstance> bindings;
	
	/** 
	 * List of attribute instances, it is.
	 */
	protected List<AttributeInstance> attributes;

	/**
	 * Creates a new instance of InterfaceInstance and places it under given parent component.
	 * 
	 * @param copy instance to be copied
	 */
	public InterfaceInstance(InterfaceType type, ComponentInstance parent,
			InterfaceDefinition definition, boolean userAdded) {
		super(parent, userAdded);
		this.definition = definition;
		this.type = type;
		this.parent = parent;
		this.provides = definition.isProvides();
		this.grounded = definition.isGrounded();
		this.bindings = new ArrayList<BindingInstance>();
		this.attributes = new ArrayList<AttributeInstance>();
		
		initialiseAttributes();
	}


	/** 
	 * This method initialises attributes in the composable instance.
	 * Is called from constructor.
	 * Note: this method must be protected and separate from constructor
	 * in order to allow subtypes to implement custom attributes.
	 */
	protected void initialiseAttributes() {
		for (Iterator iter = type.getAttributes(); iter.hasNext(); ) {
			AttributeDefinition attrDef = (AttributeDefinition) iter.next();
			AttributeInstance attr = new AttributeInstance(this, attrDef, false);
			attributes.add(attr);
		}		
	}	
	
	public String getName() {
		return definition.getName();
	}

	public String getDisplayName() {
		return definition.getDisplayName();
	}

	public String toString() {
		return getDisplayName() + ":" + type.getDisplayName() ;
	}

	public String getDescription() {
		return type.getName();
	}

	public String getInstanceKind() {
		return "interface";
	}

	public boolean isProvided() {
		return provides;
	}

	public boolean isRequired() {
		return !provides;
	}

	public boolean isOptional() {
		return definition.isOptional();
	}

	public boolean isGrounded() {
		return grounded;
	}

//	public ComponentInstance getParent() {
//		return parent;
//	}

	public InterfaceDefinition getInterfaceDefinition() {
		return definition;
	}

	/**
	 * @return an InterfaceType object that is the type of this instance
	 */
	public ElementaryType getType() {
		return type;
	}

	/**
	 * @return an InterfaceType object that is the type of this instance
	 */
	public InterfaceType getInterfaceType() {
		return type;
	}

	public Iterator getFunctions() {
		return type.getFunctions();
	}

//	public boolean equals(Object o) {
//		if (!(o instanceof InterfaceInstance))
//			return false;
//
//		InterfaceInstance comp = (InterfaceInstance) o;
//		return comp.grounded == grounded && comp.definition == definition && comp.type == type
//				&& comp.provides == provides && comp.grounded == grounded && comp.parent == parent;
//	}

	public Iterator<BindingInstance> getBindings() {
		return bindings.iterator();
	}

	public boolean hasBindings() {
		if (bindings.size() > 0)
			return true;
		else
			return false;
	}

	/**
	 * Adds a binding to this interface. Does not modify the binding itself.
	 * 
	 * @param binding a binding to be added
	 */
	public void addBinding(BindingInstance binding) {
		if (!bindings.contains(binding))
			bindings.add(binding);
	}

	/**
	 * Removes a binding from this interface. Does not modify the binding itself. Does nothing if
	 * the binding is not connected to this interface.
	 * 
	 * @param binding a binding to be removed
	 */
	public void removeBinding(BindingInstance binding) {
		bindings.remove(binding);
	}

	public void remove() {
		getComponent().removeInterface(this);
	}

	public void restore() {
		parent.addInterface(this);
	}

	/**
	 * A method for testing if a specific binding exists. This could be done more elegantly.
	 * 
	 * TODO Do this more elegantly. -toasikai 3.8.2006
	 * 
	 * @param target
	 * @return
	 */
	public boolean isBoundTo(InterfaceInstance target) {
		for (Iterator it = bindings.iterator(); it.hasNext(); ) {
			BindingInstance b = (BindingInstance) it.next();
			if (b.getSourceInterface() == this && b.getTargetInterface() == target)
				return true;
		}
		
		return false;
	}
	
	public ComponentInstance getComponent() {
		return parent;
	}
	
    /**
	 * Returns the binding between this and the given interface, or null
	 * if there is no such binding.
	 * 
	 * NOTE: can there be many bindings between same interfaces?
	 * 
	 * @param interf other interface
	 * @return BindingInstance binding between instances, or null if there
	 *         is no such binding
	 */
	public BindingInstance getBindingWith(InterfaceInstance interf) {
		Iterator it = bindings.iterator();
		while (it.hasNext()) {
			BindingInstance bind = (BindingInstance) it.next();
			if (bind.getSourceInterface().equals(interf)
					|| bind.getTargetInterface().equals(interf)) {
				return bind;
			}
		}

		return null;
	}

	@Override
	public Reference getPath() {
		return new InterfaceReference((ComposableReference) getParent().getPath(), definition
				.getName());
	}

	public AttributeInstance getAttribute(String attributeName) {
		for (AttributeInstance attr : attributes) {
			if (attr.getName().equals(attributeName)) {
				return attr;
			}
		}
		return null;
	}

	public Iterator<AttributeInstance> getAttributes() {
		return attributes.iterator();
	}

	public int getNumberOfAttributes() {
		return attributes.size();
	}

	public boolean hasAttributes() {
		return !attributes.isEmpty();
	}

	/*
	 * vmyllarn
	 * This is a stupid hack that can be avoided if interfaces are used instead of
	 * inheritance in creating type hierarchy for instances.
	 */
	public TypedInstance getAsTypedInstance() {
		return this;
	}
}
