/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ConfigurationResponse.java
 *
 * Created on 22. kes�kuuta 2004, 16:15
 */

package kumbang.core.configuration.task;

import java.io.Serializable;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;


/**
 * A response from the configuration engine that tells whether the configuration was consistent
 * (valid) or not, what steps were commited before the configuration became invalid, and what are
 * the consequences of these modifications. Consequences are either positive (an instance must exist
 * in the configuration) or negative (an instance must not be in the configuration).
 * 
 * @author vmyllarn
 */
public class ConfigurationResponse implements Serializable {

	private boolean consistent;

	private boolean complete;

	private String msg;

	private SortedSet positiveConsequences;

	private SortedSet negativeConsequences;
	
	/**
	 * Describes how the configuration needs to be updated to based on the server output.
	 * 
	 * Before applying this task to the current configuration, the task generated based on the user
	 * input needs to be applied first. Alternatively, this task can be applied to the modified
	 * configuration sent as an input to the server. -toasikai 31.7.2006
	 */
	private ClientTask task;

	/** Creates a new instance of ConfigurationResponse */
	public ConfigurationResponse() {
		consistent = true;
		complete = true;
		positiveConsequences = new TreeSet();
		negativeConsequences = new TreeSet();
	}

	public boolean isConsistent() {
		return consistent;
	}

	public boolean isComplete() {
		return complete;
	}

	public void setInconsistent(String msg) {
		this.consistent = false;
		this.msg = msg;
	}

	public void setComplete(boolean complete) {
		this.complete = complete;
	}

	public void addPositiveConsequence(ConfigurationConsequence cnsq) {
		positiveConsequences.add(cnsq);
	}

	public void addNegativeConsequence(ConfigurationConsequence cnsq) {
		negativeConsequences.add(cnsq);
	}

	public Iterator getPositiveConsequences() {
		return positiveConsequences.iterator();
	}

	public Iterator getNegativeConsequences() {
		return negativeConsequences.iterator();
	}

	public String getMessage() {
		return msg;
	}
	
	public void setTask(ClientTask task) {
		this.task = task;
	}

	public ClientTask getTask() {
		return task;
	}
	
	public String toString() {
		String ret = "Positive Consequences\n";
		for (Iterator it = positiveConsequences.iterator(); it.hasNext(); ) {
			ret += it.next().toString() + " ";
		}
		ret += "\nNegative Consequences\n";
		for (Iterator it = negativeConsequences.iterator(); it.hasNext(); ) {
			ret += it.next().toString() + " ";
		}
		ret += "\n";
		return ret;
	}

}
