/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * ModelParser.java
 *
 * Created on 25. maaliskuuta 2004, 13:13
 */

package kumbang.core.language.parser;

import kumbang.core.model.core.Model;
import kumbang.core.model.core.ComponentedModel;
import kumbang.core.model.core.FeaturedModel;

import kumbang.core.model.exception.ModelException;

import kumbang.core.language.parser.KumbangParser;
import kumbang.core.language.parser.ParseException;
import kumbang.core.language.parser.TokenMgrError;

import kumbang.core.language.syntaxtree.SyntaxModel;

import kumbang.util.ErrorHandler;

import kumbang.core.language.visitor.ConnectionVisitor;
import kumbang.core.language.visitor.TypeVisitor;
import kumbang.core.language.visitor.TaxonomyVisitor;
import kumbang.core.language.visitor.PartVisitor;
import kumbang.core.language.visitor.ConstraintVisitor;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.Reader;
import java.io.StringReader;

/**
 * NOTE: this class is not thread safe. That is, it is the responsibility
 * of caller to ensure that calls to parseModel()-methods are not interleaved.
 * 
 * @author  vmyllarn
 */
public class ModelParser implements ParserInterface {

	private static KumbangParser parser;

	/** 
	 * Creates a new empty instance of ModelParser 
	 */
	public ModelParser() {
	}

	/**
	 * Parses the kumbang model (both koalish and forfamel) from the
	 * given file.
	 * @param modelFile file containing model
	 * @return KumbangModel parsed model
	 * @throw ModelException if something goes wrong
	 */
	public Model parseModel(File modelFile)
		throws ModelException {
		try {
			return parse(new FileReader(modelFile));
		} catch (ParseException e) {
			throw new ModelException(
				"Parsing of model failed.\n" + e.getMessage());
		} catch (TokenMgrError te) {
			throw new ModelException(
				"Parsing of model failed.\n" + te.getMessage());
		} catch (FileNotFoundException fne) {
			throw new ModelException(
				"Cannot find model file.\n" + fne.getMessage());
		}
	}

	/**
	 * Parses the kumbang model (both koalish and forfamel) from the
	 * given string.
	 * @param modelString string containing model
	 * @return KumbangModel parsed model
	 * @throw ModelException if something goes wrong
	 */
	public Model parseModel(String modelString)
		throws ModelException {
		try {
			return parse(new StringReader(modelString));
		} catch (ParseException pe) {
			throw new ModelException(
				"Parsing of model failed.\n" + pe.getMessage());
		} catch (TokenMgrError te) {
			throw new ModelException(
				"Parsing of model failed.\n" + te.getMessage());
		}

	}

	/**
	 * Does the actual parsing based on the given reader. Returns the
	 * parsed model. Can only parse one model at a time (either koalish
	 * or forfamel).
	 * @param reader reader containing koalish or forfamel model
	 * @return Model parsed model
	 * @throws ParseException if something goes wrong
	 * @throws TokenMgrError if something goes wrong
	 */
	private Model parse(Reader reader) throws ParseException, TokenMgrError {

		if (parser == null) {
			parser = new KumbangParser(reader);
		} else {
			KumbangParser.ReInit(reader);
		}
		SyntaxModel syntaxModel = KumbangParser.SyntaxModel();

		/*	Symbol visitor reads in the names of the feature types,
		 *	and the names and values in attribute value types.
		 */
		ErrorHandler eh = new ErrorHandler();
		TypeVisitor tv = new TypeVisitor(syntaxModel, eh);

		/*	If there are errors, throw parse exception; warnings do not count. */
		if (eh.hasAnyErrors()) {
			StringBuffer buf = new StringBuffer("Could not extract types.\n");
			eh.report(buf);
			throw new ParseException(buf.toString());
		}

		Model model = tv.getModel();

		//int modelKind = tv.getModelKind();

		if (model instanceof FeaturedModel) {
			
			new TaxonomyVisitor(syntaxModel, model, eh);
			/*	If there are errors, throw parse exception; warnings do not count. */
			if (eh.hasAnyErrors()) {
				StringBuffer buf =
					new StringBuffer("Could not extract taxonomy of types.\n");
				eh.report(buf);
				throw new ParseException(buf.toString());
			}
		}

		
		new PartVisitor(syntaxModel, model, eh);
		/*	If there are errors, throw parse exception; warnings do not count. */
		if (eh.hasAnyErrors()) {
			StringBuffer buf =
				new StringBuffer("Could not extract part structure.\n");
			eh.report(buf);
			throw new ParseException(buf.toString());
		}

//		if (model instanceof ComponentedModel) {
//			new ConnectionVisitor(syntaxModel, model, eh);
//			//	If there are errors, throw parse exception; warnings do not count. 
//			if (eh.hasAnyErrors()) {
//				StringBuffer buf =
//					new StringBuffer("Could not extract connections.\n");
//				eh.report(buf);
//				throw new ParseException(buf.toString());
//			}
//		}	

		new ConstraintVisitor(syntaxModel, model, eh);
		eh.report();
		if (eh.hasAnyErrors()) {
			StringBuffer buf =
				new StringBuffer("Could not extract constraints.\n");
			eh.report(buf);
			throw new ParseException(buf.toString());
		}

		return model;
	}
}
