/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.core.language.visitor;

import java.util.ArrayList;
import java.util.Iterator;

import kumbang.core.configuration.reference.AttributeReference;
import kumbang.core.configuration.reference.BindingReference;
import kumbang.core.configuration.reference.ComposableReference;
import kumbang.core.configuration.reference.InterfaceReference;
import kumbang.core.configuration.reference.Reference;
import kumbang.core.configuration.reference.ReferenceItem;
import kumbang.core.language.syntaxtree.Node;
import kumbang.core.language.syntaxtree.NodeChoice;
import kumbang.core.language.syntaxtree.NodeOptional;
import kumbang.core.language.syntaxtree.NodeSequence;
import kumbang.core.language.syntaxtree.SyntaxConfigurationComposableReference;
import kumbang.core.language.syntaxtree.SyntaxConfigurationReference;
import kumbang.core.language.syntaxtree.SyntaxConfigurationReferenceItem;

public class ConfigurationReferenceVisitor extends ObjectDepthFirst {
	public static Reference getReference(SyntaxConfigurationReference ref) {
		return (Reference) ref.accept(new ConfigurationReferenceVisitor(ref), new Object());
	}

	protected ConfigurationReferenceVisitor(SyntaxConfigurationReference ref) {
		ref.accept(this, new Object());
	}
	
	/**
	 * f0 -> SyntaxConfigurationComposableReference() 
	 * f1 -> [ <ARROW> <IDENTIFIER> [ <COLON> <IDENTIFIER> ] 
	 *   [ <EQUALS> SyntaxConfigurationComposableReference() 
	 *      <ARROW> <IDENTIFIER> [ <COLON> <IDENTIFIER> ] ] 
	 * | <ATTR> <IDENTIFIER> [ <EQUALS> <IDENTIFIER> ] ]
	 */
	public Object visit(SyntaxConfigurationReference n, Object argu) {
		ComposableReference comp = (ComposableReference) n.f0.accept(this, argu);
		Reference ref = comp;
		
		if (n.f1.present()) {
			NodeChoice nc = (NodeChoice) n.f1.node;
			NodeSequence ns = (NodeSequence) nc.choice;
			if (nc.which == 0) {
				// Interface or binding
				String sourceInt = ns.elementAt(1).toString();
				
				String sourceType = "";
				NodeOptional no = (NodeOptional) ns.elementAt(2);
				if (no.present()) {
					NodeSequence nse = (NodeSequence) no.node;
					sourceType = nse.elementAt(1).toString();
				}
				
				no = (NodeOptional) ns.elementAt(3);
				if (no.present()) {
					ns = (NodeSequence) no.node;
					ComposableReference targetComp = (ComposableReference) ns.elementAt(1).accept(
							this, argu);
					String targetInt = ns.elementAt(3).toString();
					
					String targetType = "";
					no = (NodeOptional) ns.elementAt(4);
					if (no.present()) {
						ns = (NodeSequence) no.node;
						targetType = ns.elementAt(1).toString();
					}
					
					return new BindingReference(comp, sourceInt, sourceType, targetComp, targetInt,
							targetType);
				}
				
				return new InterfaceReference(comp, sourceInt, sourceType);
			} else if (nc.which == 1) {
				// Attribute value
				String attr = ns.elementAt(1).toString();
				
				String value = "";
				NodeOptional no = (NodeOptional) ns.elementAt(2);
				if (no.present()) {
					ns = (NodeSequence) no.node;
					value = ns.elementAt(1).toString();
				}
				
				return new AttributeReference(comp, attr, value);
			}
		}
		
		return comp;
	}

	/**
	* f0 -> ( <FEATURE> | <COMPONENT> )
	* f1 -> <MINUS>
	* f2 -> <ROOT>
	* f3 -> ( SyntaxConfigurationReferenceItem() )*
	*/
	public Object visit(SyntaxConfigurationComposableReference n, Object argu) {
		Object _ret = null;
		
		boolean isComponent = false;
		if (n.f0.which == 1)
			isComponent = true; 
		
		ArrayList items = new ArrayList();
		if (n.f3.present())
			for (Iterator it = n.f3.nodes.iterator(); it.hasNext(); )
				items.add(((SyntaxConfigurationReferenceItem) it.next()).accept(this, argu));
		
		return new ComposableReference(isComponent, items);
	}
	
	/**
	 * f0 -> <DOT>
	 * f1 -> <IDENTIFIER>
	 * f2 -> [ <LBRACKET> <INTEGER_LITERAL> <RBRACKET> ]
	 * f3 -> [ <COLON> <IDENTIFIER> ]
	 */
	public Object visit(SyntaxConfigurationReferenceItem n, Object argu) {
		String partName = n.f1.toString();
		
		int index = -1;
		if (n.f2.present()) {
			NodeSequence ns = (NodeSequence) n.f2.node;
			index = new Integer(ns.elementAt(1).toString()).intValue();
		}
		
		String typeName = "";
		if (n.f3.present()) {
			NodeSequence ns = (NodeSequence) n.f3.node;
			typeName = ns.elementAt(1).toString();
		}

		return new ReferenceItem(partName, index, typeName);
	}
}
