/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Created on Feb 27, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package kumbang.core.language.visitor;

import java.util.ArrayList;
import java.util.Enumeration;

import kumbang.core.language.syntaxtree.*;

import kumbang.core.model.constraint.ConnectedExpression;
import kumbang.core.model.constraint.OldInterfaceReference;
import kumbang.core.model.constraint.ReferenceItem;
import kumbang.core.model.constraint.StructureReference;

import kumbang.core.model.core.*;

import kumbang.core.model.type.*;

import kumbang.util.ErrorHandler;

/**
 * @author toasikai
 *
 */
public class ConnectionVisitor extends KumbangVisitor {
	/**
	 * Basically, it would be possible to create a superclass for the more
	 * developed visitors: there would be no need to rewrite the same
	 * constructor in every class.
	 * 
	 * @param st
	 * @param model
	 * @param handler
	 */
	public ConnectionVisitor(SyntaxModel st, Model model, ErrorHandler handler) {
		this.model = model;
		this.handler = handler;

		st.accept(this, new Object());
	}

	//
	// User-generated visitor methods below
	//

	/**
	 * f0 -> ModelKind()
	 * f1 -> <MODEL>
	 * f2 -> <IDENTIFIER>
	 * f3 -> ( <ROOT> ( <FEATURE> | <COMPONENT> ) <IDENTIFIER> )+
	 * f4 -> ( TypeDefinition() )*
	 */
	public Object visit(SyntaxModel n, Object argu) {
		n.f4.accept(this, argu);

		return null;
	}

	/**
	 * f0 -> FeatureTypeDefinition()
	 *       | AttributeTypeDefinition()
	 *       | InterfaceTypeDefinition()
	 *       | ComponentTypeDefinition()
	 */
	public Object visit(SyntaxTypeDefinition n, Object argu) {
		if (n.f0.which == 2)
			n.f0.accept(this, argu);

		return null;
	}

	/**
	 * f0 -> [ <ABSTRACT> ]
	 * f1 -> ( <FEATURE> | <COMPONENT> )
	 * f2 -> <TYPE>
	 * f3 -> SyntaxDisplayIdentifier()
	 * f4 -> [ <EXTENDS> SyntaxIdentifierList() ]
	 * f5 -> <LBRACE>
	 * f6 -> ( SyntaxSection() )*
	 * f7 -> <RBRACE>
	 */
	public Object visit(SyntaxComposableTypeDefinition n, Object argu) {
		Name name = (Name) n.f2.accept(this, argu);
		String typeKind = n.f1.choice.toString();
		
		if (n.f1.which == 0) {
			handler.complain(n.f2.beginLine, "Connects sections are only allowed for component types.");
			return null;
		}
		
		ComponentType ct = (ComponentType) model.getType(name.getName(), ComponentType.TYPE_KIND);
		
		if (ct == null) {
			handler.complain(n.f2.beginLine, "Undefined component type: " + name.getName() + ".");
			return null;
		}

		// The component type (ct) travels as an argument.
		n.f4.accept(this, ct);

		return null;
	}

	/**
	 * f0 -> SubfeatureSection()
	 *       | AttributeSection()
	 *       | ProvideSection()
	 *       | RequireSection()
	 *       | ConstraintSection()
	 *       | ContainSection()
	 *       | ConnectSection()
	 */
	public Object visit(SyntaxSection n, Object argu) {
		// Only interested in connects sections
		if (n.f0.which == 6)
			n.f0.accept(this, argu);

		return null;
	}

	/**
	 * f0 -> <CONNECTS>
	 * f1 -> ( Connection() )*
	 */
	public Object visit(SyntaxConnectSection n, Object argu) {
		ComponentType ct = (ComponentType) argu;

		if (!n.f1.present())
			return null;

		for (Enumeration e = n.f1.elements(); e.hasMoreElements();) {
			SyntaxConnection sc = (SyntaxConnection) e.nextElement();
//			Connection conn = (Connection) sc.accept(this, ct);
			ConnectedExpression ce = (ConnectedExpression) sc.accept(this, ct);
			
			return ce;

//			try {
////				ct.addConnection(conn);
//			} catch (Exception exc) {
//				handler.complain(n.f0.beginLine, exc.getMessage());
//			}
		}

		return null;
	}

	/**
	 * f0 -> <IDENTIFIER>
	 * f1 -> [ <DOT> <IDENTIFIER> ]
	 * f2 -> <EQUALS>
	 * f3 -> <IDENTIFIER>
	 * f4 -> [ <DOT> <IDENTIFIER> ]
	 * f5 -> <SEMICOLON>
	 */
	public Object visit(SyntaxConnection n, Object argu) {
		String sourcePart = null, sourceInt;
		String targetPart = null, targetInt;
		
		if (n.f1.present()) {
			sourceInt = ((NodeSequence) n.f1.node).elementAt(1).toString();
			sourcePart = n.f0.toString();
		} else
			sourceInt = n.f0.toString();
		
		if (n.f4.present()) {
			targetInt = ((NodeSequence) n.f4.node).elementAt(1).toString();
			targetPart = n.f3.toString();
		} else
			targetInt = n.f3.toString();
		
		ComponentType ct = (ComponentType) argu;
		ArrayList<ReferenceItem> part = null;
		if (sourcePart != null) {
			part = new ArrayList<ReferenceItem>();
			part.add(new ReferenceItem(sourcePart));
		}
		
		StructureReference from = ct.queryStructureReference(part, new ReferenceItem(sourceInt), null);
		
		part = null;
		if (targetPart != null) {
			part = new ArrayList<ReferenceItem>();
			part.add(new ReferenceItem(targetPart));
		}
		StructureReference to = ct.queryStructureReference(part, new ReferenceItem(targetInt), null);
		return new ConnectedExpression(from, to);
		
//		ComponentType ct = (ComponentType) argu;
//		OldInterfaceReference source, target;
//		PartDefinition sourcePart = null, targetPart = null;
//		ComponentType sourceType = null, targetType = null;
//		InterfaceDefinition sourceInterface, targetInterface;
//		String intName;
//
//		if (n.f1.present()) {
//			String name = n.f0.toString();
//
//			sourcePart = ct.getPart(name);
//			if (sourcePart == null) {
//				handler.complain(
//					n.f0.beginLine,
//					"Component type " + ct + " does not define a part named " + name + ".");
//				return null;
//			}
//
//			intName = ((NodeSequence) n.f1.node).elementAt(1).toString();
//			sourceType = (ComponentType) sourcePart.getPossibleTypes().next();
//			sourceInterface = sourceType.getInterface(intName);
//		} else {
//			intName = n.f0.toString();
//			sourceInterface = ct.getInterface(intName);
//		}
//			
//		if (sourceInterface == null)
//			handler.complain(n.f0.beginLine, "The specied interface (" + intName + ") does not exist.");
//			
//		if (n.f4.present()) {
//			String name = n.f3.toString();
//
//			targetPart = ct.getPart(name);
//			if (targetPart == null) {
//				handler.complain(
//					n.f0.beginLine,
//					"Component type " + ct + " does not define a part named " + name + ".");
//				return null;
//			}
//
//			/**
//			 * TODO Isn't this somewhat too restrictive?
//			 * 
//			 * Check that things are nice and simple: cardinality must be one,
//			 * and that there is only a single part type. These are kind of restrictive,
//			 * but make things easier; more complex cases could then be handled with
//			 * constraints, or by the user during the configuration process.
//			 */
//			if (targetPart.getLowerBound() != 1 || targetPart.getUpperBound() != 1)
//				handler.complain(
//					n.f3.beginLine,
//					"Only parts with cardinality one may be referred to in connections.");
//
//			if (targetPart.typeCount() > 1)
//				handler.complain(
//					n.f3.beginLine,
//					"Only one possible type allowed for part references in connections.");
//					
//			intName = ((NodeSequence) n.f4.node).elementAt(1).toString();
//			ComponentType pt = (ComponentType) targetPart.getPossibleTypes().next();
//			targetInterface = pt.getInterface(intName);
//		} else {
//			intName = n.f3.toString();
//			targetInterface = ct.getInterface(intName);
//		}
//		
//		if (targetInterface == null)
//			handler.complain(n.f0.beginLine, "The specied interface (" + intName + ") does not exist.");
//
//		source = new OldInterfaceReference(sourcePart, sourceType, sourceInterface);
//		target = new OldInterfaceReference(targetPart, targetType, targetInterface);
//		
//		return new Connection(source, target);
	}
}
