/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*s
 * Created on Feb 20, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package kumbang.core.language.visitor;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;

import kumbang.core.model.constraint.*;
import kumbang.core.model.core.*;
import kumbang.core.model.type.*;
import kumbang.core.language.syntaxtree.*;

import kumbang.util.ErrorHandler;

/**
 * @author toasikai
 *
 */
public class ConstraintVisitor extends KumbangVisitor {
	// Variables used to capture the status of the visitor
	private boolean implementationSection;
	
	// 	Is an implementation section being visited?
	private QuantificationStack quantifications;
	
	//	The current stack of quantifications.
	private int variableIndex;
	
	public static final int equals = 0, notEquals = 1;

	public ConstraintVisitor(SyntaxModel st, Model model, ErrorHandler handler) {
		this.model = model;
		this.handler = handler;

		implementationSection = false;
		quantifications = new QuantificationStack();

		st.accept(this, null);
	}
	
	public ConstraintVisitor(SyntaxConstraint sc, Model model, 
			ComposableType definingType, ErrorHandler handler, 
			boolean implementationSection) {
		this.model = model;
		this.handler = handler;

		this.implementationSection = implementationSection;
		quantifications = new QuantificationStack();

		sc.accept(this, definingType);
	}	
	
	/**
	 * f0 -> SyntaxModelKind()
	 * f1 -> <MODEL>
	 * f2 -> <IDENTIFIER>
	 * f3 -> ( <ROOT> ( <FEATURE> | <COMPONENT> ) <IDENTIFIER> )+
	 * f4 -> ( SyntaxTypeDefinition() )*
	 */
	public Object visit(SyntaxModel n, Object argu) {
		n.f4.accept(this, argu);
		
		return argu;
	}	
	
	/**
	 * f0 -> SyntaxFeatureTypeDefinition()
	 *       | SyntaxAttributeTypeDefinition()
	 *       | SyntaxInterfaceTypeDefinition()
	 *       | SyntaxComponentTypeDefinition()
	 */
	public Object visit(SyntaxTypeDefinition n, Object argu) {
		if (n.f0.which == 2)
			n.f0.accept(this, argu);
			
		return argu;
	}

	/**
	 * f0 -> [ <ABSTRACT> ]
	 * f1 -> ( <FEATURE> | <COMPONENT> )
	 * f2 -> <TYPE>
	 * f3 -> SyntaxDisplayIdentifier()
	 * f4 -> [ <EXTENDS> SyntaxIdentifierList() ]
	 * f5 -> <LBRACE>
	 * f6 -> ( SyntaxSection() )*
	 * f7 -> <RBRACE>
	 */
	public Object visit(SyntaxComposableTypeDefinition n, Object argu) {
		Name name = (Name) n.f3.accept(this, argu);
		String typeKind = n.f1.choice.toString();
		ComposableType ft = (ComposableType) model.getType(name.getName(), typeKind);
		
		/*
		 * TODO What happens here is awfully similar to, e.g., PartVisitor.
		 */
		if (ft == null) {
			handler.complain(n.f2.beginLine, "Undefined " + typeKind + " type: " + name.getName() + ".");
			return null;
		}

		n.f6.accept(this, ft);

		return argu;
	}

	/**
	 * f0 -> SubfeatureSection()
	 *       | AttributeSection()
	 *       | ProvideSection()
	 *       | RequireSection() 
	 *       | ConstraintSection() - n.f0.which == 4
	 *       | ContainSection()
	 *       | ConnectSection() - n.f0.which == 6
	 *       | ImplementationSection()  - n.f0.which == 7
	 */
	public Object visit(SyntaxSection n, Object argu) {
		if (n.f0.which == 4)
			n.f0.accept(this, argu);
		else if (n.f0.which == 7) {
			if (!model.getModelKind().equals(kumbang.core.model.core.KumbangModel.MODEL_KIND)) {
				handler.complain(
					"Implementation sections allowed in Kumbang models only.");
				return null;
			}

			implementationSection = true;
			n.f0.accept(this, argu);
			implementationSection = false;
		} else if (n.f0.which == 6)
			n.f0.accept(this, argu);

		return argu;
	}

	/**
	 * f0 -> <CONSTRAINTS>
	 * f1 -> ( Expression() )*
	 */
	public Object visit(SyntaxConstraintSection n, Object argu) {
		n.f1.accept(this, argu);

		return argu;
	}

	/**
	 * f0 -> SyntaxEquivalenceExpression()
	 * f1 -> <SEMICOLON>
	 */
	public Object visit(SyntaxConstraint n, Object argu) {
		variableIndex = 1;

		Expression exp = (Expression) n.f0.accept(this, argu);
		if (!checkBoolean(exp, n.f1.beginLine))
			return null;

		ComposableType ct = (ComposableType) argu;
		ct.addConstraint((BooleanExpression) exp);

		return argu;
	}

	/**
	 * f0 -> SyntaxImplicationExpression()
	 * f1 -> [ <EQUIVALENT> SyntaxImplicationExpression() ]
	 */
	public Object visit(SyntaxEquivalenceExpression n, Object argu) {
		if (!n.f1.present())
			return n.f0.accept(this, argu);
		
		Expression left = null, right = null;

		left = (Expression) n.f0.accept(this, argu);
		if (!checkBoolean(left, -1))
			return null;
		
		NodeSequence ns = (NodeSequence) n.f1.node;
		right = (Expression) ns.elementAt(1).accept(this, argu);
		if (!checkBoolean(right, -1))
			return null;
			
		return new EquivalenceExpression((BooleanExpression) left, (BooleanExpression) right);
	}

	/**
	 * f0 -> AndOrExpression()
	 * f1 -> [ <IMPLIES> AndOrExpression() ]
	 */
	public Object visit(SyntaxImplicationExpression n, Object argu) {
		if (!n.f1.present())
			return n.f0.accept(this, argu);
		
		Expression head = null, tail = null;

		NodeSequence nl = (NodeSequence) n.f1.node;
		head = (Expression) nl.elementAt(1).accept(this, argu);
		if (!checkBoolean(head, -1))
			return null;

		tail = (Expression) n.f0.accept(this, argu);
		if (!checkBoolean(tail, -1))
			return null;

		return new ImplicationExpression((BooleanExpression) tail, (BooleanExpression) head);
	}

	/**
	 * f0 -> PrimitiveExpression()
	 * f1 -> ( ( <AND> | <OR> ) PrimitiveExpression() )*
	 */
	public Object visit(SyntaxAndOrExpression n, Object argu) {
		if (!n.f1.present())
			return n.f0.accept(this, argu);
		
		ConjunctiveExpression cc = new ConjunctiveExpression();

		Expression exp = (Expression) n.f0.accept(this, argu);
		if (!checkBoolean(exp, -1))
			return null;
		
		BooleanExpression first = (BooleanExpression) exp;
		cc.addPrimitive(first);

		int op = -1; // and = 0, or = 1

		if (n.f1.present()) {
			NodeListOptional nlo = n.f1;
			for (Enumeration e = nlo.elements(); e.hasMoreElements();) {
				NodeSequence ns = (NodeSequence) e.nextElement();
				int opCur = ((NodeChoice) ns.elementAt(0)).which;

				if (op == -1) {
					op = opCur;
					cc.setOperator(op);
				} else if (op != opCur) { //	'and' and 'or' mixed
					NodeToken nt =
						(NodeToken) ((NodeChoice) ns.elementAt(0)).choice;
					String line = "Line " + nt.beginLine + ": ";
					handler.complain(
						line + "'and' and 'or' connectives may not be mixed.");
				}

				exp = (Expression) ns.elementAt(1).accept(this, argu);
				if (!checkBoolean(exp, -1))
					return null;

				cc.addPrimitive((BooleanExpression) exp);
			}
		}

		return cc;
	}

	/**
	 * f0 -> ( SyntaxQuantification() )*
	 * f1 -> [ <NOT> ]
	 * f2 -> ( SyntaxConfigurationPredicateExpression() | SyntaxComparisonExpression() )
	 */
	public Object visit(SyntaxQuantifiedExpression n, Object argu) {
		boolean not = false;

		if (n.f1.present())
			not = true;

		Quantification q;
		int nrQuantifications = 0;

		QuantificationStack incoming = new QuantificationStack(quantifications);
		QuantificationStack added = new QuantificationStack();
		if (n.f0.present()) {
			for (Enumeration e = n.f0.elements(); e.hasMoreElements();) {
				q =
					(Quantification)
						((SyntaxQuantification) e.nextElement()).accept(
						this,
						argu);
				
				//	TODO What does it mean that q is null?
				if (q == null)
					return null;
					
				added.push(q);
				quantifications.push(q);
				nrQuantifications++;
				//	This is kind of redundant, the number equals the size of added
			}
		}

		Expression exp = (Expression) n.f2.accept(this, argu);
		if (!checkBoolean(exp, -1))
			return null;

		BooleanExpression be = (BooleanExpression) exp;
		be.setNegated(not); 
		
		//	If there are quantifications, must wrap exp into a QuantifiedExpression
		if (n.f0.present()) {
			QuantifiedExpression qe = new QuantifiedExpression(be);
			qe.setContext(incoming, added);
			be = qe;
		}

		quantifications.pop(nrQuantifications);

		return be;
	}

	/**
	 * f0 -> <CARDINALITY>
	 * f1 -> <LPAREN>
	 * f2 -> SyntaxFeatureReference()
	 * f3 -> <RPAREN>
	 */
	public Object visit(SyntaxCardinalityExpression n, Object argu) {
		StructureReference sr = (StructureReference) n.f2.accept(this, argu);
		
		return sr; 
	}

	/**
	 * f0 -> [ <NOT> ]
	 * f1 -> ( <FORALL> | <EXISTS> )
	 * f2 -> <LPAREN>
	 * f3 -> SyntaxVariableBinding()
	 * f4 -> ( <COMMA> SyntaxVariableBinding() )*
	 * f5 -> <RPAREN>
	 */
	public Object visit(SyntaxQuantification n, Object argu) {
		boolean negated = n.f0.present();
		boolean universal = n.f1.which == 0;

		ArrayList al = new ArrayList();
		Variable var = (Variable) n.f3.accept(this, argu);
		
		// An error message has already been generated(?)
		if (var == null)
			return null;
		
		al.add(var);

		if (n.f4.present()) {
			for (Enumeration e = n.f4.elements(); e.hasMoreElements();) {
				NodeSequence ns = (NodeSequence) e.nextElement();
				var = (Variable) ns.elementAt(1).accept(this, argu);
				al.add(var);
			}
		}

		return new Quantification(universal, negated, al);
	}

	/**
	 * f0 -> <IDENTIFIER>
	 * f1 -> <COLON>
	 * f2 -> SyntaxFeatureReference()
	 */
	public Object visit(SyntaxVariableBinding n, Object argu) {
		StructureReference ref = (StructureReference) n.f2.accept(this, argu);
		
		if (ref == null)
			return null;
		
		return new Variable(
			n.f0.tokenImage,
			variableIndex++,
			ref);
	}

	/**
	 * f0 -> PresentExpression()
	 */
	public Object visit(SyntaxConfigurationPredicateExpression n, Object argu) {
		return n.f0.accept(this, argu);
	}

	/**
	 * f0 -> <PRESENT>
	 * f1 -> <LPAREN>
	 * f2 -> FeatureReference()
	 * f3 -> <RPAREN>
	 */
	public Object visit(SyntaxPresentExpression n, Object argu) {
		Object obj = n.f2.accept(this, argu);
		
		return new PresentExpression((StructureReference) obj);
	}

	/**
	 * f0 -> <INSTANCEOF>
	 * f1 -> <LPAREN>
	 * f2 -> FeatureReference()
	 * f3 -> <COMMA>
	 * f4 -> <IDENTIFIER>
	 * f5 -> <RPAREN>
	 */
	public Object visit(SyntaxInstanceOfExpression n, Object argu) {
		/*
		 * TODO This is ugly, do something about this.
		 */
		StructureReference sr = (StructureReference) n.f2.accept(this, argu);		
		String typeName = n.f4.toString();
		ElementaryType type = model.getType(typeName, ComponentType.TYPE_KIND);
		if (type != null) {
			return new InstanceOfExpression(sr, type);
		}
		type = model.getType(typeName, FeatureType.TYPE_KIND);
		if (type != null) {
			return new InstanceOfExpression(sr, type);
		}
		type = model.getType(typeName, InterfaceType.TYPE_KIND);
		if (type != null) {
			return new InstanceOfExpression(sr, type);
		}
		type = model.getType(typeName, AttributeValueType.TYPE_KIND);
		if (type != null) {
			handler.complain(
				n.f4.beginLine,
				"Instances cannot be of attribute types.");
			return null;
		}

		// If we got here, we did not find any suitable types
		handler.complain(n.f4.beginLine, "The specified type does not exist.");
		return null;
	}

	/**
	 * f0 -> <HASINSTANCES>
	 * f1 -> <LPAREN>
	 * f2 -> <IDENTIFIER>
	 * f3 -> <RPAREN>
	 */
	public Object visit(SyntaxHasInstancesExpression n, Object argu) {
		ComposableType containing = (ComposableType) argu;
		String typeName = n.f2.toString();
		
		if (containing.getTypeKind().equals(AttributeValueType.TYPE_KIND)) {
			handler.complain(n.f2.beginLine,
					"Attribute value types cannot be tested for having instances.");
			return null;			
		}
		
		ElementaryType type = model.getType(typeName, ComponentType.TYPE_KIND);
		if (type != null) {
			return new HasInstancesExpression(type);
		}
		type = model.getType(typeName, FeatureType.TYPE_KIND);
		if (type != null) {
			return new HasInstancesExpression(type);
		}
		type = model.getType(typeName, InterfaceType.TYPE_KIND);
		if (type != null) {
			return new HasInstancesExpression(type);
		}
		
		// if we got here, we did not find any suitable types
		handler.complain(n.f2.beginLine, "The specified type does not exist.");
		return null;
	}

	/**
	 * f0 -> <HASPART_OFTYPE>
	 * f1 -> <LPAREN>
	 * f2 -> [ SyntaxStructureReference() <COMMA> ]
	 * f3 -> <IDENTIFIER>
	 * f4 -> <RPAREN>
	 */
	public Object visit(SyntaxHasPartOfTypeExpression n, Object argu) {
		ComposableType containing = (ComposableType) argu;
		ComposableType type;
		String typeName = n.f3.toString();
		
		//	No reference
		if (!n.f2.present()) {
			type = (ComposableType) model.getType(typeName, containing.getTypeKind());
			if (type == null)
				handler.complain(n.f1.beginLine, "The " + containing.getTypeKind()
						+ " type specified (" + typeName + ") does not exist.");
			
			return new HasPartOfTypeExpression(type, true);
		}

		NodeSequence ns = (NodeSequence) n.f2.node;
		StructureReference sr = (StructureReference) ns.elementAt(0).accept(this, argu);

		if (sr == null) {
			handler.complain(n.f1.beginLine, "The reference given is not valid.");
			return null;
		}
		
		if (sr.isImplementation())
			type = (ComposableType) model.getType(typeName, ComponentType.TYPE_KIND);
		else
			type = (ComposableType) model.getType(typeName, containing.getTypeKind());
		
		return new HasPartOfTypeExpression(type, sr, true);
	}

	/**
	 * f0 -> <VALUE>
	 * f1 -> <LPAREN>
	 * f2 -> [ FeatureReference() <COMMA> ]
	 * f3 -> <IDENTIFIER>
	 * f4 -> <RPAREN>
	 */
	public Object visit(SyntaxAttributeValue n, Object argu) {
		String attrName = n.f3.toString();
		AttributedType at = (AttributedType) argu;

		//	This is perhaps somewhat misplaced: might be more natural
		//	to go through this in a class in the model package.
		if (!n.f2.present()) {
			AttributeDefinition ad = at.getAttributeDefinition(attrName);

			if (ad == null) {
				handler.complain(n.f0.beginLine, "Type " + at + " does not define an attribute named " + attrName + ".");
				return null;
			}
		}

		//	There is a reference to a subfeature or part
		StructureReference sr = null;
		if (n.f2.present()) {
			NodeSequence ns = (NodeSequence) n.f2.node;
			sr = (StructureReference) ns.elementAt(0).accept(this, argu);
		}

		AttributeReference ar;
		try {
			if (sr != null)
				ar = new AttributeReference(sr, attrName);
			else
				ar = new AttributeReference(at, attrName);
		} catch (Exception e) {
			handler.complain(n.f0.beginLine, e.toString());
			return null;
		}

		return ar;
	}
	
	/**
	 * Feature reference may be also a part reference, used with component types.
	 * Currently, the implementation assumes that the things are features.
	 * This is overly similar to something else, could it be an identifier list.
	 * 
	 * f0 -> [ SyntaxRootSymbol() | SyntaxStructureReferenceItem() ]
	 * f1 -> ( <DOT> SyntaxStructureReferenceItem() )*
	 * f2 -> [ <ARROW> SyntaxStructureReferenceItem() ]
	 *  
	 * Old things:
	 * 
	 * f0 -> [ ( <ROOTSYMBOL> | <IDENTIFIER> ) ]
	 * f1 -> ( <DOT> <IDENTIFIER> )*
	 * f2 -> [ <COLON> <IDENTIFIER> ]
	 * 
	 * @return a StructureReference object
     */
	public Object visit(SyntaxStructureReference n, Object argu) {
		ComposableType ct = (ComposableType) argu;

		//	Only an interface name: ':callee'
		if (!n.f0.present()) {
			//	Must be an InterfaceReference
			
			if (n.f1.present()) {
				//	An incomplete reference: .a.b[:i]
				handler.complain("An invalid reference.");
				return null;
			}
			
			if (!n.f2.present()) {
				//	An empty reference
				handler.complain("A reference may not be empty.");
				return null;
			}
			
			NodeSequence ns = (NodeSequence) n.f2.node;
			ReferenceItem intRef = (ReferenceItem) ns.elementAt(1).accept(this,
					InterfaceType.TYPE_KIND);
			return ct.queryStructureReference(null, /*null,*/ intRef, null);
		}
		
		NodeChoice nc0 = (NodeChoice) n.f0.node;

		//	An array containing the elements of the path expression, excluding the possible $.
		ArrayList<ReferenceItem> path = new ArrayList<ReferenceItem>();
		
		boolean implementation = nc0.which == 0;
		if (implementation) {
			NodeToken node = (NodeToken) ((SyntaxRootSymbol) nc0.choice).f0.choice;
			if (!node.toString().equals("component"))
				/*
				 * TODO This is really not good, the entire reference thing should
				 * be restructured: the syntax and parsing it.
				 */ 
				handler.complain(node.beginLine, "Only references to components allowed.");
		} else {
			//	If not an implementation constraint, collect the first part of the path from n.f0 
			path.add((ReferenceItem) nc0.choice.accept(this, ct.getTypeKind()));
		}

		//	Implementation constraint that is not in an implementation section.
		if (!implementationSection && implementation) {
			handler.complain("Implementation constraints allowed in implementation sections only.");
			return null;
		}

		//	The rest of the elements in the path expression
		if (n.f1.present())
			for (Enumeration e = n.f1.elements(); e.hasMoreElements();) {
				NodeSequence ns = (NodeSequence) e.nextElement();
				path.add((ReferenceItem) ns.elementAt(1).accept(this,
						implementation ? ComponentType.TYPE_KIND : ct.getTypeKind()));
			}

		Variable var = null;
		Iterator<ReferenceItem> it = path.iterator();

		if (it.hasNext()) {
			ReferenceItem first = it.next();
			String firstName = first.getPartName();
			if (firstName.length() == 1 
					&& Character.isLetter(firstName.charAt(0)) 
					&& Character.isUpperCase(firstName.charAt(0))) { //	A variable
				var = quantifications.getVariable(firstName);
				if (var == null) {
					handler.complain("The variable specified (" + first + ") does not exist.");
					return null;
				}
				
				first.setVariable(var);
			} else
				it = path.iterator();
		}

//		ArrayList<ReferenceItem> pathEnd = new ArrayList<ReferenceItem>();
		while (it.hasNext()) {
			ReferenceItem ri = it.next();
			String name = ri.getPartName();
			if (name.length() == 1
					&& Character.isLetter(name.charAt(0))
					&& Character.isUpperCase(name.charAt(0))) { //	A variable
				handler.complain("Variables may only be used as the first part of paths.");
				return null;
			}

//			pathEnd.add(ri);
		}

		if (var != null)
			implementation |= var.getReference().isImplementation();
			
		ComponentType croot = null;
		if (implementation)
			croot = ((KumbangModel) model).getComponentRoot();
		
		//	InterfaceReference
		ReferenceItem intRef = null;
		if (n.f2.present()) {
			NodeSequence ns = (NodeSequence) n.f2.node;
			intRef = (ReferenceItem) ns.elementAt(1).accept(this, InterfaceType.TYPE_KIND);
		}

		return ct.queryStructureReference(/*var, */ path /*End*/, intRef, croot);
	}
	
	/**
	 * f0 -> (<ASTERISK> | <QUESTION_MARK> | <IDENTIFIER>)
	 * f1 -> [ <COLON> <IDENTIFIER> ]
	 */
	public Object visit(SyntaxStructureReferenceItem n, Object argu) {
		//String partName;
		String typeKind = (String) argu;
		String typeName = null;
		InstantiableType type = null;
		
		if (n.f1.present()) {
			typeName = ((NodeSequence) n.f1.node).elementAt(1).toString(); 
			type = (InstantiableType) model.getType(typeName, typeKind);
		}
		
		// TODO find line numbers for error messages
		
		if (n.f0.which == 0 || n.f0.which == 1) {
			if (type == null) {
				handler.complain("A wildcard part must specify an existing type.");
			}
			if (n.f0.which == 0) {
				return new ReferenceItem(ReferenceItem.WILDCARD_TRANSITIVE_PART, type);
			} else {
				return new ReferenceItem(ReferenceItem.WILDCARD_PART, type);
			}
		} else {
			if (n.f1.present() && type == null) {
					handler.warn("The type specified (" + typeName
							+ ") does not exist. The reference will not match anything.");
					return new ReferenceItem(n.f0.choice.toString(), true);
			}
			if (type != null) {
				return new ReferenceItem(n.f0.choice.toString(), type);
			} else {
				return new ReferenceItem(n.f0.choice.toString());
			}
		}
	}
	
	/**
	 * f0 -> ( <FEATURE> | <COMPONENT> )
	 * f1 -> <MINUS>
	 * f2 -> <ROOT>
	 */
	public Object visit(SyntaxRootSymbol n, Object argu) {
		Object _ret=null;
		n.f0.accept(this, argu);
		n.f1.accept(this, argu);
		n.f2.accept(this, argu);
		return _ret;
	}
	
	/**
	 * f0 -> SyntaxAdditiveExpression()
	 * f1 -> [ SyntaxComparisonOperator() SyntaxAdditiveExpression() ]
	 */
	public Object visit(SyntaxComparisonExpression n, Object argu) {
		if (!n.f1.present())
			return n.f0.accept(this, argu);
			
		IntegerExpression left, right;
		NodeSequence nl = (NodeSequence) n.f1.node;
		int operator = ((Integer) nl.elementAt(0).accept(this, argu)).intValue();
		
		left = (IntegerExpression) n.f0.accept(this, argu);
		right = (IntegerExpression) nl.elementAt(1).accept(this, argu);
		
		if (left == null || right == null)
			return null;
		
		return new ComparisonExpression(left, right, operator);
	}

	/**
	 * f0 -> SyntaxMultiplicativeExpression()
	 * f1 -> ( ( <PLUS> | <MINUS> ) SyntaxMultiplicativeExpression() )*
	 */
	public Object visit(SyntaxAdditiveExpression n, Object argu) {
		if (!n.f1.present())
			return n.f0.accept(this, argu);	

		IntegerExpression left = (IntegerExpression) n.f0.accept(this, argu);
		
		IntegerExpression right;
		BinaryExpression ret = new BinaryExpression(left);
		BinaryExpression bottom = ret;
		int operator;
		for (Enumeration e = n.f1.elements(); e.hasMoreElements(); ) {
			NodeSequence nl = (NodeSequence) e.nextElement();
			
			//	Check the operator
			if ((int) ((NodeChoice) nl.elementAt(0)).which == 0)
				operator = BinaryExpression.opPlus;
			else
				operator = BinaryExpression.opMinus;
			
			//	Get the next term
			right = (IntegerExpression) nl.elementAt(1).accept(this, argu);
			bottom.setOperator(operator);
			
			if (e.hasMoreElements()) {
				BinaryExpression temp = new BinaryExpression(right); 
				bottom.setRight(temp);
				bottom = temp;
			}
			else
				bottom.setRight(right);
				// break;
		}
		
		return ret;
	}

	/**
	 * f0 -> SyntaxPrimitiveArithmetic()
	 * f1 -> ( ( <TIMES> | <DIV> | <MOD> ) SyntaxPrimitiveArithmetic() )*
	 */
	public Object visit(SyntaxMultiplicativeExpression n, Object argu) {
		Object left = n.f0.accept(this, argu);  
		
		if (!n.f1.present())
			return left;
			
		IntegerExpression right;
		BinaryExpression ret = new BinaryExpression((IntegerExpression) left);
		BinaryExpression bottom = ret;
		int operator;
		for (Enumeration e = n.f1.elements(); e.hasMoreElements(); ) {
			NodeSequence nl = (NodeSequence) e.nextElement();
			
			//	Check the operator
			switch ((int) ((NodeChoice) nl.elementAt(0)).which) {
			case 0:
				operator = BinaryExpression.opMult;
				break;
			case 1:
				operator = BinaryExpression.opDiv;
				break;
			case 2:
				operator = BinaryExpression.opMod;
				break;
			default :
				handler.complain("This really should not happen.");
				return null;
			}
			
			//	Get the next term
			right = (IntegerExpression) nl.elementAt(1).accept(this, argu);
			bottom.setOperator(operator);
			
			if (e.hasMoreElements()) {
				BinaryExpression temp = new BinaryExpression(right); 
				bottom.setRight(temp);
				bottom = temp;
			}
			else
				bottom.setRight(right);
				// break;
		}
		
		return ret;
	}

	/**
	 * f0 -> 
	 * 		( 
	 *			[ <MINUS> ] ( 
	 *							SyntaxCardinalityExpression() 
	 *							| SyntaxAttributeValue() 
	 *							| <LPAREN> SyntaxEquivalenceExpression() <RPAREN> 
	 *						) 
	 *			| SyntaxValue() 
	 *		)
	 */
	public Object visit(SyntaxPrimitiveArithmetic n, Object argu) {
		if (n.f0.which == 1)
			return n.f0.accept(this, argu);
		
		NodeSequence nl = (NodeSequence) n.f0.choice;
		
		//	TODO Why is this minus not read.
		boolean minus = false;
		if (((NodeOptional) nl.elementAt(0)).present())
			minus = true;
		
		Object ret;
		NodeChoice nc = (NodeChoice) nl.elementAt(1);
		if (nc.which == 2)
			ret = ((NodeSequence) nc.choice).elementAt(1).accept(this, argu);
		else
			ret = nc.choice.accept(this, argu);
		
		if (ret instanceof AttributeReference)
			return new AttributeExpression((AttributeReference) ret);
		if (ret instanceof StructureReference)
			return new CardinalityExpression((StructureReference) ret);
			
		return ret;
	}

	/**
	 * f0 -> [ <MINUS> ] <INTEGER_LITERAL>
	 *       | <IDENTIFIER>
	 */
	public Object visit(SyntaxValue n, Object argu) {
		if (n.f0.which == 0) {
			NodeSequence nl = (NodeSequence) n.f0.choice;
			NodeOptional no = (NodeOptional) nl.elementAt(0);
			
			Integer val = new Integer(nl.elementAt(1).toString());
			
			ConstantExpression ret = new ConstantExpression(val.intValue());
			ret.setNegated(no.present());
			return ret;
		}
		
		return new ConstantExpression(n.f0.choice.toString());
	}
	
	/**
	 * f0 -> <CONNECTED>
	 * f1 -> <LPAREN>
	 * f2 -> SyntaxStructureReference()
	 * f3 -> <RPAREN>
	 */
	public Object visit(SyntaxConnectedExpression n, Object argu) {
		StructureReference sr = (StructureReference) n.f2.accept(this, argu);
		
		if (!sr.isInterfaceReference()) {
			handler.complain(n.f1.beginLine, "An interface reference expected.");
			return null;
		}
		
		return new ConnectedExpression(sr);
	}

	/**
	 * f0 -> <CONNECTED_TO>
	 * f1 -> <LPAREN>
	 * f2 -> SyntaxStructureReference()
	 * f3 -> <COMMA>
	 * f4 -> SyntaxStructureReference()
	 * f5 -> <RPAREN>
	 */
	public Object visit(SyntaxConnectedToExpression n, Object argu) {
		StructureReference from = (StructureReference) n.f2.accept(this, argu);
		if (!from.isInterfaceReference()) {
			handler.complain(n.f1.beginLine, "An interface reference expected.");
			return null;
		}
		
		StructureReference to = (StructureReference) n.f4.accept(this, argu);
		if (!to.isInterfaceReference()) {
			handler.complain(n.f3.beginLine, "An interface reference expected.");
			return null;
		}
		
		return new ConnectedExpression(from, to);

	}

	/**
	 * f0 -> <EQUALS>
	 *       | <NE>
	 *       | <LT>
	 *       | <LTE>
	 *       | <GT>
	 *       | <GTE>
	 */
	public Object visit(SyntaxComparisonOperator n, Object argu) {
		return new Integer(n.f0.which);
	}
	
	/**
	 * Checks whether the Expression passed as an argument is a boolean expression or not.
	 * First checks whether exp is null.
	 * 
	 * @param exp
	 * @param lineNumber
	 * @return true if exp is a boolean expression; false otherwise
	 */
	private boolean checkBoolean(Expression exp, int lineNumber) {
		if (exp == null)
			return false;
		
		if (!exp.isBoolean()) {
			handler.complain(lineNumber, "Boolean expression expected.");
			return false;
		}
		
		return true;
	}

	/**
	 * f0 -> <CONNECTS> f1 -> ( Connection() )*
	 */
	public Object visit(SyntaxConnectSection n, Object argu) {
		ComponentType ct = (ComponentType) argu;

		if (!n.f1.present())
			return null;

		for (Enumeration e = n.f1.elements(); e.hasMoreElements();) {
			SyntaxConnection sc = (SyntaxConnection) e.nextElement();
			ConnectedExpression ce = (ConnectedExpression) sc.accept(this, ct);

			ct.addConstraint(ce);
		}

		return null;
	}

	/**
	 * f0 -> <IDENTIFIER> f1 -> [ <DOT> <IDENTIFIER> ] f2 -> <EQUALS> f3 -> <IDENTIFIER> f4 -> [
	 * <DOT> <IDENTIFIER> ] f5 -> <SEMICOLON>
	 */
	public Object visit(SyntaxConnection n, Object argu) {
		String sourcePart = null, sourceInt;
		String targetPart = null, targetInt;

		if (n.f1.present()) {
			sourceInt = ((NodeSequence) n.f1.node).elementAt(1).toString();
			sourcePart = n.f0.toString();
		} else
			sourceInt = n.f0.toString();

		if (n.f4.present()) {
			targetInt = ((NodeSequence) n.f4.node).elementAt(1).toString();
			targetPart = n.f3.toString();
		} else
			targetInt = n.f3.toString();

		ComponentType ct = (ComponentType) argu;
		ArrayList<ReferenceItem> part = new ArrayList<ReferenceItem>();
		if (sourcePart != null)
			part.add(new ReferenceItem(sourcePart));
		else
			part = null;

		StructureReference from = ct.queryStructureReference(part, new ReferenceItem(sourceInt),
				null);

		part = new ArrayList<ReferenceItem>();;
		if (targetPart != null)
			part.add(new ReferenceItem(targetPart));
		else
			part = null;

		StructureReference to = ct
				.queryStructureReference(part, new ReferenceItem(targetInt), null);
		return new ConnectedExpression(from, to);

	}
}
