/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Created on Feb 13, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package kumbang.core.language.visitor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import kumbang.core.language.syntaxtree.*;
import kumbang.core.model.core.*;
import kumbang.core.model.exception.OnesOwnPartException;
import kumbang.core.model.type.*;

import kumbang.util.ErrorHandler;

/**
 * This visitor reads in the subfeature/part and attribute definitions.
 * Should be run after TaxonomyVisitor (for features).
 * 
 * This is not really a generic class, in the sense that it includes 
 * methods that are specific to Koalish models, at least.
 * 
 * @author toasikai
 */
public class PartVisitor extends KumbangVisitor {
	/**
	 * Used to save the information about the direction of the interfaces being read in.
	 * The argument (argu) is already used.
	 */
	private boolean provides;

	public PartVisitor(SyntaxModel st, Model model, ErrorHandler handler) {
		this.model = model;
		this.handler = handler;

		st.accept(this, new Object());
	}

	//
	// User-generated visitor methods below
	//

	/**
	 * f0 -> ModelKind()
	 * f1 -> <MODEL>
	 * f2 -> <IDENTIFIER>
	 * f3 -> ( <ROOT> ( <FEATURE> | <COMPONENT> ) <IDENTIFIER> )+
	 * f4 -> ( TypeDefinition() )*
	 */
	public Object visit(SyntaxModel n, Object argu) {
		n.f4.accept(this, argu);

		return argu;
	}

	/**
	 * f0 -> FeatureTypeDefinition()
	 *       | AttributeTypeDefinition()
	 *       | InterfaceTypeDefinition()
	 *       | ComponentTypeDefinition()
	 */
	public Object visit(SyntaxTypeDefinition n, Object argu) {
		n.f0.accept(this, argu);
		return argu;
	}

	/**
	 * f0 -> [ <ABSTRACT> ]
	 * f1 -> ( <FEATURE> | <COMPONENT> )
	 * f2 -> <TYPE>
	 * f3 -> SyntaxDisplayIdentifier()
	 * f4 -> [ <EXTENDS> SyntaxIdentifierList() ]
	 * f5 -> <LBRACE>
	 * f6 -> ( SyntaxSection() )*
	 * f7 -> <RBRACE>
	 */
	public Object visit(SyntaxComposableTypeDefinition n, Object argu) {
		Name name = (Name) n.f3.accept(this, argu);
		String typeKind = n.f1.choice.toString();
		ComposableType ct = (ComposableType) model.getType(name.getName(), typeKind);
		
		if (ct == null) {
			handler.complain(n.f2.beginLine, "Undefined " + typeKind + " type: " + name.getName() + ".");
			return null;
		}

		//	The composable type travels as an argument.
		n.f6.accept(this, ct);

		return argu;
	}

	/**
	 * f0 -> <COMPONENT>
	 * f1 -> <TYPE>
	 * f2 -> <IDENTIFIER>
	 * f3 -> <LBRACE>
	 * f4 -> ( Section() )*
	 * f5 -> <RBRACE>
	 */
//	public Object visit(SyntaxComponentTypeDefinition n, Object argu) {
//		Name name = (Name) n.f2.accept(this, argu);
//		ComponentType ct = (ComponentType) model.getType(name.getName(), ComponentType.TYPE_KIND);
//		
//		if (ct == null) {
//			handler.complain(n.f1.beginLine, "Undefined component type: " + name.getName() + ".");
//			return null;
//		}
//
//		//	The component type travels as an argument.
//		n.f4.accept(this, ct);
//
//		return argu;
//	}

	/**
	 * f0 -> SyntaxSubfeatureSection()			0
	 *       | SyntaxAttributeSection()			1
	 *       | SyntaxProvideSection()			2
	 *       | SyntaxRequireSection()			3
	 *       | SyntaxConstraintSection()		4
	 *       | SyntaxContainSection()			5
	 *       | SyntaxConnectSection()			6
	 *       | SyntaxImplementationSection()	7
	 */
	public Object visit(SyntaxSection n, Object argu) {
		if (n.f0.which != 4 && n.f0.which != 6 && n.f0.which != 7)
			n.f0.accept(this, argu);

		return argu;
	}

	/**
	 * f0 -> SyntaxIdentifierList()
	 * f1 -> SyntaxDisplayIdentifier()
	 * f2 -> [ SyntaxCardinality() ]
	 * f3 -> [ <LBRACE> ( <NONE> | <SAME> | <DIFFERENT> ) <RBRACE> ]
	 * f4 -> <SEMICOLON>
	 */
	public Object visit(SyntaxPartSpecification n, Object argu) {
		List typeNames = (List) n.f0.accept(this, argu);
		List types = new ArrayList();
		
		ComposableType parent = (ComposableType) argu;
		for (int i = 0; i < typeNames.size(); i++) {
			ComposableType type = (ComposableType) model.getType((String) typeNames.get(i), parent.getTypeKind());

			if (type == null)
				handler.complain(n.f4.beginLine, "The type " + typeNames.get(i) + " does not exist.");
			else
				types.add(type);
		}

		//	Get cardinality
		int upper, lower;
		if (n.f2.present()) {
			List bounds = (List) n.f2.accept(this, argu);
			lower = ((Integer) bounds.get(0)).intValue();
			if (bounds.size() == 2)
				upper = ((Integer) bounds.get(1)).intValue();
			else
				upper = lower;
		} else {
			lower = 1;
			upper = 1;
		}

		if (lower > upper) {
			String line = "Line " + n.f4.beginLine;
			handler.warn(line + "The lower bound in the cardinality definition" + "exceeds the upper bound.");
		}
		
		//	Read the similarity definition
		//	f3 -> [ <LBRACE> ( <NONE> | <SAME> | <DIFFERENT> ) <RBRACE> ]
		int similarity = 0;
		if (n.f3.present()) {
			NodeSequence ns = (NodeSequence) n.f3.node;
			NodeChoice nc = (NodeChoice) ns.elementAt(1);
			similarity = nc.which;
		}

		PartDefinition sd = new PartDefinition((Name) n.f1.accept(this, argu), types, lower, upper, similarity);

		try {
			((ComposableType) argu).addPart(sd, true);
		} catch (OnesOwnPartException e) {
			String message = new String();

			if (e.isDirect()) {
				message = "Feature type " + argu + " defines a subfeature (" + sd.getName();
				message += ") that has the feature type itself as a possible type.";
			} else if (e.isDirectSuper()) {
				message = "Feature type " + argu + " defines a subfeature (" + sd.getName();
				message += ") that has a supertype of itself (" + e.getPartType() + ") as a possible type.";
			} else {
				message = "Feature type " + e.getWholeType() + " may have an instance of ";
				if (e.getPartType() == e.getWholeType())
					message += "itself ";
				else
					message += "its supertype " + e.getPartType();
				message += " as its transitive subfeature: " + e.getWholeType() + " -> ";

				for (Iterator f = e.types(); f.hasNext();)
					message += f.next() + " -> ";

				message += e.getPartType() + ".";
			}

			handler.complain(n.f4.beginLine, message);
		} catch (Exception e) {
			handler.complain(n.f4.beginLine, e.getMessage());
		}

		return argu;
	}

	/**
	 * f0 -> <LBRACKET>
	 * f1 -> <INTEGER_LITERAL>
	 * f2 -> [ <MINUS> <INTEGER_LITERAL> ]
	 * f3 -> <RBRACKET>
	 */
	public Object visit(SyntaxCardinality n, Object argu) {
		List values = new ArrayList();

		values.add(new Integer(n.f1.toString()));

		if (n.f2.present()) {
			NodeSequence ns = (NodeSequence) n.f2.node;
			values.add(new Integer(ns.elementAt(1).toString()));
		}

		return values;
	}

	/**
	 * f0 -> <IDENTIFIER>
	 * f1 -> SyntaxDisplayIdentifier()
	 * f2 -> [ <LBRACE> <HIDDEN> <RBRACE> ]
	 * f3 -> <SEMICOLON>
	 */
	public Object visit(SyntaxAttribute n, Object argu) {
		
		if (model instanceof AttributedModel) {
			AttributeValueType avt = ((AttributedModel) model).getAttributeValueType(n.f0.toString());
			if (avt == null) {
				handler.complain(
					n.f0.beginLine,
					"The attribute value type specified (" + n.f0.toString() + "). does not exist.");
				return null;
			}	
			Name name = (Name) n.f1.accept(this, argu);
			AttributeDefinition ad = new AttributeDefinition(name, avt, n.f2.present());
			try {
				((ComposableType) argu).addAttribute(ad);
			} catch (Exception e) {
				handler.complain(n.f0.beginLine, e.getMessage());
				return null;
			}

			return argu;		
			
		} else {
			handler.complain(n.f0.beginLine,
					"Non-attributed models cannot define attributes in types.");
			return null;
		}

	}

	/**
	 * f0 -> <PROVIDES>
	 * f1 -> ( Interface() )*
	 */
	public Object visit(SyntaxProvideSection n, Object argu) {
		provides = true;

		n.f1.accept(this, argu);

		return null;
	}

	/**
	 * f0 -> <REQUIRES>
	 * f1 -> ( SyntaxInterface() )*
	 */
	public Object visit(SyntaxRequireSection n, Object argu) {
		provides = false;

		n.f1.accept(this, argu);

		return null;
	}

	/**
	 * f0 -> SyntaxIdentifierList()
	 * f1 -> SyntaxInterfaceSpecification()
	 * f2 -> ( <COMMA> SyntaxInterfaceSpecification() )*
	 * f3 -> <SEMICOLON>
	 */
	public Object visit(SyntaxInterface n, Object argu) {
		List typeNames = (List) n.f0.accept(this, argu);
		List types = new ArrayList();

		for (int i = 0; i < typeNames.size(); i++) {
			InterfaceType it = ((ComponentedModel) model).getInterfaceType((String) typeNames.get(i));
			if (it == null)
				handler.complain(n.f3.beginLine, "The type " + typeNames.get(i) + " does not exist.");
			else
				types.add(it);
		}

		ComponentType compType = (ComponentType) argu;
		try {
			compType.addInterface((InterfaceDefinition) n.f1.accept(this, types));

			NodeListOptional nlo = n.f2;
			for (Enumeration e = nlo.elements(); e.hasMoreElements();) {
				NodeSequence ns = (NodeSequence) e.nextElement();
				compType.addInterface((InterfaceDefinition) ns.elementAt(1).accept(this, types));
			}
		} catch (Exception e) {
			handler.complain(n.f3.beginLine, e.getMessage());
		}

		return argu;
	}

	/**
	 * f0 -> SyntaxDisplayIdentifier()
	 * f1 -> [ <LBRACE> ( <OPTIONAL> | <GROUNDED> ) [ <COMMA> ( <OPTIONAL> | <GROUNDED> ) ] <RBRACE> ]
	 */
	public Object visit(SyntaxInterfaceSpecification n, Object argu) {
		//	Argu contains the interface type
		Collection types = (Collection) argu;
		boolean optional = false, grounded = false;

		Name name = (Name) n.f0.accept(this, argu);
		if (n.f1.present()) {
			NodeSequence ns = (NodeSequence) n.f1.node;
			NodeChoice nc = (NodeChoice) ns.elementAt(1);
			if (nc.which == 0)
				optional = true;
			else
				grounded = true;

			NodeOptional no = (NodeOptional) ns.elementAt(2);
			if (no.present()) {
				ns = (NodeSequence) no.node;
				nc = (NodeChoice) ns.elementAt(1);
				if (nc.which == 0)
					optional = true;
				else
					grounded = true;
			}
		}

		return new InterfaceDefinition(name, types, grounded, optional, provides);
	}
}
