/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Created on Feb 18, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package kumbang.core.language.visitor;

import java.util.Iterator;
import java.util.List;

import kumbang.core.language.syntaxtree.NodeSequence;
import kumbang.core.language.syntaxtree.SyntaxComposableTypeDefinition;
import kumbang.core.language.syntaxtree.SyntaxDisplayIdentifier;
import kumbang.core.language.syntaxtree.SyntaxIdentifierList;
import kumbang.core.language.syntaxtree.SyntaxModel;
import kumbang.core.language.syntaxtree.SyntaxTypeDefinition;
import kumbang.core.model.core.FeaturedModel;
import kumbang.core.model.core.Model;
import kumbang.core.model.exception.OnesOwnSupertypeException;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.FeatureType;
import kumbang.core.model.type.Name;
import kumbang.util.ErrorHandler;

/**
 * This visitor reads in the taxonomy of feature types. 
 * Should be run after TypeVisitor and before SubfeatureVisitor.
 * 
 * @author toasikai
 */
public class TaxonomyVisitor extends KumbangVisitor {
	
	public TaxonomyVisitor(SyntaxModel st, Model model, ErrorHandler handler) {
		this.model = model;
		this.handler = handler;
		
		st.accept(this, new Object());
	}

	//
	// User-generated visitor methods below
	//

	/**
	 * f0 -> ModelKind()
	 * f1 -> <MODEL>
	 * f2 -> <IDENTIFIER>
	 * f3 -> ( <ROOT> ( <FEATURE> | <COMPONENT> ) <IDENTIFIER> )+
	 * f4 -> ( TypeDefinition() )*
	 */
	public Object visit(SyntaxModel n, Object argu) {
		n.f4.accept(this, argu);

		return argu;
	}

	/**
	 * f0 -> FeatureTypeDefinition()
	 *       | AttributeTypeDefinition()
	 */
	public Object visit(SyntaxTypeDefinition n, Object argu) {
		n.f0.accept(this, argu);
		return argu;
	}

	/**
	 * TODO The complains in this method would more approriately be generated elsewhere;
	 * in classes of the symbol package.
	 * 
	 * f0 -> [ <ABSTRACT> ]
	 * f1 -> ( <FEATURE> | <COMPONENT> )
	 * f2 -> <TYPE>
	 * f3 -> SyntaxDisplayIdentifier()
	 * f4 -> [ <EXTENDS> SyntaxIdentifierList() ]
	 * f5 -> <LBRACE>
	 * f6 -> ( SyntaxSection() )*
	 * f7 -> <RBRACE>
	 */
	public Object visit(SyntaxComposableTypeDefinition n, Object argu) {
		Name name = (Name) n.f3.accept(this, argu);
		String typeKind = n.f1.choice.toString();
		ComposableType ct = (ComposableType) model.getType(name.getName(), typeKind);
//		FeaturedModel fmodel = (FeaturedModel) model;
//		FeatureType ft = fmodel.getFeatureType(name.getName());
		
		//	ft really should not be null
		if (ct == null)
			return null;

		//	An extended type
		if (n.f4.present()) {
			NodeSequence ns = (NodeSequence) n.f4.node;
			List superTypeNames = (List) ((SyntaxIdentifierList) ns.elementAt(1)).accept(this, argu);

			for (int i = 0; i < superTypeNames.size(); i++) {
				String superName = (String) superTypeNames.get(i); 
				ComposableType st = (ComposableType) model.getType(superName, typeKind);
				
				if (st == null)
					handler.complain(n.f2.beginLine, "The type specified as supertype does not exist.");
				else {
					try {
						ct.addSupertype(st, true);
					} catch (OnesOwnSupertypeException e) {
						handler.complain(n.f2.beginLine, e.getMessage());
					}
				}
			}
		}
	
		return argu;
	}

	/**
	 * f0 -> <IDENTIFIER>
	 * f1 -> [ <STRING_LITERAL> ]
	 */
	public Object visit(SyntaxDisplayIdentifier n, Object argu) {
		if (n.f1.present())
			return new Name(n.f0.toString(), n.f1.node.toString());
		
		return new Name(n.f0.toString());
	}
}
