/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Created on Feb 20, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package kumbang.core.model.constraint;

import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.InstantiableType;
import kumbang.core.model.type.InterfaceDefinition;
import kumbang.core.model.type.InterfaceType;
import kumbang.core.model.type.PartDefinition;

import java.io.Serializable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Iterator;
import java.util.Vector;

/**
 * This is really not much more than a list; the type of origin is added.
 * This should really be applicable also with components; there, this should
 * be termed a part reference.
 * 
 * @author toasikai
 *
 */
public class StructureReference implements Serializable {
	/**
	 * The type from which the reference originates.
	 * 
	 * If the reference is an implementation reference, the type 
	 * is the root component type.
	 */
	private ComposableType origin; 
	
	/**
	 * The references path (as a string). May be null.
	 * 
	 * Either path or intName must not be null.
	 * 
	 * Path refers to the part reference thing, i.e., it does not contain an interface name.
	 */
	private List<ReferenceItem> path = new ArrayList<ReferenceItem>();
	
	/**
	 * The interface possibly referenced. May be null.
	 * 
	 * Either part or intName must not be null.
	 */
	private ReferenceItem intRef;
	
	/**	
	 * True if the reference is made from a feature to components 
	 */
	private boolean implementation;
	
	/**
	 * The number of composable instances that is referred to by this reference.
	 * 
	 * TODO What does cardinality mean if the reference is an interface reference.
	 * TODO Attribute references could be handled using the same class. Maybe not, as they are not instances.
	 */
	private int cardinality = -1;
	
	/**
	 * The index of this relation; in the translated model, this relation is
	 * represented by a domain predicate named as r o index (r0, r1, etc.).
	 * 
	 * TODO Not really part of the pure model, but more or less WCRL specific.
	 */
	private int index;
	
	/**
	 * The variable in the beginning of this StructureReference. May be null.
	 */
//	private Variable variable;
	
	/**
	 * The set of types that the referred instances may have.
	 */
	private Vector possibleTypes = new Vector();
	
	/**
	 * True if the cardinality of this reference is referred to
	 */
	private boolean cardinalityRef = false;
	
	private static int nextIndex = 1;
	
	protected StructureReference() {
		index = nextIndex++;
	}
	
	public StructureReference(ComposableType type, /*Variable var,*/ ArrayList<ReferenceItem> path,
			ReferenceItem intRef, ComponentType croot) {
		this();
		
		origin = type;
		implementation = croot != null;
		this.path = path;
		this.intRef = intRef;
//		variable = var;
		
		//	TODO (cardinality == 0) is a problem.
		cardinality = calculateCardinality(isImplementation() ? croot : origin, getFullPath(), 0);
	}
		
	/* TODO 
	 * Does this work with wildcard references?
	 * Two references that are of different form may refer to same instance.
	 */
	public boolean equals(Object obj) {
		if (!(obj instanceof StructureReference))
			return false;

		StructureReference comp = (StructureReference) obj;
			
		if (this.origin != comp.origin)
			return false;
		
		if ((path == null) != (comp.path == null))
			return false;
			
		if (path != null) {
			if (path.size() != comp.path.size())
				return false;
			
			for (int i = 0; i < path.size(); i++)
				if (!path.get(i).equals(comp.path.get(i)))
					return false;
		}
		
		if ((intRef != null && comp.intRef == null) || (intRef == null && comp.intRef != null))
			return false;
		
		return intRef == null && comp.intRef == null || (intRef.equals(comp.intRef));
	}
	
//	public boolean isDirect() {
//		if (path == null)
//			return true;
//		
//		return path.size() == 1;
//	}
	
	public int getIndex() {
		return index;
	} 

	public boolean isImplementation() {
		return implementation;
	}
	
	public ComposableType getOrigin() {
		return origin;
	}
	
	public String getName(int level) {
//		if (path == null)
//			return null;
	
		return path.get(level).getPartName();
	}
	
	public InstantiableType getType(int level) {
//		if (path == null)
//			return null;
		
		return path.get(level).getType();
	}
	
	public int getDepth() {
		return path.size();
	}
	
	public boolean isLocal() {
		return path == null || path.isEmpty();
	}
	
	public String toString() {
		String ret = "";
		
//		if (variable != null) {
//			ret += variable.getName();
//			
//			if (!path.isEmpty())
//				ret += ".";
//		}		
		
		if (path != null) {
			Iterator it = path.iterator();
			while (it.hasNext()) { 
				ret += it.next();
				
				if (it.hasNext())
					ret += ".";
			}
		}
		
		if (isInterfaceReference())
			ret += "->" + getInterface();
		
		return ret;
	}
	
	private ArrayList<ReferenceItem> getFullPath() {
		ArrayList<ReferenceItem> fullPath = new ArrayList<ReferenceItem>();

		if (getVariable() != null) {
			fullPath.addAll(getVariable().getReference().getFullPath());
			
			Iterator<ReferenceItem> it = path.iterator();
			it.next();	// Skip the ReferenceItem (first, containing the variable)
			while (it.hasNext())
				fullPath.add(it.next());
		} else if (path != null)
			fullPath.addAll(path);
		
		return fullPath;
	}
	
	public Collection getPath() {
		return path; 
	}

	public Variable getVariable() {
		if (path == null)
			return null;
		
		return path.get(0).getVariable();
	}
	
	public int getCardinality() {
		return cardinality;
	}
	
	
	public Iterator possibleTypes() {
		return possibleTypes.iterator();
	}

	public boolean appears(Variable var) {
//		if (variable == null)
		if (getVariable() == null)
			return false;
		
		return getVariable() == var || getVariable().containsVariable(var);
//		return variable == var || variable.containsVariable(var);
	}
	
	public void setCardinality() {
		cardinalityRef = true;
	}
	
	public boolean isCardinality() {
		return cardinalityRef;
	}
	
	public boolean isInterfaceReference() {
		return intRef != null;
	}
	
	public String getInterfaceName() {
		if (intRef == null)
			return null;
		
		return intRef.getPartName();
	}
	
	public InterfaceType getInterfaceType() {
		if (intRef == null)
			return null;
		
		return (InterfaceType) intRef.getType();
	}
	
	public ReferenceItem getInterface() {
		if (intRef == null)
			return null;
		
		return intRef;
	}

	public String getModelString(boolean implementation) {
		String ret = "";
		
//		if (variable != null) {
//			ret += variable.getName();
		if (getVariable() != null) {
			ret += getVariable().getName();

			if (!path.isEmpty())
				ret += ".";
		}		
		
		Iterator it = path.iterator();
		
		if (implementation && isImplementation())
			ret += "component-root" +  (it.hasNext() ? "." : "");
		
		
		while (it.hasNext()) { 
			ret += it.next();
			
			if (it.hasNext())
				ret += ".";
		}
		
		if (isInterfaceReference()) {
			if (!ret.equals(""))
				ret +="->";
			
			ret += getInterfaceName();
			
		}
		
		return ret;
	}
	
	/**
	 * Returns true if this ReferenceItem cannot refer to any instances.
	 * 
	 * TODO Does this thing work correctly, check.
	 * 
	 * @return see above
	 */
	public boolean isVoid() {
		if (path == null)
			return (intRef != null) ? intRef.isVoid() : true;
		
		for (Iterator<ReferenceItem> it = path.iterator(); it.hasNext(); )
			if (it.next().isVoid())
				return true;
		
		return false;
	}
	
	
	private int calculateCardinality(ComposableType type, List<ReferenceItem> path, int level) {
		if (level > 0 && level >= path.size())
			return 0;
		
		if (level == 0 && path.isEmpty()) {
			InterfaceDefinition id = ((ComponentType) type).getInterface(getInterfaceName());
			if (id != null) {
				for (Iterator itypes = id.getPossibleTypes(); itypes.hasNext(); ) {
					InterfaceType it = (InterfaceType) itypes.next();
					if (!possibleTypes.contains(it))
						possibleTypes.add(it);
				}
				
				return possibleTypes.size();
			}			
		}
		
		List<PartDefinition> parts = new ArrayList<PartDefinition>();
		
		String partName = path.get(level).getPartName();
		if (partName.equals(ReferenceItem.WILDCARD_PART) || partName.equals(ReferenceItem.WILDCARD_TRANSITIVE_PART)) {
			for (Iterator it = type.getParts(); it.hasNext(); ) {
				parts.add((PartDefinition) it.next());
			}
		} else {
			PartDefinition part = type.getPart(partName);
			if (part == null) {
				return 0;
			} else {
				parts.add(part);
			}
		}

		//	The path is terminated here
		if (level == path.size() - 1) {
			int card = 0;
			for (PartDefinition part : parts) {
				card += calculateCardinalityFromPart(part);
			}
			return card;
		}

		// The path is not terminated here
		int totalCard = 0;
		for (PartDefinition part : parts) {
			int maxCardinality = -1;
			//	Iterate through the possible types in a part/subfeature definition (part)
			for (Iterator types = part.getPossibleTypes(); types.hasNext();) {
				ComposableType possibleType = (ComposableType) types.next();
				if (!possibleType.isAbstract()) {
					int card = calculateCardinality(possibleType, path, level + 1);
					if (card > maxCardinality)
						maxCardinality = card;
				}
	
				//	TODO There should be a way to get a flat list of the possible types and their subtypes.
				//	Iterate through the subtypes of the current type (possibleType)
				Iterator subTypes = possibleType.getSubtypes();
				while (subTypes.hasNext()) {
					ComposableType subType = (ComposableType) subTypes.next();
					if (!subType.isAbstract()) {
						int card = calculateCardinality(possibleType, path, level + 1);
						if (card > maxCardinality)
							maxCardinality = card;
					}
				}
			}
			totalCard += maxCardinality * part.getUpperBound();
		}
		return totalCard;
	}
	
	private int calculateCardinalityFromPart(PartDefinition part) {
		int partCardinality = part.getUpperBound();
		for (Iterator types = part.getPossibleTypes(); types.hasNext(); ) {
			ComposableType cand = (ComposableType) types.next();

			//	TODO Abstractness in here? Interfaces could be abstract, or not?
			//	Probably not. Should have been implemented before writing.
			if (!isInterfaceReference()) {
				if (!possibleTypes.contains(cand))
					possibleTypes.add(cand);
			} else {
				ComponentType cc = (ComponentType) cand;
				InterfaceDefinition id = cc.getInterface(getInterfaceName());
				if (id != null) {
					for (Iterator itypes = id.getPossibleTypes(); itypes.hasNext(); ) {
						InterfaceType it = (InterfaceType) itypes.next();
						if (!possibleTypes.contains(it))
							possibleTypes.add(it);
					}
				}
			}
				
			Iterator subTypes = cand.getSubtypes();
			while (subTypes.hasNext()) {
				ComposableType subCand = (ComposableType) subTypes.next();
				if (!possibleTypes.contains(subCand))
					possibleTypes.add(subCand);
				else if (cand instanceof ComponentType) {
					ComponentType cc = (ComponentType) cand;
					InterfaceDefinition id = cc.getInterface(getInterfaceName());
					if (id != null) {
						for (Iterator itypes = id.getPossibleTypes(); itypes.hasNext(); ) {
							InterfaceType it = (InterfaceType) itypes.next();
							if (!possibleTypes.contains(it))
								possibleTypes.add(it);
						}
					}
				}
			}
		}
		
		return partCardinality;		
	}
}
