/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/**
 * 
 */
package kumbang.core.model.core;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import kumbang.core.model.exception.RootAlreadyDefinedException;
import kumbang.core.model.exception.TypeAlreadyDefinedException;
import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.InterfaceType;

/**
 * @author vmyllarn
 *
 */
public class ComponentedModelFactory {

	public static ComponentedModel getComponentedModel() {
		return new ComponentedModelImpl();
	}
	
	private static class ComponentedModelImpl implements ComponentedModel {
		
		private ComponentType root; 
		
		private Map<String, ComponentType> componentTypes;
		
		private Map<String, InterfaceType> interfaceTypes;
		
		public ComponentedModelImpl() {
			componentTypes = new HashMap<String, ComponentType>();
			interfaceTypes = new HashMap<String, InterfaceType>();
		}
		
		public ComponentType getComponentType(String name) {
			return componentTypes.get(name);
		}

		public void addComponentType(ComponentType type) throws TypeAlreadyDefinedException {
			String typeName = type.getName();
			if (componentTypes.containsKey(typeName)) {
				throw new TypeAlreadyDefinedException(type);
			}		
			componentTypes.put(type.getName(), type);
		}

		public Iterator<ComponentType> getComponentTypes() {
			return componentTypes.values().iterator();
		}		

		public int getComponentCount() {
			return componentTypes.size();
		}
	
		public void removeComponentType(String name) {
			componentTypes.remove(name);
		}
		
		public ComponentType getComponentRoot() {
			return root;
		}		
		
		public void setComponentRoot(ComponentType type) throws RootAlreadyDefinedException {
			if (root != null) {
				throw new RootAlreadyDefinedException(type);
			}
			this.root = type;
		}

		public void addInterfaceType(InterfaceType type) throws TypeAlreadyDefinedException {
			String typeName = type.getName();
			if (interfaceTypes.containsKey(typeName)) {
				throw new TypeAlreadyDefinedException(type);
			}		
			interfaceTypes.put(type.getName(), type);
			
			// Count the subtyping relations.
			for (InterfaceType itType : interfaceTypes.values()) {
				if (type.isSubtype(itType)) {
					itType.addSubtype(type);
				}
			}			
		}

		public int getInterfaceCount() {
			return interfaceTypes.size();
		}

		public InterfaceType getInterfaceType(String name) {
			return interfaceTypes.get(name);
		}

		public Iterator<InterfaceType> getInterfaceTypes() {
			return interfaceTypes.values().iterator();
		}

		public void removeInterfaceType(String name) {
			interfaceTypes.remove(name);
		}

		public void removeComponentRoot() {
			root = null;
			
		}

		public void updateTypeName(ComponentType type, String newName) {
			componentTypes.remove(type.getName());
			type.setName(newName);
			componentTypes.put(newName, type);
			
		}

		public void updateTypeName(InterfaceType type, String newName) {
			interfaceTypes.remove(type.getName());
			type.setName(newName);
			interfaceTypes.put(newName, type);
			
		}		
	}

}
