/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * KumbangModel.java
 *
 * Created on 21. huhtikuuta 2004, 21:39
 */

package kumbang.core.model.core;

import kumbang.core.model.exception.RootAlreadyDefinedException;
import kumbang.core.model.exception.TypeAlreadyDefinedException;

import kumbang.core.model.type.AttributeValueType;
import kumbang.core.model.type.ComposableType;
import kumbang.core.model.type.ComponentType;
import kumbang.core.model.type.ElementaryType;
import kumbang.core.model.type.FeatureType;
import kumbang.core.model.type.InterfaceType;

import java.util.Iterator;

/**
 * @author toasikai
 * @author vmyllarn
 */
public class KumbangModel extends Model implements FeaturedModel, ComponentedModel, AttributedModel {
    
	public static final String MODEL_KIND = "Kumbang";
	
	private FeaturedModel featuredModelImpl;
	
	private ComponentedModel componentedModelImpl;
	
	private AttributedModel attributedModelImpl;	

	public KumbangModel(String name) {
		super(name);
		featuredModelImpl = FeaturedModelFactory.getFeaturedModel();
		componentedModelImpl = ComponentedModelFactory.getComponentedModel();
		attributedModelImpl = AttributedModelFactory.getAttributedModel();
	}
	
	/**
	 * Returns an upper-case string representation of this
	 * model and its kind, i.e., "Kumbang".
	 * @return String "Kumbang"
	 */	
	public String getModelKind() {
		return KumbangModel.MODEL_KIND; 
		// note that we cannot refer to this class implicitly,
		// since other model classes may inherit KumbangModel (i.e. KumbangSecModel)
	}

	/**
	 * Returns a type with the given name and given type kind.
	 * Returns null if no such type exists, or the model does not
	 * support this particular type kind.
	 * @param name
	 * @param typeKind
	 * @return ElementaryType
	 */
	public ElementaryType getType(String name, String typeKind) {
		if (typeKind.equals(FeatureType.TYPE_KIND)) {
		 return featuredModelImpl.getFeatureType(name);
		} else if (typeKind.equals(ComponentType.TYPE_KIND)) {
			return componentedModelImpl.getComponentType(name);
		} else if (typeKind.equals(InterfaceType.TYPE_KIND)) {
			return componentedModelImpl.getInterfaceType(name);
		} else if (typeKind.equals(AttributeValueType.TYPE_KIND)) {
			return attributedModelImpl.getAttributeValueType(name);
		} else {
			return null;
		}
	}	
	
	public void addFeatureType(FeatureType type) throws TypeAlreadyDefinedException {
		featuredModelImpl.addFeatureType(type);
	}

	public int getFeatureCount() {
		return featuredModelImpl.getFeatureCount();
	}

	public FeatureType getFeatureRoot() {
		return featuredModelImpl.getFeatureRoot();
	}

	public FeatureType getFeatureType(String name) {
		return featuredModelImpl.getFeatureType(name);
	}

	public Iterator<FeatureType> getFeatureTypes() {
		return featuredModelImpl.getFeatureTypes();
	}

	public void removeFeatureType(String name) {
		featuredModelImpl.removeFeatureType(name);
	}

	public void setFeatureRoot(FeatureType type) throws RootAlreadyDefinedException {
		featuredModelImpl.setFeatureRoot(type);
	}	
	
	public void addComponentType(ComponentType type) throws TypeAlreadyDefinedException {
		componentedModelImpl.addComponentType(type);
	}

	public void addInterfaceType(InterfaceType type) throws TypeAlreadyDefinedException {
		componentedModelImpl.addInterfaceType(type);
	}

	public int getComponentCount() {
		return componentedModelImpl.getComponentCount();
	}

	public ComponentType getComponentRoot() {
		return componentedModelImpl.getComponentRoot();
	}

	public ComponentType getComponentType(String name) {
		return componentedModelImpl.getComponentType(name);
	}

	public Iterator<ComponentType> getComponentTypes() {
		return componentedModelImpl.getComponentTypes();
	}

	public int getInterfaceCount() {
		return componentedModelImpl.getInterfaceCount();
	}

	public InterfaceType getInterfaceType(String name) {
		return componentedModelImpl.getInterfaceType(name);
	}

	public Iterator<InterfaceType> getInterfaceTypes() {
		return componentedModelImpl.getInterfaceTypes();
	}

	public void removeComponentType(String name) {
		componentedModelImpl.removeComponentType(name);
		
	}

	public void removeInterfaceType(String name) {
		componentedModelImpl.removeInterfaceType(name);
	}

	public void setComponentRoot(ComponentType type) throws RootAlreadyDefinedException {
		componentedModelImpl.setComponentRoot(type);
	}

	public void addAttributeValueType(AttributeValueType type) throws TypeAlreadyDefinedException {
		attributedModelImpl.addAttributeValueType(type);
	}

	public int getAttributeValueCount() {
		return attributedModelImpl.getAttributeValueCount();
	}

	public AttributeValueType getAttributeValueType(String name) {
		return attributedModelImpl.getAttributeValueType(name);
	}

	public Iterator<AttributeValueType> getAttributeValueTypes() {
		return attributedModelImpl.getAttributeValueTypes();
	}

	public void removeAttributeValueType(String name) {
		attributedModelImpl.removeAttributeValueType(name);
	}

	public void setFeaturedModel(FeaturedModel model) {
		featuredModelImpl = model;
		
	}

	public void setComponentedModel(ComponentedModel model) {
		componentedModelImpl = model;	
	}
	
	public void setAttributedModel(AttributedModel model) {
		attributedModelImpl = model;	
	}

	public void removeComponentRoot() {
		componentedModelImpl.removeComponentRoot();
		
	}

	public void removeFeatureRoot() {
		featuredModelImpl.removeFeatureRoot();
		
	}

	public void updateTypeName(FeatureType type, String newName) {
		featuredModelImpl.updateTypeName(type, newName);
		
	}

	public void updateTypeName(ComponentType type, String newName) {
		componentedModelImpl.updateTypeName(type, newName);
		
	}
	
	public void updateTypeName(InterfaceType type, String newName) {
		componentedModelImpl.updateTypeName(type, newName);
		
	}

	
	public void updateTypeName(AttributeValueType type, String newName) {
		attributedModelImpl.updateTypeName(type, newName);
		
	}

	
//	// These methods should be reasonably implemented.
//	// I think that a Kumbang model must be a Model, although this is not quite clear.
//	// The three methods below are not reasonable: which root/kind?
//	
//	public ComposableType getComposableType(String name) {
//		ComposableType ret = koalishModel.getComposableType(name); 
//		if (ret != null)
//			return ret;
//			
//		ret = forfamelModel.getComposableType(name);
//		
//		//	May be null: neither component nor feature type with that name.
//		return ret;
//	}
//
//	public void setRoot(ComposableType type) throws RootAlreadyDefinedException {
//		if (type instanceof ComponentType)
//			setComponentRoot((ComponentType) type);
//		else if (type instanceof FeatureType)
//			setFeatureRoot((FeatureType) type);
//		else
//			throw new Error("The type provided as parameter is neither a feature nor a component type.");
//	}

}
