/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


package kumbang.core.model.type;

import java.io.Serializable;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import kumbang.core.model.exception.PropertyAlreadyDefinedException;

public class AttributedTypeFactory {

	private AttributedTypeFactory() { }
	
	public static AttributedType getAttributedTypeImplementation(ElementaryType definingType) {
		return new AttributedTypeImpl(definingType);
	}
	
	static class AttributedTypeImpl implements AttributedType, Serializable {

		private ElementaryType definingType;
		private List<AttributeDefinition> attributes;
		private List<AttributeDefinition> directAttributes;
		
		public AttributedTypeImpl(ElementaryType definingType) {
			this.definingType = definingType;
			this.attributes = new ArrayList<AttributeDefinition>();
			this.directAttributes = new ArrayList<AttributeDefinition>();
		}
		
		public void addAttribute(AttributeDefinition attr) throws PropertyAlreadyDefinedException {
			String attrName = attr.getName();
			for (AttributeDefinition adt : attributes) {
				if (adt.getName().equals(attrName)) {
					throw new PropertyAlreadyDefinedException(definingType, "attribute", attrName);
				}
			}			
			this.attributes.add(attr);
			this.directAttributes.add(attr);
			
			if (definingType instanceof ComposableType) {
				// TODO change check to InheritedType when interface has been created
				// Add the attributes to direct subtypes.
				Iterator<ComposableType> it = ((ComposableType) definingType).getDirectSubtypes();
				while (it.hasNext()) {
					it.next().addInheritedAttribute(attr);
				}
			}
		}
		
		public void addInheritedAttribute(AttributeDefinition attr) throws PropertyAlreadyDefinedException {
			if (attributes.contains(attr)) {
				return; // do not add same attribute twice (multiple inheritance)
			}		
			this.attributes.add(attr);
			
			if (definingType instanceof ComposableType) {
				// TODO change check to InheritedType when interface has been created
				// Add the attributes to direct subtypes.
				Iterator<ComposableType> it = ((ComposableType) definingType).getDirectSubtypes();
				while (it.hasNext()) {
					it.next().addInheritedAttribute(attr);
				}
			}		
		}
		
		public AttributeDefinition getAttributeDefinition(String name) {
			for (AttributeDefinition adt : attributes) {
				if (adt.getName().equals(name)) {
					return adt;
				}
			}
			return null;
		}

		public Iterator<AttributeDefinition> getAttributes() {
			return attributes.iterator();
		}
		
		public Iterator<AttributeDefinition> getDirectAttributes() {
			return directAttributes.iterator();
		}

		public void removeAttribute(AttributeDefinition attr) {
			directAttributes.remove(attr);
			attributes.remove(attr);
			if (definingType instanceof ComposableType) {
				// TODO change check to InheritedType when interface has been created
				// Add the attributes to direct subtypes.
				Iterator<ComposableType> it = ((ComposableType) definingType).getDirectSubtypes();
				while (it.hasNext()) {
					it.next().removeAttribute(attr);
				}
			}			
		}
	}
	

}
