/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Created on Feb 25, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package kumbang.core.model.type;

import kumbang.core.model.constraint.AttributeReference;
import kumbang.core.model.constraint.BooleanExpression;
import kumbang.core.model.constraint.ReferenceItem;
import kumbang.core.model.constraint.StructureReference;
import kumbang.core.model.exception.OnesOwnPartException;
import kumbang.core.model.exception.OnesOwnSupertypeException;
import kumbang.core.model.exception.PropertyAlreadyDefinedException;

import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;

/**
 * ComposableType is a generalisation of FeatureType and ComponentType. They
 * may be defined attributes, a compositional structure using part definitions
 * (subfeature definitions in feature types and part definitions in component types),
 * and constraints. 
 * 
 * @author toasikai
 */
public abstract class ComposableType extends InstantiableType implements AttributedType {
	private List directParts = new ArrayList();
	//private List directAttributes = new ArrayList();
	private List constraints = new ArrayList();

	/**
	 * Lists of parts and attributes that contain also the parts and attributes of supertypes.
	 */
	private List parts = new ArrayList();
	//private List attributes = new ArrayList();

	/** Attribute implementation */
	private AttributedType attributeImpl;
	
	/**
	 * The references (subfeature or part) used in the constraints in this type.
	 * 
	 * Even feature types may have interface references, implementation ones.
	 */
	private ArrayList structureReferences = new ArrayList();
	private ArrayList attributeReferences = new ArrayList();
	private ArrayList interfaceReferences = new ArrayList();


	//	Compositional hierarchy
	private List partTypes = new ArrayList(); //	Also transitive part types?
	private List wholeTypes = new ArrayList(); //	only direct whole types

	public ComposableType(String name) {
		this(new Name(name));
	}

	public ComposableType(Name name) {
		super(name);
		attributeImpl = AttributedTypeFactory.getAttributedTypeImplementation(this);
	}

	private boolean isAbstract;

	//	Taxonomy
	protected List<ComposableType> supertypes = new ArrayList<ComposableType>();
	private List<ComposableType> directSubtypes = new ArrayList<ComposableType>();
	protected List<ComposableType> subtypes = new ArrayList<ComposableType>();
	private List<ComposableType> directSupertypes = new ArrayList<ComposableType>();

	public void addSupertype(ComposableType supertype, boolean direct) throws OnesOwnSupertypeException {
		if (supertype == this)
			throw new OnesOwnSupertypeException(this);

		supertypes.add(supertype);
		if (direct)
			directSupertypes.add(supertype);

		//	Add the supertype as an indirect supertype of all the subtypes.
		for (int i = 0; i < directSubtypes.size(); i++) {
			try {
				((ComposableType) directSubtypes.get(i)).addSupertype(supertype, false);
			} catch (OnesOwnSupertypeException e) {
				e.addType(this);
				throw e;
			}
		}

		if (direct)
			supertype.addSubtype(this, true);
	}

	private void addSubtype(ComposableType subtype, boolean direct) {
		subtypes.add(subtype);
		if (direct)
			directSubtypes.add(subtype);

		for (int i = 0; i < supertypes.size(); i++) {
			((ComposableType) supertypes.get(i)).addSubtype(subtype, false);
		}
	}

	public boolean isSubtype(ComposableType type) {
		return equals(type) || subtypes.contains(type);
	}

	public void setAbstract() {
		isAbstract = true;
	}
	
	public void setAbstract(boolean isAbstract) {
		this.isAbstract = isAbstract;
		
	}

	public Iterator<ComposableType> getSubtypes() {
		return subtypes.iterator();
	}
	
	public Iterator<ComposableType> getDirectSubtypes() {
		return directSubtypes.iterator();
	}

	public Iterator<ComposableType> getSupertypes() {
		return directSupertypes.iterator();
	}

	public boolean isAbstract() {
		return isAbstract;
	}



	public PartDefinition getPart(String name) {
		PartDefinition sd;
		for (int i = 0; i < directParts.size(); i++) {
			sd = (PartDefinition) directParts.get(i);
			if (sd.getName().equals(name))
				return sd;
		}

		return null;
	}

	public void addPart(PartDefinition sub, boolean direct) throws OnesOwnPartException, PropertyAlreadyDefinedException {
		//	Check if a part (or subfeature) with the given name already exists.
		String partName = sub.getName();
		for (Iterator e = directParts.iterator(); e.hasNext();) {
			PartDefinition pd = (PartDefinition) e.next();
			if (pd.getName().equals(partName)) {
				throw new PropertyAlreadyDefinedException(this, "part", partName);
			}
		}

		//	Add the part.
		if (direct)
			directParts.add(sub);

		parts.add(sub);

		//	Mark all the possible types of the part as part types of this, and vice versa.
		for (Iterator e = sub.getPossibleTypes(); e.hasNext();) {
			ComposableType ct = (ComposableType) e.next();

			addPartType(ct);
			ct.addWholeType(this);
		}

		//	Add the direct parts of the supertype to the subtype.
		for (ComposableType subtype : subtypes)
			subtype.addPart(sub, false);
	}

	private void addPartType(ComposableType ft) throws OnesOwnPartException {
		//	Check for loops in the composition hierarchy.
		if (ft == this || supertypes.contains(ft)) {
			throw new OnesOwnPartException(this, ft);
		}

		if (!partTypes.contains(ft))
			partTypes.add(ft);

		for (Iterator f = wholeTypes.iterator(); f.hasNext();)
			try {
				((ComposableType) f.next()).addPartType(ft);
			} catch (OnesOwnPartException exc) {
				exc.addType(this);
				throw exc;
			}
	}

	private void addWholeType(ComposableType ft) {
		if (!wholeTypes.contains(ft))
			wholeTypes.add(ft);
	}

	
	
//	public void addAttribute(AttributeDefinition attr, boolean direct) throws PropertyAlreadyDefinedException {
//		String attrName = attr.getName();
//		for (Iterator e = directAttributes.iterator(); e.hasNext();) {
//			AttributeDefinition adt = (AttributeDefinition) e.next();
//			if (adt.getName().equals(attrName)) {
//				throw new PropertyAlreadyDefinedException(this, "attribute", attrName);
//			}
//		}
//
//		if (direct)
//			directAttributes.add(attr);
//
//		// Avoid the attributes getting added multiple times.
//		// NOTE: == will be used for comparison, alright in this case.
//		if (attributes.contains(attr))
//			return;
//		
//		attributes.add(attr);
//
//		//	Add the direct parts of the supertype to the subtype.
//		for (ComposableType subtype : directSubtypes)
//			subtype.addAttribute(attr, false);
//	}	
//	
//	public AttributeDefinition getAttributeDefinition(String name) {
//		AttributeDefinition ad;
//		for (int i = 0; i < attributes.size(); i++) {
//			ad = (AttributeDefinition) attributes.get(i);
//			if (ad.getName().equals(name))
//				return ad;
//		}
//
//		return null;
//	}
//
//	public Iterator getDirectAttributes() {
//		return directAttributes.iterator();
//	}
//
//	public Iterator getAttributes() {
//		return attributes.iterator();
//	}
//
//	
//	public Object[] getAttributesArray() {
//		return attributes.toArray();
//	}
//	
//	public void removeAttribute(AttributeDefinition definition) {
//		removeAttribute(definition, true);
//		
//	}
//	
//	public void removeAttribute(AttributeDefinition attr, boolean direct){
//		if (direct)
//			directAttributes.remove(attr);
//
//		attributes.remove(attr);
//
//		//	Remove the direct parts of the supertype to the subtype.
//		for (Iterator e = subtypes.iterator(); e.hasNext();) {
//			((ComposableType) e.next()).removeAttribute(attr, false);
//		}
//	}
	
	public void addAttribute(AttributeDefinition attr) throws PropertyAlreadyDefinedException {
		attributeImpl.addAttribute(attr);
	}

	public void addInheritedAttribute(AttributeDefinition attr) throws PropertyAlreadyDefinedException {
		attributeImpl.addInheritedAttribute(attr);
	}

	public AttributeDefinition getAttributeDefinition(String name) {
		return attributeImpl.getAttributeDefinition(name);
	}

	public Iterator<AttributeDefinition> getAttributes() {
		return attributeImpl.getAttributes();
	}

	public Iterator<AttributeDefinition> getDirectAttributes() {
		return attributeImpl.getDirectAttributes();
	}

	public void removeAttribute(AttributeDefinition attr) {
		attributeImpl.removeAttribute(attr);
	}
	
	public Iterator getDirectParts() {
		return directParts.iterator();
	}

	public Iterator getParts() {
		return parts.iterator();
	}

	public Iterator getConstraints() {
		return constraints.iterator();
	}

	public void addConstraint(BooleanExpression ic) {
		constraints.add(ic);
	}

	public Iterator getStructureReferences() {
		return structureReferences.iterator();
	}

	public Iterator getAttributeReferences() {
		return attributeReferences.iterator();
	}
	
	public Iterator getInterfaceReferences() {
		return interfaceReferences.iterator();
	}

	/**
	 * Returns a lower-case string that represents name of the entities
	 * of the type. For example "component", "feature", "interface" etc.
	 * 
	 * @return String a lower-case entity name
	 */
	abstract public String getTypeKind();

	/**
	 * When presented a list of part/subfeature names, checks whether there already
	 * exists a reference with the same path; if yes, returns the existing reference,
	 * otherwise creates a new StructureReference object.
	 * 
	 * @param path
	 * @return
	 */
	public StructureReference queryStructureReference(ArrayList<ReferenceItem> path,
			ReferenceItem intRef, ComponentType croot) {
		StructureReference ret = new StructureReference(this, path, intRef, croot);

		for (Iterator it = structureReferences.iterator(); it.hasNext();) {
			StructureReference sr = (StructureReference) it.next();
			if (sr.equals(ret))
				return sr;
		}

		// not found
		structureReferences.add(ret);

		return ret;
	}

	public AttributeReference queryAttributeReference(String attrName, StructureReference ref) {
		AttributeReference ret = new AttributeReference(ref, attrName);
		
		for (Iterator it = attributeReferences.iterator(); it.hasNext();) {
			AttributeReference ar = (AttributeReference) it.next();
			if (ret.equals(ar))
				return ar;
		}
		
		attributeReferences.add(ret);

		return ret;
	}

	public Object[] getPartsArray() {
		return parts.toArray();
	}


	
	public void removePart(PartDefinition sub) {
	
		for (Iterator e = sub.getPossibleTypes(); e.hasNext();) {
			ComposableType ft = (ComposableType) e.next();

			removePartType(ft);
			ft.removeWholeType(this);
		}

		//	Add the direct parts of the supertype to the subtype.
		for (Iterator e = subtypes.iterator(); e.hasNext();) {
			((ComposableType) e.next()).removePart(sub);
		}
		
		directParts.remove(sub);
		parts.remove(sub);
	}

	private void removeWholeType(ComposableType type) {
		wholeTypes.remove(type);
	}

	private void removePartType(ComposableType ft) {
		
		for (Iterator f = wholeTypes.iterator(); f.hasNext();)
			((ComposableType) f.next()).removePartType(ft);
		
		partTypes.remove(ft);
	}
	
	public Object[] getConstraintsArray() {
		return constraints.toArray();
	}


	
	/**
	 * Returns the number of instances that are instantiated when an this type is
	 * instantiated.
	 * 
	 * @return see above
	 */
	public int getSize() {
		int size = 1;
		
		// Calculate the sizes of parts.
		for (Iterator it = getParts(); it.hasNext(); ) {
			PartDefinition def = (PartDefinition) it.next();
			size += def.getSize();
		}
		
		return size;
	}

	/*
	 * TODO All functions of this kind should go to Translator, that's the class that knows
	 * about translating to lparse.
	 */
	public int getPartDefinitionOffset(PartDefinition def) {
		int ret = 0;
		
		for (Iterator it = getParts(); it.hasNext(); ) {
			PartDefinition comp = (PartDefinition) it.next();
			if (comp == def)
				return ret;
			
			ret += comp.getSize();
		}
		
		// Error, definition was not found.
		return -1;
	}
	
	public void removeConstraint(BooleanExpression expression) {
		constraints.remove(expression);
	}
	
	public List getAllWholeTypes() {
		List list = new ArrayList(wholeTypes);
		Iterator it = wholeTypes.iterator();
		while (it.hasNext()) {
			List l = ((ComposableType)it.next()).getAllWholeTypes();
			for (int i = 0; i < l.size(); i++)
				if (!list.contains(l.get(i)))
					list.add(l.get(i));		
		}
		return list;
	}
	
	public boolean equals(Object o) {
		if (!(o instanceof ComposableType))
			return false;
		
		/*
		 * This may be a bit weak: types are only identified by their names.
		 */
		ComposableType comp = (ComposableType) o;
		if (getName().equals(comp.getName()))
			return true;
		
		return false;
	}
	
	public void setSupertypes(ArrayList newSupertypes) throws OnesOwnSupertypeException  {
		for (int i = 0; i < supertypes.size(); i++) {
			removeSupertype((ComposableType)supertypes.get(i));
		}
		
		for (int i = 0; i < newSupertypes.size(); i++) {
			addSupertype((ComposableType) newSupertypes.get(i), true);
		}
	}

	public void removeSupertype(ComposableType type) {
		supertypes.remove(type);
		directSupertypes.remove(type);

		//	Remove the supertype from all direct subtypes.
		for (int i = 0; i < directSubtypes.size(); i++) {
			((ComposableType) directSubtypes.get(i)).removeSupertype(type);
		}
		type.removeSubtype(this);
	}

	private void removeSubtype(ComposableType subtype) {
		subtypes.remove(subtype);
		directSubtypes.remove(subtype);

		for (int i = 0; i < supertypes.size(); i++) {
			((ComposableType) supertypes.get(i)).removeSubtype(subtype);
		}
	}

	public List getPartTypes() {
		return partTypes;
	}

	public boolean hasConstraints() {
		return constraints.size() > 0;
	}
}
