/*
This file is part of KumbangTools

    KumbangTools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KumbangTools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KumbangTools; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*
 * Created on Feb 12, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package kumbang.core.model.type;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

/**
 * Represents both part and subfeature definitions.
 * 
 * Could add a string stating whether this is a subfeature or a part
 * definition: would be useful when constructing error messages.
 *
 * I think it is better to use lists over vectors. Some reasons:
 *  - they are cheaper to serialize
 *  - they have better performance, especially for iterators 
 *    (vectors are synchronized, which is not needed here)
 *
 * @author toasikai
 *
 */
public class PartDefinition extends ElementaryDefinition {
	public static final int simNone = 0, simSame = 1, simDifferent = 2;
	
	private Cardinality cardinality;
	private List possibleTypes;
	private int similarity;
	
	/**
	 * 
	 * @param name As defined in the original description
	 * @param types An array of possible part types (FeatureTypes or ComponentTypes)
	 * @param lower The lower bound of cardinality
	 * @param upper The upper bound of cardinality
	 */
	public PartDefinition(Name name, Collection types, int lower, int upper, int similarity) {
		super(name);
		
		cardinality = new Cardinality(lower, upper);
		possibleTypes = new ArrayList(types);
		this.similarity = similarity;
	}
	
	/**
	 * 
	 * @param name As defined in the original description
	 * @param types An array of possible part types (FeatureTypes or ComponentTypes)
	 * @param lower The lower bound of cardinality
	 * @param upper The upper bound of cardinality
	 */
	public PartDefinition(String name, Collection types, int lower, int upper, int similarity) {
		super(name);
		cardinality = new Cardinality(lower, upper);
		possibleTypes = new ArrayList(types);
		this.similarity = similarity;
	}
	
	/**
	 * Construct a part definition with no similarity restrictions.
	 * @param name As defined in the original description
	 * @param types An array of possible part types (FeatureTypes or ComponentTypes)
	 * @param lower The lower bound of cardinality
	 * @param upper The upper bound of cardinality
	 */
	public PartDefinition(String name, Collection types, int lower, int upper) {
		this(name, types, lower, upper, simNone);
	}	
	
	public Iterator getPossibleTypes() {
		return possibleTypes.iterator();
	}
	
	public int getLowerBound() {
		return cardinality.lower();
	}
	
	public void setLowerBound(int lower) {
		cardinality.setLowerBound(lower);
	}
	
	public int getUpperBound() {
		return cardinality.upper();
	}
	
	public void setUpperBound(int upper) {
		cardinality.setUpperBound(upper);
	}
	
	public int typeCount() {
		return possibleTypes.size();
	}
	
	public int getSimilarity() {
		return similarity;
	}

	public void addPossibleType(ComposableType composableType) {
		possibleTypes.add(composableType);
	}
	
	public void removePossibleType(ComposableType composableType) {
		possibleTypes.remove(composableType);
	}
	
	public void setPossibleTypes(List types) {
		possibleTypes = types;
	}	
	
	/**
	 * Returns a list containing all the possible types that may occur as instances of the
	 * part.
	 * 
	 * @param isAbstractIncluded if true, abstract types are included, otherwise not
	 * @return an iterator ranging over all possible types that may occur as instances of the part
	 */
	public List getAllPossibleTypes(boolean isAbstractIncluded) {
		ArrayList allTypes = new ArrayList();
		
		for (Iterator types = getPossibleTypes(); types.hasNext();) {
			ComposableType possibleType = (ComposableType) types.next();
			if (!possibleType.isAbstract() || isAbstractIncluded) {
				if (!allTypes.contains(possibleType))
					allTypes.add(possibleType);
				else
					continue;	//	All the types under possibleType are already there.
			}

			// Iterate through the subtypes of the current type (possibleType)
			Iterator subTypes = possibleType.getSubtypes();
			while (subTypes.hasNext()) {
				possibleType = (ComposableType) subTypes.next();
				if (!possibleType.isAbstract() || isAbstractIncluded) {
					if (!allTypes.contains(possibleType))
						allTypes.add(possibleType);
					else
						continue;
				}
			}
		}
		
		return allTypes;
	}
	
	/**
	 * Returns an iterators over the list returned by getAllPossibleTypes(isAbstractIncluded).
	 * 
	 * @param isAbstractIncluded see getAllPossibleTypes
	 * @return see getAllPossibleTypes
	 */
	public Iterator getAllPossibleTypesIterator(boolean isAbstractIncluded) {
		return getAllPossibleTypes(isAbstractIncluded).iterator();
	}

	/**
	 * Checks whether the given type is a possible type in this part.
	 * @param type
	 * @return boolean
	 */
	public boolean isPossibleType(ComposableType type) {
		return possibleTypes.contains(type);
	}
	
	/**
	 * Returns the size of this part definition, i.e., the number of instances that may occur as
	 * part defined by this definition, under a given parent instance. 
	 * 
	 * @return see above
	 */
	public int getSize() {
		int ret = 0;
		for (Iterator it = getAllPossibleTypesIterator(false); it.hasNext(); ) {
			ComposableType type = (ComposableType) it.next();
			ret += type.getSize() * (similarity == simDifferent ? 1 : getUpperBound()); 
		}
		
		return ret;
	}
	
	/**
	 * 
	 * 
	 * @return
	 */
	public int getTypeOffset(ComposableType type) {
		int ret = 0;
		
		for (Iterator it = getAllPossibleTypesIterator(false); it.hasNext(); ) {
			ComposableType comp = (ComposableType) it.next();
			if (comp.equals(type))
				return ret;
			
			ret += comp.getSize() * (similarity == simDifferent ? 1 : getUpperBound());
		}
		
		// Error, definition was not found.
		return -1;
	}
	
	public void setSimilarity(int similarity) {
		this.similarity = similarity;
	}
	
	public boolean equals(Object o) {
		if (!(o instanceof PartDefinition))
			return false;
		
		return getName().equals(((PartDefinition) o).getName()); 
	}

	public int getPossibleTypesCount() {
		return possibleTypes.size();
	}

	public List getPossibleTypesList() {
		return new ArrayList(possibleTypes);
	}
}
